Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude pcm unionmap heap codes feaps state zigzag worlds.
Require Import inject.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive. 

(************************************)
(* Actions over a partitioned state *)
(************************************)

Section Defs.
Variables (coh : cohpred) (guar : guar coh) (V : Type).
(* guar - the guarantee under which the action lives *)
(* V - type of return values *)

(* Each action comes with a safe relation, which is its precondition. *)

Variable safe : Pred state.

(* An action is only safe in some coherent states; that is safe ->
coherent. *)

Definition safe_coh := forall s, s \In safe -> s \In coh. 

(* If an action is safe in [lp, sh, ep \* h], then it has to be safe
in [lp \* h, sh, ep]. Why? For parallel composition to work. i.e., so
that the thread containing this action (and owning lp) can be composed
in parallel with other threads (owning h). 

Note 1: this shows that safety can't depend on ep, since ep can be
emptied by zipping.

Note 1': thus, maybe we'd need safety monotonicity property: 
  forall s p, s \In safe -> s \> p \In coh -> s \> p \In safe.

Well, not quite; ep can't be fully emptied; it can only be shuffled. 
In particular, actions that transfer stuff won't satisfy the property 1'
because adding heap h to s may be safe, but adding h to s \> p need not
be safe, if p already adds h. But again, it's not clear that transfer actions
are all that important.

Note 2: safe_zagzig is unidirectional property, unlike coh_zig which
was if-and-only-if. OTOH, coh_zagzig was also a unidirectional property.

Note 3: we can't make safety a bidirectional property, as then some local state
which was essential for the action (e.g., a pointer being read), may be pushed
onto ep, and thus become out of reach.
 *)

Definition safe_zagzig := forall s p, s \> p \In safe -> s \< p \In safe. 

(* some actions, however, do satisfy the opposite property *)
(* (and it's important for modularity reasoning to export that) *)
(* these are actions which don't depend on the local state, because their *)
(* footprint is empty, or because they depend only on the shared state *)
(* typical examples are "half-actions" for acquiring and releasing *)
Definition safe_zigzag := forall s p, s \< p \In safe -> s \> p \In safe.

(* Safe should be local. If an action is safe for s1 and we get an s
such that (s1 \+ s) is coherent, then the action is safe for (s1 \+ s)
too. Although the status of locality with the example worlds we
currently have is very unclear, due to the "tight" interpretation of
coherence (cf. Ruy's email). 

The coherence always seems to determine the lables of s uniquely. Moreover
because of safe_coh, a safe state has to have those very same labels. 
Thus, s1 and s1 \+ s can only both be safe if s is not really adding
any new labels (ie., s is empty). So this seems like a dubious 
property.
*)

(* in principle, i think safe_local should not be a property required *)
(* of actions; it's just, when two modules are combined, the combination *)
(* functor should also lift the underlying actions to the combined module *)

(*
Definition safe_local := forall s1 s, 
    s1 \+ s \In coherent -> s1 \In safe -> s1 \+ s \In safe. 
*)

(* Each action comes with a step relation, which describes how the
action modifies the state. *)

Variable step : state -> state -> V -> Prop.

(* stepping is only considered in safe states *)
Definition step_safe := forall s1 s2 v, step s1 s2 v -> s1 \In safe. 

(* Step relation should be a subset of the guarantee. This makes it
immediately apparent that step does not touch env, and that step only
relates coherent states (i.e. this gives us step equivalents of
guar_inv and guar_coh) *)

Definition step_guar := forall s1 s2 v, step s1 s2 v -> guar s1 s2. 

(* We also need the equivalents of guar_zagzig and guar_local, as these
don't follow from step -> guar (they are more of "commuting diagram"
properties). *)

Definition step_zagzig := forall s s' p v, 
  env s = env s' -> step (s \> p) (s' \> p) v -> step (s \< p) (s' \< p) v.

(* the next one also seems useless, like safe_local *)
(*
Definition step_local := forall s1 s2 s v, 
  s1 \+ s \In coherent -> step s1 s2 v -> step (s1 \+ s) (s2 \+ s) v.
*)

Definition step_frame := forall s s' p v, 
  s \> p \In safe -> step (s \< p) s' v -> 
  exists s'', s' = s'' \< p /\ step (s \> p) (s'' \> p) v.

(*
it seems the following are not needed

Step relation is a total function on safe states 
*)

(* i.e., it produces a result *)
Definition step_total := 
  forall s1, s1 \In safe -> exists s2, exists v, step s1 s2 v.

(* and the result is unique *)
Definition step_det := forall s s1 s2 v1 v2, 
  step s s1 v1 -> step s s2 v2 -> s2 = s1 /\ v2 = v1. 

(*
The above will have to be generalized when we introduce
non-determinism, in the following way. We will parametrize actions
wrt. an external choice. E.g. if the action is supposed to allocate
some location, the address of the location is given in the
parameter. These external parameters are passed in by the schedule
(see file schedule.v). The schedule ChoiceAct will take this choice
as a parameter and pass it to the action for execution.

Then an action may be safe for some state, but still not be total for that
state, because the schedule may pass it a bad external choice, for which
the action diverges. In that case we have two options:

(1) Drop the step_total requirement (but not step_det one). Instead,
what will suffice to get the theorems through is the framing property
as proved below.

(2) Consider schedules which are not bad, by defining a predicate
"sensible" on shedules. Then the always predicate is proved only for
sensible schedules. The framing property, proved below, is basically a
statement of monotonicity of sensible schedules (as I show further).

The options (1) and (2) are essentially the same. To see this, consider
implementing (2). Then the following has to be resolved as a first step.

The main question with defining sensible schedules is, when is a
schedule sensible for an action? I guess, only when the schedule is of
the form ChoiceAct u, and the parameter u is such that the action
terminates for it.  So, a parameter is sensible if-and-only-if the
action terminates for it.

Then the step_det lemma would say: 
 
  step u s s1 v1 -> step u s s2 v2 -> s1 = s2 /\ v1 = v2

where u, by dint of being a parameter for which step terminates,
must be sensible.

Now, a frame lemma says -- among other things -- that if we have a sensible 
parameter for a larger state, than that parameter is sensible for a 
smaller state (asuming safety for the smaller state).

  safe (s1 \> p) ->
  step u (s1 \< p) s2 v -> 
  exists s2', s2 = s2' \< p /\ step u (s1 \> p) (s2' \> p).

Strictly speaking, the monotonicity of sensibility would have been
expressed by

  safe (s1 \> p) ->
  step u (s1 \< p) s2 v ->
  exists s2' v', step u (s1 \> p) s2' v'

so frame adds a few more things relating s2 s2' v and v'. In
particular, it adds that v = v' and s2 and s2' are a zig and a zag of
the same state.  These seem quite useful for expressing the invariance
property, namely that the execution in the larger state leaves the
extra chunk invariant.

These invariances don't seem derivable from the other properties that
we may assume of actions, e.g. determinism and zagzig (as zagzig goes
the other way).  So, the frame property is useful in the current form,
and we should keep it. In particular, we can't replace it in favor of
totality and determinacy, as totality will have to be replaced by
exactly the frame property as soon as we go into non-deterministic
schedules.  

However, for the current set of actions, we can reduce proving framing
to proving totality and determinacy, which seems to be simpler. This
is shown by a lemma below. We don't assume totality and determinacy
among the properties of action, but we can use them as a specific
patterns for proving framing. 

frame is derivable from totality, determinism, zagzig and guar lemmas *)

Lemma total_det_frame : 
        step_zagzig -> step_guar -> step_total -> step_det ->
        step_frame.
Proof.
move=>Sz Sg St Sd s s' p v; case/St=>s1 [w] H.
move/Sg/guar_env: H (H)=>/= E H.
have [s1' [E1 E2]]: exists s1', s1 = s1' \> p /\ env s = env s1'.
- by case: s1 E H=>lp sh ep /= <- H; exists [lp, sh, env s]; rewrite spP. 
rewrite {s1}E1 in E H E2 *.
by move/Sd; case/(_ _ _ (Sz _ _ _ _ E2 H))=><-<-; exists s1'. 
Qed.


(* Can we prove step_zagzig out of the step_frame and safe_zagzig? *)
(* In Abstract SL, we could, but here it seems we can't. How do we explain *)

Lemma step_zagzig_provable : 
        step_total -> 
        step_safe -> step_frame -> safe_zagzig -> step_zagzig.
Proof.
move=>T Ss F Sz s s' p v E.
rewrite /step_frame in F.
rewrite /safe_zagzig in Sz.
move=>St.
move: (Ss _ _ _ St). move=>H1. 
move: (Sz _ _ H1). 
move/T. case=>t [w] H2.  move/F: H2 (H2). move/(_ H1). case=>s''. case.  
move=>->. move=>S.
(* Now don't see how to derive that s'' = s' *)
Abort.

End Defs.


Module Action.
Section Defs.
Variables (coh : cohpred) (guar : guar coh) (V : Type).

Structure action := Action {
  safe : Pred state;
  step : state -> state -> V -> Prop;
  _ : safe_coh coh safe; 
  _ : safe_zagzig safe;
(*  _ : safe_local W safe; *)
  _ : step_safe safe step;
  _ : step_guar guar step;
(*  _ : step_zagzig step; *)
(*  _ : step_local W step; *)
  _ : step_frame safe step;
  _ : step_total safe step}.

End Defs.
End Action.

Notation action := Action.action.
Notation Action := Action.Action.

(*****************************************)
(* Repackaging of definitions and lemmas *)
(*****************************************)

Section Repack.
Variables (coh : cohpred) (guar : guar coh) (V : Type).

Definition asafe (a : action guar V) := Action.safe a.
Definition astep s1 (a : action guar V) v s2 := Action.step a s1 v s2. 

Variable (a : action guar V).

Lemma asf_coh s : s \In asafe a -> s \In coh. 
Proof. by case: a=>sf st sfc sfz stsf stg stf tot; apply: sfc. Qed.

Lemma asf_zagzig s p : s \> p \In asafe a -> s \< p \In asafe a.
Proof. by case: a=>sf st sfc sfz stsf stg stf tot; apply: sfz. Qed.

(*
Lemma asf_local s1 s : 
        s1 \+ s \In coherent -> s1 \In asafe a -> s1 \+ s \In asafe a. 
Proof. 
by case: a=>sf st sfc sfz sfl stsf stg stz stl stf ste; apply: sfl. 
Qed.
*)

Lemma ast_safe s1 s2 v : astep s1 a s2 v -> s1 \In asafe a.
Proof. by case: a=>sf st sfc sfz stsf stg stf tot; apply: stsf. Qed.

Lemma ast_guar s1 s2 v : astep s1 a s2 v -> guar s1 s2.
Proof. by case: a=>sf st sfc sfz stsf stg stf tot; apply: stg. Qed.

(*
Lemma ast_zagzig s s' p v : 
        env s = env s' -> 
        astep (s \> p) a (s' \> p) v -> 
        astep (s \< p) a (s' \< p) v. 
Proof. by case: a=>sf st sfc sfz stsf stg stz stf; apply: stz. Qed.
*)

(*
Lemma ast_local s1 s2 s v :
        s1 \+ s \In coherent -> astep s1 a s2 v -> 
        astep (s1 \+ s) a (s2 \+ s) v.
Proof. 
by case: a=>sf st sfc sfz sfl stsf stg stz stl stf ste; apply: stl. 
Qed.
*)

Lemma ast_frame s s' p v : 
        s \> p \In asafe a -> 
        astep (s \< p) a s' v -> 
        exists s'', s' = s'' \< p /\ astep (s \> p) a (s'' \> p) v. 
Proof. by case: a=>sf st sfc sfz stsf stg stf tot; apply: stf. Qed.

Lemma ast_total s1 : s1 \In asafe a -> exists s2, exists v, astep s1 a s2 v. 
Proof. by case: a=>sf st sfc sfz stsf stg stf tot; apply: tot. Qed.

(*
Lemma ast_det s s1 s2 v1 v2 : 
        astep s a s1 v1 -> astep s a s2 v2 -> s2 = s1 /\ v2 = v1.
Proof. 
by case: a=>sf st sfc sfz sfl stsf stg stz stl stt std ste; apply: std. 
Qed.
*)

(* additional lemmas *)

Lemma ast_coh s1 s2 v : astep s1 a s2 v -> s1 \In coh /\ s2 \In coh.
Proof. 
move=>H; split; first by move/ast_safe/asf_coh: H. 
by case/ast_guar/guar_coh: H. 
Qed.

Lemma ast_env s1 s2 v : astep s1 a s2 v -> env s1 = env s2.
Proof. by move/ast_guar/guar_env. Qed.

End Repack.

Prenex Implicits asf_coh asf_zagzig ast_safe ast_guar 
                 ast_frame ast_coh.

(*
Implicit Arguments ast_zagzig_joinL.
Implicit Arguments ast_zagzig_joinR.
*)

(*********************************)
(* internal vs. transfer actions *)
(*********************************)

Section Defs2.
Variables (coh : cohpred) (A : Type).

(* stepping only depends on the concrete heap *)
(*
Definition step_erase := forall s1 v1 t1 s2 v2 t2, 
  flats s1 = flats s2 -> step s1 t1 v1 -> step s2 t2 v2 -> 
  v1 = v2 /\ flats t1 = flats t2.
*)

(* a different definition of step_erase, to allow stepping in larger heaps *)
(* At first sight, this looks like some kind of framing property -- it seems to explicitly say *)
(* that no new memory is produced by stepping *)
(* But it's not really true; it's simply saying that heaps are unchanged *)
(* by the action. The fact that no new memory is produced is given by guar_footH *)
(* and may be frazed by valid (flats s1 \+ h1) -> step s1 t1 v1 -> valid (flats t1 \+ h1) *)

Definition int_step_erase (guar : int_guar coh) (a : action guar A) := 
  forall s h t v s' h' t' v', 
    valid (flats s \+ h) -> 
    flats s \+ h = flats s' \+ h' -> 
    astep s a t v -> astep s' a t' v' ->
    v = v' /\ flats t \+ h = flats t' \+ h'.

Definition acq_step_erase (guar : acq_guar coh) 
                          (a : forall k, action (guar k) A) :=
  forall s h t v k s' h' t' v' k', 
    valid (flats s \+ (k \+ h)) -> 
    flats s \+ (k \+ h) = flats s' \+ (k' \+ h') -> 
    astep s (a k) t v -> astep s' (a k') t' v' ->
    v = v' /\ flats t \+ h = flats t' \+ h'.

Definition rel_step_erase (guar : rel_guar coh) 
                          (a : forall k, action (guar k) A) :=
  forall s h t v k s' h' t' v' k', 
    valid (flats s \+ h) -> 
    flats s \+ h = flats s' \+ h' -> 
    astep s (a k) t v -> astep s' (a k') t' v' ->
    v = v' /\ flats t \+ (k \+ h) = flats t' \+ (k' \+ h').

(* the above erasure lemmas could be generalized to the case when we *)
(* transfer some other type of data, and not just heaps. basically *)
(* we have to change k and k' into flat k and flat k'. in the new setting *)
(* k and k' will be of arbitrary ghost type, but flat will determine the *)
(* concrete heap for that type. in the case of heaps, flat is identity *)
(* in the case of other ghosts, it's most likely just the empty function *)

(* transfer actions uniquely determine what's to be transfered *)

(*
Definition acq_safe_prec (guar : acq_guarclass coh) k k' 
                         (a : action (guar k) A) (a' : action (guar k') A) :=
  forall s h h', 
    valid (k \+ h) ->
    k \+ h = k' \+ h' ->
    s \In asafe a -> s \In asafe a' -> k = k'.
*)

(*
(* in fact, this only matters for release actions *)
(* they are the ones having ownership of the transfered datum *)
(* so it's easier to impose on them that they should uniquely determine *)
(* this datum *)
Definition rel_safe_prec (guar : rel_guarclass coh) 
                         (a : forall k, action (guar k) A) :=
  forall s k k', s \In asafe (a k) -> s \In asafe (a k') -> k = k'.
*)

(* release actions own the heap they release *)
Definition rel_safe_owns (guar : rel_guar coh) 
                         (a : forall k, action (guar k) A) := 
  forall s k, s \In asafe (a k) -> exists h, flats s = k \+ h.

(* acquire actions when done, own the heap they acquired *)
Definition acq_step_owns (guar : acq_guar coh)
                         (a : forall k, action (guar k) A) :=
  forall s s' v' k, astep s (a k) s' v' -> exists h, flats s' = k \+ h.

End Defs2.

Structure int_action coh (guar : int_guar coh) A := IntAction {
  int_action_of :> action guar A; 
  int_ast_erase : int_step_erase int_action_of}.
  
Structure acq_action coh (guar : acq_guar coh) A := AcqAction {
  acq_action_of :> forall h, action (guar h) A; 
(*  asf_acq_prec : forall k k', 
    acq_safe_prec (acq_action_of k) (acq_action_of k'); *)
  acq_ast_erase : acq_step_erase acq_action_of; 
  acq_ast_owns : acq_step_owns acq_action_of}.

Structure rel_action coh (guar : rel_guar coh) A := RelAction {
  rel_action_of :> forall h, action (guar h) A;
(*  rel_asf_prec : rel_safe_prec rel_action_of; *)
  rel_ast_erase : rel_step_erase rel_action_of;
  rel_asf_owns : rel_safe_owns rel_action_of}.

(*
(* IMPORTANT: should we also require that safety is not changed by erasure? *)
(* that is, that safety also only depends on the underlying heap *)
Definition safe_erase := forall s1 s2, 
  flats s1 = flats s2 -> s1 \In safe -> s2 \In safe.
(* What about safety monotonicity? Maybe the precondition should be *)
(* flats s1 \+ h1 = flats s2 *)
(* this property is obviously false for the half-actions, because *)
(* their safety also depends on the acquired heap *)
(* but then, maybe those actions are not real anyway *)

(* counter-example: l \In loc s1, l \In env s2. flats s1 = flats s2, but
s1 is safe for reading l, and s2 is not *)
*)

(* the following is needed for ghost functions *)
Definition local (U : encoded_pcm) (f : U -> U) := 
  forall x y, valid (x \+ y) -> valid (f x) -> f (x \+ y) = f x \+ y.

(*
(* the following for lifting of actions *)
(* the second requirement that s1 \+ s2 \In coh' *)
(* is introduced so that the same property can be used for acquire actions too *)
(* without the second property, the acquire action need not be *)
(* framable, as we don't know that the s2 portion won't overlap *)
(* with the acquired heap *)
Definition guar_frame coh coh' (guar : guarclass coh) (guar' : guarclass coh') := 
  forall s1 s2 s1', 
    s1 \+ s2 \In coh' -> s1' \+ s2 \In coh' -> guar s1 s1' -> 
    guar' (s1 \+ s2) (s1' \+ s2).
*)

(*
Definition acq_guar_frame coh coh' (guar : acq_guarclass coh) (guar' : acq_guarclass coh') :=
  forall h s1 s2 s1', s1 \+ s2 \In coh' -> valid (flats (s1 \+ s2) \+ h) -> 
    guar h s1 s1' -> guar' h (s1 \+ s2) (s1' \+ s2).
*)

(* some example actions for specific worlds *)
(* for a PRIV world, we'd have one set of actions; probably allowing *)
(* reading and writing. for Lock we'd have another set, locking and unlocking *)
(* but it's not clear what spec we need to give to an entanglement of the two *)
(* so that we can expose that an entangled module can support read/write/lock/unlock *)
(* probably none; we'd just have to have the actions be defined for a module, not for *)
(* a signature of a module *)

(* Definitions for transfering *)
Section Xfers.
Variables (A B : Type) (priv lock : mod).
Variables (ag : acq_guar (Mod.coh priv)) (rg : rel_guar (Mod.coh lock)).
Variable  (acq : acq_action ag A).
Variable  (rel : rel_action rg B).

Definition safe_xfer s h := 
  exists s1 s2, 
    [/\ s = s1 \+ s2, validH s &
        s1 \In asafe (acq h) /\ s2 \In asafe (rel h)]. 

Definition uniq_xfer :=
  forall s h h', safe_xfer s h -> safe_xfer s h' -> h = h'.

End Xfers.


(* Framing of actions wrt. additional modules *)
(* Simplifies the definitions of future actions, by allowing us *)
(* to define the actions using small *module* footprint *)

Module FrameAct.
Section FrameAct.
Variables (A : Type) (coh coh' : cohpred). 
Variables (g : int_guar coh) (g' : int_guar coh').
Variable act : action g A.

(* we should require that guar is a subpiece of guar' *)

Variable (pf : leq_guar g g').

Definition act_safe s := 
  [/\ exists s1 s2, s = s1 \+ s2 /\ s1 \In asafe act &
      s \In coh'].

Definition act_step s s' (v : A) := 
  [/\ exists s1 s2 s1', 
        [/\ s = s1 \+ s2, s' = s1' \+ s2 &
            Action.step act s1 s1' v],
      s \In coh' & s' \In coh']. 

Lemma act_safe_coh : safe_coh coh' act_safe.
Proof. by move=>s []. Qed.

Lemma act_safe_zagzig : safe_zagzig act_safe.
Proof.
move=>s p [[s1][s2] [E Sf1]] C; split; last by apply/coh_zigzag.
case/zag_plusf: E Sf1 (asf_coh Sf1) (C); first by case/cohV/and3P: C.
move=>s1' [s2'][p1][p2][->->-> _] {C} Sf1 C1 C.
exists (s1' \< p1), (s2' \< p2); split; last by apply: asf_zagzig.
case/coh_zigzag/cohV/and3P: C=>C _ _; case/coh_zigzag/cohV/and3P: C1=>C1 _ _.
by apply: zig_plus=>//; rewrite !spP in C C1 *; move/validZUn: C=><-.
Qed.

Lemma act_step_safe : step_safe act_safe act_step.
Proof. 
move=>s1 v s2 [[s1'][s2'][s']][->-> St] C _; split=>//.
by exists s1', s2'; split=>//; apply: ast_safe St. 
Qed.

Lemma act_step_guar : step_guar g' act_step.
Proof.
move=>s1 s2 v [[s1'][s2'][s']][->-> St] C1 C2.
by apply: (proj2 pf)=>//; apply: ast_guar St. 
Qed.

(*
Lemma act_step_zagzig : step_zagzig act_step.
Proof.
move=>s s' p v E [[t][r][t'][H H' S] C C']; split; try by apply/coh_zigzag.
case: pf; case=>c' ? _; subst coh'; case: (ast_coh S)=>Ct Ct'.
case: (envzag_plusf C C' Ct Ct' E H H')=>t1' [t2'][r1'][r2'][pt][pr][].
case=>??? [??? X][Et Er] E1 E2; subst s s' p t r t'.
move/(zagzig_eq Er): X=>?; subst r2'.
exists (t1' \< pt), (r1' \< pr), (t2' \< pt); split=>//.
by apply/ast_zagzig.
Qed.
*)
 
Lemma act_step_frame : step_frame act_safe act_step.
Proof.
move=>s s' p v [[t][r][H] S1 C]; move: (asf_coh S1)=>C1. 
have V2 : valid r.
- by move/cohV: C; rewrite H; move/validR. 
have Vso2 : validSO r. 
- by move/cohSO: C; rewrite H; apply: validSOK; move/cohSO: C1. 
have Ve : valid (env (s \> p)) by case/cohV/and3P: C.
case/(zag_plusf Ve): H=>{Ve} s1 [s2][p1][p2][F1 F2 F3 F4]; subst s p t r.
case; case=>s1' [s2'][r][F1 F2 S _ C']; subst s'.
have [Cs1' C1'] : s1' \In coh /\ r \In coh by case/ast_coh: S. 
have Ve1 : valid (env (s1 \> p1)) by case/cohV/and3P: C1.
have Ve2 : valid (env (s2 \> p2)) by case/and3P: V2.
move/coh_zigzag: C=>C; move/coh_zigzag: C1=>C1.
have Vl1 : valid (loc (s1 \< p1)) by case/cohV/and3P: C1.
have {Vso2 V2} Vl2 : valid (loc (s2 \< p2)).
- rewrite -validSO_zigzag in Vso2.
  by move: V2; rewrite -(valid_zigzag Vso2); case/and3P.
rewrite zig_plus // in C F1.
have [E1' E2'] : s1' = s1 \< p1 /\ s2' = s2 \< p2.
- move/(coh_prec (cohV C) F1 C1): Cs1' (F1)=><-.
  by move/(joinsK (cohV C))=><-. 
rewrite {F1 Cs1' s1'}E1' {s2'}E2' in S C' *.
case/(ast_frame S1): S C' C1'=>s1' [->{r}] S C' C1'.
have Vl1' : valid (loc (s1' \< p1)) by case/cohV/and3P: C1'. 
have Ve1' : valid (env (s1' \> p1)) by case/coh_zigzag/cohV/and3P: C1'. 
exists (s1' \+ s2).
split; first by rewrite zig_plus.
split; try by apply/coh_zigzag; rewrite zig_plus.
by exists (s1 \> p1), (s2 \> p2), (s1' \> p1); rewrite !zag_plus.
Qed.

Lemma act_step_total : step_total act_safe act_step.
Proof.
move=>s [[s1]][s2][->{s}] Sf C.
case/ast_total: Sf=>s1' [v] St.
exists (s1' \+ s2), v. 
split=>//; first by exists s1, s2, s1'. 
by case/guar_coh: (leq_guar_internal pf C (ast_guar St)).
Qed.

Definition FrameAct_pre := 
  Action act_safe_coh act_safe_zagzig act_step_safe act_step_guar 
         act_step_frame act_step_total. 

End FrameAct.

Section Actions.
Variables (A : Type) (c1 c2 : cohpred).

Lemma int_step_erase (g1 : int_guar c1) (g2 : int_guar c2)
                     (act : int_action g1 A) (pf : leq_guar g1 g2) :
        int_step_erase (FrameAct_pre act pf). 
Proof.
move=>w1 h1 t1 v1 w2 h2 t2 v2 V E S1 S2; move: S1 S2 V E.
case; case=>s1 [y1][s1'][->-> S1 /cohV C1 /cohV C1'].
case; case=>s2 [y2][s2'][->-> S2 /cohV C2 /cohV C2'].
rewrite !flatsUn // -4!joinA => V E.
by apply: int_ast_erase V E S1 S2.
Qed.

(*
Section AcqLemmas.
Variables (g1 : acq_guar c1) (g2 : acq_guar c2).
Variables (act : acq_action g1 A) (pf : forall k, leq_guar (g1 k) (g2 k)).

Lemma acq_step_erase : acq_step_erase (fun k => FrameAct_pre (act k) (@pf k)). 
Proof.
move=>w1 h1 t1 v1 k1 w2 h2 t2 v2 k2 V E S1 S2; move: S1 S2 V E.
case; case=>s1 [y1][s1'][->-> S1 /cohV C1 /cohV C1'].
case; case=>s2 [y2][s2'][->-> S2 /cohV C2 /cohV C2'].
rewrite !flatsUn // -4!joinA -(joinCA k1) -(joinCA k2) => V E.
by apply: acq_ast_erase V E S1 S2.
Qed.

Lemma acq_step_owns : acq_step_owns (fun k => FrameAct_pre (act k) (@pf k)).
Proof.
move=>_ _ v k [[s1][s2][s1'][->-> St] _ C']; rewrite (flatsUn (cohV C')).
by case: (acq_ast_owns St)=>h ->; exists (h \+ flats s2); rewrite joinA.
Qed.

End AcqLemmas.

Section RelLemmas.
Variables (g1 : rel_guar c1) (g2 : rel_guar c2).
Variables (act : rel_action g1 A) (pf : forall k, leq_guar (g1 k) (g2 k)).

Lemma rel_step_erase : rel_step_erase (fun k => FrameAct_pre (act k) (@pf k)). 
Proof.
move=>w1 h1 t1 v1 k1 w2 h2 t2 v2 k2 V E S1 S2; move: S1 S2 V E.
case; case=>s1 [y1][s1'][->-> S1 /cohV C1 /cohV C1'].
case; case=>s2 [y2][s2'][->-> S2 /cohV C2 /cohV C2'].
rewrite !flatsUn // -4!joinA -(joinCA k1) -(joinCA k2) => V E.
by apply: rel_ast_erase V E S1 S2.
Qed.

Lemma rel_safe_owns : rel_safe_owns (fun k => FrameAct_pre (act k) (@pf k)).
Proof.
move=>s k [[s1][s2][-> Sf] C]; rewrite (flatsUn (cohV C)).
by case: (rel_asf_owns Sf)=>h ->; exists (h \+ flats s2); rewrite joinA.
Qed.

End RelLemmas. 
*)

End Actions.

Module Exports.

Definition FrameAct A c1 c2 g1 g2 act pf := 
  IntAction (@int_step_erase A c1 c2 g1 g2 act pf).

(*
Definition AcqFrameAct A c1 c2 g1 g2 act pf := 
  AcqAction (@acq_step_erase A c1 c2 g1 g2 act pf) 
            (@acq_step_owns A c1 c2 g1 g2 act pf).

Definition RelFrameAct A c1 c2 g1 g2 act pf := 
  RelAction (@rel_step_erase A c1 c2 g1 g2 act pf) 
            (@rel_safe_owns A c1 c2 g1 g2 act pf).
*)

End Exports.

End FrameAct.

Export FrameAct.Exports.

(* one of the important pieces of the lifting is the idle action *)
(* of any priv module, not just Priv hp *)

Module IdleAction.
Section IdleAction.
Variable W : mod.
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W).

Definition idle_safe s := s \In coh.

Definition idle_step s s' (r : unit) := s = s' /\ s \In coh.

Lemma idle_safe_coh : safe_coh coh idle_safe.
Proof. by []. Qed.

Lemma idle_safe_zagzig : safe_zagzig idle_safe.
Proof. by move=>s p; move/coh_zigzag. Qed.

Lemma idle_step_safe : step_safe idle_safe idle_step.
Proof. by move=>s s' v [->]. Qed.

Lemma idle_step_guar : step_guar guar idle_step.
Proof. by move=>s s' v [->]; apply: refl_guar_refl. Qed.

(*
Lemma idle_step_zagzig : step_zagzig idle_step.
Proof. by move=>s s' p v E [] /(zagzig_eq E)-> /coh_zigzag. Qed.
*)
 
Lemma idle_step_frame : step_frame idle_safe idle_step.
Proof. by move=>s _ p v C [<- _]; exists s. Qed.

Lemma idle_step_total : step_total idle_safe idle_step.
Proof. by move=>s; exists s, tt. Qed.

Definition Idle_pre := 
  Action idle_safe_coh idle_safe_zagzig idle_step_safe idle_step_guar
         idle_step_frame idle_step_total.

End IdleAction.

Lemma idle_step_erase W : int_step_erase (Idle_pre W). 
Proof. by move=>s1 h1 t1 [] s2 h2 t2 [] V E [<- _][<- _]. Qed.

Module Exports. 

Definition Idle W := IntAction (@idle_step_erase W). 

End Exports.
End IdleAction.

Export IdleAction.Exports.



(* Entanglement of PrivAcq and LockRel actions *)

Module EntangleAcqRel. 
Section EntangleAcqRel.
Variables (A B C : Type) (f : A * B -> C) (WA WB : mod).
Variables (ca : chan_core (Mod.coh WA)) (cb : chan_core (Mod.coh WB)).
Variables (pfa : ca \In Mod.cores WA) (pfb : cb \In Mod.cores WB).
Variable (acq : acq_action (Chan.acq ca) A).
Variable (rel : rel_action (Chan.rel cb) B).
Variable uniqxfer_pf : uniq_xfer acq rel.
Notation W := (entangle WA WB). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Definition act_safe s := exists h, safe_xfer acq rel s h. 

Definition act_step s s' (v : C) :=
  exists s1 s2 s1' s2' v1 v2, 
    [/\ s = s1 \+ s2, s' = s1' \+ s2', v = f (v1, v2), validH s &
        exists h, astep s1 (acq h) s1' v1 /\ astep s2 (rel h) s2' v2].

Lemma act_safe_coh : safe_coh coh act_safe.
Proof. 
move=>_ [h][s1][s2][-> Vh][/asf_coh C1 /asf_coh C2].
by split=>//; exists s1, s2.
Qed.

Lemma act_safe_zagzig : safe_zagzig act_safe.
Proof. 
move=>s p S; move: S (S)=>/act_safe_coh/cohSO Vs [h][s1][s2][E Vh]. 
case/zag_plusf: E; first by case/validHV/and3P: Vh.
move=>s1' [s2'][p1][p2][????]; subst s p s1 s2.
case=>[/asf_zagzig S1 /asf_zagzig S2]. 
have [V1' V2'] :  valid (loc (s1' \< p1)) /\ valid (loc (s2' \< p2)).
- by case/asf_coh/cohV/and3P: S1; case/asf_coh/cohV/and3P: S2.
exists h, (s1' \< p1), (s2' \< p2); split=>//.
- by rewrite zig_plus.
by rewrite validH_zigzag // validSO_zigzag. 
Qed.

Lemma act_step_safe : step_safe act_safe act_step.
Proof.
move=>_ _ _ [s1][s2][s1'][s2'][_][_][-> _ _ Vh][h][/ast_safe S1 /ast_safe S2].
by exists h, s1, s2.
Qed.

Lemma act_step_guar : step_guar guar act_step.
Proof.
move=>_ _ _ [s1][s2][s1'][s2'][v1][v2][->-> _ Vh][h][/= S1 S2].
have E : s1 \+ s2 \In Mod.coh (entangle WA WB).
- by split=>//; exists s1, s2; case/ast_coh: S1; case/ast_coh: S2. 
right; exists (Chan ca), (Chan cb); split; try by apply/Mem_map.
left; split=>//; exists s1, s2, s1', s2'; split=>//.
by exists h; move/ast_guar: S1; move/ast_guar: S2. 
Qed.

(*
Lemma act_step_zagzig : step_zagzig act_step.
Proof.
move=>s s' p [v1 v2] E St; case/guar_coh: (act_step_guar St)=>C C'.
case: St=>s1 [s2][s1'][s2'][E1 E2 Vh][h][/= St1 St2].
case: (ast_coh St1)=>Cs1 Cs1'.
case: (envzag_plusf C C' Cs1 Cs1' E E1 E2)=>t1' [t2'][r1'][r2'][pt][pr][].
case=>??? [????][Et Er] F1 F2; subst s s' p s1 s2 s1' s2'.
exists (t1' \< pt), (r1' \< pr), (t2' \< pt), (r2' \< pr); split=>//.
- by rewrite validH_zigzag //; move/coh_zigzag/cohSO: C. 
by exists h; split; apply/ast_zagzig.
Qed.
*)

Lemma act_step_frame : step_frame act_safe act_step.
Proof.
move=>s s' p b.
case=>h [r1][r2][E Vh][Sf1 Sf2]. 
case/zag_plusf: (E); first by case/validHV/and3P: Vh.
move=>s1 [s2][p1][p2][????] {Vh}; subst s p r1 r2.
move=>St; case: St (St)=>r1 [r2][r1'][r2'][v1][v2]. 
case=>E' ->{s'} ->{b} Vh [h'][St1 St2].
move/act_step_safe/act_safe_coh/cohSO=>Vso.
case/zig_plusf: E'; first by case/validHV/and3P: Vh.
move=>s1' [s2'][p1'][p2'][Es Ep ??]; subst r1 r2.
have [Vs Vp]: valid (s1 \+ s2) /\ valid (p1 \+ p2).
- by move/validHV/valid_zig: Vh.
have [{Es Ep} ????] : [/\ s1 = s1', s2 = s2', p1 = p1' & p2 = p2'].
- move/ast_coh/proj1: St1; move/asf_coh/coh_zigzag: Sf1. 
  by apply: coh_prec_zigL.
subst s1' s2' p1' p2'. 
move: (join_zagzig (cohSO (asf_coh Sf1)) (cohSO (asf_coh Sf2)) E)=>E'.
have ? : h = h'.
- apply: (@uniqxfer_pf ((s1 \< p1 \+ s2 \< p2)));
  rewrite /safe_xfer -{2}E'; eexists _, _; split=>//.
  - by split; apply: asf_zagzig. 
  by move/ast_safe: St1; move/ast_safe: St2.
subst h'.
case/(ast_frame Sf1): St1=>s1' [->{r1'}] St1.
case/(ast_frame Sf2): St2=>s2' [->{r2'}] St2.
exists (s1' \+ s2').
have [C1' C2'] : s1' \> p1 \In Mod.coh WA /\ s2' \> p2 \In Mod.coh WB. 
- by case/ast_coh: St1; case/ast_coh: St2. 
split. 
- rewrite zig_plus //.
  - by case/coh_zigzag/cohV/and3P: C1'. 
  by case/coh_zigzag/cohV/and3P: C2'. 
exists (s1 \> p1), (s2 \> p2), (s1' \> p1), (s2' \> p2), v1, v2.
split=>//; last first. 
- by exists h. 
- by rewrite -validH_zigzag.
rewrite zag_plus //.
- by case/cohV/and3P: C1'. 
by case/cohV/and3P: C2'. 
Qed.

Lemma act_step_total : step_total act_safe act_step.
Proof.
move=>s [h][s1][s2][->{s}] Vh [].
case/ast_total=>s1' [v1] St1.
case/ast_total=>s2' [v2] St2.
exists (s1' \+ s2'), (f (v1, v2)).
by exists s1, s2, s1', s2', v1, v2; split=>//; exists h. 
Qed.

Definition EntAcqRel := 
  Action act_safe_coh act_safe_zagzig act_step_safe act_step_guar
         act_step_frame act_step_total.

Lemma act_step_erase : int_step_erase EntAcqRel.
Proof.
move=>s h t v s' h' t' v' V E St St'.
have Vt : valid t by move/act_step_guar/guar_coh/proj2/cohV: St. 
have Vt' : valid t' by move/act_step_guar/guar_coh/proj2/cohV: St'. 
case: St St' V E=>s1 [s2][t1][t2][v1][v2][??? Vh] H; subst s t v.
case=>s1' [s2'][t1'][t2'][v1'][v2'][??? Vh'] H' Vf E; subst s' t' v'. 
have Vs : valid (s1 \+ s2) by move/validHV: Vh.
have Vs' : valid (s1' \+ s2') by move/validHV: Vh'.
move: Vf E; rewrite !flatsUn //.
case: H; case: H'=>k' /= [S1' S2'] k [S1 S2]. 
case: (acq_ast_owns S1)=>h1 F1; case/rel_asf_owns: (ast_safe S2)=>h2 F2.
case: (acq_ast_owns S1')=>h1' F1'; case/rel_asf_owns: (ast_safe S2')=>h2' F2'.
rewrite F2 F2' -!joinA -(joinCA (flats t2)) -(joinCA (flats t2'))=>V E.
case: (acq_ast_erase V E S1 S1')=>->{E}.
have {V}: valid (flats t1 \+ (h2 \+ h)).
- rewrite -(footeqH_validUnL (acq_guar_footH (ast_guar S1))).
  by rewrite -joinA.
rewrite F1 F1' -!joinA (joinCA h1) (joinCA h1') (joinA k) (joinA k') -F2 -F2'.
by move=>V E; case: (rel_ast_erase V E S2 S2')=>->.
Qed.

End EntangleAcqRel.

Module Exports.

Definition EntAcqRel A B C f WA WB ca cb pfa pfb acq rel xfer_pf := 
  IntAction (@act_step_erase A B C f WA WB ca cb pfa pfb acq rel xfer_pf). 

End Exports.
End EntangleAcqRel.

Export EntangleAcqRel.Exports.




