Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype.
Require Import Eqdep prelude idynamic pcm unionmap heap codes feaps state zigzag.
Require Import worlds.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(*********************************************************)
(*********************************************************)
(* splitting state expression according to module labels *)
(*********************************************************)
(*********************************************************)

(* The main specification predicates. *)

Definition disj (p1 p2 : pred nat) := forall x, x \in p1 -> ~~ (x \in p2). 

Definition filter_spec pp (s : state) s1 s2 := 
  disj pp.1 pp.2 -> valid s -> 
  {in sdom s, [predU pp.1 & pp.2] =i predT} ->
  [/\ s = s1 \+ s2, s1 = sfilter pp.1 s & s2 = sfilter pp.2 s].

Definition filter_spec' pp s qq ss := 
  pp = qq -> filter_spec pp s (odflt Unit ss.1) (odflt Unit ss.2).


(********************************************)
(* helper structure for searching in a list *)
(********************************************)

Module SeqFind.
Section Defs.
Variable A : Type.

Structure tagged_seq := Tag {untag : seq A}.
Local Coercion untag : tagged_seq >-> seq.

Definition recurse_tag := Tag.
Canonical found_tag s := recurse_tag s.

Definition find_spec x (t : tagged_seq) := x \In untag t.

Structure find x := Find {
  seq_of : tagged_seq;
  _ : find_spec x seq_of}.

Local Coercion seq_of : find >-> tagged_seq.

Lemma found_ax x t : find_spec x (found_tag (x :: t)).
Proof. by left. Qed.

Canonical found_form x t := Find (found_ax x t).

Lemma recurse_ax x y (f : find x) : find_spec x (recurse_tag (y :: f)).
Proof. by case: f; right. Qed.

Canonical recurse_form x y (f : find x) := Find (recurse_ax y f).

End Defs.

Module Exports.
Coercion untag : tagged_seq >-> seq.
Coercion seq_of : find >-> tagged_seq.
Canonical found_tag.
Canonical found_form.
Canonical recurse_form.
End Exports. 

End SeqFind.

Export SeqFind.Exports.


(**************************************************************************)
(* Helper structure for searching a label in an expression built up       *)
(* from unions and mem_seqs. Needed because we keep the module footprints *)
(* in a form of a predicate, rather than in a form of a seq. The latter   *)
(* is so as to make equality of cohpreds insensitive to element ordering  *)
(**************************************************************************)

Definition mem_seq' (x : seq nat) := nosimpl (mem_seq x).

Module PredFind.
Section Defs.

Structure tagged_pred := Tag {untag : pred nat}.
Local Coercion untag : tagged_pred >-> pred.

Definition mem_tag := Tag.
Definition right_tag p := mem_tag p.
Canonical left_tag p := right_tag p.

Definition find_spec x (p : pred nat) := x \in p.

Structure find x := Find {
  pred_of : tagged_pred;
  _ : find_spec x pred_of}.

Local Coercion pred_of : find >-> tagged_pred.

Lemma left_ax x (p1 : find x) p2 : 
        find_spec x (left_tag (predU p1 p2)).
Proof. by case: p1=>p1; rewrite /find_spec /= -!topredE /= => ->. Qed.

Canonical left_form x (p1 : find x) p2 := Find (left_ax p1 p2).

Lemma right_ax x p1 (p2 : find x) : 
        find_spec x (right_tag (predU p1 p2)).
Proof. 
by case: p2=>p2; rewrite /find_spec /= -!topredE /= => ->; rewrite orbT. 
Qed.

Canonical right_form x p1 (p2 : find x) := Find (right_ax p1 p2).

Lemma mem_ax x (s : SeqFind.find x) : find_spec x (mem_tag (mem_seq' s)).
Proof. 
case: s; case=>s /=.
rewrite /= /SeqFind.find_spec -toPredE /find_spec -topredE /= /mem_seq'.
elim: s=>[|a s IH] //=; case.
- by move=>->; rewrite eq_refl. 
by move/IH; rewrite orbC => ->. 
Qed.

Canonical mem_form x (s : SeqFind.find x) := Find (mem_ax s). 

End Defs.

Module Exports.
Coercion pred_of : find >-> tagged_pred.
Canonical left_tag.
Canonical left_form.
Canonical right_form.
Canonical mem_form.
End Exports.

End PredFind.

Export PredFind.Exports.


(***********************************************************************)
(* Helper structure for deciding base cases. Given a singleton state,  *)
(* based at a label x, and two sets/predicates p1 and p2, we decide to *)
(* push the state to the left or to the right, as follows:             *)
(*                                                                     *)
(* (1) to the left if it matches p1                                    *)
(* (2) to the right if it matches p2                                   *)
(* (3) if it matches neither, we have to be conservative and put the   *)
(* filters of the state into both left and right component.            *)
(***********************************************************************)

Module Singleton.
Section Defs.

Structure outcome_tp := to_left | to_right | to_neither. 

Variable x : nat. 

Structure tagged_preds := Tag {untag : prod (pred nat) (pred nat)}.
Local Coercion untag : tagged_preds >-> prod.

Definition neither_tag := Tag.
Definition right_tag p := neither_tag p.
Canonical left_tag p := right_tag p.

Definition divide_spec (pp : tagged_preds) (d : outcome_tp) := 
  match d with 
    to_left => x \in pp.1
  | to_right => x \in pp.2 
  | to_neither => true
  end.

Structure divide := Divide {
  pp_of : tagged_preds; 
  _ : {d : outcome_tp | divide_spec pp_of d}}. 

Definition outcome_val (d : divide) : outcome_tp := 
  let: Divide _ (exist d _) := d in d. 

Definition outcome (f : divide) s := 
  match outcome_val f with
    to_left => (Some s, None)  
  | to_right => (None, Some s) 
  | to_neither => 
      (Some (sfilter (pp_of f).1 s), 
       Some (sfilter (pp_of f).2 s))
  end.

Lemma left_spec (p1 : PredFind.find x) p2 : 
        divide_spec (left_tag (PredFind.untag p1, p2)) to_left.
Proof. by case: p1; case. Qed.

Canonical left_form (p1 : PredFind.find x) p2 :=
  @Divide (left_tag (PredFind.untag p1, p2)) 
          (exist _ to_left (left_spec p1 p2)). 

Lemma right_spec p1 (p2 : PredFind.find x) :
        divide_spec (right_tag (p1, PredFind.untag p2)) to_right.
Proof. by case: p2; case. Qed.

Canonical right_form p1 (p2 : PredFind.find x) := 
  @Divide (right_tag (p1, PredFind.untag p2)) 
          (exist _ to_right (right_spec p1 p2)). 

Lemma neither_spec pp : divide_spec (neither_tag pp) to_neither.
Proof. by []. Qed.

Canonical neither_form pp := Divide (existT _ to_neither (neither_spec pp)). 

Lemma interp_ax pp (f : divide) s : 
        sdom s =i pred1 x -> filter_spec' pp s (pp_of f) (outcome f s).
Proof.
move=>E1 ->{pp}; case: f=>pp [d] /=; case: d=>H D /= F E2 /=;
rewrite /divide_spec in H; last 1 first.
- split=>//; rewrite -{1}[s]sfilt_predT -(eq_in_sfilter E2) sfilt_predU.
  congr (_ \+ _); apply: eq_in_sfilter=>z.
  rewrite E1; move/eqP ->; rewrite inE /=.
  by case E: (x \in pp.1)=>//=; rewrite (negbTE (D _ E)).
- rewrite unitR; split=>//; last first.
  - rewrite -(sfilt_pred0 F); apply: eq_in_sfilter=>z.
    by rewrite E1; move/eqP=>->; rewrite (negbTE (D _ H)).
  rewrite -{1}[s]sfilt_predT -(eq_in_sfilter E2).
  apply: eq_in_sfilter=>z.
  by rewrite E1; move/eqP=>->; rewrite inE /= H. 
rewrite unitL; split=>//.
- rewrite -(sfilt_pred0 F); apply: eq_in_sfilter=>z.     
  rewrite E1; move/eqP ->; case E: (x \in pp.1)=>//.
  by rewrite (negbTE (D _ E)) in H.
rewrite -{1}[s]sfilt_predT -(eq_in_sfilter E2).
apply: eq_in_sfilter=>z. 
by rewrite E1; move/eqP=>->; rewrite inE /= H orbT. 
Qed.

End Defs.

Module Exports.
Coercion pp_of : divide >-> tagged_preds.
Canonical left_tag.
Canonical left_form. 
Canonical right_form. 
Canonical neither_form. 
End Exports.

End Singleton.

Export Singleton.Exports.


(***************************************************************)
(* Helper structure for iterating the decision over a union of *)
(* singletons.                                                 *)
(***************************************************************)

Module Filter.
Section Defs.

Structure tagged_state (s : state) := Tag {untag : state}.
Local Coercion untag : tagged_state >-> state.

Definition notfound_tag s := Tag s s. 
Definition basefull_tag := notfound_tag. 
Definition baseother_tag := basefull_tag.
Definition basejoint_tag := baseother_tag.
Definition baseself_tag := basejoint_tag. 
Canonical union_tag := baseself_tag.

(* Structure for filtering according to sets p1 and p2.  We need
additional helper set q1 and q2 which will always be instantiated with
p1 and p2. The helpers are required in order to set up the recursion
correctly in the canonical structures. For example, recursion in the
union_form requires that the p1 and p2 appear *before* the state
s. But the recusion in the base case structures requires that s
appears *before* the sets so that we can pick up the label x from s
before we recurse to look for x in the sets. To satisfy both
requirements, we duplicate p1 and p2 into q1 and q2, but in uses
always instantiate q1 = p1 and q2 = p2 *)

(* We keep the result in fields lefts, rights : option state.  We use
option to recognize when we the element was skipped. The alternative
is to return Unit elements, but then Units profilerate in the
result. We package the result into a separate structure to avoid
naming the result and generating projections for it
*)

Structure filter' pp s qq := Filter {
  state_of : tagged_state s; 
  _ : {oss | filter_spec' pp state_of qq oss}}. 

Local Coercion state_of : filter' >-> tagged_state.

Definition lefts pp s qq (f : filter' pp s qq) :=
  let: Filter _ (exist (l, _) _) := f in l.

Definition rights pp s qq (f : filter' pp s qq) :=
  let: Filter _ (exist (_, r) _) := f in r.

Local Notation filter pp s := (filter' pp s pp).

Definition sum_up := 
  [fun os1 os2 : option state =>
    if os1 is Some s1 then 
      if os2 is Some s2 then Some (s1 \+ s2) else os1
    else os2].

Lemma odflt_sumup os1 os2 : 
        odflt Unit (sum_up os1 os2) = odflt Unit os1 \+ odflt Unit os2.
Proof. by case: os1=>[s1|]; case: os2=>[s2|] //=; rewrite ?unitL ?unitR. Qed.

Lemma union_spec pp s s' (f : filter pp s) (f' : filter pp s') :
        filter_spec' pp (untag f \+ untag f') pp
           (sum_up (lefts f) (lefts f'), 
            sum_up (rights f) (rights f')).
Proof.
move=>_ D V H; rewrite !odflt_sumup !(sfiltUn _ V).
case: f V H=>x [[l r]] /= /(_ (erefl _)) X V H. 
case: f' V H=>x' [[l' r']] /= /(_ (erefl _)) X' V H. 
have H1 : {in sdom x, [predU pp.1 & pp.2] =i predT}.
- by move=>y /= S; apply: H; rewrite sdomUn // inE /= S.
have H2 : {in sdom x', [predU pp.1 & pp.2] =i predT}.
- by move=>y /= S; apply: H; rewrite sdomUn // inE /= S orbT.
case/(_ D (validL V) H1): X=>X <-<-; case/(_ D (validR V) H2): X'=>X' <-<-.
by split=>//; apply/esym; rewrite -!joinA -(joinCA (odflt Unit r)) joinA X X'.
Qed.

Canonical union_form pp s s' (f : filter pp s) (f' : filter pp s') :=
  @Filter pp (untag f \+ untag f') pp
          (union_tag _) 
          (exist _ _ (@union_spec pp s s' f f')).

Section BaseCaseSelfState.
Variables (U : encoded_pcm) (lp : U).

Lemma baseself_spec pp x (f : Singleton.divide x) : 
        filter_spec' pp (x@sf->lp) (Singleton.untag f)
                     (@Singleton.outcome x f (x@sf->lp)).
Proof.
apply: Singleton.interp_ax=>z.
by rewrite sfE /sdom /= inE fdomPt inE orbF eq_sym.  
Qed.

Canonical baseself_form pp x (f : Singleton.divide x) := 
  @Filter pp (x@sf->lp) (Singleton.untag f)
          (baseself_tag _)
          (exist _ _ (@baseself_spec pp x f)).

End BaseCaseSelfState.

Section BaseCaseJointState.
Variables (V : encoded_type) (sh : V).

Lemma basejoint_spec pp x (f : Singleton.divide x) : 
        filter_spec' pp (x@jn->sh) (Singleton.untag f)
                     (@Singleton.outcome x f (x@jn->sh)).
Proof.
apply: Singleton.interp_ax=>z.
by rewrite jnE /sdom /= inE fdomPt inE orbF eq_sym.
Qed.

Canonical basejoint_form pp x (f : Singleton.divide x) := 
  @Filter pp (x@jn->sh) (Singleton.untag f)
          (basejoint_tag _)
          (exist _ _ (@basejoint_spec pp x f)).

End BaseCaseJointState.

Section BaseCaseOtherState.
Variables (U : encoded_pcm) (ep : U).

Lemma baseother_spec pp x (f : Singleton.divide x) : 
        filter_spec' pp (x@ot->ep) (Singleton.untag f)
                     (@Singleton.outcome x f (x@ot->ep)).
Proof.
apply: Singleton.interp_ax=>z.
by rewrite otE /sdom /= inE /=.
Qed.

Canonical baseother_form pp x (f : Singleton.divide x) := 
  @Filter pp (x@ot->ep) (Singleton.untag f)
          (baseother_tag _)
          (exist _ _ (@baseother_spec pp x f)).

End BaseCaseOtherState.

Section BaseCaseFullState. 
Variable (U : encoded_pcm) (V : encoded_type) (lp : U) (sh : V) (ep : U).

Lemma basefull_spec pp x (f : Singleton.divide x) : 
        filter_spec' pp (x ->> [lp, sh, ep]) (Singleton.untag f)
                     (@Singleton.outcome x f (x ->> [lp, sh, ep])).
Proof.
apply: Singleton.interp_ax=>z.
by rewrite fsE /sdom /= inE fdomPt inE /= !orbb eq_sym.  
Qed.

Canonical basefull_form pp x (f : Singleton.divide x) :=
  @Filter pp (x ->> [lp, sh, ep]) (Singleton.untag f) 
          (basefull_tag _) 
          (exist _ _ (@basefull_spec pp x f)).

End BaseCaseFullState.

Lemma notfound_spec pp s : 
        filter_spec' pp s pp (Some (sfilter pp.1 s), Some (sfilter pp.2 s)).
Proof.
move=>_ D V E; split=>//=.
rewrite -{1}[s]sfilt_predT -(eq_in_sfilter E) sfilt_predU.
congr (_ \+ _); apply: eq_in_sfilter=>x /= X; rewrite inE /=.
by case R: (x \in pp.1)=>//=; rewrite (negbTE (D _ R)).
Qed.

Canonical notfound_form pp s :=
  @Filter pp s pp
          (notfound_tag s)
          (exist _ _ (@notfound_spec pp s)).
  
End Defs.

Module Exports.
Notation filter_state ps s := (filter' ps s ps). 
Coercion untag : tagged_state >-> state.
Coercion state_of : filter' >-> tagged_state.
Canonical union_tag.
Canonical union_form.
Canonical baseself_form.
Canonical basejoint_form. 
Canonical baseother_form. 
Canonical basefull_form. 
Canonical notfound_form. 

End Exports.
End Filter.

Export Filter.Exports.


(**************************************************************************)
(* Helper structure for unfolding the definition of fV. For some reason,  *)
(* the unification fails when unifying against fV, after which fV is      *)
(* unfolded and unification succeeds. Now, we want the unfolding and      *)
(* backtracking to happen ASAP, before we start the filtering algorithm   *)
(* (because we don't want the failure to happen during filtering).        *)
(*                                                                        *)
(* Thus, we build a separate clause that matches fV expressions against   *)
(* their possible heads (predU and mem_seq'), thus forcing the unfolding. *)
(**************************************************************************)

(* This also means that the automation won't work if the fV is not
concrete. I.e., if we have a coh variable, the automation will fail. *)

Module UnfoldfV.
Section Defs.

Structure tagged_pred := Tag {untag : pred nat}.
Local Coercion untag : tagged_pred >-> pred.

Definition union_tag := Tag.
Canonical base_tag p := union_tag p. 

Definition unfold_spec xs (p : pred nat) := xs = p. 

Structure unfolding xs := Unfold {
  pred_of : tagged_pred;
  _ : unfold_spec xs pred_of}.

Local Coercion pred_of : unfolding >-> tagged_pred.

Lemma union_spec p1 p2 : unfold_spec (predU p1 p2) (union_tag (predU p1 p2)).
Proof. by []. Qed.

Canonical union_form p1 p2 := Unfold (union_spec p1 p2).

Lemma base_spec s : unfold_spec (mem_seq' s) (base_tag (mem_seq' s)).
Proof. by []. Qed.

Canonical base_form s := Unfold (base_spec s). 

End Defs.

Module Exports.
Coercion pred_of : unfolding >-> tagged_pred.
Canonical base_tag.
Canonical union_form.
Canonical base_form.
End Exports.

End UnfoldfV.

Export UnfoldfV.Exports.


(*******************************************************************)
(* Helper structure for picking up all the arguments from the goal *)
(*******************************************************************)

Structure equals_cohpred (c : cohpred) := 
  Pack {unpack :> cohpred; _ : c = unpack}.

Canonical equate (c : cohpred) := @Pack c c (erefl _).

Structure pick_up (c1 c2 : cohpred) (xs1 xs2 : pred nat) (s : state)
                  (pf1 : xs1 = fV c1) (pf2 : xs2 = fV c2) := PickUp {
  c2_of : equals_cohpred c2;
  lefts : state;
  rights : state;
  _ : filter_spec (xs1, xs2) s lefts rights}.

Section StartStruct.
Variables (c1 c2 : cohpred) (xs1 xs2 : pred nat).
Variables (p1 : UnfoldfV.unfolding xs1) (p2 : UnfoldfV.unfolding xs2).
Variables (pf1 : UnfoldfV.untag p1 = fV c1) (pf2 : UnfoldfV.untag p2 = fV c2).
Variables (s : state) (f : filter_state (xs1, xs2) s).

Lemma start_ax : filter_spec (UnfoldfV.untag p1, UnfoldfV.untag p2) f
                             (odflt Unit (Filter.lefts f)) 
                             (odflt Unit (Filter.rights f)).
Proof.
case: p1 pf1; case=>p1' /= H1 pf1'; case: p2 pf2; case=>p2' /= H2 pf2'.
rewrite {p1'}pf1' {p2'}pf2' in H1 H2 *.
by case: f; case=>t /= [[l r]]; rewrite H1 H2; apply.
Qed.

Canonical start := PickUp pf1 pf2 (equate c2) start_ax.

End StartStruct.

Notation pick c1 c2 s := (@pick_up c1 c2 (fV c1) (fV c2) s (erefl _) (erefl _)).

Lemma split_coh (s : state) (c1 c2 : cohpred) (x : pick c1 c2 s) :
        s \In star_coh c1 (unpack (c2_of x)) <->
        [/\ s = lefts x \+ rights x, validH s, 
            lefts x \In c1 & rights x \In c2].
Proof.
case: x; case=>_ /= <- lft rgh; rewrite /filter_spec /= => H.
split=>[C|]; last by case=>-> Vh C1 C2; split=>//; exists lft, rgh. 
have D : disj (fV c1) (fV c2).
- move=>x; case: C=>Vh [s1][s2][?] C1 C2; subst s.
  case/cohfV: C1=>E1 _ _; case/cohfV: C2=>E2 _ _.
  case/and3P: (validHV Vh)=>/= V _ _.
  case: fvalidUn V=>// _ _ H1 _. 
  by rewrite -E1 -E2; apply: H1.
have T : {in sdom s, [predU fV c1 & fV c2] =i predT}.
- move=>x S; case: C=>Vh [s1][s2][?] C1 C2; subst s.
  case/cohfV: C1=>E1 E2 E3; case/cohfV: C2=>F1 F2 F3.
  rewrite !inE; case/or3P: S=>/=.
  - by rewrite fdomUn !inE /= E1 F1; case/andP. 
  - by rewrite fdomUn !inE /= E2 F2; case/andP. 
  by rewrite fdomUn !inE /= E3 F3; case/andP. 
case/(_ D (cohV C) T): H=>E1 E2 E3; split=>//.
- by move/cohH: C.
- by rewrite E2; apply/sfilt_star: C.
by rewrite E3; rewrite starC in C; apply/sfilt_star: C.
Qed.

(*
(*******************)
(* Testing example *)
(*******************)

Variables (hp hp' : nat) (lkT : lockT) (hL hE : heap) (s : state).
Notation l := (@label_of lkT).

Lemma xx : hp@sf->hL \+ hp@jn->tt \+ l@sf->tt \+ l@ot->tt \+ hp@ot->tt \+ s \In 
             Mod.coh (entangle (Lock lkT) (Priv hp)) -> True.
Proof.
case/split_coh=>/= _ _.
Abort.
*)
