Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype finmap.
Require Import Eqdep prelude idynamic pcm codes unionmap heap feaps state.
Require Import zigzag worlds inject rely retract actions auto_split.
Require Import privstate footprint model lemmas locks.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.


(**********************************)
(* Module for lock-based resource *)
(**********************************)

Structure lockT := LockT {
  lockish :> lockLike;
  lock_of : ptr 
}.

(* Precision won't become important until the unlock action *)
(* but I package it here, for simplicity *)

Module LockCoh.
Section LockCoh.
Variable lkT : lockT.          (* lock package *)
Notation l := (@label_of lkT). (* feap label *)
Notation lk := (@lock_of lkT). (* lock *)
Notation U := (@pcm_of lkT).   (* pcm *) 
Notation I := (@inv_of lkT).   (* resource invariant *)

(* Aleks: the initial idea was that gL and gE be represented by *)
(* their own feap branch: g ->> [gL, tt, gE] *)
(* For now, I'm instead pairing them up with mL and mE *)
(* This is in order to simplify dealing with flats *)
(* The way flats is defined right now, the separate-branch idea *)
(* would have added gL and gE into the footprint of a module, in *)
(* case gL, gE are heaps. This complicate matters, is not really what we *)
(* want, and should hopefully be resolved once we redesign feaps. *)
(* But, for now, I'm going with the expedient hack *)

Section ValidRes.
Variable P : lift mutexUnlifted -> U -> bool -> heap -> Prop.

Definition validRes s :=
  exists (mL mE : lift mutexUnlifted) (gL gE : U) (b : bool) h,
    s = l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] /\ 
    P (mL \+ mE) (gL \+ gE) b h.

End ValidRes.

Definition cohRes P := 
  [Pred s : state | [/\ validH s, validSO s & validRes P s]].

Definition locked m (g : U) b (h : heap) :=  
  [/\ m = up own, b = true & h = Unit].
Definition unlocked m g b h := [/\ m = up nown, b = false & I g h].
Definition either m g b h := locked m g b h \/ unlocked m g b h.

Definition coh := cohRes either. 

Lemma locked_either s : s \In cohRes locked -> s \In coh.
Proof. 
case=>Vh Vso [mL][mE][gL][gE][b][h][E] H; split=>//. 
by exists mL, mE, gL, gE, b, h; split=>//; left.
Qed.

Lemma unlocked_either s : s \In cohRes unlocked -> s \In coh.
Proof.
case=>Vh Vso [mL][mE][gL][gE][b][h][E] H; split=>//. 
by exists mL, mE, gL, gE, b, h; split=>//; right.
Qed.

(* Note: the lemmas are not actually proved for *)
(* coherence, but for cohRes P, where P varies. *)
(* This indicates we may want to change the signature *)
(* to expose that cohRes will always be instantiated *)
(* by locked, unlocked, or either *)

Lemma cohSO P s : s \In cohRes P -> validSO s. 
Proof. by case. Qed.

Lemma cohH P s : s \In cohRes P -> validH s. 
Proof. by case. Qed.

Lemma cohfV P s : s \In cohRes P -> footV_eq s (mem_seq' [:: l]).
Proof. by case=>_ _ [mL][mE][gL][gE][b][h][->]; rewrite fsE. Qed.

Lemma coh_weq P s s' : 
        s \In cohRes P -> s' \In cohRes P -> weq typeCode (shr s) (shr s').
Proof.
case=>_ _ [mL][mE][gL][gE][b][h][->] _ [_ _][mL'][mE'][gL'][gE'][b'][h'][->] _. 
by rewrite !fsE /=. 
Qed.

Lemma coh_sig P s : 
        s \In cohRes P -> 
        l \-> (Unit : lift mutexUnlifted * U) = [zunit loc s].
Proof. by case=>_ _ [mL][mE][gL][gE][b][h][->] _; rewrite fsE zunitPt. Qed.

Lemma coh_zigzag P s p : (s \< p) \In cohRes P <-> (s \> p) \In cohRes P.
Proof.
suff L : validRes P (s \< p) <-> validRes P (s \> p).
- split; case=>Vh Vso Vp.
  - by split; rewrite -?validH_zigzag // -?validSO_zigzag -?L.
  by split; rewrite ?validH_zigzag // ?validSO_zigzag ?L.
split; case=>mL [mE][gL][gE][b][hM][]; rewrite -[s]st_eta zigE zagE fsE.
- case; case/zipPtf=>[[m1 g1]][[m2 g2]][[->->->->->->]] H; rewrite zipPt -fsE.
  exists m2, (mE \+ m1), g2, (gE \+ g1), b, hM.
  by split=>//; rewrite !joinA -(joinC m1) -(joinC g1) !joinA.
case=>->->; case/zipPtf=>[[m1 g1]][[m2 g2]][[->->->->]] H; rewrite zipPt -fsE.
exists (m2 \+ mL), m1, (g2 \+ gL), g1, b, hM.
by split=>//; rewrite -!joinA (joinC m2) (joinC g2) -!joinA.
Qed.

End LockCoh.

Module Exports.

Section Exports.
Variable lkT : lockT. 
Notation l := (@label_of lkT). 
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation I := (@inv_of lkT).   
Notation either := (@either lkT).

Definition LockCohMixin := 
  CohPredMixin (@cohSO lkT either) (@cohH lkT either) (@cohfV lkT either) 
               (@coh_weq lkT either) (@coh_sig lkT either) 
               (@coh_zigzag lkT either).
Definition lock_coh := CohPred (coh lkT) LockCohMixin.

Lemma lock_flatsE (mgL mgE : lift mutexUnlifted * U) (h : heap) :
        flats (l ->> [mgL, h, mgE]) = h.
Proof. by rewrite fsE flatsE !pcm_flatPt tp_flatPtE unitL unitR. Qed. 

Lemma lockVM (mL mE : lift mutexUnlifted) (gL gE : U) (h : heap) : 
        l ->> [(mL, gL), h, (mE, gE)] \In lock_coh -> valid (mL \+ mE).
Proof. by move/cohSO; rewrite /validSO fsE /= zipPt allvalidPt; case/andP. Qed.

Lemma lockVG (mL mE : lift mutexUnlifted) (gL gE : U) (h : heap) : 
        l ->> [(mL, gL), h, (mE, gE)] \In lock_coh -> valid (gL \+ gE).
Proof. by move/cohSO; rewrite /validSO fsE /= zipPt allvalidPt; case/andP. Qed.

Lemma lockI (mgL mgE : lift mutexUnlifted * U) (h : heap) : 
        l ->> [mgL, h, mgE] \In lock_coh -> 
        exists (b : bool) h', h = lk :-> b \+ h'. 
Proof.
case=>_ _ [mL][mE][gL][gE][b][k][]; rewrite !fsE; case/stE=>/= _ H _.
by move/(fcancelPt (C:=typeCode)): H; exists b, k.
Qed.

Lemma lock_prec (mL mL' mE mE' : lift mutexUnlifted) (gL gL' gE gE'  : U) 
                (b b' : bool) (h h' : heap) : 
        l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] \In lock_coh ->
        l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] = 
        l ->> [(mL', gL'), lk :-> b' \+ h', (mE', gE')] ->
        [/\ (mL, gL) = (mL', gL'), (mE, gE) = (mE', gE'), b = b' & h = h'].
Proof.
move/cohH; rewrite /validH lock_flatsE !fsE => V /stE /= []. 
case/(fcancelPt (C:=pcmCode))=>->-> /(fcancelPt (C:=typeCode)).
by case/(hcancelV V)=>{V} -> _ -> /(fcancelPt (C:=pcmCode)) [->->]. 
Qed.

Lemma lock_prec_emp (mL mL' mE mE' : lift mutexUnlifted) (gL gL' gE gE'  : U) 
                    (b b' : bool) : 
        l ->> [(mL, gL), lk :-> b, (mE, gE)] \In lock_coh ->
        l ->> [(mL, gL), lk :-> b, (mE, gE)] = 
        l ->> [(mL', gL'), lk :-> b', (mE', gE')] ->
        [/\ (mL, gL) = (mL', gL'), (mE, gE) = (mE', gE') & b = b'].
Proof. 
by rewrite -[lk :-> b]unitR -[lk :-> b']unitR => C; case/(lock_prec C). 
Qed.

Lemma lock_inv (mL mE : lift mutexUnlifted) b (gL gE : U) (h : heap) :
        l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] \In lock_coh ->
        if b then h = Unit else I (gL \+ gE) h.
Proof.
move=>C; case: (C)=>Vh Vs [mL'][mE'][gL'][gE'][b'][h'][].
by case/(lock_prec C); case=><-<- [<-<-] <-<-; case; case=>_ ->. 
Qed.

Lemma lock_inv' (mL mE : lift mutexUnlifted) b (gL gE : U) (h : heap) :
        l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] \In lock_coh ->
        mL \+ mE = if b then up own else up nown. 
Proof.
move=>C; case: (C)=>Vh Vs [mL'][mE'][gL'][gE'][b'][h'][].
by case/(lock_prec C); case=><-<- [<-<-] <-<-; case; case=>->->. 
Qed.

(* for entangling *)
Lemma priv_lock_prec hp (hL hL' hE hE' : heap) 
                     (mL mL' mE mE' : lift mutexUnlifted) 
                     (gL gL' gE gE'  : U) (b b' : bool) (h h' : heap) :
        hp ->> [hL, tt, hE] \+ 
         l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] 
                \In star_coh (priv_coh hp) lock_coh ->
        hp ->> [hL, tt, hE] \+ 
         l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] = 
        hp ->> [hL', tt, hE'] \+ 
         l ->> [(mL', gL'), lk :-> b' \+ h', (mE', gE')] ->
        [/\ hL = hL', hE = hE', b = b', (mL, gL) = (mL', gL'), 
            (mE, gE) = (mE', gE') & h = h'].
Proof.
move=>C; case/split_coh: (C)=>/= _ _ Ch Ct.
by case/(scancel (cohV C)); case=>->-> _; case/(lock_prec Ct).
Qed.

End Exports.
End Exports.

End LockCoh.

Export LockCoh.Exports.

(*************************)
(* Guarantee transitions *)
(*************************)

Module LockGuar.
Section LockGuar.
Variable lkT : lockT. 
Notation l := (@label_of lkT). 
Notation lk := (@lock_of lkT).
Notation I := (@inv_of lkT).
Notation U := (@pcm_of lkT).
Notation coh := (lock_coh lkT).

(*** internal transition ***)
 
Definition int_rel s s' := s = s' /\ s \In coh.

Lemma intguar_coh s s' : int_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case=>->. Qed.

Lemma intguar_env s s' : int_rel s s' -> env s = env s'.
Proof. by case=>->. Qed.

Lemma intguar_zagzig s s' p : 
        env s = env s' ->
        int_rel (s \> p) (s' \> p) -> int_rel (s \< p) (s' \< p).
Proof.
move=>E [E' C]; split; last by apply/coh_zigzag.
by move: E'; rewrite !stE !spP; case=>->->. 
Qed.

Lemma intguar_footH s s' : int_rel s s' -> footeqH (flats s) (flats s').
Proof. by case=>->; move/cohH. Qed.

Lemma intguar_refl s : s \In coh -> int_rel s s. 
Proof. by []. Qed.

(*** acquire transition ***)

Section Acquire.
Variable h : heap.

Definition acq_rel s s' :=  
  exists gL gL' gE : U, 
    [/\ s  = l ->> [(up own, gL), lk :-> true, (up nown, gE)],
        s' = l ->> [(up nown, gL'), lk :-> false \+ h, (up nown, gE)],
        valid (gL' \+ gE), valid (lk :-> false \+ h) &
        I (gL' \+ gE) h] /\ s \In coh.

Lemma acqguar_coh s s' : acq_rel s s' -> s \In coh /\ s' \In coh.
Proof.
case=>gL [gL'][gE][[->-> V1 V2 inv]] C; split; first by apply: C.
split.
- by rewrite /validH lock_flatsE. 
- by rewrite fsE /validSO /= zipPt allvalidPt.  
by exists (up nown), (up nown), gL', gE, false, h; split=>//; right.  
Qed.

Lemma acqguar_env s s' : acq_rel s s' -> env s = env s'.
Proof. by case=>gL [gL'][gE][[->->]]; rewrite !fsE. Qed.

Lemma acqguar_zagzig s s' p : 
        env s = env s' -> 
        acq_rel (s \> p) (s' \> p) -> acq_rel (s \< p) (s' \< p).
Proof.
move=>H; case=>gL [gL'][gE][[H1 H2 V1 V2 inv]].
move/coh_zigzag; move: H1 H2 H V1 inv.
rewrite -[s]st_eta -[s']st_eta !zigE !zagE !fsE.
case=>->->; case/zipPtf=>[[m1 g1]][[m2 g2]][[]].
case/sym/mtx00E=>->->->{gE}->->.
case=>->->; case/zipPtf=>[[m1' g1']][[m2' g2']][[]].
case/sym/mtx00E=>->-> _ -> _. 
case/(fcancelPt (C:=pcmCode))=><-{g1'}.
rewrite !zipPt joinC -!fsE => V1 inv C. 
exists (g2 \+ gL), (g2 \+ gL'), g1. 
by do ![split=>//]; rewrite joinC joinA.  
Qed.

Lemma acqguar_footH s s' : acq_rel s s' -> footeqH (flats s \+ h) (flats s').
Proof.
case=>gL [gL'][gE][[->-> _ V _]] _; rewrite !lock_flatsE.
split=>//; first by rewrite !hvalidPtUn in V *.
by move=>x; rewrite !hdomPtUn !inE /= !hvalidPtUn. 
Qed.

End Acquire.

(*** release transition ***)

Section Release.
Variable h : heap.

Definition rel_rel s s' :=  
  exists gL gE : U, 
    [/\ s  = l ->> [(up nown, gL), lk :-> false \+ h, (up nown, gE)],
        s' = l ->> [(up own, gL), lk :-> true, (up nown, gE)] &
        s \In coh].

Lemma relguar_coh s s' : rel_rel s s' -> s \In coh /\ s' \In coh.
Proof.
case=>gL [gE][->->] H; split; first by apply: H.
split.
- case: H; rewrite /validH !lock_flatsE /=.
  by move/validL; rewrite !hvalidPt. 
- by move/cohSO: H; rewrite !fsE /validSO /= !zipPt !allvalidPt. 
exists (up own), (up nown), gL, gE, true, Unit; rewrite unitR. 
by split=>//; left. 
Qed.

Lemma relguar_env s s' : rel_rel s s' -> env s = env s'.
Proof. by case=>gL [gE][->->]; rewrite !fsE. Qed.

Lemma relguar_zagzig s s' p : 
        env s = env s' -> 
        rel_rel (s \> p) (s' \> p) -> rel_rel (s \< p) (s' \< p).
Proof.
move=>H; case=>gL [gE][H1 H2]; move/coh_zigzag; move: H1 H2 H.
rewrite -[s]st_eta -[s']st_eta !zigE !zagE !fsE.
case=>->->; case/zipPtf=>[[m1 g1]][[m2 g2]][[]].
case/sym/mtx00E=>->->->{gE}->-> /=.
case=>->->; case/zipPtf=>[[m1' g1']][[m2' g2']][[]].
case/sym/mtx00E=>->-> _ -> _.
case/(fcancelPt (C:=pcmCode))=><-{g1'}.
by rewrite !zipPt -!fsE; exists (g2 \+ gL), g1.
Qed.

Lemma relguar_footH s s' : rel_rel s s' -> footeqH (flats s) (h \+ flats s').
Proof.
case=>gL [gE][->->] /cohH; rewrite /validH !lock_flatsE (joinC h). 
move=>V; split; first by rewrite !hvalidPtUn in V *.
by move=>x; rewrite !hdomPtUn !inE /= !hvalidPtUn. 
Qed.

End Release.

End LockGuar.

Module Exports.
Section Exports.
Variable lkT : lockT. 

Definition LockGuarMixin := 
  GuarMixin (@intguar_coh lkT) (@intguar_env lkT) (@intguar_zagzig lkT).
Canonical lock_int_guar := Eval hnf in Guar (int_rel lkT) LockGuarMixin.

Definition LockIntGuarMixin := IntGuarMixin (@intguar_footH lkT).
Canonical LockIntGuar := Eval hnf in IntGuar lock_int_guar LockIntGuarMixin.

Definition LockReflGuarMixin := ReflGuarMixin (@intguar_refl lkT).
Canonical LockReflGuar := Eval hnf in ReflGuar lock_int_guar LockReflGuarMixin.

Definition baseLockAcqGuarMixin h := 
  GuarMixin (@acqguar_coh lkT h) (@acqguar_env lkT h) (@acqguar_zagzig lkT h).
Canonical lock_acq_guar h := 
  Eval hnf in Guar (acq_rel lkT h) (baseLockAcqGuarMixin h).

Definition LockAcqGuarMixin := AcqGuarMixin (@acqguar_footH lkT).
Canonical LockAcqGuar := Eval hnf in AcqGuar lock_acq_guar LockAcqGuarMixin.

Definition baseLockRelGuarMixin h := 
  GuarMixin (@relguar_coh lkT h) (@relguar_env lkT h) (@relguar_zagzig lkT h).
Canonical lock_rel_guar h := 
  Eval hnf in Guar (rel_rel lkT h) (baseLockRelGuarMixin h).

Definition LockRelGuarMixin := RelGuarMixin (@relguar_footH lkT).
Canonical LockRelGuar := Eval hnf in RelGuar lock_rel_guar LockRelGuarMixin.

Definition LockChanCore := ChanCore LockAcqGuar LockRelGuar.
Canonical LockCore : mod_core := @Mod.Core (lock_coh lkT) LockReflGuar. 
Canonical CLock : mod := @Mod.Make LockCore [:: LockChanCore]. 

End Exports.
End Exports.

End LockGuar.

Export LockGuar.Exports.


(*************************************)
(* SCSL module iterates entanglement *)
(*************************************)

Definition SCSL hp (ls : seq lockT) := 
  entangle (Priv hp) (apart_iter (map CLock ls)).


(**************************************************************)
(* Instances of decorate and mod_unit structures for caslocks *)
(**************************************************************)

Module CASLockDecorate.
Section CASLockDecorate.
Variable lkT : lockT.
Notation l := (@label_of lkT). 
Notation lk := (@lock_of lkT).
Notation I := (@inv_of lkT).
Notation U := (@pcm_of lkT).
Notation coh := (lock_coh lkT).
Notation T := ((lift mutexUnlifted * U)%type).

Definition caslock_dec (g : T) s := 
  s = l ->> [g, flats s, Unit] /\ s \In lock_coh lkT.

Lemma dec_prec g h1 h2 s1 s2 :
        valid (flats s1 \+ h1) -> 
        flats s1 \+ h1 = flats s2 \+ h2 ->
        caslock_dec g s1 -> caslock_dec g s2 ->
        s1 = s2.
Proof. 
case: g=>g1 g2 V E [-> /= C1][-> /= C2].
case: C1=>_ _ [mL1][mE1][gL1][gE1][b1][x1][] E1 H1. 
case: C2=>_ _ [mL2][mE2][gL2][gE2][b2][x2][] E2 H2.
case/scancelPt: E1 E2 V E H1 H2=>->->-><-<- /scancelPt [->->->->->] V.
rewrite -!joinA in V *.
case/(hcancelV V)=><-{b2} {V} V E H1 H2.
case: H1 H2 V E.
- case=>M1 ->->; case; last by case. 
  by case=>M2 _ -> _. 
case=>M1 -> inv1; case; first by case.
by case=>M2 _ inv2 V E; rewrite (precise_of V E inv1 inv2).
Qed.

Lemma dec_inj g1 g2 s : 
        caslock_dec g1 s -> caslock_dec g2 s -> g1 = g2.
Proof. 
case: g1 g2=>g11 g12 [g21 g22][-> /= _][]. 
by case/scancelPt=>->->->.
Qed.

Lemma dec_coh g s : caslock_dec g s -> s \In Mod.coh (CLock lkT).
Proof. by case. Qed.

Lemma dec_onto g1 s1 s2 : 
        caslock_dec g1 s1 -> s2 \In lock_coh lkT -> 
        env s1 = env s2 -> exists g2, caslock_dec g2 s2.
Proof. 
case=>-> /= C1 C2. 
case: C2 (C2)=>_ _ [mL][mE][gL][gE][b][h][->{s2} _] C2 H.
rewrite !fsE /= in H. 
move/(fcancelPt (C:=pcmCode)): H C2 =><- C2.
exists (mL, gL). 
split=>//.
by rewrite lock_flatsE.
Qed.

(*
Lemma dec_flat g h s : caslock_dec g h s -> flats s = h.
Proof. by case=>->; rewrite lock_flatsE. Qed.
*)


Lemma dec_env g1 g2 s1 s2 :
        caslock_dec g1 s1 -> caslock_dec g2 s2 -> env s1 = env s2.
Proof. 
by rewrite /caslock_dec !fsE; case=>-> C1 [-> C2]; split=>[->|] //=.
Qed.

End CASLockDecorate.

Module Exports.
Notation casloc_dec := caslock_dec.

Section Exports.
Variable lkT : lockT.
Notation l := (@label_of lkT). 
Notation lk := (@lock_of lkT).
Notation I := (@inv_of lkT).
Notation U := (@pcm_of lkT).
Notation coh := (lock_coh lkT).
Notation T := ((lift mutexUnlifted * U)%type).

Definition CASLockDecorateMixin := 
  DecorateMixin (@dec_prec lkT) (@dec_inj lkT) (@dec_coh lkT) 
                (@dec_onto lkT) (@dec_env lkT).
Canonical CASLockDecorate := 
  Eval hnf in Decorate (@caslock_dec lkT) CASLockDecorateMixin.

End Exports.
End Exports.

End CASLockDecorate.

Export CASLockDecorate.Exports.


(**********************)
(* Actions with locks *)
(**********************)

(* there's no primitive internal actions *)
(* so we do the transfer actions first *)

Module LockReleaseAct.
Section LockReleaseAct.
Variable lkT : lockT. 
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).

Section Internal.
Variable (h : heap). (* the resource heap *)
Notation W := (CLock lkT). 
Notation coh := (lock_coh lkT).
Notation guar := (lock_rel_guar lkT h).

(* the action is safe only if we actually have the heap h to give away *)
(* that is, the lock is unlocked. *)
(* this is clashing a bit with the naming, as we're actually *)
(* acquiring the lock, though releasing the heap *)
Definition rel_safe s := 
  exists gL gE : U,
    s = l ->> [(up nown, gL), lk :-> false \+ h, (up nown, gE)] /\ 
    s \In coh.
    
Definition rel_step (s s' : state) (r : unit) := 
  exists gL gE : U, 
    [/\ s  = l ->> [(up nown, gL), lk :-> false \+ h, (up nown, gE)], 
        s' = l ->> [(up own, gL), lk :-> true, (up nown, gE)] & 
        s \In coh].

Lemma rel_safe_coh : safe_coh coh rel_safe.
Proof. by move=>s [hL][hE][]. Qed.

Lemma rel_safe_zagzig : safe_zagzig rel_safe.
Proof.
move=>s p; rewrite -[s]st_eta; case=>gL [gE][] /stE /=.
rewrite fsE /=; case=>->-> /zipPtf [[m1 g1]][[m2 g2]][[]].
move/sym/mtx00E=>[->->] _ ->-> /coh_zigzag C.
by exists (g2 \+ gL), g1; split=>//; rewrite fsE zigE zipPt.
Qed.

Lemma rel_step_safe : step_safe rel_safe rel_step.
Proof. by move=>s s' [][gL][gE][-> _ C]; exists gL, gE. Qed.

Lemma rel_step_guar : step_guar guar rel_step. 
Proof. by move=>s s' [][gL][gE][->-> C]; exists gL, gE. Qed.

Lemma rel_step_frame : step_frame rel_safe rel_step.
Proof.
move=>s s' p []; rewrite -[s]st_eta -[s']st_eta /=.
case=>gL [gE][/stE /=]; rewrite fsE /=. 
case=>->-> /zipPtf [[m1 g1]][[m2 g2]][[]].
case/sym/mtx00E=>->-> _ ->-> C; rewrite zigE zipPt -fsE.
case=>_ [_][/scancelPt [<-<-]] /stE /= [->->-> _].
exists (l ->> [(up own, gL), lk :-> true, (up nown, g1)]).
rewrite !fsE /=; split; first by rewrite zigE zipPt. 
by exists gL, (g1 \+ g2); split=>//; rewrite fsE zagE zipPt.
Qed.

Lemma rel_step_total : step_total rel_safe rel_step.
Proof.
move=>s [gL][gE][->] C.
exists (l ->> [(up own, gL), lk :-> true, (up nown, gE)]), tt.
by exists gL, gE. 
Qed.

Definition LockRel_pre :=
  Action rel_safe_coh rel_safe_zagzig rel_step_safe rel_step_guar
         rel_step_frame rel_step_total.

End Internal. 

Lemma rel_st_erase : rel_step_erase LockRel_pre.
Proof.
move=>s1 h1 t1 [] k1 s2 h2 t2 [] k2 V E St1 St2.
case: St1 St2 V E=>gL1 [gE1][->-> C1][gL2][gE2][->-> C2].
rewrite !lock_flatsE -!joinA.
by move=>V; move/(joinfK V)=>->.
Qed.

Lemma rel_sf_owns : rel_safe_owns LockRel_pre.
Proof.
move=>s k [gL][gE][-> C]; exists (lk :-> false).
by rewrite lock_flatsE joinC.
Qed.

End LockReleaseAct.

Module Exports.

Definition LockRel lkT := RelAction (@rel_st_erase lkT) (@rel_sf_owns lkT).

End Exports.

End LockReleaseAct.

Export LockReleaseAct.Exports. 



Module LockAcquireAct.
Section LockAcquireAct.
Variable lkT : lockT.
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation I := (@inv_of lkT).
Variables (f : U -> U) (pf_loc : local f).

Section Internal.
Variable h : heap. (* the resource heap *)
Notation W := (Lock lkT). 
Notation coh := (lock_coh lkT). 
Notation guar := (lock_acq_guar lkT h).

(* the action is safe only if the lock is owned and the heap h *)
(* to release satisfies the invariant of the future ghost state *)
Definition acq_safe s := 
  exists gL gE : U, 
    [/\ s = l ->> [(up own, gL), lk :-> true, (up nown, gE)], 
        valid (f gL \+ gE), valid (lk :-> false \+ h),
        I (f gL \+ gE) h & s \In coh].
    
(* see if the proofs become shorter if acq_step uses acq_safe *)
(* instead of repeating all the validity conditions *)
Definition acq_step s s' (r : unit) := 
  exists gL gE : U, 
    [/\ s  = l ->> [(up own, gL), lk :-> true, (up nown, gE)], 
        s' = l ->> [(up nown, f gL), lk :-> false \+ h, (up nown, gE)], 
        valid (f gL \+ gE), valid (lk :-> false \+ h),
        I (f gL \+ gE) h & s \In coh].

Lemma acq_safe_coh : safe_coh coh acq_safe.
Proof. by move=>s [gL][gE][]. Qed.

Lemma acq_safe_zagzig : safe_zagzig acq_safe.
Proof.
move=>s p; rewrite -[s]st_eta; case=>gL [gE][] /stE /=.
rewrite fsE /=; case=>->-> /zipPtf [[m1 g1]][[m2 g2]][[]].
move/sym/mtx00E=>[->->] ->->->.
move=>Vf Vh inv /coh_zigzag C.
have V1 : valid (f gL) by move/validL: Vf. 
have V2 : valid (gL \+ g2).
- move/cohSO: C; rewrite /validSO zigE !zipPt allvalidPt.
  by case/validL/andP; rewrite joinC. 
exists (gL \+ g2), g1; rewrite {1}zigE zipPt joinC.
by rewrite fsE; split=>//; rewrite pf_loc 1?joinAC -1?joinA.
Qed.

Lemma acq_step_safe : step_safe acq_safe acq_step.
Proof. by move=>s s' [][gL][gE][-> _ Vf Vh inv]; exists gL, gE. Qed.

Lemma acq_step_guar : step_guar guar acq_step. 
Proof. by move=>_ _ [][gL][gE][->-> Vf Vh inv] C; exists gL, (f gL), gE. Qed.

Lemma acq_step_frame : step_frame acq_safe acq_step.
Proof.
move=>s s' p []; rewrite -[s]st_eta -[s']st_eta /=.
case=>gL [gE][/stE /=]; rewrite fsE /=. 
case=>->-> /zipPtf [[m1 g1]][[m2 g2]][[]].
case/sym/mtx00E=>->->->->-> Vf Vh inv C; case=>gL' [gE'][]. 
rewrite {1}zigE zipPt -fsE; case/scancelPt=><-<-{gL' gE'}.
move=>-> Vf2 Vh2 I2 /coh_zigzag C2.
have V1 : valid (f gL) by move/validL: Vf. 
have V2 : valid (gL \+ g2).
- move/cohSO: C2; rewrite /validSO /= !zipPt allvalidPt joinA joinAC.  
  by case/validL/andP. 
exists (l ->> [(up nown, f gL), lk :-> false \+ h, (up nown, g1)]).
split; first by rewrite !fsE zigE zipPt joinC pf_loc // joinC.  
by exists gL, (g1 \+ g2); rewrite !fsE; split=>//; rewrite zagE zipPt. 
Qed.

Lemma acq_step_total : step_total acq_safe acq_step.
Proof.
move=>s [gL][gE][->] V1 V2 I C.
exists (l ->> [(up nown, f gL), lk :-> false \+ h, (up nown, gE)]), tt.
by exists gL, gE.
Qed.

Definition LockAcq_pre := 
  Action acq_safe_coh acq_safe_zagzig acq_step_safe acq_step_guar
         acq_step_frame acq_step_total.

End Internal.

Lemma acq_st_erase : acq_step_erase LockAcq_pre.
Proof.
move=>s1 h1 t1 [] k1 s2 h2 t2 [] k2 V E St1 St2.
case: St1 St2 V E=>gL1 [gE1][->-> _ _ inv1 _][gL2][gE2][->-> _ _ inv2 _].
rewrite !lock_flatsE -!joinA.
by move=>V; case/(hcancelV V)=>_ _ ->. 
Qed.

Lemma acq_st_owns : acq_step_owns LockAcq_pre.
Proof.
move=>s s' [] k [gL][gE][_ -> _ _ _ _]; exists (lk :-> false).
by rewrite lock_flatsE joinC.
Qed.

End LockAcquireAct.

Module Exports.

Definition LockAcq lkT f pf := 
  AcqAction (@acq_st_erase lkT f pf) (@acq_st_owns lkT f pf).

End Exports.
End LockAcquireAct.

Export LockAcquireAct.Exports.

(* the real unlocking action *)

(* eventually i should add a special action for changing *)
(* the ghost state related to the lock, without actually unlocking *) 
(* then the unlock action should just check that the ghosts *)
(* satisfy the invariant, before unlocking; but without changing *)
(* the ghosts any further *)
(* but i don't want to get ahead of myself here, so let's for now *)
(* just try to recover the old SCSL behavior *)

Module UnlockAct.
Section UnlockAct.
Variables (hp : nat) (lkT : lockT). 
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation I := (@inv_of lkT).
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (CLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Definition unlock_safe s := 
  (exists (hL hE h : heap) (gL gE : U), 
     [/\ s = hp ->> [hL \+ h, tt, hE] \+
              l ->> [(up own, gL), lk :-> true, (up nown, gE)], 
         valid (f gL \+ gE) & I (f gL \+ gE) h]) /\ s \In coh.

Definition unlock_step s s' (r : unit) := 
  [/\ exists (hL hE h : heap) (gL gE : U), 
        s = hp ->> [hL \+ h, tt, hE] \+
             l ->> [(up own, gL), lk :-> true, (up nown, gE)] /\
        s' = hp->> [hL, tt, hE] \+
              l->> [(up nown, f gL), lk :-> false \+ h, (up nown, gE)] &
      s \In coh /\ s' \In coh].

Lemma uniq_xfer_pf : uniq_xfer (LockAcq pf) (PrivRel hp). 
Proof.
move=>s h h' [s1][s2][-> Vh][Sf1 Sf2][s1'][s2'][E _][Sf1' Sf2'].
have V: valid (s1 \+ s2) by apply: validHV Vh.
move: (coh_prec V E (asf_coh Sf1) (asf_coh Sf1'))=>E'.
rewrite -{s1'}E' in E Sf1' *; rewrite -{E s2'}(joinsK V E) in Sf2'.
case: Sf1 Sf1'=>gL [gE][-> _ _ inv _][_][_][/scancelPt [<-<-] _ _ inv' _]. 
move: (validHR Vh); rewrite /validH.
case/rel_asf_owns: Sf2'; case/rel_asf_owns: Sf2=>h2 -> h2' E' V'.
apply: pf_prec V' E' inv inv'.
Qed.

Definition unlock_ent := 
  @EntAcqRel unit unit unit (fun _ => tt) (CLock lkT) (Priv hp)
             (LockChanCore lkT) (PrivChanCore hp) 
             (or_introl (erefl _)) (or_introl (erefl _)) _ _  
             uniq_xfer_pf. 

(* change the pre into a simpler one *)
(* this is not simple weakening, but needs to be iff *)
(* in order to avoid proving that the new pre satisfies the *)
(* zigzaging properties *)
(* so, strictly speaking, it makes no sense to have a consequence *)
(* constructor for building actions *)
Lemma unlockP s : unlock_safe s <-> s \In asafe unlock_ent. 
Proof.
split; last first.
- move=>S; split; last by move/asf_coh: S; rewrite /= starC.
  case: S=>h [s1][s2][->{s} _][Sf1 Sf2]; rewrite joinC.
  case: Sf1 Sf2=>gL [gE][-> V _ inv _][hL][hE][-> _]. 
  by eexists _, _, _, _, _.
case; case=>hL [hE][h][gL][gE][-> V inv C]; rewrite joinC in C *.
set lock := l ->> [_, _, _] in C *; set priv := hp ->> [_, _, _] in C *.
case: C=>Vh [s1][s2][E C1 C2]; rewrite (joinC s1) in E.
have [E2 E1] : s2 = lock /\ s1 = priv. 
- case: C2 E=>_ _ [mL][mE][ggL][ggE][b][hM][->] _. 
  by case/(scancel (validHV Vh)); case=><-<-<-<-<-.
rewrite {E s1}E1 {s2}E2 in C1 C2.
eexists h, _, _; split=>//; split; eexists _, _; split=>//. 
move: (Vh); rewrite /validH (flatsUn (validHV Vh)). 
rewrite lock_flatsE priv_flatsE -joinA joinCA (joinA _ h). 
by move/validL; rewrite !hvalidPtUn. 
Qed.

(* the simplified postcondition *)
Lemma unlockQ s s' : unlock_step s s' tt <-> astep s unlock_ent s' tt.
Proof.
split.
- case; case=>hL [hE][h][gL][gE][->->{s s'}][].
  set s1 := (hp ->> [_, _, _]); set s2 := (l ->> [_, _, _]).
  set s1' := (hp ->> [_, _, _]); set s2' := (l ->> [_, _, _]).
  case/split_coh=>/= _ Vh Cp Ct; case/split_coh=>/= _ Vh' Cp' Ct'.
  exists s2, s1, s2', s1', tt, tt; rewrite (joinC s2) (joinC s2').
  split=>//; exists h; split; last by exists hL, hE.
  exists gL, gE; split=>//.
  - by move/lockVG: Ct'.
  - by move/cohH: Ct'; rewrite /validH lock_flatsE. 
  by move/lock_inv: Ct'.
case=>s1 [s2][s1'][s2'][[]][[]][->->{s s'} _] Vh [h][St1 St2]; split.
- case: St1 Vh=>gL [gE][->->{s1 s1'}] Vf Vlk inv C1 Vh. 
  case: St2 Vh=>hL [hE][->->{s2 s2'}] C2 Vh.
  by exists hL, hE, h, gL, gE; split; rewrite joinC.
split. 
- split=>//; exists s2, s1; rewrite joinC.
  by case/ast_coh: St1; case/ast_coh: St2.
split; last first.
- exists s2', s1'; rewrite joinC.
  by case/ast_coh: St1; case/ast_coh: St2.
move: Vh; rewrite !validHUn. 
case: St1=>gL [gE][->->] _.
case: St2; move=>hL [hE][->->] _.
rewrite !lock_flatsE !priv_flatsE -!joinA (joinCA h).
by rewrite !hvalidPtUn !pcm_validE /= !fsE /= !fvalidPtUn. 
Qed.

(* change the return type to unit, instead of unit * unit *)

Lemma unlock_safe_coh : safe_coh coh unlock_safe.
Proof. by move=>s /unlockP /asf_coh; rewrite /= starC. Qed.

Lemma unlock_safe_zagzig : safe_zagzig unlock_safe.
Proof. by move=>s p /unlockP /asf_zagzig /unlockP. Qed.

Lemma unlock_step_safe : step_safe unlock_safe unlock_step.
Proof. by move=>s1 s2 [] /unlockQ /ast_safe /unlockP. Qed.

Lemma unlock_step_guar : step_guar guar unlock_step.
Proof. by move=>s1 s2 [] /unlockQ /ast_guar; rewrite entangle_intC. Qed.

Lemma unlock_step_frame : step_frame unlock_safe unlock_step.
Proof. 
move=>s s' p [] /unlockP H1 /unlockQ. 
by case/(ast_frame H1)=>s'' [->] /unlockQ; exists s''. 
Qed.

Lemma unlock_step_total : step_total unlock_safe unlock_step.
Proof.
by move=>s /unlockP /ast_total; case=>s' [[]] /unlockQ; exists s', tt.
Qed.

Definition Unlock_pre := 
  Action unlock_safe_coh unlock_safe_zagzig unlock_step_safe unlock_step_guar 
         unlock_step_frame unlock_step_total.

Lemma unlock_step_erase : int_step_erase Unlock_pre.
Proof. 
move=>s h t [] s' h' t' [] V E /unlockQ St /unlockQ. 
by move/(int_ast_erase V E St).
Qed.

End UnlockAct.

Module Exports.

Definition unlock hp lkT f pf := IntAction (@unlock_step_erase hp lkT f pf).

End Exports. 
End UnlockAct.

Export UnlockAct.Exports.

(* another useful action is testing the lock status *)
(* we can only test shared stuff; as the other things are subject to zigging *)
(* and if we fix them, we can't prove the zigging lemmas *)

Module ReadLock.
Section ReadLock.
Variables (lkT : lockT) (b : bool). 
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation coh := (Mod.coh (CLock lkT)). 
Notation guar := (Mod.internal (CLock lkT)). 

Definition readlock_safe s := 
  exists (mL mE : lift mutexUnlifted) (gL gE : U) h,
     [/\ s = l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] & s \In coh].

Definition readlock_step s s' (v : unit) := s = s' /\ readlock_safe s.

Lemma readlock_safe_coh : safe_coh coh readlock_safe.
Proof. by move=>s [mL][mE][gL][gE][hM][]. Qed.

Lemma readlock_safe_zagzig : safe_zagzig readlock_safe.
Proof. 
move=>s p; rewrite -[s]st_eta.
case=> mL [mE][gL][gE][hM]; rewrite fsE /=; case=>E1 /coh_zigzag C.
case: E1 C=>->-> /zipPtf [[m1 g1]][[m2 g2]][_ ->->].
rewrite zigE zipPt => C; exists (m2 \+ mL), m1, (g2 \+ gL), g1, hM.
by split=>//; rewrite -fsE.
Qed.

Lemma readlock_step_safe : step_safe readlock_safe readlock_step.
Proof. by move=>s s' [][]. Qed.

Lemma readlock_step_guar : step_guar guar readlock_step.
Proof. by move=>s s' [][->][mL][mE][gL][gE][hM][->]; apply: refl_guar_refl. Qed.

Lemma readlock_step_frame : step_frame readlock_safe readlock_step.
Proof. by move=>s s' p [Sf][<-{s'} _]; exists s. Qed.

Lemma readlock_step_total : step_total readlock_safe readlock_step.
Proof. by move=>s C; exists s, tt. Qed.

Definition ReadLock_pre := 
  Action readlock_safe_coh readlock_safe_zagzig readlock_step_safe 
         readlock_step_guar readlock_step_frame readlock_step_total.

End ReadLock.

Lemma readlock_step_erase lkT b : int_step_erase (@ReadLock_pre lkT b).
Proof. by move=>s1 h1 t1 [] s2 h2 t2 [] _ E [<- _][<- _]. Qed.

Module Exports.

Definition ReadLock lkT b := IntAction (@readlock_step_erase lkT b).

Lemma ReadLock_prec lkT b1 b2 s : 
        s \In asafe (@ReadLock lkT b1) -> 
        s \In asafe (@ReadLock lkT b2) -> b1 = b2.
Proof.
case=>mL1 [mE1][gL1][gE1][hM1][-> /cohH V].
case=>mL2 [mE2][gL2][gE2][hM2][E _].
rewrite /validH lock_flatsE in V.
by case/scancelPt: E=>_ _ /(hcancelV V) [->].
Qed.

End Exports.
End ReadLock.

Export ReadLock.Exports.


(* this action erases to CAS on the lock of the LockRes module *)
(* but it also manipulates the ghosts in the appropriate way *)
(* to change the resource ownership *)
Module CASAct.
Section CASAct.
Variables (hp : nat) (lkT : lockT).
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation I := (@inv_of lkT).
Notation W := (entangle (Priv hp) (CLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

(* we can't insist that cas is safe if mL = up nown *)
(* because then we can't prove that cas_safe satisfies zig-zag lemmas *)
Definition cas_safe s := s \In coh. 

Definition cas_step s s' (r : bool) := 
  [/\ exists (hL hE : heap) (mL mE : lift mutexUnlifted) 
             (gL gE : U) (b : bool) h,
        s = hp ->> [hL, tt, hE] \+
             l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] /\
        if ~~ b then
           [/\ s' = hp ->> [hL \+ h, tt, hE] \+
                     l ->> [(up own, gL), lk :-> true, 
                            (up nown, gE)] & r = true]
        else s' = s /\ r = false &
        s \In coh /\ s' \In coh].

Lemma uniq_xfer_pf : uniq_xfer (PrivAcq hp) (LockRel lkT).
Proof.
move=>s h h' [s1][s2][-> Vh][Sf1 Sf2][s1'][s2'][E _][Sf1' Sf2'].
have V: valid (s1 \+ s2) by apply: validHV Vh.
move: (coh_prec V E (asf_coh Sf1) (asf_coh Sf1'))=>E'.
rewrite -{s1'}E' in E Sf1' *; rewrite -{E s2'}(joinsK V E) in Sf2'.
case: Sf2 Sf2' (validHR Vh)=>gL [gE][-> _][_][_][/scancelPt [_ E _] _]. 
by rewrite /validH lock_flatsE=>V'; case/(hcancelV V'): E.
Qed.

Definition cas_ent := 
  @EntAcqRel unit unit unit (fun _ => tt) (Priv hp) (CLock lkT) 
             (PrivChanCore hp) (LockChanCore lkT) 
             (or_introl (erefl _)) (or_introl (erefl _)) _ _  
             uniq_xfer_pf.

Definition check_lock b := 
  FrameAct (ReadLock lkT b) (leq_guar_entangle_inr (Priv hp) _). 

(* simplification lemmas for helper functions *)
Lemma check_lockP b s : 
        s \In asafe (check_lock b) <->
        [/\ exists (hL hE : heap) (mL mE : lift mutexUnlifted) (gL gE : U) h, 
              s = hp ->> [hL, tt, hE] \+
                   l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] &
              s \In coh].
Proof.
split; last first.
- case; case=>hL [hE][mL][mE][gL][gE][h] ->{s} C.
  split=>//; case/split_coh: C=>/= _ _ C1 C2.
  by rewrite joinC; eexists _, _; do !split=>//; eexists _, _, _, _, _.
move=>S; split; last by move/asf_coh: S.
case: S; case=>s1 [s2][->{s}] S1 C; rewrite starC in C.
move/(star_coh_prec C): (asf_coh S1)=>C2.
case: C2 C=>_ [hL][hE] ->; case: S1=>mL [mE][gL][gE][h][->].
by exists hL, hE, mL, mE, gL, gE, h; rewrite joinC.
Qed.

(* and for postcondition *)  
Lemma check_lockQ b s s' : 
        astep s (check_lock b) s' tt <->
        s' = s /\ s \In asafe (check_lock b).
Proof.
split.
- case; case=>s1 [s2][s1'][->->{s s'}][<-{s1'}] H C _.
  by do !split=>//; exists s1, s2. 
case=><- S; move: (asf_coh S)=>C; split=>//.
by case: S; case=>s1 [s2][->] H _; exists s1, s2, s1. 
Qed.

(* and for cas's pre *)
Lemma entP s : s \In asafe cas_ent <-> s \In asafe (check_lock false). 
Proof.
split. 
- move=>G; split; last by move/asf_coh: G.
  case: G=>h [s1][s2][->{s} Vh][St1 St2].
  exists s2, s1; rewrite joinC; split=>//. 
  by case: St2=>gL [gE][-> C]; eexists _, _, _, _, _.
case/check_lockP; case=>hL [hE][mL][mE][gL][gE][h] -> C.
case/split_coh: C=>/= _ Vh C1 C2; eexists h, _, _; split=>//.
split; last first. 
- by exists gL, gE; split=>//; case/lock_inv'/mtx00E: C2=>->->.
exists hL, hE; split=>//. 
move: Vh; rewrite validHUn priv_flatsE lock_flatsE.
case/andP=>_; rewrite -(joinC h) -!joinA (joinCA hE) 3!joinA.  
by move/validL. 
Qed.


Lemma check_lock_prec s1 s2 b1 b2 h1 h2 :  
        valid (flats s1 \+ h1) ->
        flats s1 \+ h1 = flats s2 \+ h2 ->
        s1 \In asafe (check_lock b1) -> 
        s2 \In asafe (check_lock b2) -> b1 = b2.
Proof.
move=>V E Sf1 Sf2; move: Sf1 Sf2 V E.
case; case=>t1 [t1'][-> Sf1] C1; case; case=>t2 [t2'][-> Sf2] C2.
rewrite (flatsUn (cohV C1)) (flatsUn (cohV C2)) {C1 C2}.
case: Sf1=>[mL1][mE1][gL1][gE1][hM1][-> _].
case: Sf2=>[mL2][mE2][gL2][gE2][hM2][-> _].
by rewrite !lock_flatsE -!joinA => V; case/(hcancelV V). 
Qed.

Lemma check_lock_eq s b1 b2 :  
        s \In asafe (check_lock b1) -> 
        s \In asafe (check_lock b2) -> b1 = b2.
Proof.
move=>S1 S2.
apply: (check_lock_prec (h1:=Unit) (h2:=Unit)) (S1) S2; rewrite unitR //.
by move/asf_coh/cohH: S1. 
Qed.

(* now simplifying lemmas for the main action *)
Lemma casP s : cas_safe s <-> exists b, s \In asafe (check_lock b).
Proof.
split; last first.
- by case=>b /check_lockP []; case=>hL [hE][mL][mE][gL][gE][h] ->.
move=>S; case: S (S)=>_ [s1][s2][-> C1 C2].
case: C2 (C2)=>_ _ [mL][mE][gL][gE][b][hM][-> _] C2 S.
exists b; split=>//.
by rewrite joinC; eexists _, _; split=>//; exists mL, mE, gL, gE, hM.
Qed.

Lemma casQ s s' r : 
        cas_step s s' r <-> 
        if r then astep s cas_ent s' tt 
        else astep s (check_lock true) s' tt.
Proof.
split.
- case; case=>hL [hE][mL][mE][gL][gE][b][h][->{s}]; case: b=>/=;
  case=>->->{s' r}[C C'].
  - by apply/check_lockQ; split=>//; apply/check_lockP; eauto 10. 
  eexists _, _, _, _, tt, tt; split=>//; first by move/cohH: C.
  exists h; split; first by exists hL, hE; split=>//; case/split_coh: C'.
  case/split_coh: C=>/=_ _ _ C; case/lock_inv'/mtx00E: C (C)=>->->.   
  by exists gL, gE.
case: r; last first.
- case/check_lockQ=>-> Sf; split; last by case/asf_coh: Sf.
  case/check_lockP: Sf; case=>hL [hE][mL][mE][gL][gE][h] -> C.
  by eexists _, _, _, _, _, _, _, _.
move=>St; split; last by case/ast_coh: St. 
case: St=>s1 [s2][s1'][s2'][[]][[]][->->{s s'}] _ Vh [h][St1 St2]. 
case: St1 St2 Vh=>hL [hE][->->{s1 s1'}] Cp.
by case=>gL [gE][->->{s2 s2'}] Ct Vh; eexists _, _, _, _, _, _, _, _.
Qed.

Lemma cas_safe_coh : safe_coh coh cas_safe.
Proof. by []. Qed.

Lemma cas_safe_zagzig : safe_zagzig cas_safe.
Proof. by move=>s p; move/coh_zigzag. Qed.

Lemma cas_step_safe : step_safe cas_safe cas_step.
Proof. 
move=>s s' b /casQ G; apply/casP.
case: b G; last by move/ast_safe; exists true. 
by move/ast_safe/entP; exists false. 
Qed.

Lemma cas_step_guar : step_guar guar cas_step.
Proof. by move=>s s' b /casQ; case: b; move/ast_guar. Qed.

Lemma cas_step_frame : step_frame cas_safe cas_step.
Proof.
move=>s s' p v /casP [b] H /casQ.
case: v=>H2; last first.
- move: (check_lock_eq (asf_zagzig H) (ast_safe H2))=>?; subst b. 
  case/(ast_frame H): H2=>s'' [->{s'}] St.
  by exists s''; split=>//; apply/casQ.
move/entP: (ast_safe H2). 
move/(check_lock_eq (asf_zagzig H))=>?; subst b.
move/entP: H=>H; case/(ast_frame H): H2=>s'' [->{s'}] St.
by exists s''; split=>//; apply/casQ.
Qed.

Lemma cas_step_total : step_total cas_safe cas_step.
Proof.
move=>s /casP []; case.
- by case/ast_total=>s' [[]]; eexists s', false; apply/casQ.
by case/entP/ast_total=>s' [[]]; exists s', true; apply/casQ.
Qed.

Definition CAS := 
  Action cas_safe_coh cas_safe_zagzig cas_step_safe cas_step_guar
         cas_step_frame cas_step_total.

Lemma cas_step_erase : int_step_erase CAS.
Proof.
move=>s h t b s' h' t' b' V E.
(move/casQ; case: b=>[H1| /check_lockQ [-> H1]]);
(move/casQ; case: b'=>[H2| /check_lockQ [-> H2]])=>//.
- by case: (int_ast_erase V E H1 H2). 
- by move/ast_safe/entP: H1=>H1; move/(check_lock_prec V E H1): H2. 
by move/ast_safe/entP/(check_lock_prec V E H1): H2.
Qed.

End CASAct.

Module Exports.

Definition CAS hp lkT := IntAction (@cas_step_erase hp lkT).

End Exports.
End CASAct.

Export CASAct.Exports.

(*************)
(* Stability *)
(*************)

(* the mutual exclusion lemmas for locks *)
Section LockStableBase.
Variable (hp : nat) (lkT : lockT).
Notation l := (@label_of lkT).       (* feap label *)
Notation lk := (@lock_of lkT).       (* lock *)
Notation U := (@pcm_of lkT).         (* pcm *) 
Notation I := (@inv_of lkT).         (* resource invariant *)
Notation V := (CLock lkT).
Notation W := (entangle (Priv hp) V).

Lemma mtx_lock_mrely (mL mE : lift mutexUnlifted) (gL gE : U) 
                     (b : bool) (h : heap) s : 
        mrely V (l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)]) s -> 
        exists (b' : bool) h' mE' gE', 
          s = l ->> [(mL, gL), lk :-> b' \+ h', (mE', gE')] /\
          (mL = up own -> [/\ b' = b, h' = h, mE' = up nown & gE' = gE]).
Proof.
rewrite -[s]st_eta; move=>M; case: (mrely_coh M)=>C _; case: M=>M. 
- case: M; move/transp_coh: C; rewrite /transp !fsE /= -{1}fsE => C. 
  case=><-<-<- _; exists b, h, mE, gE; rewrite -fsE; split=>//.
  by move=>H; move/lockVM: C; rewrite H; case: mE=>//; case.
case: M=>_ [h'][/Mem_map_inv][ct][->]; rewrite InE /= => ->{ct}.
case. 
- case=>gL' [gE'][gE''][]. 
  move/transp_coh: C; rewrite /transp !fsE /= -2!fsE => C [].  
  rewrite -[lk :-> true]unitR.
  case/(lock_prec C)=>[[??]][??]??; subst mE gL' mL gE'' b h. 
  by case=>->->-> _ _ _ _; exists false, h', (up nown), gE'; rewrite fsE.
case=>gL' [gE'][]. 
move/transp_coh: C; rewrite /transp !fsE /= -2!fsE => C.  
case/(lock_prec C)=>[[??]][??]??; subst mE gL' mL gE' b h. 
case=>->->-> _; exists true, Unit, (up own), gE.
by rewrite -fsE unitR.
Qed.

(* mutual exclusion *)
Lemma mtx_lock (mL mE : lift mutexUnlifted) (gL gE : U) 
               (b : bool) (h : heap) s : 
        menv_steps V (l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)]) s ->
        exists (b' : bool) h' mE' gE', 
          s = l ->> [(mL, gL), lk :-> b' \+ h', (mE', gE')] /\ 
          (mL = up own -> [/\ b' = b, h' = h, mE' = up nown & gE' = gE]).
Proof.
case=>n; elim: n b h mE gE=>[|n IH] /= b h mE gE.
- case=><- C; exists b, h, mE, gE; split=>// H.
  by move/lockVM: C; rewrite H; case: mE=>//; case.
case=>s'' [] /mtx_lock_mrely [b'][h'][mE'][gE'][->{s''}] T.
case/IH=>b'' [h''][mE''][gE''][->{s IH}] T'.
exists b'', h'', mE'', gE''; split=>// H.
by case: (T H) (T' H)=>->->_ ->.
Qed.

(* for entangled actions *)
Lemma mtx_priv_lock (hL hE : heap) (mL mE : lift mutexUnlifted) 
                    (gL gE : U) (b : bool) (h : heap) s : 
        menv_steps W 
          (hp ->> [hL, tt, hE] \+
            l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)]) s ->
        exists hE' (b' : bool) h' mE' gE', 
          s = hp ->> [hL, tt, hE'] \+
               l ->> [(mL, gL), lk :-> b' \+ h', (mE', gE')] /\
          (mL = up own -> [/\ b' = b, h' = h, mE' = up nown & gE' = gE]).
Proof.
move=>M; case: (menvs_coh M); case/split_coh=>/= _ _ Cp Ct C.
case: C M=>_ [s1][s2][->{s}] C1 _. 
case/(menvs_split (entangle_extL _ _) Cp C1)=>Cp' /= Cl' {C1}.
case/stb_ex1: Cp'; case; case=>hE' ->{s1}. 
case/mtx_lock: Cl'=>b' [h'][mE'][gE'][->{s2}] T.
by eexists _, _, _, _, _. 
Qed.

(* a simplification for the common special case *)
Lemma mtx_priv_lock' (hL hE : heap) (gL gE : U) s : 
        menv_steps W 
          (hp ->> [hL, tt, hE] \+
            l ->> [(up own, gL), lk :-> true, (up nown, gE)]) s -> 
        exists hE', 
          s = hp ->> [hL, tt, hE'] \+
               l ->> [(up own, gL), lk :-> true, (up nown, gE)].
Proof.
rewrite -[lk :-> _]unitR.
case/mtx_priv_lock=>hE' [b'][h'][mE'][gE'][->{s}] /(_ (erefl _)) [->->->->].
by exists hE'.
Qed.

End LockStableBase.


Section UnlockStable.
Variables (hp : nat) (lkT : lockT).
Notation V := (entangle (Priv hp) (CLock lkT)). 
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation I := (@inv_of lkT).
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).

Lemma unlock_pre_stable s s' : 
        s \In asafe (unlock hp pf) ->
        menv_steps V s s' -> 
        s' \In asafe (unlock hp pf). 
Proof.
case; case=>hL [hE][h][gL][gE][->{s}] Vl inv C M.
split; last by case/menvs_coh: M.
by case/mtx_priv_lock': M=>hE' ->{s'}; eexists _, _, _, _, _. 
Qed.

Let Q s (r : unit) s' := 
  forall (hL hE : heap) h (gL gE : U), 
    s = hp ->> [hL \+ h, tt, hE] \+ 
         l ->> [(up own, gL), lk :-> true, (up nown, gE)] -> 
    I (f gL \+ gE) h ->
    exists hE' (b' : bool) h' mE' gE',
      s' = hp ->> [hL, tt, hE'] \+
            l ->> [(up nown, f gL), lk :-> b' \+ h', (mE', gE')]. 

Lemma unlock_post_conseq s s' v : 
        astep s (unlock hp pf) s' v -> Q s v s'.
Proof.
case; case=>hL [hE][h][gL][gE][->->{s s'}][C C'].
move=>hL' hE' h' gL' gE'; rewrite -[lk :-> _]unitR in C *. 
case/(priv_lock_prec C)=>E1 _ _ [<-{gL'}][<-{gE'}] _ inv. 
exists hE, false, h, (up nown), gE.
have V1 : valid (hL \+ h).
- case/split_coh: C=>/= _ _ /cohH; rewrite /validH priv_flatsE.
  by move/validL=>->.
have inv' : I (f gL \+ gE) h by case/split_coh: C'=>/= _ _ _ /lock_inv.
rewrite -(joinC h) -(joinC h') in E1 V1.
move: (pf_prec V1 E1 inv' inv)=>?; subst h'.
by rewrite -{E1}(joinfK V1 E1).
Qed.

Lemma unlock_post_stable s s' s'' r : 
        Q s r s' -> menv_steps V s' s'' -> Q s r s''.
Proof.
move=>H M hL hE h gL gE E inv; rewrite {s}E in H.
move: (H hL hE h gL gE (erefl _) inv). 
case=>hE' [b'][h'][mE'][gE'] E'; rewrite {s'}E' in H M.
case/mtx_priv_lock: M=>hE'' [b''][h''][mE''][gE''][->{s''}] T.
by eexists _, _, _, _, _.
Qed.

Lemma val_unlock (hL hE h : heap) (gL gE : U) (r : cont unit):
      valid (f gL \+ gE) ->
      I (f gL \+ gE) h -> 
      (forall hE' (b' : bool) h' mE' gE', 
          hp ->> [hL, tt, hE'] \+ 
           l ->> [(up nown, f gL), lk :-> b' \+ h', 
                  (mE', gE')] \In Mod.coh V -> 
          r tt (hp ->> [hL, tt, hE'] \+ 
                 l ->> [(up nown, f gL), lk :-> b' \+ h', 
                        (mE', gE')])) ->
       verify (hp ->> [hL \+ h, tt, hE] \+
                l ->> [(up own, gL), lk :-> true, (up nown, gE)]) 
              (act (unlock hp pf)) r.
Proof.
move=>V inv H; apply: val_act=>// s' M; split. 
- apply: unlock_pre_stable (M); split; last by case: (menvs_coh M). 
  by exists hL, hE, h, gL, gE. 
case=>s'' m /unlock_post_conseq /unlock_post_stable H' M'.
move/(_ _ M'): {M'} H' (proj2 (menvs_coh M')).
case/mtx_priv_lock': M=>hE' ->{s'}; case/(_ _ _ _ _ _  (erefl _) inv). 
by move=>hE'' [b''][h''][mE''][gE''] ->{m}; apply: H.
Qed.

End UnlockStable.


Section CASStable.
Variables (hp : nat) (lkT : lockT).
Notation V := (entangle (Priv hp) (CLock lkT)). 
Notation l := (@label_of lkT).
Notation lk := (@lock_of lkT).
Notation U := (@pcm_of lkT).
Notation I := (@inv_of lkT).
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).

Lemma cas_pre_stable s s' : 
        menv_steps V s s' -> s' \In asafe (CAS hp lkT). 
Proof. by case/menvs_coh. Qed.

Let Q s r s' := 
  forall (hL hE : heap) (mL mE : lift mutexUnlifted) 
         (gL gE : U) (b : bool) h, 
    s = hp ->> [hL, tt, hE] \+ 
         l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)] -> 
    if r then 
      exists hE' h' gE',
        s' = hp ->> [hL \+ h', tt, hE'] \+
              l ->> [(up own, gL), lk :-> true, (up nown, gE')] /\
        I (gL \+ gE') h'
    else 
      exists hE' (b' : bool) h' mE' gE',
        s' = hp ->> [hL, tt, hE'] \+
              l ->> [(mL, gL), lk :-> b' \+ h', (mE', gE')].

Lemma cas_post_conseq s r s' : 
        astep s (CAS hp lkT) s' r -> Q s r s'.
Proof.
case; case=>hL [hE][mL][mE][gL][gE][b][h][->{s}] H [C _]. 
move=>_ _ _ _ _ _ _ _ /(priv_lock_prec C) [<-] _ _ [<-<-] _ _.
case: ifP H=>H /=; case=>->->.
- exists hE, h, gE; split=>//.
  by case/split_coh: C=>/= _ _ _ /lock_inv; case: b H.
by exists hE, b, h, mE, gE. 
Qed.

Lemma cas_post_stable s s' s'' r : 
        Q s r s' -> menv_steps V s' s'' -> Q s r s''.
Proof.
move=>H M hL hE mL mE gL gE b h E; rewrite {s}E in H.
move: (H hL hE mL mE gL gE b h (erefl _))=>{H} H. 
case: ifP H M=>_. 
- case=>hE' [h'][gE'][->{s'}] inv /mtx_priv_lock' [hE''] ->{s''}.
  by exists hE'', h', gE'. 
case=>hE' [b'][h'][mE'][gE'] ->{s'}.
case/mtx_priv_lock=>hE'' [b''][h''][mE''][gE''][->{s''}] T.
by eauto 15.
Qed.

(* the lemma insists that mL = up nown, though that's not necessary *)
Lemma val_cas (hL hE : heap) (mE : lift mutexUnlifted) (gL gE : U) 
              (b : bool) h (r : cont bool) : 
        (* if we succeed *)
        (forall hE' gE' h', 
           valid (gL \+ gE') ->
           I (gL \+ gE') h' ->
           hp ->> [hL \+ h', tt, hE'] \+ 
            l ->> [(up own, gL), lk :-> true, (up nown, gE')] \In Mod.coh V ->
           r true (hp ->> [hL \+ h', tt, hE'] \+ 
                    l ->> [(up own, gL), lk :-> true, (up nown, gE')])) ->
        (* if we fail *)
        (forall hE' mE' gE' (b' : bool) h', 
           hp ->> [hL, tt, hE'] \+ 
            l ->> [(up nown, gL), lk :-> b' \+ h', (mE', gE')] \In Mod.coh V ->
           r false (hp ->> [hL, tt, hE'] \+ 
                     l ->> [(up nown, gL), lk :-> b' \+ h', (mE', gE')])) ->
        verify (hp ->> [hL, tt, hE] \+ 
                 l ->> [(up nown, gL), lk :-> b \+ h, (mE, gE)])
               (act (CAS hp lkT)) r.
Proof.
move=>H1 H2; apply: val_act=>// s' M; split; first by apply: cas_pre_stable M. 
move=>y s'' m /cas_post_conseq /cas_post_stable H' M'.
move/(_ _ M'): {M'} H' (proj2 (menvs_coh M')).
case/mtx_priv_lock: M (M)=>hE' [b'][h'][mE'][gE'][->{s'}] _ M.
move/(_  _ _ _ _ _ _ _ _ (erefl _)); case: ifP=>E; last first.
- by case=> hE'' [b''][h''][mE''][gE''] ->{m}; apply: H2.
case=>hE'' [h''][gE''][->{m}] inv C.
by apply: H1=>//; case/split_coh: C=>/= _ _ _ /lockVG.
Qed.

Lemma val_cas1 (hL hE : heap) (mL mE : lift mutexUnlifted) (gL gE : U) 
              (b : bool) h (r : cont bool) : 
        (* if we succeed *)
        (forall hE' gE' h', 
           valid (gL \+ gE') ->
           I (gL \+ gE') h' ->
           hp ->> [hL \+ h', tt, hE'] \+ 
            l ->> [(up own, gL), lk :-> true, (up nown, gE')] \In Mod.coh V ->
           r true (hp ->> [hL \+ h', tt, hE'] \+ 
                    l ->> [(up own, gL), lk :-> true, (up nown, gE')])) ->
        (* if we fail *)
        (forall hE' mE' gE' (b' : bool) h', 
           hp ->> [hL, tt, hE'] \+ 
            l ->> [(mL, gL), lk :-> b' \+ h', (mE', gE')] \In Mod.coh V ->
           r false (hp ->> [hL, tt, hE'] \+ 
                     l ->> [(mL, gL), lk :-> b' \+ h', (mE', gE')])) ->
        verify (hp ->> [hL, tt, hE] \+ 
                 l ->> [(mL, gL), lk :-> b \+ h, (mE, gE)])
               (act (CAS hp lkT)) r.
Proof.
move=>H1 H2; apply: val_act=>// s' M; split; first by apply: cas_pre_stable M. 
move=>y s'' m /cas_post_conseq /cas_post_stable H' M'.
move/(_ _ M'): {M'} H' (proj2 (menvs_coh M')).
case/mtx_priv_lock: M (M)=>hE' [b'][h'][mE'][gE'][->{s'}] _ M.
move/(_  _ _ _ _ _ _ _ _ (erefl _)); case: ifP=>E {M}.
- case=>hE'' [h''][gE''][->{m}] inv C.
  by apply: H1=>//; case/split_coh: C=>/= _ _ _ /lockVG.
- by case=> hE'' [b''][h''][mE''][gE''] ->{m} C; apply: H2. 
Qed.

End CASStable.


Section CASLockProperties. 
Variables (hp : nat) (lkT : lockT).
Notation l := (@label_of lkT). (* feap label *)
Notation flag := (@lock_of lkT). (* lock *)
Notation U := (@pcm_of lkT).   (* pcm *) 
Notation I := (@inv_of lkT).   (* resource invariant *)
Notation W := (entangle (Priv hp) (CLock lkT)). 
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).

Definition c_is_lock (gl ge : U) s := 
  exists (ml me: lift mutexUnlifted) (b: bool) (h: heap),
    s = l ->> [(ml, gl), flag :-> b \+ h, (me, ge)] /\
    s \In Mod.coh (CLock lkT).

Lemma c_is_lock_coh gl ge s : c_is_lock gl ge s -> s \In Mod.coh (CLock lkT).
Proof. by case=>ml[me][b][h]; case. Qed.

Lemma c_islock_sym gl ge s : c_is_lock gl ge s <-> c_is_lock ge gl (transp s).
Proof. 
split.
* case=>ml[me][b][h][->{s}] C; exists me, ml, b, h.
  move/transp_coh: C=> C; split=>//; last by rewrite fsE /transp /= -fsE.
* case=>ml[me][b][h][T]C. exists me, ml, b, h.
  move/transp_coh: C=> C; split=>//{C}. 
  rewrite fsE /transp /= in T. rewrite -[s]st_eta. 
  by case: T=>->->->; rewrite -fsE.
Qed.

Definition c_locked (gl ge : U) s :=
  s = l ->> [(up own, gl), flag :-> true, (up nown, ge)] /\
  s \In Mod.coh (CLock lkT).

Lemma c_locked_is_lock s gl ge : c_locked gl ge s -> c_is_lock gl ge s.
Proof. 
by case=>H1 H2'; exists (up own), (up nown), true, Unit; rewrite unitR.
Qed.
 
(* Mutual exclusion lemma *) 
Lemma c_mtx s (gl ge : U) :
        c_locked gl ge s -> c_locked ge gl (transp s) -> False.
Proof.
case=>->{s} C [/= E C']; rewrite !fsE /transp /= -!fsE in C' E.
by move/scancelPt: E. 
Qed.

Lemma val_unlock' (hL hE h : heap) (gL gE : U) s (r : cont unit):
      valid (f gL \+ gE) ->
      I (f gL \+ gE) h -> 
      c_locked gL gE s ->
      (forall i hE' gE',
          c_is_lock (f gL) gE' i -> 
          r tt (hp ->> [hL, tt, hE'] \+ i)) ->
       verify (hp ->> [hL \+ h, tt, hE] \+ s) 
              (act (unlock hp pf)) r.
Proof.
move=>V inv [->{s} C] H; apply: val_unlock=>// he' b' h' me' ge' C'.
apply: (H _ he' ge'); exists (up nown), me', b', h'; split=>//; last by case/split_coh: C'.
Qed.

Lemma val_cas' (hl he : heap) (gl ge : U)  i (r : cont bool) : 
        c_is_lock gl ge i ->
        (* if we succeed *)
        (forall (ge': U) he' h' s,
           valid (gl \+ ge') ->
           I (gl \+ ge') h' ->
           c_locked gl ge' s ->
           hp->>[hl \+ h', tt, he'] \+ s \In Mod.coh W ->
           r true (hp->>[hl \+ h', tt, he'] \+ s)) ->
        (* if we fail *)
        (forall j he' ge',
           c_is_lock gl ge' j ->
           hp->>[hl, tt, he'] \+ j \In Mod.coh W ->
           r false (hp->>[hl, tt, he'] \+ j)) ->        
        verify (hp->>[hl, tt, he] \+ i) 
               (act (CAS hp lkT)) r.
Proof.
case=>ml[me][b][h][->{i}] C.
case: (C)=> _ _[ml'][me'][gl'][ge'][b'][h']. 
case; case/lock_prec=>//[[<-][<-]][[<-][<-]]<-<-{ml' me' gl' ge' b' h'} _ H1 H2.
apply: val_cas1. 
- by move=> he' ge' h' V inv C'; case/split_coh: (C')=>/= _ _ _ L; apply: H1 V inv _ C'.
- move=> he' me' ge' b' h0 C'.
by apply: (H2 _ he' ge')=>//; exists ml, me', b', h0; split=>//; last by case/split_coh: C'.
Qed.

Definition cas_lock_spec : spec unit := @lock_spec hp lkT c_is_lock c_locked.

Program Definition cas_lock_proc : STsep W (cas_lock_spec) :=
 Do (ffix (fun (loop : unit -> STsep W cas_lock_spec) (_ : unit) =>
       Do (b <-- act (CAS hp lkT);
           if b then ret tt else loop tt)) tt).
Next Obligation. 
rewrite !ghs; apply: ghC=>i[[[[hl he] gl0] ge0] m] /=[->{i} T]; apply: step. 
apply: (@val_cas' hl he gl0 ge0)=>//=[|j he' ge' Il C].
- move=> ge'' he' h' j V inv L C; apply: val_ret=>m' M.
  case: (L)=>D _; rewrite D in M; case/mtx_priv_lock': M (M)=> hE'->{m'} _. 
  by exists h', ge'', hE', j; split=>//; rewrite D.
apply: val_do=>//=; first by exists hl, he', gl0, ge', j.
by case=>m' H1 M; apply: (H1 hl he' gl0 ge' j).
Qed.
Next Obligation. 
rewrite !ghs; apply: ghC=>i[[[[hl he] gl0] ge0] m] /=[->{i} T]. 
apply: val_do=>//; first by exists hl, he, gl0, ge0, m.
by case=>m' H1 M; apply: (H1 hl he gl0 ge0 m).
Qed.

(* Unlocking procedure specification *)
Definition cas_unlock_spec : spec unit := @unlock_spec hp lkT f c_is_lock c_locked.

(* Unlocking procedure itself *)
Program Definition cas_unlock : STsep W (cas_unlock_spec) :=
  Do (act (unlock hp pf)).
Next Obligation.
rewrite !ghs /=; apply: ghC=>i[[[[[gl ge] hl] he] h] s] /=.
case=> V inv L ->{i}; apply: (val_unlock' V)=>// i he' ge' T. 
by exists he', ge', i.
Qed.

End CASLockProperties. 

Section CASLockSpec.
Variables (hp : nat) (lkT : lockT). 
Notation W := (entangle (Priv hp) (CLock lkT)). 

Definition cas_lock := 
  Lockish (@c_is_lock_coh lkT) (@c_islock_sym lkT) (@c_locked_is_lock lkT)
          (@c_mtx lkT) (@cas_lock_proc hp lkT) (@cas_unlock hp lkT).

End CASLockSpec.

