Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq.
Require Import Eqdep rels ordtype prelude idynamic domains pcm unionmap heap.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Module Code.

Structure mixin_of (T : Type) := Mixin {
  code_dom_op : eqType;
  encode_op : code_dom_op -> T; 
  to_type_op : T -> Type}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure type : Type := Pack {sort : Type; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack T c T.
Definition clone := fun c & cT -> T & phant_id (pack c) cT => pack c. 

Definition code_dom := code_dom_op class.
Definition encode := @encode_op _ class. 
Definition to_type := to_type_op class.

End ClassDef.

Module Exports.
Coercion sort : type >-> Sortclass.
Notation code := type.
Notation codeMixin c i := (@Mixin _ c i).
Notation Code T m := (@pack T m).

Notation code_dom := code_dom.
Notation encode := encode.
Notation to_type := to_type.

Implicit Arguments to_type [cT].
Implicit Arguments encode [cT].

Prenex Implicits to_type encode.

Notation "[ 'codeMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'codeMixin'  'of'  T ]") : form_scope.
Notation "[ 'code' 'of' T 'for' C ]" := (@clone T C _ idfun id)
  (at level 0, format "[ 'code'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'code' 'of' T ]" := (@clone T _ _ id id)
  (at level 0, format "[ 'code'  'of'  T ]") : form_scope.

End Exports.

End Code.

Export Code.Exports.

(* codes for types *)

Structure tp_code := 
  tp_unit | tp_nat | tp_bool | tp_ptr | tp_mtx | tp_heap | (* nullary *)
  tp_seq of tp_code | (* unary *)
  tp_prod of tp_code & tp_code | tp_arr of tp_code & tp_code. (* binary *)

Fixpoint tp_sort (x : tp_code) : Type :=
  match x with 
  | tp_unit => unit
  | tp_nat => nat
  | tp_bool => bool
  | tp_ptr => ptr
  | tp_mtx => mutex 
  | tp_heap => heap
  | tp_seq y => seq (tp_sort y)
  | tp_prod y z => (tp_sort y * tp_sort z)%type
  | tp_arr y z => tp_sort y -> tp_sort z
  end.

(* tp_code is an equality type *)

Fixpoint tp_eq x y := 
  match x, y with 
  | tp_unit, tp_unit => true
  | tp_nat, tp_nat => true
  | tp_bool, tp_bool => true
  | tp_ptr, tp_ptr => true
  | tp_mtx, tp_mtx => true
  | tp_heap, tp_heap => true
  | tp_seq x1, tp_seq y1 => tp_eq x1 y1
  | tp_prod x1 y1, tp_prod x2 y2 => 
      tp_eq x1 x2 && tp_eq y1 y2
  | tp_arr x1 y1, tp_arr x2 y2 => 
      tp_eq x1 x2 && tp_eq y1 y2
  | _, _ => false
  end.

Lemma tp_eqP : Equality.axiom tp_eq.
Proof.
elim; 
(* nullaries *)
try by [case; constructor];
(* unaries *)
try by 
  [move=>x IH; case; try by [constructor];
   move=>y; apply: iffP (IH y) _ _; [move=>-> | case]];
(* binaries *)
by [move=>x1 IHx y1 IHy; case; try by [constructor];
    move=>x2 y2 /=; apply: iffP andP _ _;
    case; move/IHx=>->; move/IHy=>->]. 
Qed.
 
Definition tp_code_eqMixin := EqMixin tp_eqP.
Canonical tp_code_eqType := EqType tp_code tp_code_eqMixin.
Definition type_codeMixin := codeMixin tp_code_eqType tp_sort id.
Canonical typeCode := Code Type type_codeMixin.

(* now the same for pcm's *)

Structure pcm_code := 
  pcm_unit | pcm_nat | pcm_mtx | pcm_heap | pcm_natmap of tp_code |
  pcm_prod of pcm_code & pcm_code.

Fixpoint pcm_sort (x : pcm_code) : pcm := 
  match x with 
  | pcm_unit => [pcm of unit]
  | pcm_nat => [pcm of nat]
  | pcm_mtx => [pcm of lift [unlifted of mutex]]
  | pcm_heap => [pcm of heap]
  | pcm_natmap t2 => [pcm of union_map nat_ordType (tp_sort t2)]
  | pcm_prod y z => [pcm of pcm_sort y * pcm_sort z]
  end.

Fixpoint pcm_eq x y := 
  match x, y with 
  | pcm_unit, pcm_unit => true 
  | pcm_nat, pcm_nat => true 
  | pcm_mtx, pcm_mtx => true 
  | pcm_heap, pcm_heap => true
  | pcm_natmap t1, pcm_natmap t2 => t1 == t2
  | pcm_prod x1 y1, pcm_prod x2 y2 =>
      pcm_eq x1 x2 && pcm_eq y1 y2
  | _, _ => false
  end.

Lemma pcm_eqP : Equality.axiom pcm_eq.
Proof.
elim;
try by [case; constructor].
- move=>t; case=>/=; try constructor=>//; move=>y.
  by case: eqP=>[->|H]; constructor=>//; case. 
by [move=>x1 IHx y1 IHy; case; try by [constructor];
    move=>x2 y2 /=; apply: iffP andP _ _;
    case; move/IHx=>->; move/IHy=>->].
Qed.

Definition pcm_code_eqMixin := EqMixin pcm_eqP.
Canonical pcm_code_eqType := EqType pcm_code pcm_code_eqMixin.
Definition pcm_codeMixin := codeMixin pcm_code_eqType pcm_sort PCM.sort.
Canonical pcmCode := Code pcm pcm_codeMixin.

(*********************************************)
(* Computing code instance for a PCM or type *) 
(*********************************************)

(* this is a canonical structure setup that inverts the interp function. *)
(* important so that we avoid excesive pcm_code or tp_code annotations. *)

Module Encoded. 
Section Encoded.
Variable C : code.

Structure mixin_of (v : C) := Mixin {
  decode_op : code_dom C; 
  _ : encode decode_op = v}.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure type : Type := Pack {sort : C; _ : class_of sort; _ : C}.
Local Coercion sort : type >-> Code.sort.

Variables (v : C) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack v c v.

Definition clone := 
  fun c & to_type cT -> to_type v & phant_id (pack c) cT => pack c. 

Definition decode := decode_op class.

End ClassDef.
End Encoded.

Module Exports.
Coercion sort : type >-> Code.sort.
Notation encoded := type.
Notation encodeMixin := Mixin.
Notation Encode C v m := (@pack C v m).

Notation decode := decode.

Implicit Arguments encodeMixin [].

Section Lemmas.
Variable C : code.
Implicit Type U : encoded C.

(* the next lemma is problematic to use in practice *)
(* as the rewriting gets in trouble with dependencies *)
(* the solution is to break up U, just like in proof below *)

Lemma encdecE U : encode (decode U) = U. 
Proof. by case: U=>tp []. Qed.

Lemma decode_inj U1 U2 : decode U1 = decode U2 -> U1 = U2 :> C.
Proof. 
case: U1=>s1 [dec1 pf1 t1]; case: U2=>s2 [dec2 pf2 t2] /=.
by rewrite /decode /= -pf1 -pf2=>->.
Qed.

Definition code_dyn U (v : to_type U) := 
  @idyn (code_dom C) (to_type \o encode) (decode U) 
                     (icoerce to_type v (esym (encdecE U))). 

(* every idynamic can be recast into a code_dyn form *)

Section CodeEta.
Variable u : idynamic (to_type \o @encode C).

Definition code_enc : encoded C :=
  let: v := encode (idyn_tp u) in 
  let: m := @encodeMixin C v (idyn_tp u) (erefl _) in
    Encode C v m.

Definition code_val : to_type code_enc := idyn_val u.

Lemma code_eta : u = code_dyn code_val.
Proof. by rewrite -{1}(idyn_eta u); apply: idynE=>// pf; rewrite !ieqc. Qed.

End CodeEta.

End Lemmas.

Notation "[ 'encodedMixin' C 'for' v ]" := (@class C _ : @mixin_of C v)
  (at level 0, format "[ 'encodedMixin'  C  'for'  v ]") : form_scope.
Notation "[ 'encoded' C 'of' v 'for' R ]" := (@clone C v R _ idfun id)
  (at level 0, format "[ 'encoded'  C  'of'  v  'for'  R ]") : form_scope.
Notation "[ 'encoded' C 'of' v ]" := (@clone C v _ _ idfun id)
  (at level 0, format "[ 'encoded'  C  'of'  v ]") : form_scope.

End Exports.

End Encoded.

Export Encoded.Exports.


Definition encoded_type := encoded typeCode.
Definition encodeMixinTP := encodeMixin typeCode.
Definition EncodeTP := @Encoded.pack typeCode.
Implicit Arguments encodeMixinTP [].
Implicit Arguments EncodeTP [].

Definition encoded_pcm := encoded pcmCode.
Definition encodeMixinPCM := encodeMixin pcmCode.
Definition EncodePCM := @Encoded.pack pcmCode.
Implicit Arguments encodeMixinPCM [].
Implicit Arguments EncodePCM [].


(* add the explicit coercions in the two special cases *)

Definition encoded_sort_type : encoded_type -> Type :=
  fun U => Encoded.sort U.
Coercion encoded_sort_type : encoded_type >-> Sortclass.

Definition encoded_sort_pcm : encoded_pcm -> pcm := 
  fun U => Encoded.sort U.
Coercion encoded_sort_pcm : encoded_pcm >-> pcm.

(* special notation *)

Definition pcm_dyn (U : encoded_pcm) (v : U) := @code_dyn pcmCode U v.

Section PCMEta.
Variable u : idynamic (to_type \o pcm_sort). 

Definition pcm_enc : encoded_pcm := @code_enc pcmCode u.
Definition pcm_val : pcm_enc := @code_val pcmCode u.
Definition pcm_eta : u = pcm_dyn pcm_val := @code_eta pcmCode u.

End PCMEta.


(*****************************)
(* inversion table for types *)
(*****************************)

Definition unit_tp_encodedMixin := encodeMixinTP unit tp_unit (erefl _).
Canonical unit_tp_Encoded := EncodeTP unit unit_tp_encodedMixin.

Definition nat_tp_encodedMixin := encodeMixinTP nat tp_nat (erefl _).
Canonical nat_tp_Encoded := EncodeTP nat nat_tp_encodedMixin.

Definition bool_tp_encodedMixin := encodeMixinTP bool tp_bool (erefl _).
Canonical bool_tp_Encoded := EncodeTP bool bool_tp_encodedMixin.

Definition ptr_tp_encodedMixin := encodeMixinTP ptr tp_ptr (erefl _).
Canonical ptr_tp_Encoded := EncodeTP ptr ptr_tp_encodedMixin.

Definition mtx_tp_encodedMixin := encodeMixinTP mutex tp_mtx (erefl _).
Canonical mtx_tp_Encoded := EncodeTP mutex mtx_tp_encodedMixin.

Definition heap_tp_encodedMixin := encodeMixinTP heap tp_heap (erefl _).
Canonical heap_tp_Encoded := EncodeTP heap heap_tp_encodedMixin.

Lemma seq_tp_pf (U : encoded_type) : 
        @encode typeCode (tp_seq (decode U)) = seq U.
Proof. by rewrite /decode; case: U=>tp [/= dp <-]. Qed.

Definition seq_tp_encodedMixin (U : encoded_type) := 
  encodeMixin typeCode (seq U) (tp_seq (decode U)) (seq_tp_pf U).
Canonical seq_tp_Encoded (U : encoded_type) := 
  EncodeTP (seq U) (seq_tp_encodedMixin U).

Lemma prod_tp_pf (U1 U2 : encoded_type) : 
        @encode typeCode (tp_prod (decode U1) (decode U2)) = prod U1 U2. 
Proof. by rewrite /decode; case: U1 U2=>t1 [/= ? <-] _ [t2 [/= ? <-]]. Qed.

Definition prod_tp_encodedMixin (U1 U2 : encoded_type) := 
  encodeMixinTP (prod U1 U2)
                (tp_prod (decode U1) (decode U2)) (prod_tp_pf U1 U2).
Canonical prod_tp_Encoded (U1 U2 : encoded_type) := 
  EncodeTP (prod U1 U2) (prod_tp_encodedMixin U1 U2).

(* this one is kinda pointless since we can't trigger inference of lambda's *)
(* so functions will have to be decorated with their types by hand *)

Lemma arr_tp_pf (U1 U2 : encoded_type) : 
        @encode typeCode (tp_arr (decode U1) (decode U2)) = (U1 -> U2). 
Proof. by rewrite /decode; case: U1 U2=>t1 [/= ? <-] _ [t2 [/= ? <-]]. Qed.

Definition arr_tp_encodedMixin (U1 U2 : encoded_type) := 
  encodeMixinTP (U1 -> U2) (tp_arr (decode U1) (decode U2)) (arr_tp_pf U1 U2).
Canonical arr_tp_Encoded (U1 U2 : encoded_type) := 
  EncodeTP (U1 -> U2) (arr_tp_encodedMixin U1 U2).

Notation "[ 'encoded_type' 'of' v ]" := 
  ([encoded typeCode of v] : encoded_type)
  (at level 0, format "[ 'encoded_type'  'of'  v ]") : form_scope.

(****************************)
(* inversion table for pcms *)
(****************************)

Definition unit_pcm_encodedMixin := encodeMixinPCM unitPCM pcm_unit (erefl _).
Canonical unit_pcm_Encoded := EncodePCM unitPCM unit_pcm_encodedMixin.

Definition nat_pcm_encodedMixin := encodeMixinPCM natPCM pcm_nat (erefl _).
Canonical nat_pcm_Encoded := EncodePCM natPCM nat_pcm_encodedMixin.

Definition mtx_pcm_encodedMixin := 
  encodeMixinPCM (liftPCM mutexUnlifted) pcm_mtx (erefl _).
Canonical mtx_pcm_Encoded := 
  EncodePCM (liftPCM mutexUnlifted) mtx_pcm_encodedMixin.

Definition heap_pcm_encodedMixin := encodeMixinPCM heapPCM pcm_heap (erefl _).
Canonical heap_pcm_Encoded := EncodePCM heapPCM heap_pcm_encodedMixin.

Lemma natmap_pcm_pf (V : encoded_type) : 
        @encode pcmCode (pcm_natmap (decode V)) = union_mapPCM nat_ordType V.
Proof. by rewrite /decode; case: V=>_ [/= ? <-]. Qed.

Definition natmap_pcm_encodedMixin (V : encoded_type) := 
  encodeMixinPCM (union_mapPCM nat_ordType V) (pcm_natmap (decode V)) 
                 (natmap_pcm_pf V).
Canonical natmap_pcm_Encoded (V : encoded_type) := 
  EncodePCM (union_mapPCM nat_ordType V) (natmap_pcm_encodedMixin V).

Lemma prod_pcm_pf (U1 U2 : encoded_pcm) : 
        @encode pcmCode (pcm_prod (decode U1) (decode U2)) = prodPCM U1 U2. 
Proof. by rewrite /decode; case: U1 U2=>t1 [/= ? <-] _ [t2 [/= ? <-]]. Qed.

Definition prod_pcm_encodedMixin (U1 U2 : encoded_pcm) := 
  encodeMixinPCM (prodPCM U1 U2) 
                 (pcm_prod (decode U1) (decode U2)) (prod_pcm_pf U1 U2).
Canonical prod_pcm_Encoded (U1 U2 : encoded_pcm) := 
  EncodePCM (prodPCM U1 U2) (prod_pcm_encodedMixin U1 U2).

Notation "[ 'encoded_pcm' 'of' v ]" := 
  ([encoded pcmCode of [pcm of v]] : encoded_pcm)
  (at level 0, format "[ 'encoded_pcm'  'of'  v ]") : form_scope.

(*
Check [encoded_type of nat].
Check [ecnoded_pcm of nat].
Check [encoded pcmCode of natPCM].
Check (@code_dyn pcmCode _ 3).
Check (@code_dyn typeCode _ [:: 3]).
Check (@code_dyn pcmCode _ (3, 4)).
*)


