Load loadpath.
Require Import ssreflect ssrfun ssrnat div ssrbool seq path eqtype.
Require Import idynamic Eqdep rels ordtype pcm finmap unionmap. 
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive. 

(*************)
(* Locations *)
(*************)

Inductive ptr := ptr_nat of nat.

Definition null := ptr_nat 0.

Definition nat_ptr (x : ptr) := let: ptr_nat y := x in y.

Definition eq_ptr (x y : ptr) : bool := 
  match x, y with ptr_nat m, ptr_nat n => m == n end.

Lemma eq_ptrP : Equality.axiom eq_ptr.
Proof. by case=>x [y] /=; case: eqP=>[->|*]; constructor=>//; case. Qed.

Definition ptr_eqMixin := EqMixin eq_ptrP. 
Canonical ptr_eqType := EqType ptr ptr_eqMixin.

(* some pointer arithmetic: offsetting from a base *)

Definition ptr_offset x i := ptr_nat (nat_ptr x + i).

Notation "x .+ i" := (ptr_offset x i) 
  (at level 3, format "x .+ i"). 

Lemma ptrE x y : (x == y) = (nat_ptr x == nat_ptr y).
Proof. by move: x y=>[x][y]. Qed.

Lemma ptr0 x : x.+0 = x.
Proof. by case: x=>x; rewrite /ptr_offset addn0. Qed.

Lemma ptrA x i j : x.+i.+j = x.+(i+j).
Proof. by case: x=>x; rewrite /ptr_offset addnA. Qed.

Lemma ptrK x i j : (x.+i == x.+j) = (i == j).
Proof. by case: x=>x; rewrite ptrE eqn_add2l. Qed.

Lemma ptr_null x m : (x.+m == null) = (x == null) && (m == 0).
Proof. by case: x=>x; rewrite !ptrE addn_eq0. Qed.

Lemma ptrT x y : {m : nat | (x == y.+m) || (y == x.+m)}.
Proof. 
case: x y=>x [y]; exists (if x <= y then (y - x) else (x - y)).
rewrite !ptrE leq_eqVlt /=.
by case: (ltngtP x y)=>/= E; rewrite subnKC ?(ltnW E) ?eq_refl ?orbT // E.
Qed.

Definition ltn_ptr (x y : ptr) := 
  match x, y with ptr_nat m, ptr_nat n => m < n end.

Lemma ltn_ptr_irr : irreflexive ltn_ptr.
Proof. by case=>x /=; rewrite ltnn. Qed.

Lemma ltn_ptr_trans : transitive ltn_ptr.
Proof. by case=>x [y][z]; apply: ltn_trans. Qed.

Lemma ltn_ptr_total : forall x y : ptr, [|| ltn_ptr x y, x == y | ltn_ptr y x].
Proof. by case=>x [y]; rewrite ptrE /=; case: ltngtP. Qed.

Definition ptr_ordMixin := OrdMixin ltn_ptr_irr ltn_ptr_trans ltn_ptr_total.
Canonical ptr_ordType := OrdType ptr ptr_ordMixin.

(*********)
(* Heaps *)
(*********)

Module Heap.

Local Notation idyn v := (@idyn _ id _ v).

Inductive heap := 
  Undef | Def (finmap : {finMap ptr -> idynamic id}) of 
               null \notin supp finmap.

Section NullLemmas.
Variables (f g : {finMap ptr -> idynamic id}) (x : ptr) (v : idynamic id).

Lemma upd_nullP : 
        x != null -> null \notin supp f -> null \notin supp (ins x v f).
Proof. by move=>H1 H2; rewrite supp_ins negb_or /= inE /= eq_sym H1. Qed.

Lemma free_nullP : null \notin supp f -> null \notin supp (rem x f). 
Proof. by move=>H; rewrite supp_rem negb_and /= H orbT. Qed.

Lemma un_nullP : 
        null \notin supp f -> null \notin supp g -> 
          null \notin supp (fcat f g).
Proof. by move=>H1 H2; rewrite supp_fcat negb_or H1 H2. Qed.

Lemma filt_nullP (q : pred ptr) : 
        null \notin supp f -> null \notin supp (kfilter q f).
Proof. by move=>H; rewrite supp_kfilt mem_filter negb_and H orbT. Qed.

Lemma heap_base : null \notin supp f -> all (fun k => k != null) (supp f).
Proof. by move=>H; apply/allP=>k; case: eqP H=>// -> /negbTE ->. Qed.

Lemma base_heap : all (fun k => k != null) (supp f) -> null \notin supp f.
Proof. by move/allP=>H; apply: (introN idP); move/H. Qed.

Lemma heapE (h1 h2 : heap) : 
        h1 = h2 <-> 
        match h1, h2 with
          Def f' pf, Def g' pg => f' = g'
        | Undef, Undef => true
        | _, _ => false
        end.
Proof.
split; first by move=>->; case: h2.
case: h1; case: h2=>// f1 pf1 f2 pf2 E. 
move: pf2; rewrite {f2}E in pf1 *; move=>pf2.
by congr Def; apply: bool_irrelevance.
Qed.

End NullLemmas.


(* methods *)

Definition def h := if h is Def _ _ then true else false.

Definition empty := @Def (finmap.nil _ _) is_true_true.

Definition upd k v h := 
  if h is Def hs ns then 
    if decP (@idP (k != null)) is left pf then 
      Def (@upd_nullP _ _ v pf ns)
    else Undef
  else Undef.

Definition dom h : pred ptr := 
  if h is Def f _ then [fun x => x \in supp f] else pred0.

Definition dom_eq h1 h2 :=
  (def h1 == def h2) && 
  match h1, h2 with
    Def f1 _, Def f2 _ => supp f1 == supp f2
  | Undef, Undef => true
  | _, _ => false
  end.

Definition free x h := 
  if h is Def hs ns then Def (free_nullP x ns) else Undef.

Definition find (x : ptr) h := 
  if h is Def hs _ then fnd x hs else None.

Definition union h1 h2 := 
  if (h1, h2) is (Def hs1 ns1, Def hs2 ns2) then  
    if disj hs1 hs2 then 
       Def (@un_nullP _ _ ns1 ns2)
    else Undef
  else Undef.

Definition um_filter q f := 
  if f is Def fs pf then Def (@filt_nullP fs q pf) else Undef.

Definition pts (x : ptr) v := upd x v empty. 

Definition empb h :=
  if h is Def hs _ then supp hs == [::] else false. 

Local Notation base := 
  (@UM.base ptr_ordType (idynamic id) (fun k => k != null)).

Definition from (f : heap) : base :=
  if f is Def hs ns then UM.Def (heap_base ns) else UM.Undef _ _.

Definition to (b : base) : heap := 
  if b is UM.Def hs ns then Def (base_heap ns) else Undef.

Lemma ftE b : from (to b) = b. 
Proof. by case: b=>// f H; rewrite UM.umapE. Qed.

Lemma tfE f : to (from f) = f. 
Proof. by case: f=>// f H; rewrite heapE. Qed.

Lemma undefE : Undef = to (@UM.Undef _ _ _). 
Proof. by []. Qed.

Lemma defE f : def f = UM.valid (from f). 
Proof. by case: f. Qed.

Lemma emptyE : empty = to (@UM.empty _ _ _). 
Proof. by rewrite heapE. Qed.

Lemma updE k v f : upd k v f = to (UM.upd k v (from f)). 
Proof. by case: f=>[|f H] //=; case: decP=>// H1; rewrite heapE. Qed.

Lemma domE f : dom f = UM.dom (from f). 
Proof. by case: f. Qed.

Lemma dom_eqE f1 f2 : dom_eq f1 f2 = UM.dom_eq (from f1) (from f2). 
Proof. by case: f1 f2=>[|f1 H1][|f2 H2]. Qed.

Lemma freeE k f : free k f = to (UM.free k (from f)). 
Proof. by case: f=>[|f H] //; rewrite heapE. Qed.

Lemma findE k f : find k f = UM.find k (from f).
Proof. by case: f. Qed.

Lemma unionE f1 f2 : union f1 f2 = to (UM.union (from f1) (from f2)).
Proof. 
case: f1 f2=>[|f1 H1][|f2 H2] //; rewrite /union /UM.union /=. 
by case: ifP=>D //; rewrite heapE.
Qed.

Lemma umfiltE q f : um_filter q f = to (UM.um_filter q (from f)).
Proof. by case: f=>[|f H] //; rewrite heapE. Qed.

Lemma empbE f : empb f = UM.empb (from f). 
Proof. by case: f. Qed.

Lemma ptsE k (v : idynamic id) : pts k v = to (@UM.pts _ _ _ k v).
Proof. 
by rewrite /pts /UM.pts /UM.upd /=; case: decP=>// H; rewrite heapE. 
Qed.

Module Exports.

(* the inheritance from PCMs *)

Definition heapUMCMixin := 
  UMCMixin ftE tfE defE undefE emptyE updE domE dom_eqE freeE
           findE unionE umfiltE empbE ptsE.
Canonical heapUMC := Eval hnf in UMC heap heapUMCMixin.

(* Can I get heap's own PCM? *)

Definition heapPCMMixin := union_map_classPCMMixin heapUMC.
Canonical heapPCM := Eval hnf in PCM heap heapPCMMixin.

End Exports.

End Heap.

Export Heap.Exports.

Notation heap := Heap.heap.

Definition heap_pts A (x : ptr) (v : A) := 
  @UMC.pts heapUMC x (@idyn _ id _ v). 

(*
Notation "x :-> v" := (@heap_pts _ x v) (at level 30).
*)

(*
Notation "x :-> v" := (@UMC.pts heapUMC x (idyn v)) (at level 30).
*)

(* The above notation is not getting printed *)
(* I think something is overriding it; possibly *)
(* that it's head is the same as in other notations *)
(* namely, UMC.pts, and it's only types that differentiate *)
(* which notation should be printed *)

(* if I introduce an explicit definition to override any other *)
(* conflicting notation, it is printed *)
Notation "x :-> v" := (@heap_pts _ x v) (at level 30).

(*****************************)
(* Specific points-to lemmas *)
(*****************************)

Section HeapPointsToLemmas.
Implicit Types (x : ptr) (h : heap).
(* A: upd requires explicit annotation; change that *)
(* why is notation not being displayed? *)
Local Notation idyn v := (@idyn _ id _ v).

Lemma hptsU A x (v : A) : x :-> v = upd x (idyn v) (Unit : heap).
Proof. exact: gen_ptsU. Qed.

Lemma hdomPt A x (v : A) : dom (x :-> v) =i [pred y | x != null & x == y].
Proof. exact: gen_domPt. Qed.

Lemma hvalidPt A x (v : A) : valid (x :-> v) = (x != null).
Proof. exact: gen_validPt. Qed.

Lemma hdomeqPt A1 A2 x (v1 : A1) (v2 : A2) : 
        dom_eq (x :-> v1) (x :-> v2).
Proof. exact: gen_domeqPt. Qed.

Lemma hcancelPt A1 A2 x (v1 : A1) (v2 : A2) : 
        valid (x :-> v1) -> x :-> v1 = x :-> v2 -> idyn v1 = idyn v2.
Proof. exact: gen_cancelPt. Qed.

Lemma hcancelPtT A1 A2 x (v1 : A1) (v2 : A2) : 
        valid (x :-> v1) -> x :-> v1 = x :-> v2 -> A1 = A2.
Proof. by move=>V; move/(hcancelPt V); move/idyn_injT. Qed.

Lemma hcancelPtV A x (v1 v2 : A) : 
        valid (x :-> v1) -> x :-> v1 = x :-> v2 -> v1 = v2.
Proof. by move=>V; move/(hcancelPt V)/idyn_inj. Qed.

Lemma hupdPt A1 A2 x (v1 : A1) (v2 : A2) : 
        upd x (idyn v1) (x :-> v2) = x :-> v1.
Proof. exact: gen_updPt. Qed.

Lemma hempbPt A x (v : A) : empb (x :-> v) = false.
Proof. exact: gen_empbPt. Qed.

(* valid *)

Lemma hvalidPtUn A x (v : A) h : 
        valid (x :-> v \+ h) = [&& x != null, valid h & x \notin dom h].
Proof. exact: gen_validPtUn. Qed.

Lemma hvalidPt_cond A x (v : A) h : valid (x :-> v \+ h) -> x != null.
Proof. exact: gen_validPt_cond. Qed.

Lemma hvalidPtV A x (v : A) h : valid (x :-> v \+ h) -> valid h.
Proof. exact: gen_validPtV. Qed.

Lemma hvalidPtD A x (v : A) h : valid (x :-> v \+ h) -> x \notin dom h.
Proof. exact: gen_validPtD. Qed.

(* dom *)

Lemma hdomPtUn A x (v : A) h : 
        dom (x :-> v \+ h) =i 
        [pred y | valid (x :-> v \+ h) & (x == y) || (y \in dom h)].
Proof. exact: gen_domPtUn. Qed.

(* find *)

Lemma hfindPtUn A x (v : A) h : 
        valid (x :-> v \+ h) -> find x (x :-> v \+ h) = Some (idyn v).
Proof. exact: gen_findPtUn. Qed.

(* free *)

Lemma hfreePtUn A x (v : A) h : 
        valid (x :-> v \+ h) -> free x (x :-> v \+ h) = h.
Proof. exact: gen_freePtUn. Qed.

(* upd *)

Lemma hupdPtUn A1 A2 x (v1 : A1) (v2 : A2) h : 
        upd x (idyn v1) (x :-> v2 \+ h) = x :-> v1 \+ h.
Proof. exact: gen_updPtUn. Qed.

Lemma heap_eta x h : 
        x \in dom h -> exists A (v : A), 
        find x h = Some (idyn v) /\ h = x :-> v \+ free x h.
Proof. by case/gen_eta; case=>A v H; exists A, v. Qed.

Lemma hcancel A1 A2 x (v1 : A1) (v2 : A2) h1 h2 : 
        valid (x :-> v1 \+ h1) -> 
        x :-> v1 \+ h1 = x :-> v2 \+ h2 -> 
       [/\ idyn v1 = idyn v2, valid h1 & h1 = h2].
Proof. exact: gen_cancel. Qed.

Lemma hcancelT A1 A2 x (v1 : A1) (v2 : A2) h1 h2 : 
        valid (x :-> v1 \+ h1) -> 
        x :-> v1 \+ h1 = x :-> v2 \+ h2 -> A1 = A2.
Proof. by move=>V; case/(hcancel V); move/idyn_injT. Qed.

Lemma hcancelV A x (v1 v2 : A) h1 h2 : 
        valid (x :-> v1 \+ h1) -> 
        x :-> v1 \+ h1 = x :-> v2 \+ h2 -> [/\ v1 = v2, valid h1 & h1 = h2].
Proof. by move=>V; case/(hcancel V); move/idyn_inj. Qed.

Lemma humfiltPt A p x (v : A) : 
        um_filter p (x :-> v) = 
        if p x then x :-> v else if x != null then Unit else um_undef.
Proof. exact: gen_umfiltPt. Qed.

Lemma heap_ind' (P : heap -> Prop) : 
         P um_undef -> P Unit ->
         (forall A x (v : A) f, P f -> valid (x :-> v \+ f) -> P (x :-> v \+ f)) ->
         forall f, P f.
Proof. 
move=>H1 H2 H3; apply: gen_ind' H1 H2 _. 
by move=>/= k [A v] f; apply: H3.
Qed.

End HeapPointsToLemmas.


(******************************************)
(* additional stuff, on top of union maps *)
(* mostly random stuff, kept for legacy   *)
(* should be removed/redone eventually    *)
(******************************************)

Definition fresh (h : heap) := 
  (if h is Heap.Def hs _ then last null (supp hs) else null) .+ 1.

Definition subdom (h1 h2 : heap) := 
  if (h1, h2) is (Heap.Def hs1 _, Heap.Def hs2 _) then 
    all (fun x => x \in supp hs2) (supp hs1)
  else false.

(* Definition subheap h1 h2 p :=  *)
(*   if (h1, h2) is (Def hs1 _, Def hs2 _) then  *)
(*     all (fun x => p (look x h1) (look x h2)) (supp hs1) *)
(*   else false. *)

(* this is not a good definition; the result should be *)
(* undefined if h2 is not a subset of h1 *)
(* then subdom predicate can be defined as def (h1 :- h2) *)
(* this will make the theorems about subtraction much nicer *)
Definition subtract (h1 h2 : heap) := 
  if h1 is (Heap.Def hs1 _) then 
    foldl (fun h x => free x h) h2 (supp hs1)
  else Heap.Undef.

Definition pick (h : heap) := 
  if h is Heap.Def hs _ then head null (supp hs) else null.

Definition subheap (h1 h2 : heap) :=
  subdom h1 h2 /\ forall x, x \in dom h1 -> find x h1 = find x h2.

Definition stores A (h : heap) (v : A) := 
  [Pred x | x \in dom h /\ find x h = Some (@idyn _ id _ v)].


Notation "h2 :- h1" := (subtract h1 h2) (at level 43, left associativity).

  
(*********)
(* fresh *)
(*********)

Lemma path_last n s x : path ord x s -> ord x (last x s).+(n+1).
Proof.
move: n s x.
suff L: forall s x, path ord x s -> ord x (last x s).+(1).
- elim=>[|n IH] // s x; move/IH=>E; apply: trans E _. 
  by case: (last x s)=>m; rewrite /ord /= addSn (addnS m).
elim=>[|y s IH x] /=; first by case=>x; rewrite /ord /= addn1.
by case/andP=>H1; move/IH; apply: trans H1.
Qed.

Lemma path_filter (A : ordType) (s : seq A) (p : pred A) x : 
        path ord x s -> path ord x (filter p s).
Proof.
elim: s x=>[|y s IH] x //=.
case/andP=>H1 H2.
case: ifP=>E; first by rewrite /= H1 IH.
apply: IH; elim: s H2=>[|z s IH] //=.
by case/andP=>H2 H3; rewrite (@trans _ y).
Qed.

Lemma dom_fresh h n : (fresh h).+n \notin dom h.
Proof.
suff L2: forall h x, x \in dom h -> ord x (fresh h).
- by apply: (contra (L2 _ _)); rewrite -leqNgt leq_addr. 
case=>[|[s H1]] //; rewrite /supp => /= H2 x.
rewrite /dom2 /dom' /fresh /supp -topredE /=. 
elim: s H1 null H2 x=>[|[y d] s IH] //= H1 x.
rewrite inE negb_or; case/andP=>H3 H4 z; rewrite inE.
case/orP; first by move/eqP=>->{z}; apply: (path_last 0).
by apply: IH; [apply: path_sorted H1 | apply: notin_path H1].
Qed.

Lemma fresh_null h : fresh h != null.
Proof. by rewrite -lt0n addn1. Qed.

Opaque fresh.

Hint Resolve dom_fresh fresh_null.

(********)
(* pick *)
(********)

Lemma emp_pick (h : heap) : (pick h == null) = (~~ valid h || empb h).
Proof.
rewrite /empb; case: h=>[|h] //=; case: (supp h)=>[|x xs] //=.
by rewrite inE negb_or eq_sym; case/andP; move/negbTE=>->.
Qed.

Lemma pickP h : valid h && ~~ empb h = (pick h \in dom h).
Proof. 
rewrite /dom2 /dom' /empb; case: h=>[|h] //=.
by case: (supp h)=>// *; rewrite inE eq_refl. 
Qed.

(**********)
(* subdom *)
(**********)

Lemma subdom_def h1 h2 : subdom h1 h2 -> valid h1 && valid h2. 
Proof. by case: h1 h2=>[|h1 H1] // [|h2 H2]. Qed.

Lemma subdomR h1 h2 : 
        valid h1 -> valid h2 -> 
        reflect (forall x, x \in dom h1 -> x \in dom h2) 
                (subdom h1 h2).
Proof. by case: h1 h2=>[|h1 H1] // [|h2 H2] //= _ _; apply: allP. Qed. 

Lemma subdomP h1 h2 : 
        valid h1 -> ~~ empb h1 -> 
        reflect (forall x, x \in dom h1 -> x \in dom h2) 
                (subdom h1 h2). 
Proof.
rewrite /empb.
case: h1 h2=>[|h1 H1] // [|h2 H2] //= _ H3; last by apply: allP.
apply: ReflectF.
suff H: head null (supp h1) \in supp h1 by move/(_ _ H).
by case: (supp h1) H1 H3=>[|x xs] //=; rewrite !inE eq_refl.
Qed.

Lemma subdomQ x h1 h2 : subdom h1 h2 -> x \in dom h1 -> x \in dom h2.
Proof.
move=>S H; case: subdomP S=>//. 
- by apply: dom_valid H.
- by case: empbP=>// E; rewrite E dom0 in H.
by move=>H2 _; apply: H2.
Qed.

Lemma subdom_refl h : valid h -> subdom h h.
Proof. by case: h=>[//|h H _]; apply/allP. Qed.

Lemma subdomD h1 h2 h : subdom h1 h2 -> valid (h2 \+ h) -> valid (h1 \+ h).
Proof.
case: h1 h2 h=>[|h1 H1]; case=>[|h2 H2]; case=>[|h H] //=.
rewrite /subdom !pcmE /= /UMC.defined /UMC.union /= /Heap.def /Heap.union /=.
case: ifP=>E1 //; case: ifP=>E2 // E _.
case: disjP E2=>// x H3 H4 _; case: disjP E1=>// X1 _.
by case: (allP (s := supp h1)) E=>//; move/(_ _ H3); move/X1; rewrite H4.
Qed.
  
Lemma subdomE h1 h2 h : 
        valid (h2 \+ h) -> subdom h1 h2 -> subdom (h1 \+ h) (h2 \+ h).
Proof.
case: h1 h2 h=>[|h1 H1]; case=>[|h2 H2]; case=>[|h H] //=.
rewrite /subdom !pcmE /= /UMC.defined /UMC.union /= /Heap.def /Heap.union /=.
case: ifP=>E1 // _; case: ifP=>E2; 
case: (allP (s:=supp h1))=>// E _; last first.
- case: disjP E2=>// x H3 H4; move/E: H3.
  by case: disjP E1=>// X _; move/X; rewrite H4.
case: (allP (s:=supp (fcat h1 h)))=>//; case=>x.
rewrite !supp_fcat !inE /= inE/=.
by case/orP; [move/E=>->| move=>->; rewrite orbT].
Qed.

Lemma subdomUE h1 h2 h1' h2' : 
        valid (h2 \+ h2') -> subdom h1 h2 -> subdom h1' h2' ->
        subdom (h1 \+ h1') (h2 \+ h2').
Proof.
case: h1 h2 h1' h2'=>[|h1 H1]; case=>[|h2 H2]; 
case=>[|h1' H1']; case=>[|h2' H2'] //.
rewrite /subdom !pcmE /= /UMC.defined /UMC.union /= /Heap.def /Heap.union /=.
case: ifP=>E1 // _; case: ifP=>E2 // T1 T2; last first.
- case: disjP E2=>// x; case: allP T1=>// X _; move/X=>{X}. 
  case: disjP E1=>// X _; move/X=>{X}.
  by case: allP T2=>// X _ H3 H4; move/X: H4 H3=>->.
case: allP=>//; case=>x.
rewrite !supp_fcat !inE; case/orP=>E.
- by case: allP T1=>//; move/(_ _ E)=>->.
by case: allP T2=>//; move/(_ _ E)=>->; rewrite orbT.
Qed.

Lemma subdom_emp h : valid h -> subdom Unit h.
Proof. by case: h. Qed.

Lemma subdom_emp_inv h : subdom h Unit -> h = Unit.
Proof. 
case: h=>[|h H] //; rewrite /subdom /=.
case: (allP (s:=supp h))=>// E _; apply/Heap.heapE; apply: fmapP=>x.
case: suppP (E x)=>// v E2; move/(_ (erefl _)).
by rewrite supp_nil.
Qed.

Lemma subdomPE A B x (v1 : A) (v2 : B) : 
        x != null -> subdom (x :-> v1) (x :-> v2).
Proof. 
move=>H; rewrite /subdom /heap_pts /pts /= /Heap.pts /Heap.upd /Heap.empty.
case: decP=>//= _.
rewrite (_ : FinMap _ = ins x (@idyn _ id _ v2) (finmap.nil _ _)) //.
by rewrite supp_ins inE /= eq_refl.
Qed.

Lemma subdom_trans h2 h1 h3 : subdom h1 h2 -> subdom h2 h3 -> subdom h1 h3.
Proof.
move=>H1 H2; move: (subdom_def H1) (subdom_def H2).
case/andP=>D1 _; case/andP=>_ D2. 
case E: (empb h1).
- by move/empbP: E => ->; rewrite subdom_emp.
apply/subdomP=>[//||x in1]; first by apply negbT.
by apply: (subdomQ H2) (subdomQ H1 in1).
Qed.

Hint Resolve subdom_emp subdomPE.

(***********)
(* subheap *)
(***********)

Lemma subheap_refl h : valid h -> subheap h h.
Proof. by move=>D; split=>//; apply: subdom_refl. Qed.

Lemma subheapE h : valid h -> subheap Unit h.
Proof. by split=>//; apply: subdom_emp. Qed.

Lemma subheapUn h1 h2 h1' h2' : 
        valid (h2 \+ h2') -> subheap h1 h2 -> subheap h1' h2' -> 
        subheap (h1 \+ h1') (h2 \+ h2').
Proof.
move=>defs [Sd1 Sl1] [Sd2 Sl2].
split=>[|x]; first by apply: subdomUE.
rewrite domUn inE /=; case/andP=>D; case/orP=>H.
- by rewrite !findUnL // H Sl1 // (subdomQ Sd1 H).
by rewrite !findUnR // H Sl2 // (subdomQ Sd2 H).
Qed.

Lemma subheapUnl h1 h2 : valid (h1 \+ h2) -> subheap h1 (h1 \+ h2).
Proof.
move=>D; rewrite -{1}[h1]unitR; apply: subheapUn=>//.
- by apply: subheap_refl; apply: validL D.
by apply: subheapE; apply: validR D.
Qed. 

Lemma subheapUnr h1 h2 : valid (h1 \+ h2) -> subheap h2 (h1 \+ h2).
Proof. by rewrite joinC; apply: subheapUnl. Qed.

Lemma subheap_def h1 h2 : subheap h1 h2 -> valid h1 /\ valid h2.
Proof. by case=>[subdm _]; move/andP: (subdom_def subdm). Qed.

Lemma subheap_trans h2 h1 h3 : subheap h1 h2 -> subheap h2 h3 -> subheap h1 h3.
Proof.
move=>[S12 E12] [S23 E23].
split=> [|x in1]; first by apply: (subdom_trans S12 S23).
by rewrite (E12 x in1); apply: (E23 x (subdomQ S12 in1)).
Qed.

Lemma subheap_id hp1 hp2: subheap hp1 hp2 -> subheap hp2 hp1 -> hp1 = hp2.
Proof.
move=>S12; move: (S12) => [D12 _].
move/andP: (subdom_def D12) S12=>{D12} [D1 D2].
case: hp1 D1=>[//=|fm1 pf1].
case: hp2 D2=>[//=|fm2 pf2] _ _ [S12 L12] [S21 L21].
rewrite Heap.heapE; apply: fmapP => k.
move: (@subdomQ k _ _ S12) (@subdomQ k _ _ S21) => S'12 S'21.
move: (L12 k) (L21 k).
case H1: (k \in dom (Heap.Def pf1)).
- move: (S'12 H1)=> H2.
  case F1: (fnd k fm1)=> [d1|]; case F2: (fnd k fm2)=> [d2|] //=; 
  rewrite /UMC.find /= F1 F2.
  - by move=>H; rewrite (H is_true_true).
  - by move: (fnd_supp_in H2); rewrite F2=> [[v]]. 
  - by move: (fnd_supp_in H1); rewrite F1=> [[v]]. 
case H2 : (k \in dom (Heap.Def pf2)).
- by rewrite (S'21 H2) in H1.
move => _ _; rewrite /dom2 /dom' -topredE in H2.
by rewrite (fnd_supp (negbT H1)) (fnd_supp (negbT H2)).
Qed.

Lemma subdom_subheap h1 h2 k1 k2: 
        valid (h1 \+ k1) -> subdom h1 h2 -> 
        h1 \+ k1 = h2 \+ k2 -> subheap h1 h2.
Proof.
move=>D1 H1 H2; split=>// x H3; move/(subdomQ H1): (H3)=>H4.
rewrite (_ : find x h1 = find x (h1 \+ k1)); last by rewrite findUnL // H3.
rewrite (_ : find x h2 = find x (h2 \+ k2)); 
  last by rewrite findUnL 1?H4 // -H2.
by rewrite H2.
Qed.

Lemma subheapL x h1 h2 : 
        x \in dom h1 -> subheap h1 h2 -> find x h1 = find x h2.
Proof. by move=>D [_]; move/(_ x D)=>->. Qed.

(************)
(* subtract *)
(************)

Lemma defS h1 h2 : valid (h1 :- h2) = valid h1 && valid h2.
Proof.
case: h2=>[|s2 N2] /=; first by rewrite andbF.
by elim: (supp s2) h1=>[|x xs IH] h1; [rewrite andbT | rewrite IH validF].
Qed.

Lemma domS (h1 h2 : heap) : 
        dom (h1 :- h2) =i 
        [pred x | valid h2 && (x \in [predD dom h1 & dom h2])].
Proof.
case: h2=>[//|s2 N2] /= x; rewrite !inE /= {2}/dom2 /dom'.
rewrite /UMC.dom_op /=.
elim: (supp s2) N2 h1=>[|y ys IH] //. 
rewrite !inE !negb_or /=; case/andP=>H1 H2 h1. 
rewrite IH //= domF -!topredE /= eq_sym.
by case: eqP=>//= ->; rewrite andbF.
Qed.

Lemma freeS h1 h2 x : free x (h1 :- h2) = free x h1 :- h2.
Proof.
case: h2=>[|s2 N2] //=; elim: (supp s2) h1=>[|y ys IH] h1 //=.
by rewrite IH freeF; case: eqP=>[->|//]; rewrite freeF eq_refl.
Qed.

Lemma lookS h1 h2 x : 
        find x (h1 :- h2) = 
        if valid h2 && (x \notin dom h2) then find x h1 else None.
Proof.
case: h2=>[|s2 N2] //=; rewrite /UMC.dom_op /= /dom2 /dom' /=. 
elim: (supp s2) N2 h1=>[//|y ys IH].
rewrite !inE !negb_or; case/andP=>H1 H2 h1.
by rewrite IH // findF; case: eqP=>//; case: ifP.
Qed.

Lemma defShUnh h1 h2 : valid (h1 :- h2 \+ h2) = valid h1 && valid h2.
Proof.
case: validUn=>/=. 
- by rewrite defS andbC; move/negbTE=>->.
- by move/negbTE=>->; rewrite andbF.
- by move=>x; rewrite domS !inE; case/and3P=>->; move/negbTE=>->.
by rewrite defS; case/andP=>->->.
Qed.

Lemma defSUn h1 h2 h3 : valid (h1 \+ h2) -> valid h3 -> valid (h1 :- h3 \+ h2).
Proof.
move=>D1 D2.
case: validUn=>//.
- by rewrite defS (validL D1) D2.
- by rewrite (validR D1).
move=>x; rewrite domS !inE D2 /=; case/andP=>_.
by case: validUn D1=>// _ _ L _; move/L; move/negbTE=>->.
Qed.

Lemma subhh h : valid h -> h :- h = Unit.
Proof.
move=>D.
apply: subheap_id; last by apply: subheapE; rewrite defS D.
split=>[|x]; last by rewrite domS !inE D /= andNb.
case: subdomR=>//; first by rewrite defS D.
by case=>x; rewrite domS !inE D /= andNb.
Qed.

Lemma sub_undef h : um_undef :- h = um_undef.
Proof.
case: h=>//= f; elim: {f} (finmap.supp f)=>[|t s IH] //=.
by rewrite inE /= negb_or; case/andP=>H1; apply: IH.
Qed.

Lemma subD h1 h2 : valid (h1 \+ h2) -> h1 :- h2 = h1.
Proof.
move=>D.
have [def1 def2] : valid h1 /\ valid h2.
- by rewrite (validL D) (validR D).
have def3 : valid (h1 :- h2) by rewrite defS def1 def2. 
case: validUn D=>// _ _ L _.
have dom1 : dom (h1 :- h2) =i dom h1.
- move=>x; rewrite domS !inE def2 /= andbC.
  by case E: (x \in dom h1)=>//=; apply: L E.
apply: subheap_id.
- split=>[|x]; first by case: subdomR=>//; case=>x; rewrite dom1.
  rewrite domS !inE def2 /=; case/andP=>H1.
  by rewrite lookS def2 H1.
split=>[|x Dom]; last by rewrite lookS def2 /= (L _ Dom).
case: subdomR=>//; case=>x Dom.
by rewrite domS !inE /= def2 /= (L _ Dom) Dom.
Qed.

(* :- is right distributive over :+ *)
Lemma subUnr h1 h2 h3 : 
        valid (h1 \+ h2) -> h1 \+ h2 :- h3 = (h1 :- h3) \+ (h2 :- h3).
Proof.
move=>D.
case D3 : (valid h3); last first.
- have []: ~~ valid (h1 \+ h2 :- h3) /\ ~~ valid (h1 :- h3).
  - by rewrite !defS D D3 !andbF.
  by do 2![move/invalidE=>->].
have D1 : valid (h1 \+ h2 :- h3) by rewrite defS D D3.
have D2 : valid (h1 :- h3 \+ (h2 :- h3)).
- case: validUn=>//=.
  - by rewrite defS (validL D) D3.
  - by rewrite defS (validR D) D3.
  move=>x; rewrite !domS !inE /= !inE /= D3 /=.
  case/andP=>->/=; case: validUn D=>// _ _ L _.
  by move/L; move/negbTE=>->.
have dom1 : dom (h1 \+ h2 :- h3) =i dom (h1 :- h3 \+ (h2 :- h3)).
- move=>x.
  rewrite domS !inE !domUn !inE D D2 D3 !domS !inE /= !inE D3 /=.
  by case: (x \in dom h3).
apply: subheap_id.
- split=>[|x]; first by case: subdomR=>//; case=>x; rewrite dom1.
  rewrite domS !inE D3 /= domUn !inE /= D /=.
  case/andP=>H1; case/orP=>H2.
  - rewrite lookS D3 H1 findUnL //= H2. 
    by rewrite findUnL // domS !inE /= D3 H1 H2 lookS D3 H1.
  rewrite lookS D3 H1 findUnR // H2 findUnR //.
  by rewrite domS !inE /= D3 H1 H2 lookS D3 H1.
split=>[|x].
- case: subdomR=>//; case=>x.
  rewrite domUn !inE D2 /= !domS !inE D3 /=.
  rewrite !inE /= !inE /= domUn inE /= D.
  by case: (x \in dom h3).
rewrite -dom1 domS !inE D3 /= domUn !inE D /=.
case/andP=>H1; case/orP=>H2.
- rewrite findUnL // domS !inE /= D3 H1 H2.
  by rewrite !lookS D3 H1 /= findUnL // H2.
rewrite findUnR // domS !inE /= D3 H1 H2.
by rewrite !lookS D3 H1 /= findUnR // H2.
Qed.

Lemma subUnC h h1 h2 : valid (h1 \+ h2) -> h \+ h1 :- h2 = h :- h2 \+ h1.
Proof.
move=>D1; case D2: (valid (h \+ h1)); first by rewrite subUnr // (subD D1). 
move/invalidE: (negbT D2)=>->; rewrite sub_undef.
suff: ~~ (valid (h :- h2 \+ h1)) by move/invalidE=>->.
case: validUn=>//; rewrite defS; case/andP=>H1 H2 H3.
case: validUn D2=>//; first by [rewrite H1]; first by rewrite H3.
move=>x Dom Dom1 _; move/(_ x).
rewrite domS !inE H2 Dom Dom1 andbT /=; apply.
by case: validUn D1=>// _ _; move/(_ _ Dom1).
Qed.

Lemma unhSh h1 h2 : valid (h1 \+ h2) -> h1 \+ h2 :- h2 = h1.
Proof. by move=>D; rewrite subUnr // subhh ?(validR D) // unitR subD. Qed.

(* a lot of times, we work with the subheap predicate *)
(* so some lemmas should be specialized for that *)
 
Lemma subheapK h1 h2 : subheap h1 h2 -> h2 :- h1 \+ h1 = h2.
Proof.
case=>H1 H2.
have subdomM: dom (h2 :- h1 \+ h1) =i dom h2.
- move=>x; move/subdom_def: (H1); case/andP=>D1 D2.
  rewrite domUn !inE domS !inE defShUnh D1 D2 /=.
  case E: (_ \in dom h1); last by rewrite orbF.
  by apply/esym; case: subdomR H1=>// H _; apply: H.
move/subdom_def: (H1); case/andP=>D1 D2.
have D3: valid (h2 :- h1 \+ h1) by rewrite defShUnh D1 D2.
apply: subheap_id.
- split=>[|x]; first by case: subdomR=>//; case=>x; rewrite subdomM.
  rewrite subdomM // findUnL // domS !inE lookS D1 /= =>->.  
  by case E: (x \in dom h1)=>//=; apply: H2.
split=>[|x H].
- by case: subdomR=>//; case=>x; rewrite subdomM.
rewrite findUnL // domS !inE D1 H andbT /=.
case: ifP=>E; first by rewrite lookS D1 E.
by rewrite H2 // (negbFE E).
Qed.

Lemma subheapX h1 h2 : 
        subheap h1 h2 <-> h2 = h1 \+ (h2 :- h1) /\ valid h2.
Proof.
split=>[S|[->]]; last by apply: subheapUnl.
by rewrite joinC subheapK //; case/subheap_def: S.
Qed.

(* some lemmas regarding symmetric differences *)
(* not sure how useful these are, but let's leave them in for now *)

Lemma subheapS sh1 sh2 ep1 ep2 : 
        valid (sh1 \+ ep1) ->
        sh1 \+ ep1 = sh2 \+ ep2 -> 
        subheap ep2 sh1 -> subheap ep1 sh2 -> 
        sh1 :- ep2 = sh2 :- ep1.
Proof.
move=>D1 E1; case/subheapX=>L1 _; case/subheapX=>L2 _.
rewrite (_ : sh1 = sh2 \+ ep2 :- ep1); last by rewrite -E1 (unhSh D1).
rewrite subUnC; last by rewrite L1 joinAC in D1; apply: validL D1.
rewrite unhSh //; apply: defSUn; first by rewrite -E1.
by apply: validR D1.
Qed.

(* sh1 :- ep2 is the intersection of sh1 and sh2 *)

(* it is an upper bound of subheaps of sh1 and sh2 *)

Lemma subheapB sh1 sh2 ep1 ep2 : 
        valid (sh1 \+ ep1) ->
        sh1 \+ ep1 = sh2 \+ ep2 ->
        subheap ep2 sh1 -> subheap ep1 sh2 -> 
        forall h, subheap h sh1 -> subheap h sh2 -> subheap h (sh1 :- ep2).
Proof.
move=>D1 E S1 S2 h H1 H2; have D2 : valid (sh2 \+ ep2) by rewrite E in D1.
have [Dh Ds1 Ds2 De1 De2] : 
  [/\ valid h, valid sh1, valid sh2, valid ep1 & valid ep2].
- rewrite (validL D1) (validR D1) (validL D2) (validR D2).
  by case: H1; move/subdom_def; case/andP.
move: H1 H2=>[R1 P1][R2 P2].
split.
- case: subdomR=>//; first by rewrite defS Ds1 De2.
  case=>x Domh; rewrite domS !inE De2 (subdomR Dh Ds1 R1 _ Domh) andbT /=.
  by case: validUn D2=>// _ _; move/(_ _ (subdomR Dh Ds2 R2 _ Domh)). 
move=>x Domh; rewrite lookS De2 /=.
case: validUn D2=>// _ _; move/(_ _ (subdomR Dh Ds2 R2 _ Domh))=>-> _.
by apply: P1 Domh.
Qed.

(* and it is the least upper bound *)

Lemma subheapM sh1 sh2 ep1 ep2 m : 
        valid (sh1 \+ ep1) -> 
        sh1 \+ ep1 = sh2 \+ ep2 -> 
        subheap ep2 sh1 -> subheap ep1 sh2 ->
        (forall h, subheap h sh1 -> subheap h sh2 -> subheap h m) -> 
        subheap (sh1 :- ep2) m.
Proof.
move=>D1 E S1 S2 H; have D2 : valid (sh2 \+ ep2) by rewrite E in D1.
move: (subheapS D1 E S1 S2)=>F.
case/subheapX: S1=>S1 _; case/subheapX: S2=>S2 _.
apply: H; apply/subheapX; [move: S1 D1 D2|move: S2 D2 D1; rewrite F];
by move=>S L1 L2; rewrite {4}S unhSh joinC ?defShUnh (validL L1) // (validR L2).
Qed.

(**********)
(* stores *)
(**********)

Lemma stores_injT A1 A2 h x (v1 : A1) (v2 : A2) :
        x \In stores h v1 -> x \In stores h v2 -> 
        A1 = A2 /\ ijmeq (sort:=id) v1 v2.
Proof.
case=>_ E [_]; move: E=>-> [E]; rewrite -{A2}E in v2 *.
by move/(inj_pair2 _ _ _ _ _)=>->.
Qed.

Lemma stores_inj A h x (v1 v2 : A) :
        x \In stores h v1 -> x \In stores h v2 -> v1 = v2.
Proof. by case=>_ E [_]; rewrite E; case; move/(inj_pair2 _ _ _ _ _). Qed.

Lemma storesD A h x (v : A) : x \In stores h v -> x \in dom h.
Proof. by case. Qed.

Lemma storesU A1 A2 h (v1 : A1) (v2 : A2) x : 
        x \In stores h v1 -> x \In stores (upd x (@idyn _ id _ v2) h) v2.
Proof.
case=>D L; split.  
- by rewrite domU /= inE eq_refl (dom_cond D) (dom_valid D).
by rewrite findU eq_refl (dom_valid D) (dom_cond D).
Qed.

Lemma storesUnl A h1 h2 x (v : A) :
        valid (h1 \+ h2) -> x \In stores h1 v -> x \In stores (h1 \+ h2) v.
Proof.
move=>D [Dom L]; split; first by rewrite domUn !inE /= D Dom.
by rewrite findUnL // Dom.
Qed.

Lemma storesUnr A h1 h2 x (v : A) :
        valid (h1 \+ h2) -> x \In stores h2 v -> x \In stores (h1 \+ h2) v.
Proof. by rewrite joinC; apply: storesUnl. Qed.

Lemma stores_mono A h1 h2 (v : A) : 
        valid (h1 \+ h2) -> stores h1 v <=p stores (h1 \+ h2) v.
Proof.
move=>D x; rewrite !InE /=; case=>D1 E.
split=>//; first by rewrite domUn !inE /= D D1.
by rewrite findUnL // D1.
Qed.

Lemma stores_sub A h1 h2 (v : A) : 
        {in dom h1, stores (h1 \+ h2) v <=p stores h1 v}.
Proof.
move=>x D /=; rewrite !InE /= /stores domUn !inE.
by case; case/andP=>D1 _; rewrite findUnL // D.
Qed.

(***********************)
(* Some derived lemmas *)
(***********************)

Lemma dom_agree (h1 h2 k1 k2 : heap) : 
        valid (h1 \+ k1) -> dom h1 =i dom h2 -> h1 \+ k1 = h2 \+ k2 -> h1 = h2.
Proof.
move=>D1 H1 H2; have D2: valid (h2 \+ k2) by rewrite -H2. 
apply: subheap_id.
- apply: subdom_subheap H2=>//.
  apply/subdomR; [by apply: validL D1 | by apply: validL D2 |].
  by move=>x; rewrite (H1 x).
apply: subdom_subheap (esym H2)=>//.
apply/subdomR; [by apply: validL D2 | by apply: validL D1 |].
by move=>x; rewrite (H1 x).
Qed.

Lemma noalias (h1 h2 : heap) x1 x2 : 
        x1 \in dom h1 -> x2 \in dom h2 -> valid (h1 \+ h2) -> x1 != x2.
Proof.
by case: validUn=>// H1 H2 H; move/H; case: eqP=>// ->; move/negbTE=>->.
Qed.

Lemma domPtUnX A (v : A) x i : valid (x :-> v \+ i) -> x \in dom (x :-> v \+ i).
Proof. by move=>D; rewrite hdomPtUn inE /= D eq_refl. Qed.

Lemma domPtX A (v : A) x : valid (x :-> v) -> x \in dom (x :-> v).
Proof. by move=>D; rewrite -(unitR (x :-> v)) domPtUnX // unitR. Qed.

Lemma dom_notin_notin (h1 h2 : heap) x :
        valid (h1 \+ h2) -> x \notin dom (h1 \+ h2) -> x \notin dom h1.
Proof. by move=>D; rewrite domUn inE /= negb_and negb_or /= D; case/andP. Qed.

Lemma dom_in_notin (h1 h2 : heap) x : 
       valid (h1 \+ h2) -> x \in dom h1 -> x \notin dom h2.
Proof. by case: validUn=>// D1 D2 H _; apply: H. Qed.

Lemma storesPt A (v : A) x h : valid (x :-> v \+ h) -> x \In stores (x :-> v \+ h) v.
Proof.
move=>D; split=>//=; first by apply/domPtUnX.
by rewrite hfindPtUn.
Qed.

Lemma storesPtX A (v : A) x : valid (x :-> v) -> x \In stores (x :-> v) v.
Proof. by rewrite -(unitR (_ :-> _)); apply: storesPt. Qed.


(*************************************)
(* the automation of the PtUn lemmas *)
(*************************************)

(* First, the mechanism for search-and-replace for the overloaded lemas, *)
(* pattern-matching on heap expressions.                                 *)

Structure tagged_heap := Tag {untag :> heap}.

Definition right_tag := Tag.
Definition left_tag := right_tag.
Canonical found_tag i := left_tag i.

Definition partition_axiom k r (h : tagged_heap) := untag h = k \+ r.

Structure partition (k r : heap) :=
  Partition {heap_of :> tagged_heap; 
             _ : partition_axiom k r heap_of}.

Lemma partitionE r k (f : partition k r) : untag f = k \+ r.
Proof. by case: f=>[[j]] /=; rewrite /partition_axiom /= => ->. Qed.

Lemma found_pf k : partition_axiom k Unit (found_tag k). 
Proof. by rewrite /partition_axiom unitR. Qed.

Canonical found_struct k := Partition (found_pf k).

Lemma left_pf h r (f : forall k, partition k r) k : 
        partition_axiom k (r \+ h) (left_tag (untag (f k) \+ h)).
Proof. by rewrite partitionE /partition_axiom /= joinA. Qed.

Canonical left_struct h r (f : forall k, partition k r) k :=
  Partition (left_pf h f k).

Lemma right_pf h r (f : forall k, partition k r) k :
        partition_axiom k (h \+ r) (right_tag (h \+ f k)).
Proof. by rewrite partitionE /partition_axiom /= joinCA. Qed.

Canonical right_struct h r (f : forall k, partition k r) k := 
  Partition (right_pf h f k).

(* now the actual lemmas *)

Lemma defPtUnO A h x (v : A) (f : partition (x :-> v) h) : 
        valid (untag f) = [&& x != null, valid h & x \notin dom h].
Proof. by rewrite partitionE hvalidPtUn. Qed.

Implicit Arguments defPtUnO [A h v f].

Lemma defPt_nullO A h x (v : A) (f : partition (x :-> v) h) : 
        valid (untag f) -> x != null.
Proof. by rewrite partitionE; apply: hvalidPt_cond. Qed.

Implicit Arguments defPt_nullO [A h x v f].

Lemma defPt_defO A h x (v : A) (f : partition (x :-> v) h) : 
        valid (untag f) -> valid h.
Proof. by rewrite partitionE; apply: hvalidPtV. Qed.

Implicit Arguments defPt_defO [A h v f].

Lemma defPt_domO A h x (v : A) (f : partition (x :-> v) h) : 
        valid (untag f) -> x \notin dom h.
Proof. by rewrite partitionE; apply: hvalidPtD. Qed.

Implicit Arguments defPt_domO [A h v f].

Lemma domPtUnO A h x (v : A) (f : partition (x :-> v) h) : 
        dom (untag f) =i 
        [pred y | valid (untag f) & (x == y) || (y \in dom h)].
Proof. by rewrite partitionE; apply: hdomPtUn. Qed.

Implicit Arguments domPtUnO [A h v f].

Lemma lookPtUnO A h x (v : A) (f : partition (x :-> v) h) : 
        valid (untag f) -> find x (untag f) = Some (@idyn _ id _ v).
Proof. by rewrite partitionE; apply: hfindPtUn. Qed.

Implicit Arguments lookPtUnO [A h x v f].

Lemma freePtUnO A h x (v : A) (f : partition (x :-> v) h) : 
        valid (untag f) -> free x (untag f) = h.
Proof. by rewrite partitionE; apply: hfreePtUn. Qed.

Implicit Arguments freePtUnO [A h x v f].

Lemma updPtUnO A1 A2 x i (v1 : A1) (v2 : A2) 
               (f : forall k, partition k i) : 
        upd x (@idyn _ id _ v2) (untag (f (x :-> v1))) = f (x :-> v2). 
Proof. by rewrite !partitionE; apply: hupdPtUn. Qed.

Implicit Arguments updPtUnO [A1 A2 x i v1 v2].

Lemma cancelTO A1 A2 h1 h2 x (v1 : A1) (v2 : A2) 
               (f1 : partition (x :-> v1) h1) 
               (f2 : partition (x :-> v2) h2) : 
        valid (untag f1) -> f1 = f2 :> heap -> A1 = A2. 
Proof. by rewrite !partitionE; apply: hcancelT. Qed.

Implicit Arguments cancelTO [A1 A2 h1 h2 v1 v2 f1 f2].

Lemma cancelO A h1 h2 x (v1 v2 : A) 
              (f1 : partition (x :-> v1) h1) 
              (f2 : partition (x :-> v2) h2) : 
        valid (untag f1) -> f1 = f2 :> heap -> 
        [/\ v1 = v2, valid h1 & h1 = h2].
Proof. by rewrite !partitionE; apply: hcancelV. Qed.

Implicit Arguments cancelO [A h1 h2 v1 v2 f1 f2].

Lemma domPtUnXO A (v : A) x i (f : partition (x :-> v) i) : 
        valid (untag f) -> x \in dom (untag f).
Proof. by rewrite partitionE; apply: domPtUnX. Qed.

(******************************)
(* Properties of block update *)
(******************************)

Section BlockUpdate.
Variable (A : Type).

Fixpoint updi x (vs : seq A) {struct vs} : heap := 
  if vs is v'::vs' then x :-> v' \+ updi (x .+ 1) vs' else Unit.

Lemma updiS x v vs : updi x (v :: vs) = x :-> v \+ updi (x .+ 1) vs.
Proof. by []. Qed.   

Lemma updi_last x v vs : 
        updi x (rcons vs v) = updi x vs \+ x.+(size vs) :-> v.
Proof.
elim: vs x v=>[|w vs IH] x v /=.
- by rewrite ptr0 unitR unitL.
by rewrite -(addn1 (size vs)) addnC -ptrA IH joinA.
Qed.

Lemma updi_cat x vs1 vs2 : 
        updi x (vs1 ++ vs2) = updi x vs1 \+ updi x.+(size vs1) vs2.
Proof.
elim: vs1 x vs2=>[|v vs1 IH] x vs2 /=.
- by rewrite ptr0 unitL.
by rewrite -(addn1 (size vs1)) addnC -ptrA IH joinA.
Qed.

Lemma updi_catI x y vs1 vs2 : 
        y = x.+(size vs1) -> updi x vs1 \+ updi y vs2 = updi x (vs1 ++ vs2).
Proof. by move=>->; rewrite updi_cat. Qed.

(* helper lemma *)
Lemma updiVm' x m xs : m > 0 -> x \notin dom (updi x.+m xs).
Proof. 
elim: xs x m=>[|v vs IH] x m //= H.
rewrite ptrA hdomPtUn inE /= negb_and negb_or -{4}(ptr0 x) ptrK -lt0n H /=.
by rewrite orbC IH // addn1.
Qed.

Lemma updiD x xs : valid (updi x xs) = (x != null) || (size xs == 0).
Proof.
elim: xs x=>[|v xs IH] x //=; first by rewrite orbC.
by rewrite hvalidPtUn updiVm' // orbF IH ptr_null andbF andbC.
Qed.

Lemma updiVm x m xs : 
        x \in dom (updi x.+m xs) = [&& x != null, m == 0 & size xs > 0].
Proof.
case: m=>[|m] /=; last first.
- by rewrite andbF; apply: negbTE; apply: updiVm'.
case: xs=>[|v xs]; rewrite ptr0 ?andbF ?andbT //=.
by rewrite hdomPtUn inE /= eq_refl -updiS updiD orbF andbT /=.
Qed.

Lemma updimV x m xs : 
        x.+m \in dom (updi x xs) = (x != null) && (m < size xs).
Proof.
case H: (x == null)=>/=.
- by case: xs=>// a s; rewrite (eqP H).
elim: xs x m H=>[|v vs IH] x m H //; case: m=>[|m].
- by rewrite ptr0 /= hdomPtUn inE /= eq_refl andbT -updiS updiD H.
rewrite -addn1 addnC -ptrA updiS hdomPtUn inE /= IH; last first.
- by rewrite ptrE /= addn1.
by rewrite -updiS updiD H /= -{1}(ptr0 x) ptrA ptrK. 
Qed.

Lemma updiP x y xs : 
        reflect (y != null /\ exists m, x = y.+m /\ m < size xs) 
                (x \in dom (updi y xs)).
Proof.
case H: (y == null)=>/=.
- by rewrite (eqP H); elim: xs=>[|z xs IH] //=; constructor; case.
case E: (x \in _); constructor; last first.
- by move=>[_][m][H1] H2; rewrite H1 updimV H2 H in E.
case: (ptrT x y) E=>m; case/orP; move/eqP=>->.
- by rewrite updimV H /= => H1; split=>//; exists m.
rewrite updiVm; case/and3P=>H1; move/eqP=>-> H2.
by split=>//; exists 0; rewrite ptrA addn0 ptr0.
Qed.

(* Invertibility *)
Lemma updi_inv x xs1 xs2 : 
        valid (updi x xs1) -> updi x xs1 = updi x xs2 -> xs1 = xs2.
Proof.
elim: xs1 x xs2 =>[|v1 xs1 IH] x /=; case=>[|v2 xs2] //= D; 
[move/esym| | ]; try by rewrite empbE empbUn hempbPt.
by case/(hcancelV D)=><- {D} D; move/(IH _ _ D)=><-.
Qed.

Lemma updi_iinv x xs1 xs2 h1 h2 : 
        size xs1 = size xs2 -> valid (updi x xs1 \+ h1) ->        
        updi x xs1 \+ h1 = updi x xs2 \+ h2 -> xs1 = xs2 /\ h1 = h2.
Proof.
elim: xs1 x xs2 h1 h2=>[|v1 xs1 IH] x /=; case=>[|v2 xs2] //= h1 h2.
- by rewrite !unitL.
move=>[E]; rewrite -!joinA=>D; case/(hcancelV D)=><- {D} D.
by case/(IH _ _ _ _ E D)=>->->.
Qed.

End BlockUpdate. 

(***********************************************************)
(* Properties of block update2 -- used in information flow *)
(* to define low/high allocb.                              *)
(***********************************************************)

Section BlockUpdate2.
Variable (A : Type).

Fixpoint updi2 x (vs : seq A) {struct vs} : heap := 
  if vs is v'::vs' then (x :-> v') \+ updi2 (x .+ 2) vs' else Unit.

Lemma updi2S x v vs : updi2 x (v :: vs) = x :-> v \+ updi2 (x .+ 2) vs.
Proof. by []. Qed.   

Lemma updi2_last x v vs : 
        updi2 x (rcons vs v) = updi2 x vs \+ x.+(2*size vs) :-> v.
Proof.
elim: vs x v=>[|w vs IH] x v /=.
- by rewrite ptr0 unitR unitL. 
by rewrite -(addn1 (size vs)) addnC IH joinA ptrA mulnDr muln1.
Qed.

Lemma updi2_cat x vs1 vs2 : 
        updi2 x (vs1 ++ vs2) = updi2 x vs1 \+ updi2 x.+(2*size vs1) vs2.
Proof.
elim: vs1 x vs2=>[|v vs1 IH] x vs2 /=.
- by rewrite ptr0 unitL.
by rewrite -(addn1 (size vs1)) addnC IH joinA ptrA mulnDr muln1. 
Qed.

Lemma updi2_catI x y vs1 vs2 : 
        y = x.+(2*size vs1) -> updi2 x vs1 \+ updi2 y vs2 = updi2 x (vs1 ++ vs2).
Proof. by move=>->; rewrite updi2_cat. Qed.

(* helper lemma *)
Lemma updi2Vm' x m xs : m > 0 -> x \notin dom (updi2 x.+m xs).
Proof. 
elim: xs x m=>[|v vs IH] x m //= H.
rewrite ptrA hdomPtUn inE /= negb_and negb_or -{4}(ptr0 x) ptrK -lt0n H /=.
by rewrite orbC IH // ltn_addr.
Qed.

Lemma updi2D x xs : valid (updi2 x xs) = (x != null) || (size xs == 0).
Proof.
elim: xs x=>[|v xs IH] x //=; first by rewrite orbC.
by rewrite hvalidPtUn updi2Vm' // orbF IH ptr_null andbF andbC.
Qed.

Lemma updi2Vm x m xs : 
        x \in dom (updi2 x.+m xs) = [&& x != null, m == 0 & size xs > 0].
Proof.
case: m=>[|m] /=; last first.
- by rewrite andbF; apply: negbTE; apply: updi2Vm'.
case: xs=>[|v xs]; rewrite ptr0 ?andbF ?andbT //=.
by rewrite hdomPtUn inE /= eq_refl -updi2S updi2D orbF andbT /=.
Qed.

(* Invertibility *)
Lemma updi2_inv x xs1 xs2 : 
        valid (updi2 x xs1) -> updi2 x xs1 = updi2 x xs2 -> xs1 = xs2.
Proof.
elim: xs1 x xs2 =>[|v1 xs1 IH] x /=; case=>[|v2 xs2] //= D; 
[move/esym| | ]; try by rewrite empbE empbUn hempbPt.
by case/(hcancelV D)=><- {D} D; move/(IH _ _ D)=><-.
Qed.

Lemma updi2_iinv x xs1 xs2 h1 h2 : 
        size xs1 = size xs2 -> valid (updi2 x xs1 \+ h1) ->        
        updi2 x xs1 \+ h1 = updi2 x xs2 \+ h2 -> xs1 = xs2 /\ h1 = h2.
Proof.
elim: xs1 x xs2 h1 h2=>[|v1 xs1 IH] x /=; case=>[|v2 xs2] //= h1 h2.
- by rewrite !unitL.
move=>[E]; rewrite -!joinA=>D; case/(hcancelV D)=><- {D} D.
by case/(IH _ _ _ _ E D)=>->->.
Qed.

Lemma updi2_in_dom x b (s : seq A) :
        x \in dom (updi2 b s) -> exists m, x = b.+(2 * m).
Proof.
elim: s b=>//= a s' IH b; rewrite hdomPtUn; case/andP=> Def; case/orP. 
by move/eqP=>->; exists 0; rewrite muln0 ptr0.
by move=> H; move: (IH _ H)=>[m]->; exists (m.+1); rewrite mulnS ptrA.
Qed.

End BlockUpdate2. 

(*********************************************************************)
(* Freshness for even and odd locations -- used in information flow, *)
(* where heaps are split into low (even) and high (odd) parts.       *)
(*********************************************************************)

Definition low : pred ptr := fun x => 0 == nat_ptr x %[mod 2]. 
Definition high : pred ptr := fun x => 1 == nat_ptr x %[mod 2]. 

Definition get_lows h := 
  if h is Heap.Def hs _ then filter low (supp hs) else [::].

Definition get_highs h := 
  if h is Heap.Def hs _ then filter high (supp hs) else [::].

Definition ldom h : pred ptr := fun x => x \in get_lows h.
Definition hdom h : pred ptr := fun x => x \in get_highs h.

Lemma ldomP h x : x \in ldom h = (x \in dom h) && low x. 
Proof.
case: h=>[//|[h S]]; rewrite /ldom /= /dom2 /dom' /= /supp /= =>H.
rewrite -!topredE /=.
elim: (map key h)=>[|y s IH] //=.
case: ifP=>E; rewrite !inE IH; case: eqP=>// -> //=.
by rewrite E andbF.
Qed.

Lemma hdomP h x : x \in hdom h = (x \in dom h) && high x. 
Proof.
case: h=>[//|[h S]]; rewrite /hdom /= /dom2 /dom' /= /supp /= =>H.
rewrite -!topredE /=.
elim: (map key h)=>[|y s IH] //=.
case: ifP=>E; rewrite !inE IH; case: eqP=>// -> //=.
by rewrite E andbF.
Qed.

Lemma ldomK h1 h2 t1 t2 : 
        valid (h1 \+ t1) -> valid (h2 \+ t2) ->
        ldom h1 =i ldom h2 -> 
        ldom (h1 \+ t1) =i ldom (h2 \+ t2) -> ldom t1 =i ldom t2.
Proof.
move=>D1 D2 H1 H2 x; move: {H1 H2} (H1 x) (H2 x).
rewrite !ldomP !domUn !inE.
case: validUn D1=>// H1 H2 L1 _; case: validUn D2=>// H3 H4 L2 _.
case E1: (x \in dom t1); case E2: (x \in dom t2)=>//; rewrite orbF orbT /=;
case E3: (x \in dom h1); case E4: (x \in dom h2)=>//= _ _; 
by [move/L1: E3; rewrite E1 | move/L2: E4; rewrite E2].
Qed.

Definition lfresh h := (last null (get_lows h)) .+ 2.
Definition hfresh h := (last (null .+ 1) (get_highs h)) .+ 2.

Lemma last_inv A B (f : A -> B) (x1 x2 : A) (h : seq A) : 
        f x1 = f x2 -> f (last x1 h) = f (last x2 h).
Proof. by elim: h. Qed.

Lemma lfresh_low h n : low (lfresh h) .+ (2*n).
Proof.
rewrite /lfresh /low /get_lows. 
case: h; first by rewrite modnDl modnMr.
case; rewrite /supp /low /=.
elim=>[|[[x] v] h IH] /=; first by rewrite modnDl modnMr.
rewrite inE negb_or ptrE /=; move/path_sorted=>H1; case/andP=>H2 H3.
case: ifP=>E /=; last by apply: IH.
set f := fun x => (nat_ptr x + 2 + 2 * n) %% 2.
have F: f (ptr_nat x) = f null.
- by rewrite /f -modn_mod -addnA -modnDm -(eqP E) !modn_mod.
move: (last_inv (f := f) (x1 := (ptr_nat x)) (x2 := null))=>L.
by rewrite /f /= in L; rewrite {}L //; apply: IH.
Qed.

Lemma hfresh_high h n : high (hfresh h) .+ (2*n).
Proof.
rewrite /hfresh /high /get_highs. 
case: h n=>[n|].
- by rewrite /null /= add0n -addnA -modnDm modnDl modnMr addn0.
case; rewrite /supp /high /=.
elim=>[|[[x] v] h IH] /=.
- by move=>_ _ n; rewrite add0n -addnA -modnDm modnDl modnMr addn0.
rewrite inE negb_or ptrE /=; move/path_sorted=>H1; case/andP=>H2 H3.
case: ifP=>E n /=; last by apply: IH.
set f := fun x => (nat_ptr x + 2 + 2 * n) %% 2.
have F: f (ptr_nat x) = f (null .+ 1).
- rewrite /f -modn_mod /= add0n -addnA.
   rewrite -modnDm -(eqP E) modn_mod.
   by rewrite modnDm addnA.
move: (last_inv (f := f) (x1 := (ptr_nat x)) (x2 := null .+ 1))=>L.
by rewrite /f /= in L; rewrite {}L //; apply: IH.
Qed.

Lemma dom_lfresh h n : (lfresh h) .+ (2*n) \notin dom h. 
Proof.
suff L2: forall (h : heap) x, low x -> x \in dom h -> ord x (lfresh h).
- apply: (contra (L2 _ _ (lfresh_low h n))).
  by rewrite -leqNgt leq_addr.
case=>[|[s H1]] //; rewrite /supp => /= H2 x.
rewrite /dom2 /dom' /= /lfresh /get_lows /low /supp -topredE /=.
elim: s H1 null H2 x=>[|[y d] s IH] //= H1 x. 
rewrite inE negb_or; case/andP=>H3 H4 z /= E; rewrite inE.
case/orP=>H5.
- rewrite -!{H5 y}(eqP H5) E in H1 H3 *.
  by apply: (path_last 1); apply: path_filter.
by case: ifP=>E2; apply: IH=>//; move: H1; 
[apply: path_sorted | apply: notin_path | apply: path_sorted].
Qed.

Lemma dom_hfresh h n : (hfresh h) .+ (2*n) \notin dom h. 
Proof.
suff L2: forall (h : heap) x, high x -> x \in dom h -> ord x (hfresh h).
- apply: (contra (L2 _ _ (hfresh_high h n))).
  by rewrite -leqNgt leq_addr.
case=>[|[s H1]] //; rewrite /supp => /= H2 x.
rewrite /dom2 /dom' /= /hfresh /get_highs /high /supp -topredE /=.
elim: s H1 null H2 x=>[|[y d] s IH] //= H1 x. 
rewrite inE negb_or; case/andP=>H3 H4 z /= E; rewrite inE.
case/orP=>H5. 
- rewrite -!{H5 y}(eqP H5) E in H1 H3 *.
  by apply: (path_last 1); apply: path_filter.
case: ifP=>E2; last by apply: IH=>//; apply: path_sorted H1.
move: H1.
have [t -> H1]: exists t, y = t .+ 1. 
- case: y {H3} E2; case=>[|m] //.
  by exists (ptr_nat m); rewrite /ptr_offset /= addn1.
apply: IH=>//; first by apply: path_sorted H1.
apply: notin_path; apply: ord_path H1.
by case: t=>m; rewrite /ord /= addn1.
Qed.

Lemma lfresh_null h : lfresh h != null.
Proof. by case: h=>[//|[h H] F]; rewrite /lfresh ptrE -lt0n /= addnS. Qed.

Lemma hfresh_null h : hfresh h != null.
Proof. by case: h=>[//|[h H] F]; rewrite /lfresh ptrE -lt0n /= addnS. Qed.

Lemma high_lowD : [predI low & high] =i pred0.
Proof.
case=>x; rewrite inE /low /high /= -!topredE /=.
by case: x=>// n; case E: (0 %% 2 == _)=>//=; rewrite -(eqP E).
Qed.

Lemma modnS x1 x2 : (x1 == x2 %[mod 2]) = (x1.+1 == x2.+1 %[mod 2]).
Proof.
case E: (x1 %% 2 == _).
- by rewrite -addn1 -modnDm (eqP E) modnDm addn1 eq_refl. 
suff L: ((x1.+1) %% 2 == (x2.+1) %% 2) -> (x1 %% 2 == x2 %% 2).
- by rewrite E in L; case: eqP L=>// _; apply.
move=>{E} E; rewrite -(modnDr x1) -(modnDr x2).
by rewrite -addSnnS -modnDm (eqP E) modnDm addSnnS.
Qed.

Lemma hlE x : high x = ~~ low x.
Proof.
case: x=>n; rewrite /high /low /=.
elim: n=>[//|m IH]; apply: negb_inj.
by rewrite negbK -modnS -IH modnS modnn.
Qed.

Lemma lhE x : low x = ~~ high x.
Proof. by apply: negb_inj; rewrite negbK hlE. Qed.
 
Lemma ldomUn h1 h2 : 
        ldom (h1 \+ h2) =i 
        [pred x | valid (h1 \+ h2) && (x \in [predU ldom h1 & ldom h2])].
Proof. 
by move=>x; rewrite !inE !ldomP domUn !inE /= -andbA andb_orl. 
Qed.

Definition loweq h1 h2 := get_lows h1 == get_lows h2.

Notation "h1 =~ h2" := (loweq h1 h2) (at level 80).

Lemma low_refl h : h =~ h.
Proof. by rewrite /loweq. Qed.

Hint Resolve low_refl.

Lemma low_sym h1 h2 : (h1 =~ h2) = (h2 =~ h1).
Proof. by rewrite /loweq eq_sym. Qed.

Lemma low_trans h2 h1 h3 : h1 =~ h2 -> h2 =~ h3 -> h1 =~ h3.
Proof. by rewrite /loweq; move/eqP=>->. Qed.

Lemma loweqP h1 h2 : reflect (ldom h1 =i ldom h2) (h1 =~ h2). 
Proof.
case E: (loweq h1 h2); constructor; rewrite /loweq in E.
- by move=>x; rewrite /ldom (eqP E).
move=>F.
suff {E} : get_lows h1 = get_lows h2 by move/eqP; rewrite E.
apply: (eq_sorted_irr (leT := ord)); last by apply: F.
- by apply: trans.
- by apply: irr.
- case: h1 {F}=>// [[h S] H]. rewrite /get_lows. simpl.
  by rewrite sorted_filter //; apply: trans.
case: h2 {F}=>// [[h S] H].
by rewrite sorted_filter //; apply: trans.
Qed.

Lemma loweqK h1 h2 t1 t2 : 
        valid (h1 \+ t1) -> valid (h2 \+ t2) ->
        h1 =~ h2 -> h1 \+ t1 =~ h2 \+ t2 -> t1 =~ t2.
Proof.
move=>D1 D2.
case: loweqP=>// E1 _; case: loweqP=>// E2 _; apply/loweqP.
by apply: ldomK E2.
Qed.

Lemma loweqE h1 h2 : h1 =~ h2 -> lfresh h1 = lfresh h2.
Proof. by rewrite /loweq /lfresh; move/eqP=>->. Qed.

Lemma lowUn h1 h2 t1 t2 :
        valid (h1 \+ t1) -> 
        valid (h2 \+ t2) -> 
        h1 =~ h2 -> t1 =~ t2 -> h1 \+ t1 =~ h2 \+ t2.
Proof.
move=>D1 D2; do 2![case: loweqP=>//]=>H1 H2 _ _. 
apply/loweqP=>x; move: (H1 x) (H2 x).
by rewrite !ldomP !domUn !inE D1 D2 /= !andb_orl=>-> ->.
Qed.

Lemma lowPn A1 A2 (x : ptr) (v1 : A1) (v2 : A2) : x :-> v1 =~ x :-> v2.
Proof. by apply/loweqP=>y; rewrite !ldomP !hdomPt. Qed.

Hint Resolve lowPn.

Lemma highPn A1 A2 (x1 x2 : ptr) (v1 : A1) (v2 : A2) : 
        high x1 -> high x2 -> x1 :-> v1 =~ x2 :-> v2.
Proof.
move=>H1 H2.
apply/loweqP=>y; rewrite !ldomP !hdomPt !inE. 
case E1: (x1 == y); first by rewrite -(eqP E1) lhE H1 !andbF. 
case E2: (x2 == y); last by rewrite !andbF.
by rewrite -(eqP E2) lhE H2 !andbF.
Qed.

Lemma lowPtUn A1 A2 h1 h2 (x : ptr) (v1 : A1) (v2 : A2) : 
        valid (x :-> v1 \+ h1) ->
        valid (x :-> v2 \+ h2) -> 
        (x :-> v1 \+ h1 =~ x :-> v2 \+ h2) = (h1 =~ h2).
Proof. 
move=>D1 D2. 
case E: (h1 =~ h2); first by apply: lowUn.
move/(elimF idP): E=>E; apply: (introF idP)=>F; case: E.
by apply: loweqK F.
Qed.

Lemma highPtUn A h1 h2 (x : ptr) (v : A) : 
        valid (x :-> v \+ h1) -> high x -> 
        (x :-> v \+ h1 =~ h2) = (h1 =~ h2).
Proof.
move=>D H.
case E: (h1 =~ h2); case: loweqP E=>// L1 _; apply/loweqP.
- move=>y; rewrite !ldomP hdomPtUn !inE D.
  case: eqP=>[<-|]; last by rewrite -!ldomP L1. 
  by rewrite lhE H /= andbF.
move=>L2; case: L1 => y; move: {L2} (L2 y).
rewrite !ldomP !hdomPtUn !inE D /=.
by case: eqP=>//= <-; rewrite lhE H andbF -[x \in dom h1]negbK (hvalidPtD D).
Qed.

Lemma highPtUn2 A1 A2 h1 h2 (x1 x2 : ptr) (v1 : A1) (v2 : A2) : 
        valid (x1 :-> v1 \+ h1) -> 
        valid (x2 :-> v2 \+ h2) -> 
        high x1 -> high x2 ->
        h1 =~ h2 -> x1 :-> v1 \+ h1 =~ x2 :-> v2 \+ h2.
Proof. by move=>D1 D2 H1 H2 L; apply: lowUn=>//; apply: highPn. Qed.  

(**********************************************)
(* several basic operations on pairs of heaps *)
(**********************************************)

Definition plus2 (h1 h2 : heap * heap) : heap * heap := 
  (h1.1 \+ h2.1, h1.2 \+ h2.2).

Definition def2 (h : heap * heap) := valid h.1 && valid h.2.

Notation "h1 :++ h2" := (plus2 h1 h2) (at level 50). 

Lemma unA2 h1 h2 h3 : h1 :++ (h2 :++ h3) = h1 :++ h2 :++ h3.
Proof. by congr (_, _); rewrite /= joinA. Qed.

Lemma unC2 h1 h2 : h1 :++ h2 = h2 :++ h1.
Proof. by congr (_, _); rewrite joinC. Qed.   

Lemma unKhl2 h h1 h2 : def2 (h1 :++ h) -> h1 :++ h = h2 :++ h -> h1 = h2.
Proof.
move: h h1 h2=>[h1 h2][h11 h12][h21 h22]; case/andP=>/= [D1 D2] [E1 E2]. 
by rewrite (joinKf D1 E1) (joinKf D2 E2).
Qed.

Lemma unKhr2 h h1 h2 : def2 (h2 :++ h) -> h1 :++ h = h2 :++ h -> h1 = h2.
Proof.
move: h h1 h2=>[h1 h2][h11 h12][h21 h22]; case/andP=>/= [D1 D2] [E1 E2]. 
rewrite -E1 -E2 in D1 D2.
by rewrite (joinKf D1 E1) (joinKf D2 E2).
Qed.

Lemma unDl2 h1 h2 : def2 (h1 :++ h2) -> def2 h1.
Proof. by case/andP=>/= D1 D2; rewrite /def2 (validL D1) (validL D2). Qed.

Lemma unDr2 h1 h2 : def2 (h1 :++ h2) -> def2 h2.
Proof. by case/andP=>/= D1 D2; rewrite /def2 (validR D1) (validR D2). Qed.

Lemma un0h2 h : (Unit, Unit) :++ h = h.
Proof. by case: h=>h1 h2; rewrite /plus2 /= !unitL. Qed.

Lemma unh02 h : h :++ (Unit, Unit) = h.
Proof. by case: h=>h1 h2; rewrite /plus2 /= !unitR. Qed.

(**************************************************************************)
(* Several tactics for canceling common terms in disjoint unions          *)
(* Currently, they don't deal with weak pointers. I.e.  they only if they *)
(* see iterms like x :-> v1 and x :-> v2, they will reduce to v1 = v2     *)
(* only if v1, v2 are of the same type A more general tactic would emit   *)
(* obligation dyn v1 = dyn v2, but I don't bother with this now.          *)
(**************************************************************************)

(* First cancelation in hypotheses *)

Lemma injUh A h1 h2 x (v1 v2 : A) : 
        valid (h1 \+ (x :-> v1)) -> 
        h1 \+ (x :-> v1) = h2 \+ (x :-> v2) -> 
          valid h1 /\ h1 = h2 /\ v1 = v2. 
Proof. by rewrite -!(joinC (x :-> _))=>D; case/(hcancelV D)=><- -> ->. Qed.  

Lemma eqUh (h1 h2 h : heap) : 
        valid (h1 \+ h) -> h1 \+ h = h2 \+ h -> valid h1 /\ h1 = h2. 
Proof. by move=>D E; rewrite {2}(joinKf D E) (validL D). Qed.

Lemma exit1 (h1 h2 h : heap) : 
        valid (h1 \+ h) -> h1 \+ h = h \+ h2 -> valid h1 /\ h1 = h2.
Proof. by move=>D; rewrite (joinC h); apply: eqUh. Qed.

Lemma exit2 (h1 h : heap) : 
       valid (h1 \+ h) -> h1 \+ h = h -> valid h1 /\ h1 = Unit.
Proof. by move=>H1; rewrite -{2}(unitR h)=>H2; apply: exit1 H2. Qed.

Lemma exit3 (h1 h : heap) : 
        valid h -> h = h \+ h1 -> @valid heapPCM Unit /\ Unit = h1.
Proof.
move=>H1 H2; split=>//; rewrite -{1}(unitR h) in H2.
by apply: joinfK H2; rewrite unitR.
Qed.

Lemma exit4 (h : heap) : 
        valid h -> h = h -> valid (@Unit heapPCM) /\ Unit = (@Unit heapPCM).
Proof. by []. Qed. 

Ltac cancelator t H :=
  match goal with 
  (* we exit when we hit the terminator on the left *)
  | |- ?h1 \+ t = ?h2 -> _ => 
     let j := fresh "j" in
     set j := {1}(h1 \+ t); 
     rewrite -1?joinA /j {j};
     (move/(exit1 H)=>{H} [H] || move/(exit2 H)=>{H} [H]) 
  | |- t = ?h2 -> _ => 
     rewrite -?joinA;
     (move/(exit3 H)=>{H} [H] || move/(exit4 H)=>{H} [H])
  | |- (?h1 \+ (?x :-> ?v) = ?h2) -> _ => 
    let j := fresh "j" in 
    set j := {1}(h1 \+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(joinC (x :-> _)) -?(joinAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    (move/(injUh H)=>{H} [H []] ||
    (* if not, rotate x in the first union *)
     rewrite (joinC h1) ?joinA in H * );
    (* and proceed *)
    cancelator t H
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- (?h1 \+ ?h = ?h2) -> _ => 
    let j := fresh "j" in 
    set j := {1}(h1 \+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(joinC h) -?(joinAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (move/(eqUh H)=>{H} [H []] ||
    (* if not, rotate h in the first union *)
    rewrite (joinC h1) ?joinA in H * );
    (* and proceed *)
    cancelator t H
  | |- _ => idtac 
  end.

Ltac heap_cancel := 
  match goal with 
  | |- ?h1 = ?h2 -> ?GG =>
    let t1 := fresh "t1" in
    let t2 := fresh "t2" in 
    let t := fresh "t" in
    let H := fresh "H" in
    let G := fresh "hidden_goal"
    in
      (* generate the obligation to prove that the left heap is defined *)
      suff : valid h1; first (
       (* make sure no sharing of expressions in the goal *)
       set t1 := {1 2}h1; set t2 := {1}h2; set G := GG;
       (* introduce terminators *)
       rewrite -(unitL t1) -(unitL t2) [empty]lock;
       set t := locked Unit; rewrite /t1 /t2 {t1 t2}; 
       move=>H; 
       (* flatten the goal *)
       rewrite ?joinA in H *; 
       (* call the cancelation routine *)
       cancelator t H; 
       (* remove the terminator and push H onto the goal *)
       move: H {t}; rewrite /G {G})
  | |- _ => idtac
  end.

(* Then cancelation in conclusions *)

Lemma cexit1 (h1 h2 h : heap) : h1 = h2 -> h1 \+ h = h \+ h2.
Proof. by move=>->; rewrite joinC. Qed.

Lemma cexit2 (h1 h :  heap) : h1 = Unit -> h1 \+ h = h.
Proof. by move=>->; rewrite unitL. Qed.

Lemma cexit3 (h1 h : heap) : Unit = h1 -> h = h \+ h1.
Proof. by move=><-; rewrite unitR. Qed.

Lemma congUh A h1 h2 x (v1 v2 : A) : 
        h1 = h2 -> v1 = v2 -> h1 \+ (x :-> v1) = h2 \+ (x :-> v2).
Proof. by move=>-> ->. Qed.

Lemma congeqUh (h1 h2 h : heap) : h1 = h2 -> h1 \+ h = h2 \+ h.
Proof. by move=>->. Qed. 

Ltac congruencer t :=
  match goal with 
  | |- ?h1 \+ t = ?h2 => 
     let j := fresh "j" in
     set j := {1}(h1 \+ t); 
     rewrite -1?joinA /j {j};
     (apply: cexit1 || apply: cexit2)
  | |- t = ?h2 => 
     rewrite -1?joinA;
     (apply: cexit3 || apply: refl_equal)
  | |- (?h1 \+ (?x :-> ?v) = ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 \+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(joinC (x :-> _)) -?(joinAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    ((apply: congUh; [congruencer t | idtac]) || 
    (* if not, rotate x in the first union *)
     (rewrite (joinC h1) ?joinA; congruencer t))
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- (?h1 \+ ?h = ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 \+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(joinC h) -?(joinAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (apply: congeqUh || 
    (* if not, rotate h in the first union *)
    rewrite (joinC h1) ?joinA);
    (* and proceed *)
    congruencer t
  | |- _ => idtac 
  end.

Ltac heap_congr := 
  match goal with 
  | |- ?h1 = ?h2 =>
    let t1 := fresh "t1" in
    let t2 := fresh "t2" in 
    let t := fresh "t" in
      set t1 := {1}h1; set t2 := {1}h2; 
      (* introduce terminators *)
      rewrite -(unitL t1) -(unitL t2) [empty]lock;
      set t := locked Unit; rewrite /t1 /t2 {t1 t2}; 
      (* flatten the goal *)
      rewrite ?joinA; 
      (* call the congruence routine and remove the terminator *)
      congruencer t=>{t}
  | |- _ => idtac
  end.

(*
Lemma test h1 h2 h3 x (v1 v2 : nat) : 
        h3 = h2 -> v1 = v2 -> 
        h1 \+ (x :-> v1) \+ h3= h2 \+ h1 \+ (x :-> v2).
Proof. by move=>H1 H2; heap_congr. Qed.
*)

(* and a tactic for computing the subdom relation *)

Definition supdom h2 h1 := subdom h1 h2.

Lemma sexit1 h1 h2 h : 
        valid (h2 \+ h) -> 
          (valid h2 -> supdom h2 h1) -> supdom (h2 \+ h) (h \+ h1).
Proof.
move=>H1 H2; rewrite (joinC h); apply: subdomE=>//.
by apply: H2; apply: validL H1.
Qed.

Lemma sexit2 h1 h : 
        valid (h1 \+ h) -> (valid h1 -> supdom h1 Unit) -> 
        supdom (h1 \+ h) h.
Proof.
move=>H1 H2; rewrite -{2}(unitL h); apply: subdomE=>//.
by apply: H2; apply: validL H1.
Qed.

Lemma sexit3 h1 h : 
        valid h -> (@valid heapPCM Unit -> supdom Unit h1) -> 
        supdom h (h \+ h1).
Proof.
move=>H1 H2; rewrite joinC -{1}(unitL h).
by apply: subdomE; [rewrite unitL | apply: H2].
Qed.

Lemma sexit4 h : 
       valid h -> (@valid heapPCM Unit -> @Unit heapPCM = Unit) -> supdom h h.
Proof. by move=>*; rewrite -(unitL h); apply: subdomE=>//; rewrite unitL. Qed.

Lemma supdomUh A B h1 h2 x (v1 : A) (v2 : B) : 
        valid (h2 \+ (x :-> v2)) -> 
        (valid h2 -> supdom h2 h1) -> 
        supdom (h2 \+ (x :-> v2)) (h1 \+ (x :-> v1)).
Proof.
move=>H1 H2.
apply: subdomUE=>//; first by apply: H2; apply: validL H1.
apply: subdomPE; rewrite joinC in H1.
by apply: (hvalidPt_cond H1).
Qed.

Lemma supdomeqUh h1 h2 h : 
        valid (h2 \+ h) -> (valid h2 -> supdom h2 h1) -> supdom (h2 \+ h) (h1 \+ h).
Proof. by rewrite (joinC h1); apply: sexit1. Qed.    

Lemma sup_defdef h1 h2 : valid h2 -> supdom h2 h1 -> valid h1.
Proof. by move=>H1; rewrite /supdom; move/subdom_def; rewrite H1 andbT. Qed.

Ltac supdom_checker t H :=
  match goal with 
  | |- is_true (supdom (?h1 \+ t) ?h2) => 
     let j := fresh "j" in
     set j := {1}(h1 \+ t); 
     rewrite -1?joinA /j {j};
     (apply: (sexit1 H)=>{H} H || apply: (sexit2 H)=>{H} H)
  | |- is_true (supdom t ?h1) => 
     rewrite -1?joinA;
     (apply: (sexit3 H)=>{H} H || apply: (sexit4 H)=>{H} H)
  | |- is_true (supdom (?h1 \+ (?x :-> ?v)) ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 \+ (x :-> v));
    (* if x appears in the second union, first bring it to the back *)
    rewrite 1?(joinC (x :-> _)) -?(joinAC _ _ (x :-> _)) /j {j}; 
    (* then one of the following must apply *)
    (* if x is in the second union then cancel *)
    (apply: (supdomUh _ H)=>{H} H || 
    (* if not, rotate x in the first union *)
     (rewrite (joinC h1) ?joinA in H * )); supdom_checker t H
  (* if the heap is not a points-to relation, also try to cancel *)
  | |- is_true (supdom (?h1 \+ ?h) ?h2) => 
    let j := fresh "j" in 
    set j := {1}(h1 \+ h);
    (* if h appears in the second union, first bring it to the back *)
    rewrite 1?(joinC h) -?(joinAC _ _ h) /j {j}; 
    (* then one of the following must apply *)
    (* if h is in the second union then cancel *)
    (apply: (supdomeqUh H)=>{H} H || 
    (* if not, rotate h in the first union *)
    (rewrite (joinC h1) ?joinA in H * )); 
    (* and proceed *)
    supdom_checker t H
  | |- _ => idtac 
  end.

Ltac defcheck := 
  match goal with 
  | |- is_true (valid ?h2) -> is_true (valid ?h1) =>
    let t1 := fresh "t1" in
    let t2 := fresh "t2" in 
    let t := fresh "t" in
    let H := fresh "H" in
      set t2 := {1}h2; set t1 := {1}h1; 
      (* introduce terminators *)
      rewrite -(unitL t1) -(unitL t2) [empty]lock;
      set t := locked Unit; rewrite /t1 /t2 {t1 t2}; 
      (* flatten the goal *)
      rewrite ?joinA; 
      move=>H; 
      apply: (sup_defdef H); 
      (* call the subdom_cheker routine and remove the terminator *)
      supdom_checker t H; move: H {t}; rewrite /supdom
  | |- _ => idtac
  end.

(* this diverges in coq 8.3 
Lemma test2 h1 h2 x (v1 v2 : nat) : subdom h1 h2 ->
        def (h2 :+ (x :-> v2)) -> def (h1 :+ (x :-> v1)).
Proof. by move=>H; defcheck. Qed.
*)

Ltac hhauto := (do ?econstructor=>//; try by [heap_congr])=>//.
