Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude domains idynamic pcm unionmap heap codes feaps state.
Require Import zigzag worlds rely inject actions schedule retract. 
Require Import process always. 
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* generic semantics of programs as sets of trees *)
(* we can have specifications be predicates over states *)

(* We require that each program's denotation contains at least one
element -- the unfinished tree. This way, an infinite loop is not modeled *)
(* by an empty set, but by a set containing approximating finite trees *)

Module Prog.
Structure prog W A := Prog {
  set_of : tree W A -> Prop; 
  _ : set_of Unfinished}.
End Prog.

Notation prog := Prog.prog.
Notation Prog := Prog.Prog.
Implicit Arguments Prog [W A].

Coercion Prog.set_of : prog >-> Funclass. 
(* admit \In notation *)
Canonical prog_PredType W A := mkPredType (@Prog.set_of W A). 

Lemma progE W A (T1 T2 : prog W A) : 
        T1 = T2 <-> Prog.set_of T1 = Prog.set_of T2.
Proof.
split=>[->//|]; case: T1 T2=>m1 H1 [m2 H2] /= E.
by move: H2; rewrite -E => H2; rewrite (proof_irrelevance H1 H2).
Qed.

(******************)
(* specifications *)
(******************)

Definition pre := state -> Prop.
Definition post A := state -> A -> state -> Prop.
Definition cont A := A -> state -> Prop.
Definition spec A := prod pre (post A).

Definition has_spec W A (s : spec A) := 
  [Pred T : prog W A | forall i t, 
     s.1 i -> i \In Mod.coh W -> t \In T -> after i t (s.2 i)]. 

Module STsep.

Structure STsep W A (s : spec A) := STsep_make {
  prog_of : prog W A; 
  _ : prog_of \In has_spec W s}.

End STsep.

Notation STsep := STsep.STsep.
Notation STsep_make := STsep.STsep_make.

Coercion STsep.prog_of : STsep >-> prog.

Lemma stsepE W A (s : spec A) (e1 e2 : STsep W s) : 
        e1 = e2 <-> e1 = e2 :> prog W A.
Proof.
split=>[->//|]; case: e1 e2=>T1 H1 [T2 H2] /= E.
by rewrite -{T2}E in H2 *; rewrite (proof_irrelevance H1 H2).
Qed.

Structure ST W A := with_spec (s : spec A) of STsep W s.

Definition spec_of W A (e : ST W A) := let: with_spec s _ := e in s.
Definition pre_of W A := [fun e : ST W A => (spec_of e).1]. 
Definition post_of W A := [fun e : ST W A => (spec_of e).2]. 
Definition code_of W A (e : ST W A) := 
  let: with_spec _ c := e return STsep W (spec_of e) in c.

Implicit Arguments pre_of [W A].
Implicit Arguments post_of [W A].
Implicit Arguments with_spec [W A].
Prenex Implicits pre_of post_of.

Coercion with_spec : STsep >-> ST.

Definition verify W A (i : state) (e : ST W A) r := 
  i \In Mod.coh W -> 
  forall t, t \In STsep.prog_of (code_of e) -> after i t r. 

(*
Do I want a new type ST, which existentially hides the world, and then
use values of that type for programming? For a command to be safe in
some world W, we issue a proof obligation that W is frame-larger than
the world packaged with the command.

It doesn't seem that will quite work, for a few reasons.

First, which world do we compute as a default world for bind and par?
We need some lattice theory of worlds, so that we can take a join of
two worlds. Plus, the second world should be independent of the result
of the first command. We'd also need a bottom world of the lattice
(that's the world for ret), but I don't see one such world existing
now.

OTOH: if the worlds are always something achored by Priv, and then
SEntangled with a set of modules, it seems like the join and bottom
are easy to compute, at least classically!!! 
The bottom is Priv, the join is the union of the sets of modules.
Now, the union may be difficult to compute in Coq, as we don't have
computable equality on modules right now, so we can't necessarily
weed out duplicates.

In the case of SCSL however, we could do that, as each lock module is
uniquely identified by lock name (but this doesn't have to be the case
in general).

*)

(*********************)
(* lattice structure *)
(*********************)

Module STLattice.
Section STLattice.
Variables (W : mod) (A : Type) (s : spec A).
Notation prog A := (prog W A).
Notation STsep s := (@STsep W A s).
 
Definition leq (e1 e2 : STsep s) := 
  Prog.set_of (STsep.prog_of e1) <== Prog.set_of (STsep.prog_of e2).

Lemma leq_refl e : leq e e. 
Proof. by []. Qed.

Lemma leq_trans e1 e2 e3 : leq e1 e2 -> leq e2 e3 -> leq e1 e3.
Proof. by move=>H1 H2 p x; apply: H2; apply: H1. Qed.

Lemma leq_asym e1 e2 : leq e1 e2 -> leq e2 e1 -> e1 = e2.
Proof.
move: e1 e2=>[m1 N1][m2 N2]; rewrite /leq /= => H1 H2.
by apply/stsepE/progE; apply: poset_asym H1 H2.
Qed.

Definition bot_set t := t = @Unfinished W A. 
Definition bot_prg := Prog bot_set (erefl _).
Lemma bot_spec : bot_prg \In has_spec W s.
Proof. by move=>i H C t ->; apply: alw_unfin. Qed.
Definition bot := STsep_make bot_spec.

Lemma bot_bot e : leq bot e.
Proof. by case: e=>[[p U] H] t ->. Qed.

Definition sup_set (es : Pred (STsep s)) t := 
  t = Unfinished \/ exists e : STsep s, t \In STsep.prog_of e /\ e \In es.
Definition sup_prog es := Prog (sup_set es) (or_introl (erefl _)).

Lemma sup_spec es : sup_prog es \In has_spec W s.
Proof.
move=>i H C t [->|[e [H1 H2]]]; first by apply: alw_unfin. 
by case: e H1 H2=>p H3 /= H4 _; apply: H3 H C _ H4. 
Qed.

Definition sup es := STsep_make (@sup_spec es).

(* sup is an upper bound *)
Lemma sup_supB es e : e \In es -> leq e (sup es).
Proof. by move=>H1 x H2; right; exists e. Qed.

(* sup is the least upper bound *)
Lemma sup_supL es e : (forall c, c \In es -> leq c e) -> leq (sup es) e.
Proof. by case: e; case=>m N H /= H1 t [->//|][T][H2]; move/H1; apply. Qed.

End STLattice.

Module Exports.
Section Exports.
Variables (W : mod) (A : Type) (s : spec A).

Definition stPosetMixin := PosetMixin (@bot_bot W A s) (@leq_refl W A s) 
                                      (@leq_asym W A s) (@leq_trans W A s).
Canonical stPoset := Eval hnf in Poset (@STsep W A s) stPosetMixin.

Definition stLatticeMixin := LatticeMixin (@sup_supB W A s) (@sup_supL W A s).
Canonical stLattice := Eval hnf in Lattice (@STsep W A s) stLatticeMixin.

End Exports.
End Exports.

End STLattice.

Export STLattice.Exports.

(*************************)
(* Constuctors for STsep *)
(*************************)

Section Fix. 
Variables (W : mod) (A : Type) (B : A -> Type) (s : forall x, spec (B x)).
Notation tp := (forall x, STsep W (s x)).
Notation lat := (dfunLattice (fun x => [lattice of STsep W (s x)])).
Variable (f : tp -> tp).

(* we take a fixpoint not of f, but of its monotone completion f' *)
(* should eventually prove that f' is monotone *)

Definition f' (e : lat) := sup [Pred t : lat | exists e', e' <== e /\ t = f e'].

Definition ffix : tp := tarski_lfp f'. 

End Fix.


Section Return.
Variables (W : mod) (A : Type) (x : A).

Definition ret_set t := t = Unfinished \/ t = @Ret W A x.
Definition ret_prog := Prog ret_set (or_introl (erefl _)). 

(* the environment may interfere with ret to change the heap *)
(* in ways described by W; hence, env_steps W i m *)

Definition ret_s : spec A := 
  (fun i => True, fun i y m => menv_steps W i m /\ y = x). 

Lemma ret_has_spec : ret_prog \In has_spec W ret_s.
Proof. 
move=>i _ C t [->|->]; first by apply: alw_unfin.
by apply: alw_ret=>//= m E y [<-].
Qed.

Definition ret := with_spec _ (STsep_make ret_has_spec). 

End Return.

Implicit Arguments ret [W A].

Section Bind.
Variables (W : mod) (A B : Type).

Section Prog.
Variables (T : prog W A) (K : A -> prog W B).
Definition bnd_set t := 
  t = Unfinished \/ exists t', t \In tcat t' K /\ t' \In T.
Definition bnd_prog := Prog bnd_set (or_introl (erefl _)).
End Prog.

Variables (e1 : ST W A) (e2 : A -> ST W B).
Notation s1 := (spec_of e1).
Notation s2 := (fun x => spec_of (e2 x)).

Definition bind_s : spec B := 
  (fun i => s1.1 i /\ forall x s, s1.2 i x s -> (s2 x).1 s,
   fun i y m => exists x s, s1.2 i x s /\ (s2 x).2 s y m).

Lemma bind_has_spec : 
        bnd_prog (code_of e1) (fun x => code_of (e2 x)) \In has_spec W bind_s.
Proof.
move=>i t [S1 S2] Ci [->|[t1 [Pc P1]]] /=; first by apply: alw_unfin.  
apply: aft_bnd Pc _; case: (code_of e1) P1=>T1 H /= P1.
apply: {H P1 Ci S1 T1 t t1} aft_imp (H _ _ S1 Ci P1).
move=>x s Cs S3 t; case: (code_of (e2 x))=>T /= H P. 
by apply: aft_imp (H _ _ (S2 _ _ S3) Cs P)=>m y Cm; exists x, s.
Qed.

Definition bind := with_spec _ (STsep_make bind_has_spec).

End Bind.


Section Par.
Variables (W : mod) (A1 A2 : Type).

Section Prog.
Variables (T1 : prog W A1) (T2 : prog W A2).
Definition par_set t :=  
  t = Unfinished \/ exists t1 t2, 
  [/\ t = Par t1 t2 Ret, t1 \In T1 & t2 \In T2]. 
Definition par_prog  := Prog par_set (or_introl (erefl _)).
End Prog.

Variables (e1 : ST W A1) (e2 : ST W A2).
Notation s1 := (spec_of e1).
Notation s2 := (spec_of e2).

Definition par_s : spec (A1 * A2) := 
  (fun i => exists lp1 lp2 sh ep, 
     [/\ i = [lp1 \* lp2, sh, ep], 
         s1.1 [lp1, sh, ep \* lp2] & s2.1 [lp2, sh, ep \* lp1]],
   fun i y m => forall lp1 lp2 sh ep, 
     i = [lp1 \* lp2, sh, ep] -> 
     s1.1 [lp1, sh, ep \* lp2] -> s2.1 [lp2, sh, ep \* lp1] ->
     exists lp1' lp2' sh' ep', 
       [/\ m = [lp1' \* lp2', sh', ep'], 
           s1.2 [lp1, sh, ep \* lp2] y.1 [lp1', sh', ep' \* lp2'] & 
           s2.2 [lp2, sh, ep \* lp1] y.2 [lp2', sh', ep' \* lp1']]).

Lemma par_has_spec : par_prog (code_of e1) (code_of e2) \In has_spec W par_s.
Proof.
move=>i t [lp1][lp2][sh][ep][-> S1 S2] C [->|[t1][t2][->] P1 P2]. 
- by apply: alw_unfin.
case: (code_of e1) (code_of e2) P1 P2=>T1 H1 [T2 H2] /= P1 P2.
have {S1 S2} Ls: alwsafe [lp1 \* lp2, sh, ep] (Par t1 t2 Ret).
  have C1 : [lp1, sh, ep] \> lp2 \In Mod.coh W. 
  - by rewrite -coh_zigzag zigE zipC. 
  have C2 : [lp2, sh, ep] \> lp1 \In Mod.coh W.
  - by rewrite -coh_zigzag zigE. 
  move: (H1 _ _ S1 C1 P1) (H2 _ _ S2 C2 P2)=>Ls1 Ls2.
  by apply/aft_alwsf; apply: aft_imp (aft_par Ls1 Ls2). 
rewrite aftA // => lp1'; rewrite aftA // => lp2'.
rewrite aftA // => sh'; rewrite aftA // => ep'.
rewrite !aftA // => E S1 S2; rewrite {Ls}E in C *.
have C1 : [lp1', sh', ep'] \> lp2' \In Mod.coh W.
- by rewrite -coh_zigzag zigE zipC.
have C2 : [lp2', sh', ep'] \> lp1' \In Mod.coh W.
- by rewrite -coh_zigzag zigE.
move: (H1 _ _ S1 C1 P1) (H2 _ _ S2 C2 P2)=>{H1 P1 S1 C1 H2 P2 S2 C2 C} Ls1 Ls2.
apply: {lp1 lp2 sh ep Ls1 Ls2} aft_imp (aft_par Ls1 Ls2).
by move=>s y _ [lp1][lp2][sh][ep][->] P1 P2; exists lp1, lp2, sh, ep. 
Qed.

Definition par := with_spec _ (STsep_make par_has_spec).

End Par.


(* refinement/hiding *)
Section Hide. 
Variables (W V : mod) (U A : Type) (f : refinement U W V) (gL : U).

Section Prog.
Variable T : prog V A.

Definition hide_set t := 
  t = Unfinished \/ exists t', t' \In T /\ t = Hide (Retracts gL f) t'.
Definition hide_prog := Prog hide_set (or_introl (erefl _)).
End Prog.

Variable (e : ST V A).
Notation s := (spec_of e). 

Definition hide_s : spec A := 
  (fun i => exists i', f gL i i' /\ s.1 i', 
   fun i y m => forall i', f gL i i' -> s.1 i' -> 
      exists m' gL', f gL' m m' /\ s.2 i' y m').

Lemma hide_has_spec : hide_prog (code_of e) \In has_spec W hide_s.
Proof.
move=>i _  [i'][Rfi H1] Ci [->|[t][Ht] ->]; first by apply: alw_unfin. 
have : after i' t (s.2 i').
- by case: (code_of e) Ht=>p; apply=>//; case: (rf_coh Rfi).
move/(aft_hide Rfi); apply: aft_imp=>v m _ [m'][gL'][Rm] H2. 
by move=>_ /(rf_func Rfi) <- _; exists m', gL'. 
Qed.

Definition hide := with_spec _ (STsep_make hide_has_spec).

End Hide.

(* refinement/hiding in the common case when the decorated world *)
(*  is an extension of priv *)
Section PrivHide. 
Variables (hp : nat) (V W : mod) (A T : Type).
Variables (d : decorate W T).

Definition priv_hide gL := hide (A:=A) (PrivEntangleRefine hp V d) gL.  

End PrivHide.


Section Act.
Variables (W : mod) (A : Type).
Notation guar := (Mod.internal W).
Variable a : int_action guar A.

Definition act_set t := t = Unfinished \/ t = @Act W A a.
Definition act_prog := Prog act_set (or_introl (erefl _)).

(* action is safe to execute in all initial states *)
(* for which a number of env_steps leaves us asafe *)
(* after we're done we know that we stepped from i to j by envsteps *)
(* then atomically from j to k by a, and then from k to m by envsteps again *)

Definition act_s : spec A := 
  (fun i => forall j, menv_steps W i j -> j \In asafe a,
   fun i y m => exists j k, 
     [/\ menv_steps W i j, astep j a k y & menv_steps W k m]). 

Lemma act_has_spec : act_prog \In has_spec W act_s.
Proof. 
move=>i t S C [] -> /=; first by apply: alw_unfin.
apply: alw_act C _.
move=>j E1; split=>//; first by apply: S E1.
by move=>k y m St E2 v [<-]; exists j, k. 
Qed.

Definition act := with_spec _ (STsep_make act_has_spec).

End Act.

(*************************************************)
(* Coercions: consequence and framing wrt worlds *)
(*************************************************)

Definition conseq W A (e : ST W A) p q := 
  forall i, p i -> verify i e (q i). 

Local Notation conseq1 e s := 
     (conseq e (let 'pair x _ := s in x) 
               (let 'pair _ x := s in x)).

Lemma conseq_refl W A (e : ST W A) : conseq1 e (spec_of e).
Proof. by case: e; case=>p q [T H i] /= Hp C t; apply: H. Qed.

Hint Resolve conseq_refl.

Section Do.
Variables (W : mod) (A : Type) (s2 : spec A).
Variables (e : ST W A) (pf : conseq1 e s2).

Definition do_prog := STsep.prog_of (code_of e).

Lemma do_has_spec : do_prog \In has_spec W s2. 
Proof. 
rewrite /do_prog; case: e pf=>T Ht /=; case: s2=>p q.
by move=>H1 i t /= Hp C H; apply: H1. 
Qed.

Definition do' := STsep_make do_has_spec.

End Do.

Notation "'Do' e" := (@do' _ _ _ e _) (at level 80).

(* I will need a command from framing wrt. worlds, in order to include
programs developed in smaller worlds, into larger ones. There seems no
way to avoid this coercion. Moreover, the coercion should explicitly
present the larger world as an entanglement of the smaller with some
set of other worlds. That leads to the best verify lemmas, and allows
one to focus on a smaller world, when verifying the programs in the
smaller world. *)

Section Extend.
Variables (V W : mod) (A : Type) (w : injects V W). 
Variable (e : ST V A).
Notation W2 := (inj_ext w). 
Notation s := (spec_of e).

Section Prog.
Variable T : prog V A.

Definition extend_set t := 
  t = Unfinished \/ exists t', t' \In T /\ t = Extend w t'. 
Definition extend_prog := Prog extend_set (or_introl (erefl _)).
End Prog.

Definition extend_s : spec A :=
  (fun i => exists i1 i2, i = i1 \+ i2 /\ i1 \In Mod.coh V /\ s.1 i1, 
   fun i y m => forall i1 i2, i = i1 \+ i2 -> i1 \In Mod.coh V -> 
     exists m1 m2, [/\ m = m1 \+ m2, s.2 i1 y m1 & menv_steps W2 i2 m2]).

Lemma extend_has_spec : extend_prog (code_of e) \In has_spec W extend_s.
Proof.
move=>i /= t [i1][j1][->{i}][Ci1 H1] C.
case=>[|[t' [H2 ->{t}]]]; first by move=>->; apply: alw_unfin.
have : after i1 t' (s.2 i1) by case: (code_of e) H2=>p /= /(_ _ _ H1); apply. 
move/(aft_extend w C); apply: aft_imp=>{H1 t' H2} v m _.
case=>i2 [j2][->{m} Ci2 S2 H2] i3 j3 E Ci3.
suff [E1 E2]: i3 = i1 /\ j3 = j1.
- by rewrite {i3 E Ci3}E1 {j3}E2; exists i2, j2. 
move: (coh_prec (cohV C) E Ci1 Ci3) (E)=>{Ci3 E} <-.
by move/(joinsK (cohV C)).
Qed.

Definition extend := with_spec _ (STsep_make extend_has_spec).

End Extend.


(**********************************)
(* Conditionals for various types *)
(**********************************)

(* Do I need these? Now that things are pakcaged into ST, 
  I can just use ordinary coq conditionals *)

(*
Section CondBool.
Variables (W : MOD) (A : Type) (b : bool).

Program Definition If (e1 : ST W A) (e2 : ST W A) :=
  with_spec (if b then spec_of e1 else spec_of e2) 
            (if b then Do e1 else Do e2).

End CondBool.

Section CondDecide.
Variable (W : MOD) (A : Type) (p1 p2 : Prop) (b : {p1} + {p2}).

Program Definition Match_dec (e1 : p1 -> ST W A) (e2 : p2 -> ST W A) := 
  with_spec (match b with left x => spec_of (e1 x) 
                        | right x => spec_of (e2 x) end) 
            (match b with left x => Do (e1 x) 
                        | right x => Do (e2 x) end).
End CondDecide. 

Section CondList.
Variables (W : MOD) (A B : Type) (xs : seq A).

Program Definition Match_list xs (e1 : ST W B)  (e2 : A -> seq A -> ST W B) := 
  with_spec (match xs with cons x' xs' => spec_of (e2 x' xs') | nil => spec_of e1 end)
            (match xs with cons x' xs' => Do (e2 x' xs') | nil => Do e1 end).

End CondList.

(* add more conditionals here *)
*)

(* some notation *)

(* Notation "'Do' e" := (do' e _) (at level 80). *)
Notation "x '<--' c1 ';' c2" := (bind c1 (fun x => c2)) 
  (at level 81, right associativity).
Notation "c1 ';;' c2" := (bind c1 (fun _ => c2)) 
  (at level 81, right associativity).
(*
Notation "'!' x" := (read _ x) (at level 50) : proc_scope.
Notation "e1 '::=' e2" := (write e1 e2) (at level 60) : proc_scope.
*)
