Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude idynamic pcm unionmap heap feaps state zigzag.
Require Import worlds rely inject actions schedule retract.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(*************)
(* Processes *)
(*************)

Section TreeDef.
Notation coh W := (Mod.coh W).
Notation guar W := (Mod.internal W).

Structure tree (W : mod) A := 
    Unfinished | Ret of A | Act of int_action (guar W) A
  | Seq B of tree W B & B -> tree W A 
  | Par B1 B2 of tree W B1 & tree W B2 & B1 * B2 -> tree W A
  | Hide V of retracts W V & tree V A 
  | Extend V of injects V W & tree V A.

End TreeDef.

Implicit Arguments Unfinished [W A].
Implicit Arguments Ret [W A].
Implicit Arguments Act [W A].
Implicit Arguments Seq [W A B].
Implicit Arguments Par [W A B1 B2].
Implicit Arguments Hide [W A V].
Implicit Arguments Extend [W A V].

Prenex Implicits Unfinished Ret Act Seq Par Hide Extend.

Section Process.
Notation coh W := (Mod.coh W).
Notation guar W := (Mod.internal W).

(* Getting equality of components *)

Lemma seqE W A B1 B2 (t1 : tree W B1) (t2 : tree W B2) k1 k2 : 
        Seq t1 k1 = Seq t2 k2 <-> 
        [/\ B1 = B2, ijmeq t1 t2 &
            ijmeq (sort := fun B => B -> tree W A) k1 k2].
Proof.
split; case=>E; rewrite -{B2}E in t2 k2 *.
- by do 2![move/inj_pair2=><-].
by do 2![move/ijmE=><-].
Qed.

Lemma seqEX W A B (t1 t2 : tree W B) k1 k2 : 
        Seq t1 k1 = Seq t2 k2 :> tree W A <-> t1 = t2 /\ k1 = k2.
Proof. by rewrite seqE; split=>[[_]|[->->] //]; do 2![move/ijmE=>->]. Qed.

Lemma parE W A1 A2 B1 B2 C t1 t2 q1 q2 
             (r1 : A1 * B1 -> tree W C) (r2 : A2 * B2 -> tree W C) :
        Par t1 q1 r1 = Par t2 q2 r2 <-> 
        [/\ A1 = A2, B1 = B2, ijmeq t1 t2, ijmeq q1 q2 & 
            ijmeq (sort := fun B => B -> tree W C) r1 r2].
Proof. 
split; case=>E1 E2; rewrite -{A2}E1 -{B2}E2 in t2 q2 r1 r2 *.
- by do 2![move/inj_pair2=>->]; do 2![move/inj_pair2]; move=>->.
by do 3![move/jmE=>->].
Qed.

Lemma parEX W A1 A2 C (t1 t2 : tree W A1) (q1 q2 : tree W A2) r1 r2 :
        Par t1 q1 r1 = Par t2 q2 r2 :> tree W C <-> 
          [/\ t1 = t2, q1 = q2 & r1 = r2].
Proof. 
split; case; last by move=>->->->.
by do 2![move/inj_pair2=>->]; do 2![move/inj_pair2]=>->.
Qed.

Lemma hideE W V1 V2 A (f1 : retracts W V1) (f2 : retracts W V2)
            (t1 : tree V1 A) (t2 : tree V2 A) : 
        Hide f1 t1 = Hide f2 t2 :> tree W A <-> 
        [/\ V1 = V2, ijmeq f1 f2 &
            ijmeq (sort := tree^~A) t1 t2].
Proof. 
split; case=>E; rewrite -{V2}E in f2 t2 *.
- by do 2![move/inj_pair2=>->].
by do 2![move/ijmE=>->].
Qed.

Lemma hideEX W V A f1 f2 (t1 t2 : tree V A) : 
        Hide f1 t1 = Hide f2 t2 :> tree W A <-> f1 = f2 /\ t1 = t2.
Proof. 
split; case; last by move=>->->.
by do 2![move/inj_pair2=>->].
Qed.

(* catenating a tree with a set *)

Definition tcat W A B (t : tree W A) (k : A -> Pred (tree W B)) :=
  [Pred s | exists q, s = Seq t q /\ forall x, q x \In k x].

Fixpoint good W A (t : tree W A) sc :=
  match sc, t with 
    ChoiceAct, Act a => True
  | SeqRet, Seq _ (Ret _) _ => True
  | SeqStep sc', Seq _ t' _ => good t' sc'
  | ParRet, Par _ _ (Ret _) (Ret _) k => True
  | ParL sc', Par _ _ t' _ _ => good t' sc' 
  | ParR sc', Par _ _ _ t' _ => good t' sc'
  | HideStep sc', Hide _ (Retracts _ gL f) t' => 
      good t' sc'
  | HideRet, Hide _ (Retracts _ gL f) (Ret _) => True
  | ExtendStep sc', Extend V em t' => good t' sc'
  | ExtendRet, Extend V em (Ret _) => True
  | _, _ => False
  end.

Fixpoint safe' W A (t : tree W A) sc s :=
  match sc, t with 
    ChoiceAct, Act a => s \In asafe a 
  | SeqRet, Seq _ (Ret _) _ => True
  | SeqStep sc', Seq _ t' _ => safe' t' sc' s
  | ParRet, Par _ _ (Ret _) (Ret _) k => True
  | ParL sc', Par _ _ t' _ _ => safe' t' sc' s
  | ParR sc', Par _ _ _ t' _ => safe' t' sc' s
  | HideStep sc', Hide _ (Retracts _ gL f) t' => 
      exists s', f gL s s' /\ safe' t' sc' s'
  | HideRet, Hide _ (Retracts _ gL f) (Ret _) => 
      exists s', f gL s s'
  | ExtendStep sc', Extend V em t' =>
      exists s', extends em s s' /\ safe' t' sc' s'
  | ExtendRet, Extend V em (Ret _) =>
      exists s', extends em s s'
  | _, _ => True
  end.

Definition tsafe W A (t : tree W A) sc := 
  [Pred s | s \In coh W /\ safe' t sc s].

Lemma tsf_coh W A s (t : tree W A) sc : s \In tsafe t sc -> s \In coh W. 
Proof. by case. Qed.

Lemma tsf_zagzig W A s p (t : tree W A) sc : 
        s \> p \In tsafe t sc -> s \< p \In tsafe t sc. 
Proof. 
case; move/coh_zigzag=>C H; split=>//.
elim: sc W s p C A t H=>[||sc IH||sc IH|sc IH|sc IH||sc IH|] W s p C A;
case=>//; try by move=>*; apply: IH. 
- by move=>a; apply: asf_zagzig.
- move=>V [sT gL f] t /=; case: (rf_zagzig f p)=>q [Rz _].
  case=>_ [/Rz {Rz}][s'][-> R S].
  exists (s' \< q); split=>//. 
  by apply: IH S; case: (rf_coh R).
- move=>V [sT gL f]; case=>//= v [s'].
  case: (rf_zagzig f p)=>q [Rz _]. 
  by case/Rz=>s2 [_]; exists (s2 \< q). 
- move=>V em t /= [s'][] /ext_zagzig [s1][s2][p1][p2][->->->] H.
  by move/(IH _ _ _ (proj2 (ext_coh H)))=>S; exists (s1 \< p1). 
move=>V em; case=>//= v [s'] /ext_zagzig [s1][s2][p1][p2][->-> _]. 
by exists (s1 \< p1). 
Qed.

(*
Lemma tsf_local A s1 s (p : tree A) sc : 
        s1 \+ s \In coherent -> s1 \In tsafe p sc -> s1 \+ s \In tsafe p sc.
Proof. 
move=>C [C1 H]; split=>//.
elim: sc A p H=>[||sc IH||sc IH|sc IH] A; case=>//; try by move=>*; apply: IH.
by move=>a; apply: asf_local.
Qed.
*)

(* Stepping through a tree via a schedule. Beta redexes return the same *)
(* initial state and reduced tree. None is returned by trees that get   *)
(* stuck because of a safe, but wrong schedule. This boils down to      *)
(* actions that are safe but get stuck because of a wrong external      *)
(* choice. *)

(* In the case of HideStep, we require that gE remains unchanged or else *)
(* the step can't be made. No need to put this as a condition for safety *)
(* because it will always be satisfied; it follows from the rf_env *)
(* requirement on refinements *)

Fixpoint step' W A s1 (t : tree W A) sc s2 (q : tree W A) :=
  match (sc, t) with
    (ChoiceAct, Act a) => exists v, astep s1 a s2 v /\ q = Ret v
  | (SeqRet, Seq _ (Ret v) k) => s2 = s1 /\ q = k v
  | (SeqStep sc', Seq _ t1 k1) => 
      exists t2, step' s1 t1 sc' s2 t2 /\ q = Seq t2 k1
  | (ParRet, Par _ _ (Ret v1) (Ret v2) k) =>
      s2 = s1 /\ q = k (v1, v2)
  | (ParL sc', Par _ _ t1 t2 k) => 
      exists r1, step' s1 t1 sc' s2 r1 /\ q = Par r1 t2 k
  | (ParR sc', Par _ _ t1 t2 k) => 
      exists r2, step' s1 t2 sc' s2 r2 /\ q = Par t1 r2 k
  | (HideStep sc', Hide _ (Retracts _ gL f) t1) =>
      exists s1' s2' gL' t2,
        [/\ q = Hide (Retracts gL' f) t2,
            f gL s1 s1', f gL' s2 s2' &
            step' s1' t1 sc' s2' t2]
  | (HideRet, Hide _ (Retracts _ gL f) (Ret v)) => 
      exists s1', [/\ s2 = s1, q = Ret v & f gL s1 s1']
  (* why is the next a computable operational semantics? *)
  (* because the V-footprint of V uniquely determines s1' *)
  | (ExtendStep sc', Extend V em t1') =>
      exists s1' s2' s t2', 
        [/\ q = Extend em t2', s1 = s1' \+ s, s2 = s2' \+ s, 
            s1' \In Mod.coh V & step' s1' t1' sc' s2' t2']
  | (ExtendRet, Extend V em (Ret v)) =>
       exists s1', [/\ s2 = s1, q = Ret v & extends em s1 s1']
  | _ => False
  end.

(* repackage the definition *)

Definition tstep W A s1 (t : tree W A) sc s2 (q : tree W A) := 
  s1 \In tsafe t sc /\ step' s1 t sc s2 q.

(* views for tstep, organized to match on the tree *)
(* as we typically know the structure of the tree *)

Lemma tstUnfin W A s1 sc s2 (t : tree W A) : 
        tstep s1 Unfinished sc s2 t <-> False.
Proof. by split=>//; case; case: sc. Qed.

Lemma tstRet W A s1 v sc s2 (t : tree W A) : 
        tstep s1 (Ret v) sc s2 t <-> False.
Proof. by split=>//; case; case: sc. Qed.

Lemma tstAct W A s1 a sc s2 (t : tree W A) : 
        tstep s1 (Act a) sc s2 t <->
        exists v, [/\ sc = ChoiceAct, t = Ret v & astep s1 a s2 v].
Proof.
split; first by case; case: sc=>//= c [v [H ->]]; exists v. 
case=>v [->-> H]; split; last by exists v. 
by move/ast_safe: H=>H; split=>//; apply: asf_coh H.
Qed.

Lemma tstSeq W A B s1 (t : tree W B) k sc s2 (q : tree W A) :
        tstep s1 (Seq t k) sc s2 q <->
        (exists v, [/\ sc = SeqRet, t = Ret v, q = k v, s2 = s1 &
                       s1 \In coh W]) \/
         exists sc' p',
           [/\ sc = SeqStep sc', q = Seq p' k & tstep s1 t sc' s2 p'].
Proof.
split; last first.
- case; first by case=>v [->->->->]. 
  by case=>sc' [t'][->->][S H]; do !split=>//; exists t'. 
case; case: sc=>//. 
- by case: t=>//= v [S H][->->]; left; exists v. 
by move=>sc /= [S H][t'][S' ->]; right; exists sc, t'. 
Qed.

Lemma tstPar W A1 A2 B s1 t1 t2 sc s2 k (q : tree W B) :
        tstep s1 (Par t1 t2 k) sc s2 q <-> 
        (sc = ParRet /\ exists (v1 : A1) (v2 : A2),
           [/\ t1 = Ret v1, t2 = Ret v2, q = k (v1, v2), s2 = s1 &
               s1 \In coh W]) \/
        exists t1' t2' sc',
          q = Par t1' t2' k /\
          ([/\ sc = ParL sc', t2' = t2 & tstep s1 t1 sc' s2 t1']  \/
           [/\ sc = ParR sc', t1' = t1 & tstep s1 t2 sc' s2 t2']).
Proof.
split; last first.
- case; first by case=>-> [v1][v2][->->->->].
  case=>t1' [t2'][sc'][->]; case; case=>->->[S H]; split=>//; 
  by [exists t1' | exists t2']. 
case; case: sc=>//.
- case: t1=>// v1; case: t2=>//= v2 [S H][->->]; left; split=>//.
  by exists v1, v2.
- by move=>sc S /=[t1'][H ->]; right; exists t1', t2, sc; split=>//; left.
by move=>sc S /=[t2'][H ->]; right; exists t1, t2', sc; split=>//; right.
Qed.

Lemma tstHide W V A (F : retracts W V) 
              s1 (t : tree V A) sc s2 (q : tree W A) :
        tstep s1 (Hide F t) sc s2 q <-> 
        exists U (gL : U) f, F = Retracts gL f /\ 
          ((exists s1' v, [/\ sc = HideRet, t = Ret v, q = Ret v, s2 = s1 & 
                              f gL s1 s1']) \/
            exists s1' s2' gL' t' sc',
              [/\ sc = HideStep sc', q = Hide (Retracts gL' f) t', 
                  f gL s1 s1', f gL' s2 s2' &
                  tstep s1' t sc' s2' t']).
Proof.
split; last first.
- case=>U [gL][f][->{F}][].
  - case=>s1' [v][->->->-> R]; split; last by exists s1'.
    by split; [case/rf_coh: R | exists s1']. 
  case=>s1' [s2'][gL'][t'][sc'][->-> R1 R2][S H].
  split; last by exists s1', s2', gL', t'.
  split; first by case/rf_coh: R1. 
  by exists s1'; split=>//; case: S.
case: F=>U gL f /=.
case; case: sc=>//.
- move=>sc' S /= [s1'][s2'][gL'][t2][->R1 R2 T]. 
  exists U, gL, f; split=>//; right. 
  exists s1', s2', gL', t2, sc'. 
  case: S=>C /= [_][/(rf_func R1)<-] S; split=>//. 
  by case/rf_coh: R1. 
case=>C; case: t=>// v [s1'] R1 [_][->{s2}->{q}] _.
by exists U, gL, f; split=>//; left; exists s1', v. 
Qed.

Lemma tstHideMorph U W V A (f : refinement U W V) gL 
                   s1 (t : tree V A) sc s2 (q : tree W A) :
        tstep s1 (Hide (Retracts gL f) t) sc s2 q <-> 
        (exists s1' v, [/\ sc = HideRet, t = Ret v, q = Ret v, s2 = s1 & 
                           f gL s1 s1']) \/
         exists s1' s2' (gL' : U) t' sc',
           [/\ sc = HideStep sc', q = Hide (Retracts gL' f) t', 
               f gL s1 s1', f gL' s2 s2' &
               tstep s1' t sc' s2' t'].
Proof.
split; rewrite tstHide.
- case=>U1 [gL1][f1][].
  case=>E; rewrite -{U1}E in f1 gL1 *.
  by do 2![move/inj_pair2=><-]. 
case. 
- case=>s1' [v][->->->-> R1]. 
  by exists U, gL, f; split=>//; left; exists s1', v.  
case=>s1' [s2'][gL'][t'][sc'][->-> R1 R2 T]. 
by exists U, gL, f; split=>//; right; exists s1', s2', gL', t', sc'.
Qed.

Lemma tstExtend V W A (em : injects V W) 
                s1 (t : tree V A) sc s2 (q : tree W A) :
        tstep s1 (Extend em t) sc s2 q <->
        (exists s1' v, [/\ sc = ExtendRet, t = Ret v, q = Ret v, s2 = s1 &
                           extends em s1 s1']) \/
        exists sc' t' s1' s2' s, 
          [/\ sc = ExtendStep sc', q = Extend em t', 
              s1 = s1' \+ s, s2 = s2' \+ s, s1 \In Mod.coh W &
              tstep s1' t sc' s2' t'].
Proof.
split; last first.
- case.
  - case=>s1' [v][->->->->] E. 
    by do 2![split; last by exists s1']; case/ext_coh: E.
  case=>sc' [t'][s1'][s2'][s][->->->-> C][[C' S] T]. 
  split; last by exists s1', s2', s, t'. 
  by split=>//; exists s1'; split=>//; exists s. 
case; case: sc=>//; last first.
- case: t=>//= v [C1 S][s1'][->->{s2 q}] X.
  by left; exists s1'; exists v. 
move=>sc [C] /= [s'][X] S [s1'][s2'][t'][t2'][??? C1'] T; subst q s1 s2. 
right; exists sc, t2', s1', s2', t'; do !split=>//.
by case: X=>t'' [E] Cs' _; rewrite (coh_prec (cohV C) E C1' Cs'). 
Qed.

(* we can put these together, in a lemma that serves as a *)
(* analysis over the tree, but with more structure than ordinary casing *)

Lemma tst W A s1 (t : tree W A) sc s2 q :
        tstep s1 t sc s2 q <->
    [\/ exists a v, 
          [/\ t = Act a, sc = ChoiceAct, q = Ret v & astep s1 a s2 v], 
        exists B (p' : tree W B) k, t = Seq p' k /\ 
          ((exists v, [/\ sc = SeqRet, p' = Ret v, q = k v, s2 = s1 & 
                          s1 \In coh W]) \/
           (exists sc' q',
             [/\ sc = SeqStep sc', q = Seq q' k & tstep s1 p' sc' s2 q'])),
        exists B1 B2 (t1 : tree W B1) (t2 : tree W B2) k, 
          [/\ t = Par t1 t2 k &
              (exists v1 v2, 
                 [/\ sc = ParRet, t1 = Ret v1, t2 = Ret v2, s2 = s1, 
                     q = k (v1, v2) & s1 \In coh W]) \/
              exists t1' t2' sc', q = Par t1' t2' k /\
                ([/\ sc = ParL sc', t2' = t2 & tstep s1 t1 sc' s2 t1'] \/
                 [/\ sc = ParR sc', t1' = t1 & tstep s1 t2 sc' s2 t2'])],
        exists U V (gL : U) (f : refinement U W V) t1,
          [/\ t = Hide (Retracts gL f) t1 &
              (exists s1' v, [/\ sc = HideRet, t1 = Ret v, q = Ret v, s2 = s1 &
                                 f gL s1 s1']) \/
              (exists s1' s2' gL' t2 sc', 
                [/\ sc = HideStep sc', q = Hide (Retracts gL' f) t2, 
                    f gL s1 s1', f gL' s2 s2' & 
                    tstep s1' t1 sc' s2' t2])] | 
        exists V (em : injects V W) t1, 
          [/\ t = Extend em t1 & 
              (exists s1' v, [/\ sc = ExtendRet, t1 = Ret v, q = Ret v, s2 = s1 &
                                 extends em s1 s1']) \/
              exists sc' t' s1' s2' s, 
                [/\ sc = ExtendStep sc', q = Extend em t', 
                    s1 = s1' \+ s, s2 = s2' \+ s, s1 \In Mod.coh W &
                    tstep s1' t1 sc' s2' t']]].
Proof.
split; last first.
- case.
  - by case=>a [v][->->->] H; apply/tstAct; exists v. 
  - case=>B [t'][k][->]; case.
    - by case=>v [->->->-> C]; apply/tstSeq; left; exists v. 
    by case=>sc' [q'][->-> H]; apply/tstSeq; right; exists sc', q'. 
  - case=>B1 [B2][t1][t2][k][->]; case; first by case=>v1 [v2][->->->->->]. 
    case=>t1' [t2'][sc'][->]; case; case=>->-> [S H]; split=>//;
    by [exists t1' | exists t2'].
  - by case=>U [V][f][gL][t1][->] H; apply/tstHide; exists U, f, gL. 
  by case=>V [em][t1][->] H; apply/tstExtend. 
case: t=>[|v|a|B t k|B1 B2 t1 t2 k|V F t1|V em t].  
- by rewrite tstUnfin.
- by rewrite tstRet.
- by rewrite tstAct; case=>v [->-> H]; apply: Or51; exists a, v. 
- rewrite tstSeq; case.
  - case=>v [->->->-> C]; apply: Or52; exists B, (Ret v), k.
    by split=>//; left; exists v. 
  case=>sc' [t'][->-> H]; apply: Or52; exists B, t, k; split=>//.
  by right; exists sc', t'. 
- rewrite tstPar; case.
  - case=>-> [v1][v2][->->->->] C; apply: Or53. 
    exists B1, B2, (Ret v1), (Ret v2), k. 
    by split=>//; left; exists v1, v2. 
  case=>t1' [t2'][sc'][->]; case.
  - case=>->-> H; apply: Or53; exists B1, B2, t1, t2, k; split=>//.
    by right; exists t1', t2, sc'; split=>//; left.
  case=>->-> H; apply: Or53; exists B1, B2, t1, t2, k; split=>//.
  by right; exists t1, t2', sc'; split=>//; right. 
- rewrite tstHide; case=>U [gL][f][->] H.
  by apply: Or54; exists U, V, gL, f, t1. 
by rewrite tstExtend=>H; apply: Or55; exists V, em, t. 
Qed.

(* We may be given both the tree and the schedule *)
 
Lemma tstActS W A s1 a s2 (t : tree W A) : 
        tstep s1 (Act a) ChoiceAct s2 t <->
        exists v, t = Ret v /\ astep s1 a s2 v.
Proof.
rewrite tstAct; split; last by case=>v [->]; exists v. 
by case=>v [_ ->]; exists v. 
Qed.

Lemma tstSeqS W A B s1 (t : tree W B) k sc s2 (q : tree W A) :
        tstep s1 (Seq t k) sc s2 q <->
        match sc with 
          SeqRet => exists v, 
            [/\ t = Ret v, q = k v, s2 = s1 & s1 \In coh W]
        | SeqStep sc' => exists p', 
            [/\ q = Seq p' k & tstep s1 t sc' s2 p']
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc|sc||sc|]; try by [split=>//; case].
- split; last by case=>v [->->->]. 
  by case; case: t=>//= v [C H][->->]; exists v.
split; last by case=>t' [->][S H]; split=>//; exists t'. 
by case=>S /= [t'][H ->]; exists t'. 
Qed.

Lemma tstParS W A1 A2 B s1 t1 t2 sc s2 k (q : tree W B) :
        tstep s1 (Par t1 t2 k) sc s2 q <-> 
        match sc with 
          ParRet => exists (v1 : A1) (v2 : A2),
            [/\ t1 = Ret v1, t2 = Ret v2, q = k (v1, v2), s2 = s1 &
                s1 \In coh W]
        | ParL sc' => exists t1',
            q = Par t1' t2 k /\ tstep s1 t1 sc' s2 t1'
        | ParR sc' => exists t2',
            q = Par t1 t2' k /\ tstep s1 t2 sc' s2 t2'
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc|sc||sc|]; try by split=>//; case.
- split; last by case=>v1 [v2][->->->->].
  case; case: t1=>// v1; case: t2=>// v2. 
  by case=>C _ /= [->->]; exists v1, v2.
- split; last by case=>t1' [->][S H]; split=>//; exists t1'. 
  by case=>S /= [q'][H ->]; exists q'.
split; last by case=>t1' [->][S H]; split=>//; exists t1'.
by case=>S /= [q'][H ->]; exists q'.
Qed.

Lemma tstHideS W V A (F : retracts W V) s1 t sc s2 (q : tree W A) :
        tstep s1 (Hide F t) sc s2 q <-> 
        exists U (gL : U) f, F = Retracts gL f /\ 
        match sc with 
          HideRet => exists s1' v, 
            [/\ t = Ret v, q = Ret v, s2 = s1 & f gL s1 s1']
        | HideStep sc' => exists s1' s2' gL' q', 
            [/\ q = Hide (Retracts gL' f) q', f gL s1 s1', 
                f gL' s2 s2' & tstep s1' t sc' s2' q']
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc|sc||sc|]; try by [split=>//; case=>// U [f][gL][]];
case: F=>U gL f; split. 
- case; case=>C /= [s'][R1] S' H; exists U, gL, f; split=>//. 
  case: H=>_ [s2'][gL'][t2][->] /(rf_func R1) <- R2 H. 
  by exists s', s2', gL', t2; do !split=>//; case/rf_coh: R1. 
- case=>U1 [gL1][f1][][E]; rewrite -{U1}E in f1 gL1 *.
  do 2![move/inj_pair2=>->]; case=>s1' [s2'][gL'][q'][-> R1 R2 [S T]].
  split; last by exists s1', s2', gL', q'. 
  split; last by exists s1'; split=>//; case: S.
  by case/rf_coh: R1. 
- case; case: t=>//= v [C][s'] R1 [s1'][->{s2}->{q} R2]. 
  by exists U, gL, f; split=>//; exists s1', v. 
case=>U1 [gL1][f1][][E]; rewrite -{U1}E in f1 gL1 *.
do 2![move/inj_pair2=>->]; case=>s1' [v][->->-> R]. 
split; last by exists s1'. 
split; last by exists s1'.
by case/rf_coh: R. 
Qed.

Lemma tstExtendS V W A (em : injects V W) 
                 s1 (t : tree V A) sc s2 (q : tree W A) :
        tstep s1 (Extend em t) sc s2 q <->
        match sc with
          ExtendRet => exists s1' v, 
            [/\ t = Ret v, q = Ret v, s2 = s1 & extends em s1 s1']
        | ExtendStep sc' => exists t' s1' s2' s, 
            [/\ q = Extend em t', 
                s1 = s1' \+ s, s2 = s2' \+ s, s1 \In Mod.coh W &
                tstep s1' t sc' s2' t']
        | _ => False
        end.
Proof.
case: sc=>[||sc||sc|sc|sc||sc|]; try by [split=>//; case];
split; rewrite tstExtend.
- case; first by case=>s1' [v][].
  by case=>sc' [t'][s1'][s2'][s][[->]] ->->-> C T; exists t', s1', s2', s. 
- by case=>t' [s1'][s2'][s][->->-> C T]; right; exists sc, t', s1', s2', s. 
- case; last by case=>sc' [t'][s1'][s2'][s][].
  by case=>s1' [v][_ ->->-> X]; exists s1', v. 
by case=>s1' [v][->->-> X]; left; exists s1', v. 
Qed.

(* this view if we are given a schedule but not a tree *)
(* typically used when we have to induct on schedule *)

Lemma tstS W A s1 (t : tree W A) sc s2 q :
        tstep s1 t sc s2 q <->
        match sc with 
          ChoiceAct => exists a v, 
            [/\ t = Act a, q = Ret v & astep s1 a s2 v]
        | SeqRet => exists B (v : B) k,
            [/\ t = Seq (Ret v) k, q = k v, s2 = s1 & s1 \In coh W]
        | SeqStep sc' => exists B (t' : tree W B) q' k, 
            [/\ t = Seq t' k, q = Seq q' k & tstep s1 t' sc' s2 q']
        | ParRet => exists B1 B2 (v1 : B1) (v2 : B2) k,
            [/\ t = Par (Ret v1) (Ret v2) k, q = k (v1, v2), s2 = s1 & 
                s1 \In coh W]
        | ParL sc' => exists B1 B2 (r : tree W B2) (t' q' : tree W B1) k,
            [/\ t = Par t' r k, q = Par q' r k & tstep s1 t' sc' s2 q']
        | ParR sc' => exists B1 B2 (r : tree W B1) (t' q' : tree W B2) k,
            [/\ t = Par r t' k, q = Par r q' k & tstep s1 t' sc' s2 q']
        | HideRet => 
            exists U V (f : refinement U W V) s1' v gL,
              [/\ t = Hide (Retracts gL f) (Ret v), q = Ret v, s2 = s1 &
                  f gL s1 s1']
        | HideStep sc' => 
            exists U V (f : refinement U W V) s1' s2' gL1 gL2 t1 t2,
            [/\ t = Hide (Retracts gL1 f) t1, 
                q = Hide (Retracts gL2 f) t2,
                f gL1 s1 s1', f gL2 s2 s2' &
                tstep s1' t1 sc' s2' t2]
        | ExtendRet => exists V s1' v (em : injects V W),
            [/\ t = Extend em (Ret v), q = Ret v, s2 = s1 & extends em s1 s1']
        | ExtendStep sc' => exists V t1 t2 s1' s2' s (em : injects V W),
            [/\ t = Extend em t1, q = Extend em t2, 
                s1 = s1' \+ s, s2 = s2' \+ s, s1 \In Mod.coh W &
                tstep s1' t1 sc' s2' t2]
        end.
Proof.
case: sc=>[||sc||sc|sc|sc||sc|].
- split; last by case=>a [v][->-> H]; apply/tstAct; exists v.
  by case; case: t=>//= a [C S][v][H ->]; exists a, v.
- split; last by case=>B [v][k][->]->->.
  by case; case: t=>// B; case=>// v k [C S][->->]; exists B, v, k.
- split; last by case=>B [t'][q'][k][->->][S H]; split=>//; exists q'.
  by case; case: t=>//= B t' k [C S][q'][H ->]; exists B, t', q', k.
- split; last by case=>B1 [B2][v1][v2][k][->->-> C]. 
  case; case: t=>// B1 B2; case=>// v1; case=>//= v2 k [C S][->->]. 
  by exists B1, B2, v1, v2, k.  
- split; last by case=>B1 [B2][r][t'][q'][k][->->][S H]; split=>//; exists q'.
  case; case: t=>// B1 B2 t' r k S /= [q'][H ->].
  by exists B1, B2, r, t', q', k.
- split; last by case=>B1 [B2][r][t'][q'][k][->->][S H]; split=>//; exists q'.
  case; case: t=>// B1 B2 r t' k S /= [q'][H ->].
  by exists B1, B2, r, t', q', k.
- split; last first. 
  - case=>U [V][f][s1'][s2'][gL1][gL2][t1][t2][->-> R1 R2][S H].
    rewrite tstHideS; exists U, gL1, f; split=>//.
    by exists s1', s2', gL2, t2. 
  case; case: t=>// V [U gL f] /= t1 S [s1'][s2'][gL'][t2][-> R1 R2 H].
  exists U, V, f; exists s1', s2', gL, gL', t1, t2; do 2![split=>//].
  case: S=>C /= [_][/(rf_func R1)<-] S; split=>//.
  by case/rf_coh: R1. 
- split; last first.
  - case=>U [V][f][s1'][v][gL][->->-> R].
    by rewrite tstHideS; exists U, gL, f; split=>//; exists s1', v. 
  case; case=>C; case: t=>// V [U gL f]; case=>// v [s'] R1 [s1'][->-> R2].
  by exists U, V, f, s', v, gL. 
- split; last first.
  - case=>V [t1][t2][s1'][s2'][s][em][->->->-> C T].
    by rewrite tstExtendS; exists t2, s1', s2', s. 
  case; case: t=>// V em t S /=; case: S=>C /= S.
  case=>s1' [s2'][s][t2'][???] C1' T; subst q s1 s2. 
  exists V, t, t2', s1', s2', s, em; do !split=>//.
  by case: S=>/= [s'][][k][E Cs' _]; rewrite (coh_prec (cohV C) E C1' Cs').
split; last first.
- by case=>V [s1'][v][em][->->-> X]; rewrite tstExtendS; exists s1', v. 
case; case: t=>// V em; case=>//= v S [s1'][->-> X]. 
by exists V, s1', v, em. 
Qed.

(* some additional lemmas on par *)

Lemma tstParT W A B C s1 (t1 t2 : tree W A) sc s2 (q1 q2 : tree W B) 
              (k1 k2 : A * B -> tree W C) : 
         tstep s1 (Par t1 q1 k1) sc s2 (Par t2 q2 k2) <->
         (k2 = k1 /\ 
            exists sc', 
              [/\ sc = ParL sc', q2 = q1 & tstep s1 t1 sc' s2 t2] \/ 
              [/\ sc = ParR sc', t2 = t1 & tstep s1 q1 sc' s2 q2]) \/
         exists x, exists y, 
            [/\ sc = ParRet, t1 = Ret x, q1 = Ret y, s2 = s1,
                k1 (x, y) = Par t2 q2 k2 & s1 \In coh W].
Proof.
rewrite tstPar; split.
- case; first by case=>-> [x][y][->-> E -> C1]; right; exists x, y. 
  by case=>t1' [t2'][sc'][]; case/parEX=>->->-> H; left; split=>//; exists sc'.
case; first by case=>-> [sc' H]; right; exists t2, q2, sc'.  
by case=>x [y][->->->-> E C1]; left; split=>//; exists x, y. 
Qed.

Lemma tstParL W A B C s1 t1 t2 sc s2 q (k : A * B -> tree W C) : 
        tstep s1 t1 sc s2 t2 -> 
        tstep s1 (Par t1 q k) (ParL sc) s2 (Par t2 q k). 
Proof. by case=>S H; split=>//; exists t2. Qed.

Lemma tstParR W A B C s1 t1 t2 sc s2 q (k : A * B -> tree W C) : 
        tstep s1 t1 sc s2 t2 -> 
        tstep s1 (Par q t1 k) (ParR sc) s2 (Par q t2 k). 
Proof. by case=>S H; split=>//; exists t2. Qed.

(*************************************************************)
(* we need the same lemmas for trees that we had for actions *)
(*************************************************************)

Lemma tst_safe W A s1 (t : tree W A) sc s2 q : 
        tstep s1 t sc s2 q -> s1 \In tsafe t sc.
Proof. by case. Qed.

Lemma tst_guar W A s1 (t : tree W A) sc s2 q : 
        tstep s1 t sc s2 q -> guar W s1 s2.
Proof.
elim: sc W A s1 s2 t q=>[||sc IH||sc IH|sc IH|sc IH||sc IH|] W A s1 s2 p q;
rewrite tstS.
- by case=>a [v][_ _]; apply: ast_guar.
- by case=>B [v][k][_ _ ->]; apply: refl_guar_refl. 
- by case=>B [t'][q'][k][_ _]; apply: IH.
- by case=>B1 [B2][v1][v2][k][_ _ ->]; apply: refl_guar_refl. 
- by case=>B1 [B2][r][p'][q'][k][_ _]; apply: IH.
- by case=>B1 [B2][r][p'][q'][k][_ _]; apply: IH.
- case=>U [V][f][s1'][s2'][gL1][gL2][t1][t2][_ _ R1 R2] /IH.
  by apply: rf_guar R1 R2.
- case=>U [V][f][s1'][v][gL][_ _ -> /rf_coh [C _]]. 
  by apply: refl_guar_refl C.
- case=>V [t1][t2][s1'][s2'][s][em][_ _ ->->] C /IH. 
  by apply: (internal_extend em).
by case=>V [s1'][v][em][_ _ ->] /ext_coh [C _]; apply: refl_guar_refl.
Qed.

(* additional lemmas that follow from tst_guar *)

Lemma tst_coh W A s (t : tree W A) sc s' q : 
        tstep s t sc s' q -> s \In coh W /\ s' \In coh W.
Proof. by case/tst_guar/guar_coh. Qed.

Lemma tst_env W A s (t : tree W A) sc s' q : 
        tstep s t sc s' q -> env s = env s'.
Proof. by move/tst_guar/guar_env. Qed.

(* now tst_zagzig *)

(* is this one useful 
Lemma tst_zagzig W A s s' p (t : tree W A) sc q : 
        env s = env s' ->
        tstep (s \> p) t sc (s' \> p) q ->
        tstep (s \< p) t sc (s' \< p) q. 
Proof.
elim: sc W A s s' p t q=>[||sc IH||sc IH|sc IH|sc IH||sc IH|] 
         W A s s' p t q E T; move: T (T); case/tst_coh=>[Cs Cs']; rewrite !tstS.
- by case=>a [v][->->] /(ast_zagzig E) H; exists a, v. 
- case=>B [v][k][->->] /esym/(zagzig_eq E) -> /coh_zigzag C.
  by exists B, v, k.
- by case=>B [t'][q'][k][->-> H]; exists B, t', q', k; split=>//; apply: IH. 
- case=>B1 [B2][v1][v2][k][->->] /esym/(zagzig_eq E) -> /coh_zigzag C.
  by exists B1, B2, v1, v2, k. 
- case=>B1 [B2][r][p'][q'][k][->->] H; exists B1, B2, r, p', q', k. 
  by split=>//; apply: IH.
- case=>B1 [B2][r][p'][q'][k][->->] H; exists B1, B2, r, p', q', k.
  by split=>//; apply: IH.
- case=>U [V][f][s1][s1'][gL1][gL2][gE][t1][t1'][->-> R R' H].
  move/tst_env: (H); case/(rf_env R R')=>_ S.
  case: (rf_zagzig f p)=>p' [Rz _].
  case/Rz: R=>s2 [X R]; case/Rz: R'=>s2' [X' R'].
  rewrite {s1}X {s1' Rz}X' in H *.
  exists U, V, f, (s2 \< p'), (s2' \< p'), gL1, gL2, gE, t1, t1'. 
  split=>//; apply: IH H; rewrite -(spSenv s2 p') -(spSenv s2' p').
  by apply/(rf_env R R'); rewrite !spP. 
- case=>U [V][f][s1'][v][gL][gE][->-> /esym]. 
  case: (rf_zagzig f p)=>p' [Rz _] /(zagzig_eq E) -> /Rz [s2][X R]. 
  by exists U, V, f, (s2 \< p'), v, gL, gE. 
- case=>V [t1][t2][s1'][s2'][x][em][->-> E1 E2 C T].
  case: (tst_coh T)=>Cs1' Cs2'; case: {-}(em)=>W2 ?; subst W.
  rewrite se_coh in Cs Cs'.
  case: (envzag_plusf Cs Cs' Cs1' Cs2' E E1 E2)=>t1' [t2'][r1'][r2'][pt][pr][].
  case=>??? [??? X][Et Er] F1 F2; subst s s' p s1' x s2'.
  move/(zagzig_eq Er): X=>?; subst r2'.
  exists V, t1, t2, (t1' \< pt), (t2' \< pt), (r1' \< pr), em.
  by split=>//; [move/coh_zigzag: C | apply: IH].
case=>V [s1'][v][em][->-> /esym]/(zagzig_eq E)=><-{s' E Cs'}.
case/ext_zagzig=>s1 [s2][p1][p2][???] X; subst s p s1'.
by exists V, (s1 \< p1), v, em. 
Qed.
*)

(* this was always useless 
Lemma tst_local A s1 s2 s (p : tree A) sc q :
        s1 \+ s \In coherent -> tstep s1 p sc s2 q -> 
        tstep (s1 \+ s) p sc (s2 \+ s) q.
Proof.
move=>C; elim: sc A p q=>[||sc IH||sc IH|sc IH] A p q; rewrite !tstS.
- by case=>a [v][->-> H]; exists a, v; split=>//; apply: ast_local.
- by case=>B [v][k][->->-> C1]; exists B, v, k.
- by case=>B [t'][q'][k][->-> H]; exists B, t', q', k; split=>//; apply: IH.
- by case=>B1 [B2][v1][v2][k][->->-> C1]; exists B1, B2, v1, v2, k.
- case=>B1 [B2][r][p'][q'][k][->-> H]; exists B1, B2, r, p', q', k.
  by split=>//; apply: IH.
case=>B1 [B2][r][p'][q'][k][->-> H]; exists B1, B2, r, p', q', k.
by split=>//; apply: IH.
Qed.
*)

(*
the following not needed 

Lemma tst_det A s s1 s2 (p : tree A) sc q1 q2 : 
        tstep s p sc s1 q1 -> tstep s p sc s2 q2 -> s2 = s1 /\ q2 = q1. 
Proof.
elim: sc A p q1 q2=>[||sc IH||sc IH|sc IH] A p q1 q2; rewrite !tstS.
- by case=>a1 [v1][->-> H1][a2][v2][[<- ->]]; case/(ast_det H1)=>->->.
- case=>B1 [v1][k1][->->-> C1]; case=>B2 [v2][k2][]; case/seqE=>E. 
  by rewrite -{B2}E in v2 k2 *; case/jmE=>->; move/jmE=>->->->. 
- case=>B1 [t1'][q1'][k1][->-> H1]; case=>B2 [t2'][q2'][k2][]; case/seqE=>E.
  rewrite -{B2}E in t2' q2' k2 *; move/jmE=><-; move/jmE=><- ->.
  by case/(IH _ _ _ _ H1)=>->->.
- case=>B1 [B2][v1][v2][k1][->->-> C1][D1][D2][w1][w2][k2][]. 
  case/parE=>E1 E2; rewrite -{D1}E1 -{D2}E2 in w1 w2 k2 *.
  by do 2![case/jmE=>->]; move/jm2E=>->->->.  
- case=>B1 [B2][r][p'][q'][k][->-> H][D1][D2][w][t'][z'][h][]. 
  case/parE=>E1 E2; rewrite -{D1}E1 -{D2}E2 in t' z' h w *.
  move/jmE=><-; move/jmE=>->; move/jm2E=>->->. 
  by case/(IH _ _ _ _ H)=>->->. 
case=>B1 [B2][r][p'][q'][k][->-> H][D1][D2][w][t'][z'][h][]. 
case/parE=>E1 E2; rewrite -{D1}E1 -{D2}E2 in t' z' h w *.
move/jmE=>->; move/jmE=><-; move/jm2E=>->->. 
by case/(IH _ _ _ _ H)=>->->. 
Qed.
*)

(* The following lemma doesn't hold, because schedules that are *)
(* wrong for the term are considered safe, as they don't actually step. *)
(* The story here can be made more precise if we consider "sensible" schedules *)
(* but that seems to be beating the purpose. *)
(* It's forcing me to define when a process is steppable for a schedule *)
(* which is just duplicating the functionality of the stepping relation *)
(* In the end, I think framing is probably better taken as primitive *)
(* rather than insisting that every safe tree steps *)
(*
Lemma tst_total A s1 (p : tree A) sc : 
        s1 \In tsafe p sc -> exists s2, exists v, tstep s1 p sc s2 v.
*)

Structure erased_tree A := 
    EUnfinished | ERet of A | EAct (W : mod) of int_action (guar W) A
  | ESeq B of erased_tree B & B -> erased_tree A 
  | EPar B1 B2 of erased_tree B1 & erased_tree B2 & 
                  B1 * B2 -> erased_tree A
  | EHide of erased_tree A 
  | EExtend of erased_tree A.

Fixpoint erase_tree W A (t : tree W A) : erased_tree A := 
  match t with 
    Unfinished => EUnfinished A
  | Ret v => ERet v 
  | Act a => EAct a
  | Seq _ p k => ESeq (erase_tree p) (fun v => erase_tree (k v))
  | Par _ _ p1 p2 k => 
      EPar (erase_tree p1) (erase_tree p2) (fun v => erase_tree (k v))
  | Hide _ f p => EHide (erase_tree p)
  | Extend _ e p => EExtend (erase_tree p)
  end.

Lemma EactEX (W1 W2 : mod) A (a1 : int_action (guar W1) A) 
                             (a2 : int_action (guar W2) A) : 
        EAct a1 = EAct a2 <->
        W1 = W2 /\ ijmeq (sort := fun W => int_action (guar W) A) a1 a2.
Proof.
split; case=>E; rewrite -{W2}E in a2 *.
- by move/inj_pair2=>->. 
by move/ijmE=>->. 
Qed.

Lemma EseqE A B1 B2 (t1 : erased_tree B1) (t2 : erased_tree B2) k1 k2 : 
        ESeq t1 k1 = ESeq t2 k2 <-> 
        [/\ B1 = B2, ijmeq t1 t2 &
            ijmeq (sort := fun B => B -> erased_tree A) k1 k2].
Proof.
split; case=>E; rewrite -{B2}E in t2 k2 *.
- by do 2![move/inj_pair2=><-].
by do 2![move/ijmE=><-].
Qed.

Lemma EseqEX A B (t1 t2 : erased_tree B) k1 k2 : 
        ESeq t1 k1 = ESeq t2 k2 :> erased_tree A <-> t1 = t2 /\ k1 = k2.
Proof. by rewrite EseqE; split=>[[_]|[->->] //]; do 2![move/ijmE=>->]. Qed.

Lemma EparE A1 A2 B1 B2 C t1 t2 q1 q2 
            (r1 : A1 * B1 -> erased_tree C) 
            (r2 : A2 * B2 -> erased_tree C) :
        EPar t1 q1 r1 = EPar t2 q2 r2 <-> 
        [/\ A1 = A2, B1 = B2, ijmeq t1 t2, ijmeq q1 q2 & 
            ijmeq (sort := fun B => B -> erased_tree C) r1 r2].
Proof. 
split; case=>E1 E2; rewrite -{A2}E1 -{B2}E2 in t2 q2 r1 r2 *.
- by do 2![move/inj_pair2=>->]; do 2![move/inj_pair2]; move=>->.
by do 3![move/jmE=>->].
Qed.

Lemma EparEX A1 A2 C (t1 t2 : erased_tree A1) 
                     (q1 q2 : erased_tree A2) r1 r2 :
        EPar t1 q1 r1 = EPar t2 q2 r2 :> erased_tree C <-> 
          [/\ t1 = t2, q1 = q2 & r1 = r2].
Proof. 
split; case; last by move=>->->->.
by do 2![move/inj_pair2=>->]; do 2![move/inj_pair2]=>->.
Qed.

Lemma tst_erase W1 W2 A s1 s1' s2 s2' (t1 : tree W1 A) (t2 : tree W2 A) 
                sc q1 q2 h1 h2 : 
        valid (flats s1 \+ h1) ->
        flats s1 \+ h1 = flats s2 \+ h2 -> 
        erase_tree t1 = erase_tree t2 ->
        tstep s1 t1 sc s1' q1 -> tstep s2 t2 sc s2' q2 -> 
        erase_tree q1 = erase_tree q2 /\ 
        flats s1' \+ h1 = flats s2' \+ h2.
Proof. 
elim: sc W1 W2 A s1 s1' s2 s2' h1 h2 t1 t2 q1 q2=>
      [||sc IH||sc IH|sc IH|sc IH||sc IH|] 
         W1 W2 A s1 s1' s2 s2' h1 h2 t1 t2 q1 q2 D F X; rewrite !tstS.
- case=>a1 [v1][E1 -> H1][a2][v2][E2 ->]; move: E1 E2 X=>->-> /=.
  case/EactEX=>E; rewrite -{W2}E in a2 t2 q2 *; move/ijmE=><-.
  by case/(int_ast_erase D F H1)=>->. 
- case=>B1 [v1][k1][E1 ->-> C1][B2][v2][k2][E2 ->-> C2].
  move: E1 E2 X=>->-> /= /EseqE [E]; rewrite -{B2}E in v2 k2 *. 
  by case/jmE=>->; move/jmE/ext/(_ v2). 
- case=>B1 [t1'][q1'][k1][E1 -> H1][B2][t2'][q2'][k2][E2 -> H2]. 
  move: E1 E2 X=>->-> /= /EseqE [E]; rewrite -{B2}E in t2' q2' k2 H2 *.
  move/jmE=>E; move/jmE=>->. 
  by case: (IH _ _ _ _ _ _ _ _ _ _ _ _ _ D F E H1 H2)=>->->. 
- case=>B1 [B2][v1][v2][k1][E1 ->-> C1][D1][D2][w1][w2][k2][E2 ->-> C2]. 
  move: E1 E2 X=>->-> /= /EparE [E1 E2]; rewrite -{D1}E1 -{D2}E2 in w1 w2 k2 *.
  by do 2![case/ijmE=>->]; move/ijmE/ext/(_ (w1, w2)). 
- case=>B1 [B2][r][p'][q'][k][E1 -> H1][D1][D2][w][t'][z'][h][E2 -> H2]. 
  move: E1 E2 X=>->-> /= /EparE [E1 E2].
  rewrite -{D1}E1 -{D2}E2 in t' z' h w H2 *.
  move/ijmE=>E; do 2![move/ijmE=>->].
  by case: (IH _ _ _ _ _ _ _ _ _ _ _ _ _ D F E H1 H2)=>->->.  
- case=>B1 [B2][r][p'][q'][k][E1 -> H1][D1][D2][w][t'][z'][h][E2 -> H2]. 
  move: E1 E2 X=>->-> /= /EparE [E1 E2].
  rewrite -{D1}E1 -{D2}E2 in t' z' h w H2 *. 
  move/ijmE=>->; move/ijmE=>E; move/ijmE=>->.
  by case: (IH _ _ _ _ _ _ _ _ _ _ _ _ _ D F E H1 H2)=>->->.  
- case=>U1 [V1][f1][x1][y1][gL1'][gL1''][p1][r1][E1 -> R1 S1 T1].
  case=>U2 [V2][f2][x2][y2][gL2'][gL2''][p2][r2][E2 -> R2 S2 T2].
  move: E1 E2 X=>->-> /= [E]. 
  rewrite (rf_flats R1) (rf_flats R2) in D F.
  rewrite (rf_flats S1) (rf_flats S2).
  by case: (IH _ _ _ _ _ _ _ _ _ _ _ _ _ D F E T1 T2)=>->.
- case=>U1 [V1][f1][x1][v1][gL1][E1 ->-> R1].
  case=>U2 [V2][f2][x2][v2][gL2][E2 ->-> R2].
  by move: E1 E2 X=>->-> /= [->]. 
- move=>H1 H2; move: H1 H2 D F X.
  case=>V1 [x1][y1][t1'][r1'][z1][em1][->->->->{t1 q1 s1 s1'}] Ct1 T1. 
  case=>V2 [x2][y2][t2'][r2'][z2][em2][->->->->{t2 q2 s2 s2'}] Ct2 T2 /= D F.
  move: (ext_local em1 (tst_guar T1) Ct1) => Cr1.
  move: (ext_local em2 (tst_guar T2) Ct2) => Cr2.
  rewrite (flatsUn (cohV Ct1)) (flatsUn (cohV Ct2)) 
          (flatsUn (cohV Cr1)) (flatsUn (cohV Cr2)) -!joinA in F D *.
  by case=>E; case: (IH _ _ _ _ _ _ _ _ _ _ _ _ _ D F E T1 T2)=>->->.
case=>V1 [x1][v1][em1][???] X1; subst t1 q1 s1'.
case=>V2 [x2][v2][em2][???] X2; subst t2 q2 s2'.
by case: X=>->.
Qed.

(*
(* here i can't prove q1 = q2; what seems to holds is that *)
(* erasing seed info from q1 and q2 produces equal trees *)
(* but i don't think that seeds in q1 and q2 have to be equal *)
(* as they may be determined by the ghosts appearing in s1 and s2 *)
(* however, so far the conjunct q1 = q2 is not needed anywhere, so *)
(* I omit it. *)
(* the erasure theorems are just a way to show *)
(* that ghosts don't mess the final values and heaps produced *)
(* so they are orthogonal to everything else *)
Lemma tst_erase W A s1 s1' s2 s2' (t : tree W A) sc q1 q2 h1 h2 : 
        valid (flats s1 \+ h1) ->
        flats s1 \+ h1 = flats s2 \+ h2 -> 
        tstep s1 t sc s1' q1 -> tstep s2 t sc s2' q2 -> 
        (* q1 = q2 /\ *)
        flats s1' \+ h1 = flats s2' \+ h2.
Proof. 
move=>D F.
elim: sc W A s1 s1' s2 s2' t q1 q2 D F=>[||sc IH||sc IH|sc IH|sc IH] 
      W A s1 s1' s2 s2' t q1 q2 D F; rewrite !tstS.
- by case=>a1 [v1][->-> H1][a2][v2][[<-]] ->; case/(int_ast_erase D F H1)=>->. 
- case=>B1 [v1][k1][->->-> C1][B2][v2][k2][]; case/seqE=>E.
  by rewrite -{B2}E in v2 k2 *; case/jmE=>->; move/jmE=>->->->.
- case=>B1 [t1'][q1'][k1][->-> H1]; case=>B2 [t2'][q2'][k2][]; case/seqE=>E.
  rewrite -{B2}E in t2' q2' k2 *; move/jmE=><-; move/jmE=><- ->.
  by case/(IH _ _ _ _ _ _ _ _ _ D F H1)=>->->.
- case=>B1 [B2][v1][v2][k1][->->-> C1][D1][D2][w1][w2][k2][]. 
  case/parE=>E1 E2; rewrite -{D1}E1 -{D2}E2 in w1 w2 k2 *.
  by do 2![case/ijmE=>->]; move/ijmE=>->->->. 
- case=>B1 [B2][r][p'][q'][k][->-> H][D1][D2][w][t'][z'][h][]. 
  case/parE=>E1 E2; rewrite -{D1}E1 -{D2}E2 in t' z' h w *.
  move/ijmE=><-; move/ijmE=>->; move/ijmE=>->->. 
  by case/(IH _ _ _ _ _ _ _ _ _ D F H)=>->->. 
- case=>B1 [B2][r][p'][q'][k][->-> H][D1][D2][w][t'][z'][h][]. 
  case/parE=>E1 E2; rewrite -{D1}E1 -{D2}E2 in t' z' h w *.
  move/ijmE=>->; move/ijmE=><-; move/ijmE=>->->. 
  by case/(IH _ _ _ _ _ _ _ _ _ D F H)=>->->. 
case=>V1 [f1][x1][y1][seed1][t1][r1][->-> R1 S1 T1].
case=>V2 [f2][x2][y2][seed2][t2][r2][].
move/hideE. case=>E. 
rewrite -{V2}E in f2 seed2 t2 r2 *.
move/ijmE=>E. rewrite -{f2}E in seed2 *.
move/ijmE=><-{t2}. 
move=>->{q2}.
move=>R2 S2.
Check @refine_flats.
rewrite (refine_flats R1) in D F.
rewrite (refine_flats R2) in D F.
rewrite (refine_flats S1) (refine_flats S2).
move/(IH _ _ _ _ _ _ _ _ _ D F T1).
case.
move=>->->.
split=>//.
apply/hideE. split=>//.
move/(IH _ _ _ _ _ T1).
case=>V1' [f1'][t1'][q1'][->-> H][V2'][f2'][t2'][q2'][].
case/hideE=>E; rewrite -{V2'}E in f2' t2' q2' *.
do 2![move/ijmE=><-]; move=>->.
by case/(IH _ _ _ _ _ _ H)=>->->. 
Qed.
*)

Lemma tst_frame W A s s' p (t : tree W A) sc q : 
        s \> p \In tsafe t sc -> tstep (s \< p) t sc s' q ->
        exists s'', [/\ s' = s'' \< p, env s'' = env s &
                        tstep (s \> p) t sc (s'' \> p) q].
Proof.
move=>S H; move/tst_env: (H).
elim: sc W A p s s' t q H S=>[||sc IH||sc IH|sc IH|sc IH||sc IH|] 
         W A p s s' t q; rewrite tstS.
- case=> a [v][->-> H][C /= S]; case/(ast_frame S): H=>s'' [->] H.
  by rewrite !spP; exists s''; do !split=>//; exists v.
- by case=>B [v][k][->->->] C S; exists s.  
- case=>B [t'][q'][k][->-> H] S E; case/(IH _ _ _ _ _ _ _ H): S=>// {H}. 
  move=>s'' [->] H; exists s''; split=>//.
  by rewrite tstS; exists B, t', q', k. 
- by case=>B1 [B2][v1][v2][k][->->-> C] S; exists s.
- case=>B1 [B2][r][p'][q'][k][->-> H] S E; case/(IH _ _ _ _ _ _ _ H): S=>// {H}.
  move=>s'' [->] H; exists s''; split=>//.
  by rewrite tstS; exists B1, B2, r, p', q', k. 
- case=>B1 [B2][r][p'][q'][k][->-> H] S E; case/(IH _ _ _ _ _ _ _ H): S=>// {H}.
  move=>s'' [->] H; exists s''; split=>//.
  by rewrite tstS; exists B1, B2, r, p', q', k. 
- case=>U [V][f][s1'][s2'][gL1][gL2][t1][t2][->->{t q} R1 R2 H] S E.
  case: (rf_zagzig f p)=>q [H1 H2]. 
  case: (S)=>_ /= [x][R1']; case/H1: {H1} R1' (R1')=>[s'0][->{x}] R1' R. 
  move: {R1 R1'} (rf_func R1 R1') (R1') H=>->{s1'} R1 H Sx.
  have {Sx} Sf: s'0 \> q \In tsafe t1 sc. 
  - by split=>//; apply/coh_zigzag; case/rf_coh: R1.
  case: (IH _ _ _ _ _ _ _ H Sf (tst_env H)) R2=>x' [->{s2'} E' {H IH Sf} H].
  case/H2=>y' {H2} [?] R2'; subst s'.
  exists y'; rewrite !spP in E; rewrite E; do !split=>//.
  by exists (s'0 \> q), (x' \> q), gL2, t2; split=>//; case: H.  
- case=>U [V][f][s1'][v][gL][->->-> R1] S _.  
  by exists s; do !split=>//; case: S=>/= C [s2 R]; exists s2.
- case=>V [t1][t2][s1'][s2'][x][em][->->{t q} E ->{s'} C T].
  case: (tst_coh T)=>C1 C1'; rewrite (cohE em) in C. 
  have C2 : x \In coh (inj_ext em).
  - by rewrite E in C; move/(star_coh_prec C): C1.
  case: (cohzig_plusf C C1 E)=>r1 [r2][p1][p2][????] {E} E Sf En; 
  subst s p s1' x.
  have Sf' : r1 \> p1 \In tsafe t1 sc.
  - case: Sf=>/= _ [s'][X]; rewrite E in X; move/coh_zigzag: C1=>C1.
    by case/ext_coh: X (X)=>_ C'' /(ext_precise C1)=>->. 
  have {C En} En : env (r1 \< p1) = env s2'.
  - case/cohV/valid_zig: C; case/and3P=>_ _ V' _.
    rewrite /= !spP -!(joinC (env r2)) in En V' *.  
    by move/(funhKl V'): En. 
  case: (IH _ _ _ _ _ _ _ T Sf' En)=>{Sf' T En} s' [? En] T; subst s2'. 
  exists (s' \+ r2); rewrite /= En; split=>//.
  - apply/esym; apply: zig_plus; first by case/cohV/and3P: C1'. 
    by case/cohV/and3P: C2. 
  split=>// {Sf}; exists (r1 \> p1), (s' \> p1), (r2 \> p2), t2.
  split=>//; last first. 
  - by case: T.
  - by move/coh_zigzag: C1. 
  apply: zag_plus; first by case/coh_zigzag/cohV/and3P: C1'.
  by case/coh_zigzag/cohV/and3P: C2.
case=>V [s1'][v][em][->->->{q t s'} X] Sf _; exists s; do !split=>//.
by case/ext_zigzag: X=>s1 [s2][p1][p2][->-> _]; exists (s1 \> p1). 
Qed.

Lemma tst_progress W A s (t : tree W A) sc : 
        s \In tsafe t sc -> good t sc -> 
        exists s' (q : tree W A), tstep s t sc s' q.
Proof.
move=>Sf G. 
elim: sc W A s t G Sf.
- move=>W A s. case=>// g /=. move=>_. 
  case=>C. move/ast_total. case=>s2 [v] St.
  exists s2, (Ret v).  split. 
  - split=>//.  simpl. move/ast_safe: St. by [].
  simpl. exists v. by [].
- move=>W A s. case=>//. move=>B t q. 
  case: t=>//=. move=>v _. 
  case. move=>C. simpl. move=>_.
  exists s. exists (q v). split.
  - rewrite /tsafe. simpl. by [].
  simpl. by [].
- move=>sc IH. move=>W A s. case=>//. move=>B t q /= G. 
  case. simpl. move=>C S. 
  have: s \In tsafe t sc. by []. 
  move/IH. move/(_ G). case=>s' [q'] St.
  exists s'. exists (Seq q' q).
  split=>//.
  case: St=>Sf St. simpl. exists q'. by [].
- move=>W A s. case=>//. move=>B1 B2 t q f. 
  case: t=>// v. case: q=>// w.
  simpl. move=>_.
  move=>Sf. exists s. exists (f (v, w)). simpl. by []. 
- move=>sc IH W A s. case=>//. move=>B1 B2 t t1 t2. simpl. move=>G.
  case=>C /= Sf.
  move: (IH _ _ _ _ G (conj C Sf)). 
  case=>s' [q St]. exists s', (Par q t1 t2).
  split=>//. simpl.
  case: St=>Sf' St.
  exists q. by [].
- move=>sc IH W A s. case=>//. move=>B1 B2 t t1 t2. simpl. move=>G.
  case=>C /= Sf.
  move: (IH _ _ _ _ G (conj C Sf)).
  case=>s' [q St]. exists s', (Par t q t2).
  split=>//. simpl.
  case: St=>Sf' St.
  exists q. by [].
- move=>sc IH W A s; case=>//.
  move=>V r t. case: r=>/= T gL rf G. 
  case.  move=>C. simpl. case=>s' [H1 H2].
  have: s' \In tsafe t sc. split=>//.
  - by case/rf_coh: H1.
  move/IH.  move/(_ G). case=>s1 [q] St.
  move: (tst_guar St).
  move/(rf_guarsur H1). case=>gL' [s1'] H3.
  rewrite /tstep. simpl.
  exists s1'. 
  exists (Hide (Retracts gL' rf) q).
  split. 
  - by split=>//=; exists s'. 
  exists s'. exists s1. exists gL'. exists q.
  split=>//.
  case: St. by [].
- move=>W A s. case=>//. move=>V r t. 
  case: r=>/= T gL rf.
  case: t=>// v _.
  case.  move=>C. simpl. case=>s' H.
  exists s, (Ret v).
  split. 
  - simpl. split=>//. simpl. exists s'. by [].
  simpl. exists s'.  by [].
- move=>sc IH W A s. case=>//.
  move=>V i t. simpl. 
  move=>G.
  case. move=>C. simpl. case=>s' [E Sf].
  move/IH: G. 
  have X : s' \In tsafe t sc. 
  - split=>//.   
    case/ext_coh: E. by [].
  move/(_ _ X). case=>s1 [q] St.
  move/tst_guar: (St).
  move/ext_local. move/(_ _ i). 
  case: E=>s2 [-> C' Cw].
  move/(_ _ Cw). move=>Cw'.
  exists (s1 \+ s2). exists (Extend i q).
  split.
  - split=>//.  simpl. exists s'. split=>//.
    rewrite /extends. exists s2. by split=>//.
  simpl.
  exists s'. exists s1. exists s2. exists q. split=>//.
  case: St. by [].
move=>W A s. case=>//.
move=>V i. case=>//= v _.
case=>C /= [s' E].
exists s. exists (Ret v). 
split; last first.  
- simpl. exists s'. by [].
split=>//.
simpl.
by exists s'.
Qed.

End Process.

