Require Import ssreflect ssrfun ssrbool eqtype seq rels Setoid Morphisms. 
Set Implicit Arguments.
Unset Strict Implicit. 
Unset Printing Implicit Defensive. 

Open Scope rel_scope. 

(* infix membership predicate in Type *)

(**************************************************************************)
(* We follow ssrbool, and provide four different types of predicates.     *)
(*                                                                        *)
(* (1) tPred is the type of propositional functions                       *)
(* (2) Simpl_tPred is the type of predicates that automatically simplify  *)
(*     when used in an applicative position.                              *)
(* (3) Mem_tPred is for predicates that support infix notation x \tIn P   *) 
(* (4) tPredType is the structure for interpreting various types, such as *)
(* lists, tuples, etc. as predicates.                                     *)
(*                                                                        *)
(* Important point is that custom lemmas over predicates can be stated in *)
(* terms of tPred, while tSimpl_Pred, Mem_tPred and tPredType are for     *)
(* technical developments used in this file only. More on this point      *)
(* can be found in ssrbool.                                               *)
(**************************************************************************)

Definition tPred T := T -> Type.
Identity Coercion fun_of_tPred : tPred >-> Funclass.

Notation xtPred0 := (fun _ => False : Type).
Notation xtPred1 := (fun x y => x = y : Type).
Notation xtPredT := (fun _ => unit).
Notation xtPredI := (fun (p1 p2 : tPred _) x => (p1 x * p2 x)%type).
Notation xtPredU := (fun (p1 p2 : tPred _) x => (p1 x + p2 x)%type).
Notation xtPreim := (fun f (p : tPred _) x => p (f x)).

Section Predicates.
Variable T : Type.

(* simple predicates *)

Definition Simpl_tPred := simpl_fun T Type.
Definition SimpltPred (p : tPred T) : Simpl_tPred := SimplFun p.
Coercion tPred_of_Simpl (p : Simpl_tPred) : tPred T := p : T -> Type.

(* it's useful to declare the operations as simple predicates, so that *)
(* complex expressions automatically reduce when used in applicative   *)
(* positions *)

Definition tPred0 := SimpltPred xtPred0.
Definition tPred1 x := SimpltPred (xtPred1 x).
Definition tPredT := SimpltPred xtPredT.
Definition tPredI p1 p2 := SimpltPred (xtPredI p1 p2).
Definition tPredU p1 p2 := SimpltPred (xtPredU p1 p2).
Definition tPreim rT f (d : tPred rT) := SimpltPred (xtPreim f d).

(* membership predicates *)

CoInductive Mem_tPred : Type := MemtPred of tPred T.
Definition istMem pT totPred mem := mem = (fun p : pT => MemtPred [eta totPred p]).

(* the general structure for predicates *)

Structure tPredType : Type := TypetPredType {
  tPred_Sort :> Type; 
  totPred : tPred_Sort -> tPred T; 
  _ : {mem | istMem totPred mem}}.

Definition mktPredType pT toP := 
  TypetPredType (exist (@istMem pT toP) _ (erefl _)).

(* tPred, tSimplPred, Mem_tPred are tPredType's *)
Canonical Structure tPredtPredType := Eval hnf in @mktPredType (tPred T) id.
Canonical Structure SimpltPredtPredType := 
  Eval hnf in mktPredType tPred_of_Simpl.
Coercion tPred_of_Mem mp : tPred_Sort tPredtPredType :=
  let: MemtPred p := mp in [eta p].
Canonical Structure MemtPredType := Eval hnf in mktPredType tPred_of_Mem.

End Predicates.

Implicit Arguments tPred0 [T].
Implicit Arguments tPredT [T].
Prenex Implicits tPred0 tPredT tPredI tPredU tPreim.

Notation "[ 'tPred' : T | E ]" := (SimpltPred (fun _ : T => E))
  (at level 0, format "[ 'tPred' :  T  |  E ]") : fun_scope.
Notation "[ 'tPred' x | E ]" := (SimpltPred (fun x => E))
  (at level 0, x ident, format "[ 'tPred'  x  |  E ]") : fun_scope.
Notation "[ 'tPred' x : T | E ]" := (SimpltPred (fun x : T => E))
  (at level 0, x ident, only parsing) : fun_scope.
Notation "[ 'tPred' x y | E ]" := (SimpltPred (fun t => let: (x, y) := t in E))
  (at level 0, x ident, y ident, format "[ 'tPred'  x  y  |  E ]") : fun_scope.
Notation "[ 'tPred' x y : T | E ]" := 
  (SimpltPred (fun t : (T*T) => let: (x, y) := t in E))
  (at level 0, x ident, y ident, only parsing) : fun_scope.

Definition repack_tPred T pT :=
  let: TypetPredType _ a mP := pT 
         return {type of @TypetPredType T for pT} -> _ in
   fun k => k a mP.

Notation "[ 'tPredType' 'of' T ]" := (repack_tPred (fun a => @TypetPredType _ T a))
  (at level 0, format "[ 'tPredType'  'of'  T ]") : form_scope.

Notation tPred_Class := (tPred_Sort (tPredtPredType _)).
Coercion Sort_of_Simpl_tPred T (p : Simpl_tPred T) : tPred_Class := p : tPred T.

Definition tPredArgType := Type.
Coercion tPred_of_argType (T : tPredArgType) : Simpl_tPred T := tPredT.

(* 
Notation "{ :: T }" := (T%type : tPredArgType)
  (at level 0, format "{ ::  T }") : type_scope.
*)

(* These must be defined outside a Section because "cooking" kills the *)
(* nosimpl tag. *)
Definition tMem T (pT : tPredType T) : pT -> Mem_tPred T :=
  nosimpl (let: TypetPredType _ _ (exist mem _) := pT return pT -> _ in mem).
Definition tInMem T x mp := nosimpl tPred_of_Mem T mp x.

Prenex Implicits tMem.

(* Membership Predicates can be used as simple ones *)
Coercion tPred_of_Mem_Pred T mp := [tPred x : T | tInMem x mp].

Notation "x \tIn A" := (tInMem x (tMem A)) 
  (at level 70, no associativity) : rel_scope.
Notation "x \tNotin A" := (~ (x \tIn A)) 
  (at level 70, no associativity) : rel_scope.

(* Some notation for turning tPredTypes into tPred or Simple tPred *)
Notation "[ 'tMem' A ]" := (tPred_of_Simpl (tPred_of_Mem_Pred (tMem A)))
  (at level 0, only parsing) : fun_scope.
Notation "[ 'tPredI' A & B ]" := (tPredI [tMem A] [tMem B])
  (at level 0, format "[ 'tPredI'  A  &  B ]") : fun_scope.
Notation "[ 'tPredU' A & B ]" := (tPredU [tMem A] [tMem B])
  (at level 0, format "[ 'tPredU'  A  &  B ]") : fun_scope.
Notation "[ 'tPreim' f 'of' A ]" := (tPreim f [tMem A])
  (at level 0, format "[ 'tPreim'  f  'of'  A ]") : fun_scope.

Notation "[ 'tPred' x \tIn A ]" := [tPred x | x \tIn A]
  (at level 0, x ident, format "[ 'tPred'  x  \tIn  A ]") : fun_scope.
Notation "[ 'tPred' x \tIn A | E ]" := [tPred x | ((x \tIn A) * E)%type]
  (at level 0, x ident, format "[ 'tPred'  x  \tIn  A  |  E ]") : fun_scope.
Notation "[ 'tPred' x y \tIn A & B | E ]" :=
  [tPred x y | ((x \tIn A) * (y \tIn B) * E)%type]
  (at level 0, x ident, y ident,
   format "[ 'tPred'  x  y  \tIn  A  &  B  |  E ]") : fun_scope.
Notation "[ 'tPred' x y \tIn A & B ]" := [tPred x y | ((x \tIn A) * (y \tIn B))%type]
  (at level 0, x ident, y ident,
   format "[ 'tPred'  x  y  \tIn  A  &  B ]") : fun_scope.
Notation "[ 'tPred' x y \tIn A | E ]" := [tPred x y \tIn A & A | E]
  (at level 0, x ident, y ident,
   format "[ 'tPred'  x  y  \tIn  A  |  E ]") : fun_scope.
Notation "[ 'tPred' x y \tIn A ]" := [tPred x y \tIn A & A]
  (at level 0, x ident, y ident,
   format "[ 'tPred'  x  y  \tIn  A ]") : fun_scope.

Section Simplifications. 
Variables (T : Type) (pT : tPredType T).

Lemma tMem_totPred : forall (p : pT), tMem (totPred p) = tMem p.
Proof. by rewrite /tMem; case: pT => T1 app1 [mem1  /= ->]. Qed.

Lemma totPredE : forall x (p : pT), totPred p x = (x \tIn p).
Proof. by move=> *; rewrite -tMem_totPred. Qed.

Lemma tIn_Simpl : forall x (p : Simpl_tPred T), (x \tIn p) = p x.
Proof. by []. Qed.

Definition tInE := (tIn_Simpl).

Lemma tMem_Simpl : forall (p : Simpl_tPred T), tMem p = p :> tPred T.
Proof. by []. Qed.

Definition tMemE := tMem_Simpl. (* could be extended *)

Lemma tMem_tMem : forall p : pT, (tMem (tMem p) = tMem p) * (tMem [tMem p] = tMem p).
Proof. by move=> p; rewrite -tMem_totPred. Qed.

End Simplifications. 

