Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude idynamic pcm unionmap heap codes feaps state.
Require Import zigzag worlds inject rely privstate footprint.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* Refinement structure. World W' refines W if there is a partial
bijection between the coherent worlds, with a number of simulation
properties. The refinement structure captures the properties *)

Module Refinement.
Section Refinement.
Variables (seedT : Type) (W W' : mod).
(* seedT is the type of extra information used for refining *)
(* we need two items of seedT type; one for refining self *)
(* the other for other component *)

Structure mixin_of (func : seedT -> reltp) := Mixin {
  (* func is a partial function; refining states are uniquely determined *)
  (* but some states in W need not refine into anything in W' *)
  (* e.g. we can't create a lock out of a pointer x, if x doesn't exist *)
  _ : forall g s s1' s2', func g s s1' -> func g s s2' -> s1' = s2';
  (* refine is injective, i.e., abstracting back from a refinement is unique *)
  (* i havent used this one so far (have I) *)
  _ : forall g1 s1 g2 s2 s',   
        func g1 s1 s' -> func g2 s2 s' -> (g1, s1) = (g2, s2);
  (* refine is surjective on the coherent space of W', i.e. *)
  (* abstracting back from a refinement is always possible *)
  (* thus, safety for trees need not emit VC's upon exiting the scope of Hide *)
  (* if we wanted that, then tsafe would have to keep a stack of Hide scopes *)
  (* the property is not used so far, but it would be in an eventual proof of progress *)
  (* !!! WAIT, we need that s' is not just an arbitrary W', but actually 
     something with a unit second ghost !! How to state that? *)
  (* _ : forall s', Mod.coh W' s' -> exists g s, func g s s';  *)
  (* how about the following *)
  (* this should be combined with the property two steps below *)
  _ : forall g1 s1 s1' s2', func g1 s1 s1' -> Mod.internal W' s1' s2' ->
      exists g2 s2, func g2 s2 s2';
  (* coherence *)
  _ : forall g s s', func g s s' -> 
        s \In Mod.coh W /\ s' \In Mod.coh W';
  (* environment steps in the coarse world don't change gE in the fine world *)
  (* because steps in the coarse world don't interfere with the added gE module *)
  _ : forall g s1 s1' s2, 
        func g s1 s1' -> mrely W s1 s2 -> 
        exists s2', func g s2 s2' /\ mrely W' s1' s2';
  (* abstracting back preserves guars; like a simulation property *)
  (* the coarse module W (which has more states) simulates the finer W' *)
  _ : forall g1 s1 s1' g2 s2 s2', 
        func g1 s1 s1' -> func g2 s2 s2' -> 
        Mod.internal W' s1' s2' -> Mod.internal W s1 s2;
  (* refinement realigns loc and shr, but doesn't move things into/outof env *)
  (* this seems not provable from zagzig, which merely states *)
  (* that refine doesn't move things out of env, but not that they are unchanged *)
  _ : forall g1 s1 s1' g2 s2 s2', 
      func g1 s1 s1' -> func g2 s2 s2' -> 
      (env s1 = env s2 <-> env s1' = env s2');
  (* zagzig *)
  (* every p extension to env, is (uniquely?) refined into q *)
  (* moreover, refining direction takes stuff from loc and moves to shr *)
  (* but doesn't remove stuff from env *)
  (* the oposite direction takes from shr and moves to loc *) 
  (* that q is unique can probably be infered from functionality *)
  (* the statement of the lemma is kinda of interesting; in the second part *)
  (* we say we only care to invert q's which have already been obtained from p *)
  (* in practice, obtaining q can only be done by refining with a Unit feap *)
  (* so, the statement says that we can invert back only those q which are obtained *)
  (* by refining with unit. *)
  (* if we wanted to be more general, and allow arbitrary q's *)
  (* we will have to also existentially produce new g (in addition to s1' *)
  (* since we by zigzagging q in the opposite way, we will be zigzaging *)
  (* the l ghosts in the oposite way too *)
  (* be it as it may; let's go with this statement for now *)
  _ : forall p, exists q, 
        [/\ forall g s s', 
              func g (s \> p) s' -> 
              exists s1', s' = s1' \> q /\ func g (s \< p) (s1' \< q) &
            forall g s s', 
              func g s (s' \< q) ->
              exists s1, s = s1 \< p /\ func g (s1 \> p) (s' \> q)];
  (* refinement only deals with ghosts *)
  _ : forall g s s', func g s s' -> flats s = flats s'}.

Section ClassDef.

Notation class_of := mixin_of (only parsing). 
Notation base := (seedT -> reltp).

Structure refinement : Type := Pack {sort : base; _ : class_of sort; _ : base}.
Local Coercion sort : refinement >-> Funclass.

Variables (T : base) (cT : refinement).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack T c T.
Definition clone := fun c & T = cT & phant_id (pack c) cT => pack c.

End ClassDef.
End Refinement.

Module Exports.
Coercion sort : refinement >-> Funclass.
Notation refinement := refinement.
Notation RefinementMixin := Mixin.
Notation Refinement T m := (@pack _ _ _ T m).

(*
Notation "[ 'refinementMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'refinementMixin'  'of'  T ]") : form_scope.
Notation "[ 'refinement' 'of' T 'for' C ]" := (@clone T C _ (erefl _) id)
  (at level 0, format "[ 'refinement'  'of'  T  'for'  C ]") : form_scope.
Notation "[ 'refinement' 'of' T ]" := (@clone T _ _ (erefl _) id)
  (at level 0, format "[ 'refinement'  'of'  T ]") : form_scope.
*)

Section Repack.
Variables (T : Type) (U V : mod) (f : refinement T U V).

Lemma rf_func g s s1' s2' : f g s s1' -> f g s s2' -> s1' = s2'.
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: rf. Qed.

Lemma rf_inj g1 s1 g2 s2 s' : f g1 s1 s' -> f g2 s2 s' -> (g1, s1) = (g2, s2).
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: inj. Qed.

Lemma rf_guarsur g1 s1 s1' s2' : 
        f g1 s1 s1' -> Mod.internal V s1' s2' ->
        exists g2 s2, f g2 s2 s2'.
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: sr. Qed.

(*
Lemma rf_sur s' : Mod.coh V s' -> exists g s, f g s s'.
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: sr. Qed.
*)

Lemma rf_coh g s s' : f g s s' -> s \In Mod.coh U /\ s' \In Mod.coh V.
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: coh. Qed.

Lemma rf_mrely g s1 s1' s2 : 
        f g s1 s1' -> mrely U s1 s2 -> 
        exists s2', f g s2 s2' /\ mrely V s1' s2'.
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: rp. Qed.

Lemma rf_guar g1 s1 s1' g2 s2 s2' :
        f g1 s1 s1' -> f g2 s2 s2' -> 
        Mod.internal V s1' s2' -> Mod.internal U s1 s2. 
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: gp. Qed.

Lemma rf_env g1 s1 s1' g2 s2 s2' :
        f g1 s1 s1' -> f g2 s2 s2' -> 
        (env s1 = env s2 <-> env s1' = env s2').
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: ep. Qed.

Lemma rf_zagzig p :
        exists q, 
        [/\ forall g s s',
              f g (s \> p) s' -> 
              exists s1', s' = s1' \> q /\ f g (s \< p) (s1' \< q) &
            forall g s s',
              f g s (s' \< q) -> 
              exists s1, s = s1 \< p /\ f g (s1 \> p) (s' \> q)].
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: zg. Qed.

Lemma rf_flats g s s' : f g s s' -> flats s = flats s'. 
Proof. by case: f=>q [rf inj sr coh rp gp ep zg fl xx]; apply: fl. Qed.

End Repack.

Prenex Implicits rf_func rf_inj rf_coh rf_mrely rf_guar rf_env rf_flats. 

End Exports. 

End Refinement.

Export Refinement.Exports.

(* now we pack up the seeds with the refinement *)

Structure retracts W W' := Retracts {
  seedT : Type;
  seedL : seedT;
  ref_of : refinement seedT W W'}.

(**************************************************************************)
(* Our intended use for refinement is always a refinement of modules with *)
(* *private state*. For example, creating a lock requires taking a piece  *)
(* of private state containing a pointer for the lock, and the state of   *)
(* the resource, and moving them into the shared state (of the lock       *)
(* module). The refinement will also have to come up with auxiliary       *)
(* state, etc., but those correspond to the seed arguments.               *)
(*                                                                        *)
(* Thus, we instantiate the above general setup for refinement, to the    *)
(* specific case, when we are refining by moving private state into       *)
(* shared state.                                                          *)
(**************************************************************************)

(* We need a new structure to package the information on how to divide *)
(* into feap fealds, the heap that's transfered into the shared *)
(* state. Also, we need seed values for building the local and *)
(* environment auxiiliary fields. *)

(* We will have to build this kind of structure for any module *)
(* that we plan to use in priv-based retractions *)

Module Decorate.
Section Decorate.
Variables (W : mod) (T : Type).

Structure mixin_of (func : T -> state -> Prop) := Mixin {
  dec_prec_op : forall g h1 h2 s1 s2, 
    valid (flats s1 \+ h1) -> 
    flats s1 \+ h1 = flats s2 \+ h2 -> 
    func g s1 -> func g s2 -> 
    s1 = s2; 
  dec_inj : forall g1 g2 s, 
    func g1 s -> func g2 s -> 
    g1 = g2; 
  dec_coh : forall g s, 
    func g s -> s \In Mod.coh W;
  (* heh, a bit of a hack; we need to limit the 
     surjectivity only on stuff which has an env unit;
     but since that's a bit hard to state, we hack *)
  dec_onto : forall g1 s1 s2, 
    func g1 s1 -> s2 \In Mod.coh W -> env s1 = env s2 -> exists g2, func g2 s2;
  dec_env : forall g1 g2 s1 s2, 
    func g1 s1 -> func g2 s2 -> env s1 = env s2}.

Section ClassDef.

Notation class_of := mixin_of (only parsing). 
Notation base := (T -> state -> Prop).

Structure decorate : Type := Pack {sort : base; _ : class_of sort; _ : base}.
Local Coercion sort : decorate >-> Funclass.

Variables (B : base) (cT : decorate).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack B c B.
Definition clone := fun c & B = cT & phant_id (pack c) cT => pack c.

End ClassDef.

End Decorate.

Module Exports.
Coercion sort : decorate >-> Funclass.
Notation decorate := decorate.
Notation DecorateMixin := Mixin.
Notation Decorate B m := (@pack _ _ B m).

Section Repack.
Variables (W : mod) (T : Type) (d : decorate W T).

Lemma dec_prec g h1 h2 s1 s2 :
        valid (flats s1 \+ h1) -> 
        flats s1 \+ h1 = flats s2 \+ h2 ->
        d g s1 -> d g s2 -> s1 = s2.
Proof. by case: d=>x [dp di dc dsur de xx]; apply: dp. Qed.

Lemma dec_inj g1 g2 s : d g1 s -> d g2 s -> g1 = g2.
Proof. by case: d=>x [dp di dc ddo de xx]; apply: di. Qed.

Lemma dec_coh g s : d g s -> s \In Mod.coh W.
Proof. by case: d=>x [dp di dc ddo de xx]; apply: dc. Qed.

Lemma dec_onto g1 s1 s2 : 
        d g1 s1 -> s2 \In Mod.coh W -> env s1 = env s2 -> 
        exists g2, d g2 s2.
Proof. by case: d=>x [dp di dc ddo de xx]; apply: ddo. Qed.

(*
Lemma dec_flat g h s : d g h s -> flats s = h.
Proof. by case: d=>x [dp di dc df de xx]; apply: df. Qed.
*)

Lemma dec_env g1 g2 s1 s2 :
        d g1 s1 -> d g2 s2 -> env s1 = env s2.
Proof. by case: d=>x [dp di dc ddo de xx]; apply: de. Qed.

End Repack.

Prenex Implicits dec_coh.

End Exports.

End Decorate.

Export Decorate.Exports.


(*************************************************************************)
(* Proofs that given a module with decorate and mod_unit, we can build a *)
(* retraction wrt. priv for that module. *)
(*****************************************)

(* first, a bit of helper abstraction *)

Section CtxLemmas.
Variable (hp : nat) (coh : cohpred).
Variable (pf : leq_coh (priv_coh hp) coh). 
(* this pf uniformly includes priv_coh hp as well as all extension *)

Definition ctx (hL : heap) (hs : heap * state) := hp ->> [hL, tt, hs.1] \+ hs.2.

Lemma transp_ctxE hL hE s : transp (ctx hL (hE, s)) = ctx hE (hL, transp s).
Proof. by rewrite /ctx /= transpUn fsE /transp /= -fsE. Qed.

Lemma ctx_valid hL hs : ctx hL hs \In coh -> valid (hL \+ hs.1).
Proof.
case: hs=>hE s /cohH; rewrite validHUn priv_flatsE /=.
by move/andP/proj2/validL.
Qed.

Lemma ctx_inj hL1 hL2 hs1 hs2 :
        valid (ctx hL1 hs1) -> ctx hL1 hs1 = ctx hL2 hs2 -> 
        hL1 = hL2 /\ hs1 = hs2.
Proof.
case: hs1 hs2=>[hE1 s1][hE2 s2] V. 
by case/(scancel V)=>/=; case=>->-> _ ->. 
Qed.

Lemma ctx_onto s : s \In coh -> exists hL hs, s = ctx hL hs.
Proof.
case: pf=>c -> [_][s1][s2][->] C1 _. 
by case: C1=>_ [hL][hE] ->; exists hL, (hE, s2).
Qed.

Lemma ctx_shift coh' hL hE s s1 s2 : 
        ctx hL (hE, s) \In star_coh coh coh' -> s1 \In coh -> 
        ctx hL (hE, s) = s1 \+ s2  -> 
        exists s', s1 = ctx hL (hE, s') /\ s = s' \+ s2.
Proof.
case: pf=>c -> /cohV V [_][s3][s4][->{s1}][_][hL'][hE'] ->{s3} _. 
rewrite -joinA; case/(scancel V)=>/=; case=><-<- _ ->{hL' hE' V s}. 
by exists s4. 
Qed.

Lemma ctxeq_validV hL1 hL2 {hs} : 
        valid (ctx hL1 hs) = valid (ctx hL2 hs). 
Proof. by rewrite /valid /= !fsE /= !fvalidPtUn. Qed.

Lemma ctxeq_validH hL1 hL2 {hs} :
        footeqH hL1 hL2 ->
        validH (ctx hL1 hs) = validH (ctx hL2 hs).
Proof.
move=>F; rewrite !validHUn !(ctxeq_validV _ hL1) /=. 
by rewrite !priv_flatsE -!joinA -(footeqH_validUnL F). 
Qed.

Lemma ctxeq_validSO hL1 hL2 {hs} :
        footeqH hL1 hL2 ->
        validSO (ctx hL1 hs) = validSO (ctx hL2 hs).
Proof.
move=>F; rewrite /validSO /= !fsE /= !zipUnPt.
by rewrite !allvalidPtUn (footeqH_validUnL F). 
Qed.

Lemma ctxeq_coh hL1 hL2 hs :
        footeqH hL1 hL2 -> 
        (ctx hL1 hs \In coh <-> ctx hL2 hs \In coh). 
Proof.
suff {hL1 hL2} L hL1 hL2 : footeqH hL1 hL2 -> 
  ctx hL1 hs \In coh -> ctx hL2 hs \In coh. 
- by move=>H; split; [apply: L H | apply: L (footeqH_sym H)].
case: hs=>hE s F; case: pf=>c -> [Vh][s1][s2][E] C1 C2.
case: C1=>Vh' [hL'][hE'] E'; rewrite {s1}E' in E Vh'.
case/(scancel (validHV Vh)): E Vh' C2.
case=>/= <-<- _ <- {hL' hE' s2} Vh' C2. 
split; first by rewrite -(ctxeq_validH F).
exists (ctx hL2 (hE, Unit)), s.
rewrite /ctx /= unitR; split=>//.
split; last by exists hL2, hE. 
move: (@ctxeq_validH _ _ (hE, Unit) F). 
by rewrite /ctx /= !unitR; move=><-.
Qed.

Lemma ctxH hL hE h1 h2 s : 
        validH (ctx (hL \+ h1) (hE \+ h2, s)) ->
        validH (ctx hL (hE, s)).
Proof.
rewrite !validHUn !priv_flatsE /=.
case/andP=>V1 V2; apply/andP; split.
- by move: V1; rewrite /valid /= !fsE !fvalidPtUn.  
move: V2; rewrite -!joinA -(joinCA h1) -!(joinCA h2). 
by do 2![move/validR]. 
Qed.

Lemma ctxHL hL hE h1 s : 
        validH (ctx (hL \+ h1) (hE, s)) ->
        validH (ctx hL (hE, s)).
Proof. by rewrite -{1}[hE]unitR; move/ctxH. Qed.

Lemma ctxHR hL hE h2 s : 
        validH (ctx hL (hE \+ h2, s)) ->
        validH (ctx hL (hE, s)).
Proof. by rewrite -{1}[hL]unitR; move/ctxH. Qed.

Lemma ctxCL hL h hs : 
        ctx (hL \+ h) hs \In coh -> ctx hL hs \In coh. 
Proof.
case: pf=>c -> /get_priv [C1 C2 /ctxHL Vh]; split=>//. 
exists (hp ->> [hL, tt, hs.1]), hs.2; split=>//.
have V : valid (hL \+ hs.1). 
- by move/cohH: C1; rewrite /validH !priv_flatsE joinAC; move/validL.
split; last first.
- by exists hL, hs.1. 
by rewrite /validH priv_flatsE.
Qed.

Lemma ctx_cohL hL h hs :  
        validH (ctx (hL \+ h) hs) ->
        ctx hL hs \In coh -> ctx (hL \+ h) hs \In coh.
Proof.
case: pf=>c -> Vh /get_priv [C1 C2 Vh']; split=>//.
exists (hp ->> [hL \+ h, tt, hs.1]), hs.2; split=>//.
split; first by move/validHL: Vh. 
by exists (hL \+ h), hs.1. 
Qed.

Lemma ctxCR hL h hE s: 
        ctx hL (hE \+ h, s) \In coh -> ctx hL (hE, s) \In coh. 
Proof.
move/transp_coh=>C; apply/transp_coh.
by rewrite !transp_ctxE in C *; apply: ctxCL C.
Qed.

End CtxLemmas.

(**************************************)
(* ctx function and guar constructors *)
(**************************************)

Section CtxGuar. 
Variables (hp : nat) (coh : cohpred).
Variables (pf : leq_coh (priv_coh hp) coh) (coh' : cohpred).
Notation ctx := (@ctx hp).

Variables (G : guar coh) (G' : guar coh'). 

Lemma priv_step hL1 hL2 hE1 s1 hE2 s2 : 
        frame_rel coh coh' G (ctx hL1 (hE1, s1)) (ctx hL2 (hE2, s2)) <->
        exists s1' s2' t, 
          [/\ s1 = s1' \+ t, s2 = s2' \+ t, 
              validH (ctx hL1 (hE1, s1)) /\ validH (ctx hL2 (hE2, s2)),
              t \In coh' & G (ctx hL1 (hE1, s1')) (ctx hL2 (hE2, s2'))].
Proof.
split.
- case; case=>s1' [s][s2'][E1 E2 H] C1 C2.
  have C : s \In coh'. 
  - by rewrite E1 in C1; rewrite -(star_coh_prec C1); case/guar_coh: H.  
  case/(ctx_shift pf C1 (proj1 (guar_coh H))): E1=>t1 [??]; subst s1' s1.
  case/(ctx_shift pf C2 (proj2 (guar_coh H))): E2=>t2 [??]; subst s2' s2.
  by exists t1, t2, s; split=>//; move/cohH: C1; move/cohH: C2.
case=>s1' [s2'][t][->-> [Vh1 Vh2] C H]; case/guar_coh: (H)=>C1 C2.
do !split=>//. 
- by exists (ctx hL1 (hE1, s1')), t, (ctx hL2 (hE2, s2')); rewrite /ctx !joinA. 
- by exists (ctx hL1 (hE1, s1')), t; rewrite /ctx !joinA. 
by exists (ctx hL2 (hE2, s2')), t; rewrite /ctx !joinA. 
Qed.

Lemma other_step hL1 hL2 hE1 s1 hE2 s2 : 
        frame_rel coh' coh G' (ctx hL1 (hE1, s1)) (ctx hL2 (hE2, s2)) <->
        exists s t1 t2,
          [/\ s1 = s \+ t1, s2 = s \+ t2, hL1 = hL2 /\ hE1 = hE2, 
              validH (ctx hL1 (hE1, s1)) /\ validH (ctx hL2 (hE2, s2)),
              ctx hL1 (hE1, s) \In coh & G' t1 t2].
Proof.
split. 
- case; case=>s1' [t1][s2'][E1] E2 H C1 C2. 
  rewrite starC in C1 C2; rewrite -!(joinC t1) in E1 E2.
  have C : t1 \In coh. 
  - by rewrite E1 in C1; rewrite (star_coh_prec C1); case/guar_coh: H. 
  case/(ctx_shift pf C1 C): E1=>s [??]; subst t1 s1.
  case/(ctx_shift pf C2 C): E2=>t2 [E ?]; subst s2.
  case/(scancel (cohV C)): E C2; case=>/= <-<- _ <-{hL2 hE2 t2} C2.
  by exists s, s1', s2'; do !split=>//; [move/cohH: C1 | move/cohH: C2].
case=>s [t1][t2][->->{s1 s2}][<-<-][Vh1 Vh2] C H.  
split.
- exists t1, (ctx hL1 (hE1, s)), t2; rewrite /ctx /= !joinA.
  by split=>//; rewrite -joinAC -joinA joinCA joinA.
- split=>//; exists t1, (ctx hL1 (hE1, s)). 
  rewrite /ctx /= (joinC s) joinCA; split=>//.
  by case/guar_coh: H. 
split=>//; exists t2, (ctx hL1 (hE1, s)). 
rewrite /ctx /= (joinC s) joinCA; split=>//.
by case/guar_coh: H. 
Qed.

Variables (Ga : acq_guar coh) (Ga' : rel_guar coh').

Lemma acqrel_step hL1 hL2 hE1 hE2 s1 s2 :
        acqrel_rel coh coh' [eta Ga] [eta Ga'] (ctx hL1 (hE1, s1))
                                               (ctx hL2 (hE2, s2)) <->
        exists k t1 t2 r1 r2, 
           [/\ s1 = t1 \+ r1, s2 = t2 \+ r2, 
               validH (ctx hL1 (hE1, s1)) /\ validH (ctx hL2 (hE2, s2)), 
               Ga k (ctx hL1 (hE1, t1)) (ctx hL2 (hE2, t2)) &
               Ga' k r1 r2].
Proof. 
split. 
- move=>H; case/guar_coh: (H)=>C1 C2.
  case: H; case=>t1 [r1][t2][r2][E1 E2][k][G1 G2] _.
  exists k. 
  case/(ctx_shift pf C1 (proj1 (guar_coh G1))): E1=>t1' [??]; subst t1 s1.
  case/(ctx_shift pf C2 (proj2 (guar_coh G1))): E2=>t2' [??]; subst t2 s2.
  by exists t1', t2', r1, r2; split=>//; move/cohH: C1; move/cohH: C2.
case=>k [t1][t2][r1][r2][->-> [Vh1 Vh2] G1 G2]. 
split.
- exists (ctx hL1 (hE1, t1)), r1, (ctx hL2 (hE2, t2)), r2. 
  by split; last by [exists k]; rewrite /ctx /= joinA.
split=>//.
exists (ctx hL1 (hE1, t1)), r1; split; first by rewrite /ctx /= joinA.
- by case/guar_coh: G1. 
by case/guar_coh: G2. 
Qed.

Variables (Gr : rel_guar coh) (Gr' : acq_guar coh').

Lemma relacq_step hL1 hL2 hE1 hE2 s1 s2 :
        acqrel_rel coh' coh [eta Gr'] [eta Gr] (ctx hL1 (hE1, s1)) 
                                               (ctx hL2 (hE2, s2)) <->
        exists k t1 t2 r1 r2, 
           [/\ s1 = t1 \+ r1, s2 = t2 \+ r2, 
               validH (ctx hL1 (hE1, s1)) /\ validH (ctx hL2 (hE2, s2)), 
               Gr k (ctx hL1 (hE1, t1)) (ctx hL2 (hE2, t2)) &
               Gr' k r1 r2].
Proof.
split. 
- move=>H; case/guar_coh: (H)=>C1 C2.
  case: H; case=>t1 [r1][t2][r2][E1 E2][k][G1 G2] _.
  rewrite starC in C1 C2.
  rewrite -(joinC r1) -(joinC r2) in E1 E2.
  case/(ctx_shift pf C1 (proj1 (guar_coh G2))): E1=>t1' [??]; subst r1 s1. 
  case/(ctx_shift pf C2 (proj2 (guar_coh G2))): E2=>t2' [??]; subst r2 s2.
  exists k, t1', t2', t1, t2; split=>//.
  by move/cohH: C1; move/cohH: C2.
case=>k [t1][t2][r1][r2][->-> [Vh1 Vh2] G1 G2].
split.
- exists r1, (ctx hL1 (hE1, t1)), r2, (ctx hL2 (hE2, t2)).
  by split; last by [exists k]; rewrite /ctx /= joinA joinAC -joinA joinCA. 
split=>//; exists r1, (ctx hL1 (hE1, t1)); split.
- by rewrite /ctx /= joinA joinAC -joinA joinCA.
- by case/guar_coh: G2. 
by case/guar_coh: G1. 
Qed.

End CtxGuar.

(*************************************************************************) 
(* Properties of modules that extend Priv by entanglement.  These lemmas *)
(* make the entanglement module "behave like Priv" wrt. retraction.      *)
(*************************************************************************) 

Section PrivEntangleLemmas.
Variables (hp : nat) (V : mod).
Notation U := (entangle (Priv hp) V).
Notation ctx := (ctx hp).
Notation ctx_shift := (ctx_shift (leq_coh_refl _)).

Lemma mrely_reduce hL h hs1 hs2 :
          mrely U (ctx (hL \+ h) hs1) (ctx (hL \+ h) hs2) -> 
          mrely U (ctx hL hs1) (ctx hL hs2).
Proof.
move: hL h hs1 hs2.
pose guar_reduce_ax (M : mod) := forall hL h hE1 s1 hE2 s2,
  [/\ Mod.internal M (ctx hE1 (hL \+ h, s1)) (ctx hE2 (hL \+ h, s2)) ->
      Mod.internal M (ctx hE1 (hL, s1)) (ctx hE2 (hL, s2)) &
      [/\ forall ch k, ch \In Mod.chans M ->
            get acq ch k  (ctx hE1 (hL \+ h, s1)) (ctx hE2 (hL \+ h, s2)) ->
            get acq ch k (ctx hE1 (hL, s1)) (ctx hE2 (hL, s2)) &
          forall ch k, ch \In Mod.chans M ->
            get rel ch k  (ctx hE1 (hL \+ h, s1)) (ctx hE2 (hL \+ h, s2)) ->
            get rel ch k (ctx hE1 (hL, s1)) (ctx hE2 (hL, s2))]].
have priv_guar_reduce : guar_reduce_ax (Priv hp). 
- move=>hL h hE1 s1 hE2 s2; 
  split; [|split; move=>ch k /Mem_map_inv [ct][->]; rewrite InE =>->{ct}];
  move=>R; have [E1 E2] : s1 = Unit /\ s2 = Unit
    by [case/guar_coh: R; rewrite /= -[Mod.coh _]starC0; 
        do 2![case/get_priv=>/= _ -> _]];
  rewrite {}E1 {}E2 in R *.
  - case: R; case=>hl [hl'][he][V1 V2 V' D] C.  
    split; last by apply: ctxCR C. 
    exists hl, hl', hL; rewrite /ctx /= !unitR !fsE in V1 V2 *.
    by case/stE: V1=>/= -> _ _; case/stE: V2=>/= ->.
  - case: R; case=>hl [he][E1 E2 Vl C]. 
    split; last by apply: ctxCR C. 
    exists hl, hL; rewrite /ctx /= !unitR !fsE in E1 E2 *.
    case/stE: E1=>/= -> _ _; case/stE: E2 Vl=>/= -> _.
    by move/(fcancelPt (C:=pcmCode))=><-; rewrite joinA; move/validL. 
  case: R; case=>hl [he][E1 E2 C]. 
  split; last by apply: ctxCR C.
  exists hl, hL; rewrite /ctx /= !unitR !fsE in E1 E2 *.
  by case/stE: E1=>/= -> _ _; case/stE: E2=>/= ->.
have entangle_reduce : guar_reduce_ax U. 
- split; last first.
  - split=>_ k /Mem_map_inv [ct][->]; rewrite InE => ->{ct};
    case/priv_step=>// s1' [s2'][t][E1 E2 [/ctxHR V1 /ctxHR V2] C] H;
    apply/priv_step=>//; exists s1', s2', t; split=>//;
    case: (priv_guar_reduce hL h hE1 s1' hE2 s2')=>_ [].
    - by move/(_ (Chan (PrivChanCore hp)) k (or_introl (erefl _)) H).  
    by move=>_ /(_ (Chan (PrivChanCore hp)) k (or_introl (erefl _)) H).  
  move=>R; case: R=>R; [left | right].
  - case: R=>R; [left | right].
    - case/priv_step: R=>// s1' [s2'][t][E1 E2 [/ctxHR Vh1 /ctxHR Vh2] C] H. 
      apply/priv_step=>//; exists s1', s2', t; split=>//. 
      by case: (priv_guar_reduce hL h hE1 s1' hE2 s2'); move/(_ H). 
    case/other_step: R=>// s [t1][t2][E1 E2][<-{hE2} _]
                             [/ctxHR Vh1 /ctxHR Vh2] C G.
    by apply/other_step=>//; exists s, t1, t2; split=>//; apply: ctxCR C.
  case: R=>ch1 [ch2][Ch1 Ch2] R; exists ch1, ch2; split=>//.
  case/Mem_map_inv: Ch1 R=>ct [->{ch1}] /mem_chan_eq Ch1 R.
  case: R=>R; [left | right].
  - case/acqrel_step: R=>// k [t1][t2][r1][r2][E1 E2 
                              [/ctxHR V1 /ctxHR V2]] G1 G2.
    apply/acqrel_step=>//; exists k, t1, t2, r1, r2; split=>//.
    case: (priv_guar_reduce hL h hE1 t1 hE2 t2)=>_ [].
    by move/(_ (Chan ct) k Ch1 G1).
  case/relacq_step: R=>// k [t1][t2][r1][r2][E1 E2 [/ctxHR V1 /ctxHR V2] G1 G2].
  apply/relacq_step=>//; exists k, t1, t2, r1, r2; split=>//.
  case: (priv_guar_reduce hL h hE1 t1 hE2 t2)=>_ [_].
  by move/(_ (Chan ct) k Ch1 G1). 
move=>hL h [hE1 s1][hE2 s2]; rewrite /mrely !transp_ctxE.
case: (entangle_reduce hL h hE1 (transp s1) hE2 (transp s2))=>T1 [T2 T3].
case=>H; first by left; apply: T1 H. 
case: H=>ch [g][Ch] H; right; exists ch, g; split=>//.
by case: H=>H; [left; apply: T2 H | right; apply: T3 H].
Qed.

Lemma guar_extend hL1 hL2 h hs1 hs2 :
          ctx (hL1 \+ h) hs1 \In Mod.coh U ->
          ctx (hL2 \+ h) hs2 \In Mod.coh U ->
          Mod.internal U (ctx hL1 hs1) (ctx hL2 hs2) ->
          Mod.internal U (ctx (hL1 \+ h) hs1) (ctx (hL2 \+ h) hs2).
Proof.
move: hL1 hL2 h hs1 hs2.
pose guar_extend_ax (M : mod) := forall hL1 hL2 h hs1 hs2,
  ctx (hL1 \+ h) hs1 \In Mod.coh M ->
  ctx (hL2 \+ h) hs2 \In Mod.coh M ->
  [/\ Mod.internal M (ctx hL1 hs1) (ctx hL2 hs2) ->
      Mod.internal M (ctx (hL1 \+ h) hs1) (ctx (hL2 \+ h) hs2) &
      [/\ forall ch k, ch \In Mod.chans M -> 
            get acq ch k (ctx hL1 hs1) (ctx hL2 hs2) -> 
            get acq ch k (ctx (hL1 \+ h) hs1) (ctx (hL2 \+ h) hs2) &
          forall ch k, ch \In Mod.chans M -> 
            get rel ch k (ctx hL1 hs1) (ctx hL2 hs2) -> 
            get rel ch k (ctx (hL1 \+ h) hs1) (ctx (hL2 \+ h) hs2)]].
have priv_guar_extend : guar_extend_ax (Priv hp).
- move=>hL1 hL2 h [hE1 s1][hE2 s2] H1 H2.
  split; [|split=>_ k /Mem_map_inv [ct][->]; rewrite InE => ->]; 
  move=>H; case/guar_coh: (H)=>/cohV V1 /cohV V2.
  - case: H; case=>hL1' [hL2'][hE'][E1 E2 _ E'] _.
    rewrite -[hp ->> [_, _, hE']]unitR in E1.
    rewrite -[hp ->> [_, _, hE']]unitR in E2.
    case/(scancel V1): E1 E2 H1 H2 E';case=>/= <-<- _ ->{hL1' hE' V1 s1};
    case/(scancel V2); case=>/= <--> _ ->{hL2' hE2 V2 s2} H1 H2 E'.
    split=>//; exists (hL1 \+ h), (hL2 \+ h), hE1. 
    have [V1 V2] : valid (hL1 \+ h) /\ valid (hL2 \+ h).
    - by move/ctx_valid/validL: H1; move/ctx_valid/validL: H2.
    by rewrite /ctx /= !unitR; split=>// x; rewrite !domUn /= !inE E' V1 V2. 
  - case: H; case=>hL1' [hE'][E1 E2 _] _.
    rewrite -[hp ->> [_, _, hE']]unitR in E1.
    rewrite -[hp ->> [_, _, hE']]unitR in E2.
    case/(scancel V1): E1 E2 H1 H2; case=>/= <-<-_ ->{hL1' hE' V1 s1}. 
    case/(scancel V2); case=>/= ->-> _ ->{hL2 V2 hE2 s2} H1 H2. 
    split=>//; exists (hL1 \+ h), hE1.
    by rewrite /ctx /= !unitR -(joinAC _ k); split=>//; move/ctx_valid: H2. 
  case: H; case=>hL1' [hE'][E1 E2] _.
  rewrite -[hp ->> [_, _, hE']]unitR in E1.
  rewrite -[hp ->> [_, _, hE']]unitR in E2.
  case/(scancel V1): E1 E2 H1 H2; case=>/= -><- _ ->{hL1 V1 hE' s1}.
  case/(scancel V2); case=>/= <-<- _ ->{hL1' hE1 V2 s2} H1 H2. 
  by split=>//; exists (hL2 \+ h), hE2; rewrite (joinAC _ k) /ctx /= !unitR.
have entangle_guar_extend : guar_extend_ax U.
- move=>hL1 hL2 h [hE1 s1][hE2 s2] H1 H2. 
  split; last first.
  (* strictly speaking, this case is unnecessary, but I can't *)
  (* bother now to reformulate the statements *)
  - split=>_ k /Mem_map_inv [ct][->]; rewrite InE => ->;
    case/priv_step=>// t1 [t2][t][?? _ C G]; subst s1 s2;
    apply/priv_step=>//; exists t1, t2, t;
    (split=>//; first by move/cohH: H1; move/cohH: H2);
    have [K1 K2] : ctx (hL1 \+ h) (hE1, t1) \In Mod.coh (Priv hp) /\
                   ctx (hL2 \+ h) (hE2, t2) \In Mod.coh (Priv hp)
      by [rewrite /ctx !joinA in H1 H2;
          rewrite (star_coh_prec H1) (star_coh_prec H2)];
    case: (priv_guar_extend _ _ _ _ _ K1 K2)=>_ []; [|move=>_];
    by move/(_ (Chan (PrivChanCore hp)) k (or_introl (erefl _)) G).
  case=>H; [left | right]; case/guar_coh: (H)=>C1 C2.
  - case: H=>H; [left | right]. 
    case/priv_step: H=>// t1 [t2][t][?? _ C G]; subst s1 s2.
    apply/priv_step=>//; exists t1, t2, t.
    split=>//; first by move/cohH: H1; move/cohH: H2.
    have [K1 K2] : ctx (hL1 \+ h) (hE1, t1) \In Mod.coh (Priv hp) /\
                   ctx (hL2 \+ h) (hE2, t2) \In Mod.coh (Priv hp).
    - rewrite /ctx /= !joinA in H1 H2.
      by rewrite (star_coh_prec H1) (star_coh_prec H2). 
    by case: (priv_guar_extend _ _ _ _ _ K1 K2); move/(_ G).
  - case/other_step: H=>// s [t1][t2][?? [??] _] C G; subst s1 s2 hL2 hE2.
    apply/other_step=>//; exists s, t1, t2.
    split=>//; first by move/cohH: H1; move/cohH: H2.
    rewrite /ctx /= !joinA in H1; rewrite (star_coh_prec H1).
    by case/guar_coh: G. 
  case: H=>ch1 [ch2][Ch1 Ch2] H; exists ch1, ch2; split=>//. 
  case/Mem_map_inv: Ch1 H=>ct1 [->{ch1}] /mem_chan_eq Ch1 H.
  case/Mem_map_inv: Ch2 H=>ct2 [->{ch2}] /mem_chan_eq Ch2 H.
  case: H=>H; [left | right].
  - case/acqrel_step: H=>// k [t1][t2][r1][r2][?? _ G1 G2]; subst s1 s2.
    apply/acqrel_step=>//; exists k, t1, t2, r1, r2.
    split=>//; first by move/cohH: H1; move/cohH: H2. 
    have [K1 K2] : ctx (hL1 \+ h) (hE1, t1) \In Mod.coh (Priv hp) /\
                   ctx (hL2 \+ h) (hE2, t2) \In Mod.coh (Priv hp).
    - rewrite /ctx /= !joinA in H1 H2.
      by rewrite (star_coh_prec H1) (star_coh_prec H2); case/guar_coh: G2.
    case: (priv_guar_extend _ _ _ _ _ K1 K2)=>_ [].
    by move/(_ (Chan ct1) k Ch1 G1).
  case/relacq_step: H=>// k [t1][t2][r1][r2][?? _ G1 G2]; subst s1 s2.
  apply/relacq_step=>//; exists k, t1, t2, r1, r2.
  split=>//; first by move/cohH: H1; move/cohH: H2.
  have [K1 K2] : ctx (hL1 \+ h) (hE1, t1) \In Mod.coh (Priv hp) /\
                 ctx (hL2 \+ h) (hE2, t2) \In Mod.coh (Priv hp).
  - rewrite /ctx /= !joinA in H1 H2.
    by rewrite (star_coh_prec H1) (star_coh_prec H2); case/guar_coh: G2.
  case: (priv_guar_extend _ _ _ _ _ K1 K2)=>_ [_]. 
  by move/(_ (Chan ct1) k Ch1 G1).
move=>hL1 hL2 h hs1 hs2 C1 C2.
by case: (entangle_guar_extend _ _ _ _ _ C1 C2). 
Qed.

Lemma foot_internal hL1 hL2 hs : 
        footeqH hL1 hL2 -> 
        ctx hL1 hs \In Mod.coh U ->
        Mod.internal U (ctx hL1 hs) (ctx hL2 hs).
Proof.
move: hL1 hL2 hs.
pose foot_internal_ax (M : mod) := forall hL1 hL2 hE hs, 
  footeqH hL1 hL2 ->
  ctx hL1 (hE, hs) \In Mod.coh M ->
  Mod.internal M (ctx hL1 (hE, hs)) (ctx hL2 (hE, hs)).
have priv_foot_internal : foot_internal_ax (Priv hp).
- move=>hL1 hL2 hE hs F C. 
  split=>//; case: C=>Vh [hL'][hE'] E.
  rewrite -[hp ->> [_, _, _]]unitR in E.
  case/(scancel (validHV Vh)): E=>/= _ _ ?; subst hs. 
  exists hL1, hL2, hE; case: F=><- F.
  rewrite /ctx !unitR; split=>//; move: Vh.
  by rewrite /validH /ctx /= unitR priv_flatsE; move/validL. 
move=>hL1 hL2 [hE hs] F C.
left; left; case: (C)=>Vh [s1][s2][E] C1 C2.
case/(ctx_shift C C1): E=>s [??]; subst s1 hs.
move/(priv_foot_internal _ _ _ _ F): C1=>H; split=>//.
- exists (ctx hL1 (hE, s)), s2, (ctx hL2 (hE, s)).
  by rewrite /ctx /= !joinA. 
by apply/(ctxeq_coh _ _ F)=>//; exists (Mod.coh V).
Qed.

Lemma acq_internal hL1 hL2 h1 h2 k hs1 hs2 ch : 
        footeqH h1 (k \+ h2) -> 
        ctx (hL1 \+ h1) hs1 \In Mod.coh U ->
        ctx (hL2 \+ h2) hs2 \In Mod.coh U ->
        ch \In Mod.chans U ->
        get acq ch k (ctx hL1 hs1) (ctx hL2 hs2) ->
        Mod.internal U (ctx (hL1 \+ h1) hs1) (ctx (hL2 \+ h2) hs2).
Proof.
move: hL1 hL2 h1 h2 k hs1 hs2 ch.
pose acq_internal_ax (M : mod) := forall hL1 hL2 h1 h2 k hE1 s1 hE2 s2 ch,
  footeqH h1 (k \+ h2) -> 
  ctx (hL1 \+ h1) (hE1, s1) \In Mod.coh M ->
  ctx (hL2 \+ h2) (hE2, s2) \In Mod.coh M ->
  ch \In Mod.chans M ->
  get acq ch k (ctx hL1 (hE1, s1)) (ctx hL2 (hE2, s2)) ->
  Mod.internal M (ctx (hL1 \+ h1) (hE1, s1)) (ctx (hL2 \+ h2) (hE2, s2)).
have priv_acq_internal : acq_internal_ax (Priv hp).
- move=>hL1 hL2 h1 h2 k hE1 s1 hE2 s2 /= ch F C1 C2.
  case/Mem_map_inv=>ct [->{ch}]; rewrite InE=>->{ct} G.
  case/guar_coh: (G)=>C1' C2'.
  case: G; case=>hL [hE][E1 E2] _ _; split=>//.
  rewrite -[hp ->> [_, _, _]]unitR in E1.
  rewrite -[hp ->> [_, _, _]]unitR in E2. 
  case/(scancel (cohV C1')): E1; case=>/= ?? _ ?; subst hL hE s1. 
  case/(scancel (cohV C2')): E2; case=>/= ?? _ ?; subst hL2 hE2 s2. 
  exists (hL1 \+ h1), (hL1 \+ k \+ h2), hE1.
  have Vl : valid (hL1 \+ h1).
  - by move/cohH: C1; rewrite /validH /ctx unitR priv_flatsE; move/validL. 
  have F' : footeqH (hL1 \+ h1) (hL1 \+ (k \+ h2)).
  - by apply: footeqHUnR F.
  by case: F'=>E1 F1; rewrite /ctx !unitR /= -!joinA -E1 Vl. 
move=>hL1 hL2 h1 h2 k [hE1 s1][hE2 s2] ch F C1 C2.
case/Mem_map_inv=>ct [->{ch}]; rewrite InE=>->{ct} G. 
case/guar_coh: (G)=>C1' C2'.
case: G; case=>t1 [s][t2][E1 E2] C H1 H2.   
case/(ctx_shift C1' (proj1 (guar_coh C))): E1=>t1' [??]; subst t1 s1.
case/(ctx_shift C2' (proj2 (guar_coh C))): E2=>t2' [??]; subst t2 s2.
have [K1 K2] : ctx (hL1 \+ h1) (hE1, t1') \In Mod.coh (Priv hp) /\
               ctx (hL2 \+ h2) (hE2, t2') \In Mod.coh (Priv hp).
- rewrite /ctx /= !joinA in C1 C1' C2 C2'.
  rewrite (star_coh_prec C1) -(star_coh_prec C1').
  rewrite (star_coh_prec C2) -(star_coh_prec C2').
  by case/guar_coh: C.
left; left; split=>//.
exists (ctx (hL1 \+ h1) (hE1, t1')), s, (ctx (hL2 \+ h2) (hE2, t2')).
rewrite /ctx /= !joinA; split=>//.
by apply: priv_acq_internal F K1 K2 (or_introl (erefl _)) C. 
Qed.

Lemma rel_internal hL1 hL2 h1 h2 k hs1 hs2 ch : 
        footeqH (h1 \+ k) h2 -> 
        ctx (hL1 \+ h1) hs1 \In Mod.coh U ->
        ctx (hL2 \+ h2) hs2 \In Mod.coh U ->
        ch \In Mod.chans U ->
        get rel ch k (ctx hL1 hs1) (ctx hL2 hs2) ->
        Mod.internal U (ctx (hL1 \+ h1) hs1) (ctx (hL2 \+ h2) hs2).
Proof.
move: hL1 hL2 h1 h2 k hs1 hs2 ch.
pose rel_internal_ax (M : mod) := forall hL1 hL2 h1 h2 k hE1 s1 hE2 s2 ch,
  footeqH (h1 \+ k) h2 -> 
  ctx (hL1 \+ h1) (hE1, s1) \In Mod.coh M ->
  ctx (hL2 \+ h2) (hE2, s2) \In Mod.coh M ->
  ch \In Mod.chans M ->
  get rel ch k (ctx hL1 (hE1, s1)) (ctx hL2 (hE2, s2)) ->
  Mod.internal M (ctx (hL1 \+ h1) (hE1, s1)) (ctx (hL2 \+ h2) (hE2, s2)).
have priv_rel_internal : rel_internal_ax (Priv hp).
- move=>hL1 hL2 h1 h2 k hE1 s1 hE2 s2 ch F C1 C2.
  case/Mem_map_inv=>ct [->{ch}]; rewrite InE =>->{ct} G.
  case/guar_coh: (G)=>C1' C2'.
  case: G; case=>hL [hE][E1 E2] _; split=>//.
  rewrite -[hp ->> [_, _, _]]unitR in E1.
  rewrite -[hp ->> [_, _, _]]unitR in E2. 
  case/(scancel (cohV C1')): E1; case=>/= ?? _ ?; subst hL1 hE s1. 
  case/(scancel (cohV C2')): E2; case=>/= ?? _ ?; subst hL2 hE2 s2. 
  exists (hL \+ k \+ h1), (hL \+ h2), hE1.
  have Vl : valid (hL \+ h2).
  - by move/cohH: C2; rewrite /validH /ctx unitR priv_flatsE; move/validL. 
  have F' : footeqH (hL \+ (k \+ h1)) (hL \+ h2).
  - by apply: footeqHUnR; rewrite joinC.
  by case: F'=>_ F1; rewrite /ctx !unitR /= -!joinA Vl.
move=>hL1 hL2 h1 h2 k [hE1 s1][hE2 s2] ch F C1 C2.
case/Mem_map_inv=>ct [->{ch}]; rewrite InE=>->{ct} G. 
case/guar_coh: (G)=>C1' C2'.
case: G; case=>t1 [s][t2][E1 E2] C H1 H2.   
case/(ctx_shift C1' (proj1 (guar_coh C))): E1=>t1' [??]; subst t1 s1.
case/(ctx_shift C2' (proj2 (guar_coh C))): E2=>t2' [??]; subst t2 s2.
have [K1 K2] : ctx (hL1 \+ h1) (hE1, t1') \In Mod.coh (Priv hp) /\
               ctx (hL2 \+ h2) (hE2, t2') \In Mod.coh (Priv hp).
- rewrite /ctx /= !joinA in C1 C1' C2 C2'.
  rewrite (star_coh_prec C1) -(star_coh_prec C1').
  rewrite (star_coh_prec C2) -(star_coh_prec C2').
  by case/guar_coh: C.
left; left; split=>//.
exists (ctx (hL1 \+ h1) (hE1, t1')), s, (ctx (hL2 \+ h2) (hE2, t2')).
rewrite /ctx /= !joinA; split=>//.
by apply: priv_rel_internal F K1 K2 (or_introl (erefl _)) C. 
Qed.

End PrivEntangleLemmas.

(********************************)
(* Finally, the main properties *)
(********************************)

Module RefinePrivEntangle.
Section RefinePrivEntangle.
Variables (hp : nat) (V W : mod) (T : Type).
Variables (decor : decorate W T). (*  (e : mod_unit W). *)
Notation U := (entangle (Priv hp) V).
Notation W' := (entangle U W). 
Notation ctx := (ctx hp).

Definition ref_rel (g : T) (s1 s2 : state) :=
  exists hL hs s,
    [/\ s1 = ctx (hL \+ flats s) hs,
        s2 = ctx hL hs \+ s, decor g s, 
        s1 \In Mod.coh U & s2 \In Mod.coh W'].

Lemma refrel_func g s s1' s2' : 
        ref_rel g s s1' -> ref_rel g s s2' -> s1' = s2'.
Proof. 
case=>hL1 [t1][f1][->->{s s1'} D1 _ C1]. 
case=>hL2 [t2][f2][E ->{s2'} D2 C C2].
case/(ctx_inj (cohV C)): E C2=>E <-{t2} C2.
have Vl : valid (hL1 \+ flats f1).
- by move/ctx_valid/validL: C. 
rewrite -(joinC (flats f1)) -(joinC (flats f2)) in E Vl.
move: (dec_prec Vl E D1 D2)=>?; subst f2.
by rewrite (joinfK Vl E).
Qed.

Lemma refrel_inj g1 s1 g2 s2 s' : 
        ref_rel g1 s1 s' -> ref_rel g2 s2 s' -> (g1, s1) = (g2, s2). 
Proof.
case=>hL1 [t1][f1][/= ->-> D1 C1 C].
case=>hL2 [t2][f2][/= -> E D2 C2 _].
move/(ctxCL star_id): C1=>C1.
move/(ctxCL star_id): C2; move/(coh_prec (cohV C) E C1). 
case/(ctx_inj (cohV C1))=>??; subst hL2 t2.
move/(joinsK (cohV C)): E=>?; subst f2.
by rewrite (dec_inj D1 D2).
Qed.

(*
Lemma refrel_onto s2 : s2 \In Mod.coh W' -> exists g s1, ref_rel g s1 s2.
Proof.
move=>C; case: C (C)=>_ [t1][t2][->{s2} C1 C2] C. 
case/(ctx_onto star_id): C1 C=>hL [hs] ->{t1} C.
case: (dec_onto decor C2)=>g D.
exists g, (ctx (hL \+ flats t2) hs), hL, hs, t2; split=>//.
apply: (ctx_cohL star_id); last by apply/(star_coh_prec C).
move: (cohH C).
rewrite /validH (flatsUn (cohV C)) (flatsUn (validL (cohV C))). 
rewrite flatsUn; last first.
- by move: (cohV C); move/validL; rewrite /PCM.valid /= !fsE !fvalidPtUn. 
by rewrite !priv_flatsE -2!(joinAC _ (flats t2)).
Qed.
*)

Lemma refrel_coh g s1 s2 : 
        ref_rel g s1 s2 -> s1 \In Mod.coh U /\ s2 \In Mod.coh W'.
Proof. by case=>hL [hs][f][->->]. Qed.

Lemma refrel_mrely g s1 s1' s2 :
        ref_rel g s1 s1' -> mrely U s1 s2 -> 
        exists s2', ref_rel g s2 s2' /\ mrely W' s1' s2'.
Proof.
case=>hL [t1][r][->->{s1 s1'} D1 C1 C1'] R.
set h := flats r.
have [t2 [S2 E]] : exists t2, s2 = ctx (hL \+ h) t2 /\ footeqV t2.2 t1.2.
- case/mrely_coh: R (R) (mrely_footV R)=>_ C.
  case: C (C)=>_ [s'][s][-> C' C].
  case: C'=>_ [hL'][hE] ->{s'} /cohV/and3P [Vl _ _] /mrely_loc /esym. 
  rewrite /ctx !fsE /= in Vl *.
  case/(fcancel (C:=pcmCode) Vl)=>-> _ _ {Vl hL'}.
  rewrite -!fsE => E. 
  exists (hE, s); split=>//=; rewrite footeqV_sym.
  by apply: footeqVK (cohV C1) E _. 
rewrite {s2}S2 in R *.
set s1 := ctx (hL \+ _) t1 in C1 R *.
set s2 := ctx (hL \+ _) t2 in R *. 
set k1 := ctx hL t1 in C1' *.
set k2 := ctx hL t2.
have V2': valid (k2 \+ r). 
- move/cohV: C1'; rewrite -!joinA -(joinCA t1.2) -(joinCA t2.2).
  by rewrite !fsE (footeqV_validUnL E) !(joinCA t1.2) /valid /= !fvalidPtUn. 
have Vh2' : validH (k2 \+ r). 
- move/mrely_coh/proj2/cohH: R; rewrite !validHUn (flatsUn (validL V2')) V2' /=.
  by rewrite !priv_flatsE !(joinAC _ h); case/andP.
have C2' : k2 \+ r \In Mod.coh W'. 
- split=>//; exists k2, r; split=>//. 
  - by apply: (ctxCL star_id) (proj2 (mrely_coh R)).
  by apply/(star_coh_prec C1'); apply: (ctxCL star_id) C1.  
exists (k2 \+ r).
split; first by exists hL, t2, r; split=>//; case/mrely_coh: R.
by apply: mrely_entangle_extend C1' C2' _; apply: mrely_reduce R.
Qed.

Lemma refrel_guar g1 s1 s1' g2 s2 s2' : 
        ref_rel g1 s1 s1' -> ref_rel g2 s2 s2' -> 
        Mod.internal W' s1' s2' -> Mod.internal U s1 s2.
Proof.
case=>hL [t][r][->-> D1 C1 C2][hL'][t'][r'][->-> D1' C1' C2'].
case/entangle_split.
- by apply: (ctxCL star_id C1).
- by apply: (ctxCL star_id C1').
- case=>G ?; subst r'; move: (dec_inj D1 D1')=>?; subst g2.
  by apply: guar_extend C1 C1' G.
- case; case/(ctx_inj (cohV (ctxCL star_id C1)))=>??; subst hL' t'.
  move/int_guar_footH=>H. 
  by apply: foot_internal (footeqHUnR H) C1.
case=>ch1 [ch2][hh][Ch1 Ch2]; case; case=>Gp Gl.
- move/rel_guar_footH: Gl => H. 
  by apply: acq_internal H C1 C1' Ch1 Gp.
move/acq_guar_footH: Gl=>H. 
by apply: rel_internal H C1 C1' Ch1 Gp.
Qed.

Lemma refrel_sur g1 s1 s1' s2' : 
        ref_rel g1 s1 s1' -> Mod.internal W' s1' s2' -> 
        exists g2 s2, ref_rel g2 s2 s2'.
Proof.
case=>hL [t][r][->->{s1 s1'} D1 C1 C1'] G. 
set h := flats r in C1.
case/guar_coh: G (G). move=>_ C.
case: C (C)=>_ [s'][s][->{s2'} C' C].
move=>D.
move/entangle_split.
have H: hp ->> [hL, tt, t.1] \+ t.2 \In Mod.coh U.
- apply: ctxCL C1. simpl. by apply: star_id.
  move/(_ H C').
case.
Focus 2.
- case=>? G. subst s'.
  move: dec_onto. move/(_ _ _ _ _ _ _ D1). 
  case/guar_coh: (G). move=>_ Cs. move/(_ _ Cs).
  move/(_ (guar_env G)). case=>g2 D2.
  exists g2. exists (ctx (hL \+ flats s) t).
  rewrite /ref_rel.
  exists hL. exists t. exists s. split=>//.
  have X: footeqH (hL \+ h) (hL \+ flats s).
  - apply: footeqHUnR. apply: int_guar_footH G.
  move/ctxeq_coh: X=><- //.  
  by apply: star_id.

- case. move=>G ?. subst s.
  exists g1. 
  suff [hL' [t' X]]: exists hL' t', s' = ctx hL' t'.
  - subst s'.  
    exists (ctx (hL' \+ flats r) t').
    rewrite /ref_rel. simpl.
    exists hL'. exists t'. exists r. split=>//.
    apply: ctx_cohL=>//.
    rewrite /validH.
    rewrite flatsUn; last first.
    - have X : footeqV (hp ->> [hL', tt, t'.1]) (hp ->> [hL' \+ flats r, tt, t'.1]).
      - by apply: footeqVPt.
      move: (@footeqV_validUnL). move/(_ _ _ _ X).  move=><-. 
      by move/validL: (cohV D).
    rewrite priv_flatsE.
    move/cohH: (D). 
    rewrite /validH. 
    rewrite flatsUn; last first.
    - by apply: cohV D.
    rewrite /ctx. simpl.
    rewrite flatsUn; last first.
    - by apply: cohV C'.
    rewrite priv_flatsE. rewrite joinAC.
    by rewrite -(joinAC _ t'.1).
  case/guar_coh: G=>_.
  apply: ctx_onto.
  by apply: star_id.

case=>ch1 [ch2][hh][Ch1 Ch2]. case; case=>Gp Gl.
- move: dec_onto. move/(_ _ _ _ _ _ _ D1 C).
  move/(_ (guar_env Gl)). case=>g2 D2.
  exists g2. 
  suff [hL' [t' X]]: exists hL' t', s' = ctx hL' t'.
  - subst s'.
    exists (ctx (hL' \+ flats s) t').
    exists hL'. exists t'. exists s. split=>//.
    apply: ctx_cohL C'.
    - by apply: star_id.
    rewrite /validH.
    rewrite flatsUn; last first.
    - have X : footeqV (hp ->> [hL', tt, t'.1]) (hp ->> [hL' \+ flats s, tt, t'.1]).
      - by apply: footeqVPt.
      move: (@footeqV_validUnL). move/(_ _ _ _ X). move=><-.
      by move/validL: (cohV D).
    rewrite priv_flatsE.
    move/cohH: (D).
    rewrite /validH.
    rewrite flatsUn; last first.
    - by apply: cohV D.
    rewrite /ctx. simpl. rewrite flatsUn; last first.
    - move: (cohV D). move/validL. by [].
    rewrite priv_flatsE. rewrite joinAC. rewrite -(joinAC _ t'.1). by [].
  apply: ctx_onto C'. by apply: star_id.

move: dec_onto. move/(_ _ _ _ _ _ _ D1 C).
move/(_ (guar_env Gl)). case=>g2 D2.
exists g2. 
suff [hL' [t' X]]: exists hL' t', s' = ctx hL' t'.
- subst s'.
    exists (ctx (hL' \+ flats s) t').
    exists hL'. exists t'. exists s. split=>//.
    apply: ctx_cohL C'.
    - by apply: star_id.
    rewrite /validH.
    rewrite flatsUn; last first.
    - have X : footeqV (hp ->> [hL', tt, t'.1]) (hp ->> [hL' \+ flats s, tt, t'.1]).
      - by apply: footeqVPt.
      move: (@footeqV_validUnL). move/(_ _ _ _ X). move=><-.
      by move/validL: (cohV D).
    rewrite priv_flatsE.
    move/cohH: (D).
    rewrite /validH.
    rewrite flatsUn; last first.
    - by apply: cohV D.
    rewrite /ctx. simpl. rewrite flatsUn; last first.
    - move: (cohV D). move/validL. by [].
    rewrite priv_flatsE. rewrite joinAC. rewrite -(joinAC _ t'.1). by [].
  apply: ctx_onto C'. by apply: star_id.
Qed.


Lemma refrel_env g1 s1 s1' g2 s2 s2' :
        ref_rel g1 s1 s1' -> ref_rel g2 s2 s2' -> 
        (env s1 = env s2 <-> env s1' = env s2').
Proof.
case=>hL [t][r][->->{s1 s1'} D C1 C2] /=.
case=>hL' [t'][r'][->->{s2 s2'} D' C1' C2'].
rewrite /ctx !fsE /=.
split=>[|E]; first by move: (dec_env D D')=>->->.
have Vl : valid (env (ctx hL t \+ r)) by case/cohV/and3P: C2. 
rewrite /ctx fsE /= in Vl.
rewrite /= -(joinC (env r)) -(joinC (env r')) in Vl E. 
suff L : env r = env r'.
- by rewrite -L in E; rewrite (funhKl Vl E).
suff : footeqV r r' by case/and3P=>_ _ /fdomeqP [Vx]; apply: fdomE Vl E.
have X1: r' \In Mod.coh W by move/(star_coh_prec C2'): (ctxCL star_id C1').
have X2 : r \In Mod.coh W by move/(star_coh_prec C2): (ctxCL star_id C1).
by apply: coh_footV X2 X1.
Qed.

Lemma refrel_zagzig p : 
        exists q, 
        [/\ forall g s1 s2, 
              ref_rel g (s1 \> p) s2 -> 
              exists s2', s2 = s2' \> q /\ ref_rel g (s1 \< p) (s2' \< q) &
            forall g s1 s2, 
              ref_rel g s1 (s2 \< q) ->
              exists s1', s1 = s1' \< p /\ ref_rel g (s1' \> p) (s2 \> q)].
Proof.
set e := sig (Mod.coh W).
exists (p \+ e); split=>gL s1 s2 [hL][[hE] t][r][/=]. 
- move=>E -> D C'.
  have Vl : valid (hp \-> hE \+ env t).
  - by case/cohV/and3P: C'=>_ _; rewrite E /ctx fsE /=. 
  move: E C'; rewrite -[s1]st_eta /ctx /= !fsE; case/stE=>/= ->->. 
  case/(zipPtUnf Vl)=>u1 [u2][t1][t2] /= [->] Et ->-> {p s1} C C'. 
  have Cr : r \In Mod.coh W by move/dec_coh: D.
  have Ve1 : valid (env r \* e).
  -  by rewrite /e (sig_env Cr) zunitR; case/cohV/and3P: Cr. 
  have Ve2 : valid (e \* loc r).
  - by rewrite /e (sig_loc Cr) zunitL; case/cohV/and3P: Cr. 
  have Vt : valid (t1 \* t2) by rewrite -Et; move/validR: Vl.
  exists (ctx hL (u1, [loc t, shr t, t1]) \+ r); rewrite -!joinA.
  split. 
  - rewrite -fsE !zagUnPt /= !fsE !zagE /= zipPt -zipUn //=.
    by rewrite /e (sig_env Cr) zunitR -Et. 
  exists (u2 \+ hL), (u1, [t2 \* loc t, shr t, t1]), r. 
  split=>//; last 2 first.
  - by move/coh_zigzag: C. 
  - rewrite coh_zigzag !zagUnPt fsE !zagE !zipPt -zipUn //=.
    by rewrite -Et /e (sig_env Cr) zunitR joinA. 
  - by rewrite -joinA /ctx fsE zigE zipUnPt.  
  rewrite /= zigUnPt !fsE !zigE zipPt -zipUn //=. 
  - by rewrite -joinA /e (sig_loc Cr) zunitL -fsE.  
  suff: valid (loc t \* env t) by rewrite zipC Et -zipA; case/validZP. 
  suff: validSO t by move/allvalidV. 
  suff: t \In Mod.coh V by  move/cohSO. 
  suff: ctx hL (u1 \+ u2, t) \In Mod.coh U by case/get_priv.
  by move/(star_coh_prec C'): Cr; rewrite -fsE. 
move=>->{s1} E D C C'.
have Vpe : valid (p  \+ e) by move/cohV: C'; case/valid_zig.
rewrite E in C'; rewrite joinC in Vpe.
have Vl : valid (hp \-> hL \+ loc t \+ loc r).
- by case/cohV/and3P: C'; rewrite /ctx fsE.
rewrite -[s2]st_eta /ctx fsE in E C' *; case/stE: E C' (Vl) C (E)=>/=. 
case/(zipUnf Vl)=> f1 [g1][f2][g2][/esym]. 
case/(zipUnf (validL Vl))=>f3 [g3][f4][g4][/esym].
case/zipPtf=>u1 [u2][?->->{f3 g3}] E1 ->->{f1 g1}; subst hL. 
move=>E2 E3 ->->-> C' _ C E; rewrite (joinC p) -(joinC f2) in E3.
have Cr : r \In Mod.coh W by move/dec_coh: D.
have Vr : valid (loc r) by case/cohV/and3P: Cr. 
have Z : [zunit f2] = [zunit g2] by rewrite E2 in Vr; case/validZP: Vr. 
have X : e = f2 :> feap _.
- apply: fdomE Vpe (E3) _=>x.
  by rewrite /e (sig_loc Cr) fdomZU E2 fdomZL -E2 Vr.
subst f2; move/(funhKl Vpe): E3=>?; subst p.
rewrite /e (sig_loc Cr) zunit_zunit in E2 Z; rewrite Z zunitL in E2.
subst g2.
exists (ctx (u2 \+ flats r) (hE, [g4, shr t, env t])).
split; first by rewrite zigUnPt !fsE !zigE zipPt joinA -E1. 
have Ct: t \In Mod.coh V by case/get_priv: C.
exists u2, (hE \+ u1, [g4, shr t, env t \* f4]), r.
split=>//.
- by rewrite zagUnPt /ctx !fsE !zagE zipPt.  
- rewrite zagE -!joinA zipUnPt fsE -zipUn.
  - by rewrite /e (sig_env Cr) zunitR. 
  - by move/cohSO/allvalidV: Ct; rewrite E1 zipC zipA; case/validZP.
  by rewrite /e (sig_env Cr) zunitR; case/cohV/and3P: Cr. 
- rewrite -coh_zigzag zigUnPt /= !fsE !zigE /= !zipPt.
  by rewrite -E1 st_eta joinA -fsE.  
rewrite -coh_zigzag zigE -!joinA !zipUnPt -zipUn; last first.
- by rewrite /e (sig_loc Cr) zunitL. 
- by rewrite -E1; case/cohV/and3P: Ct. 
by rewrite /e (sig_loc Cr) zunitL -E1 !joinA. 
Qed.

Lemma refrel_flat g s1 s2 : ref_rel g s1 s2 -> flats s1 = flats s2.
Proof.
case=>hL [hs][s][->->] D C1 C2; move: (cohV C1) (cohV C2)=>V1 V2.
have V3 : valid (ctx hL hs) by move/(ctxCL star_id)/cohV: C1. 
by rewrite !flatsUn // !priv_flatsE -2!(joinAC _ (flats s)).
Qed.

End RefinePrivEntangle.

Module Exports.
Section Exports.
Variables (hp : nat) (V W : mod) (T : Type).
Variables (decor : decorate W T). (* (e : mod_unit W). *)

Definition PrivEntangleRefineMixin := 
  RefinementMixin (@refrel_func hp V W T decor)
                  (@refrel_inj hp V W T decor)
                  (@refrel_sur hp V W T decor) 
                  (@refrel_coh hp V W T decor)
                  (@refrel_mrely hp V W T decor)
                  (@refrel_guar hp V W T decor)
                  (@refrel_env hp V W T decor)
                  (@refrel_zagzig hp V W T decor)
                  (@refrel_flat hp V W T decor).
Canonical PrivEntangleRefine := 
  Eval hnf in Refinement (ref_rel hp V decor) PrivEntangleRefineMixin.

End Exports.
End Exports.
End RefinePrivEntangle.

Export RefinePrivEntangle.Exports.
