Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype.
Require Import Eqdep prelude finmap unionmap heap idynamic pcm codes feaps.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* we make state store feaps with codes, so we can implement flattening *)
(* which collects the heap components *)

Notation dpcm := (@idynamic pcm_code (to_type \o pcm_sort)). 
Notation dtp := (@idynamic tp_code (to_type \o tp_sort)). 

Structure state := 
  State {loc : feap dpcm; shr : feap dtp; env : feap dpcm}.

Notation "[ lp , sh , ep ]" := (State lp sh ep) 
  (at level 0, format "[ lp ,  sh ,  ep ]").

(* notation for various states *)

Module Type FS_sig.
Section Defs. 
Variables (U : encoded_pcm) (V : encoded_type).

Parameter fullstate : nat -> to_type U -> to_type V -> to_type U -> state.
Axiom fsE : forall x l s e, fullstate x l s e = [x \-> l, x ::-> s, x \-> e].

Parameter selfstate : nat -> to_type U -> state.
Axiom sfE : forall x l, selfstate x l = [x \-> l, Unit, Unit].

Parameter jointstate : nat -> to_type V -> state.
Axiom jnE : forall x s, jointstate x s = [Unit, x ::-> s, Unit].

Parameter otherstate : nat -> to_type U -> state.
Axiom otE : forall x e, otherstate x e = [Unit, Unit, x \-> e]. 

End Defs.
End FS_sig.

Module FS : FS_sig.
Section Defs. 
Variables (U : encoded_pcm) (V : encoded_type).
Variables (x : nat) (l : U) (s : V) (e : U).

Definition fullstate := [x \-> l, x ::-> s, x \-> e].
Definition fsE := erefl fullstate. 

Definition selfstate := [x \-> l, Unit, Unit].
Definition sfE := erefl selfstate.

Definition jointstate := [Unit, x ::-> s, Unit].
Definition jnE := erefl jointstate.

Definition otherstate := [Unit, Unit, x \-> e].
Definition otE := erefl otherstate.

End Defs.
End FS.

Export FS.

Notation "x ->> [ l , s , e ]" := (fullstate x l s e) (at level 30).
Notation "x '@sf' -> l" := (selfstate x l) (at level 40, format "x '@sf' ->  l").
Notation "x '@jn' -> l" := (jointstate x l) (at level 40, format "x '@jn' ->  l").
Notation "x '@ot' -> l" := (otherstate x l) (at level 40, format "x '@ot' ->  l").

Lemma st_eta (s : state) : [loc s, shr s, env s] = s.
Proof. by case: s. Qed.

Lemma stE (s1 s2 : state) : 
        s1 = s2 <-> [/\ loc s1 = loc s2, shr s1 = shr s2 & env s1 = env s2].
Proof. by case: s1 s2=>lp1 sh1 ep1 [lp2 sh2 ep2] /=; split; case=>->->->. Qed.

(* flattening collects all heap components from feaps in the state *)

Module Type flats_Sig.
Parameter flats : state -> heap. 
Axiom flatsE : forall s, 
  flats s = @flat pcmFlatCode (loc s) \+ 
            @flat typeFlatCode (shr s) \+ 
            @flat pcmFlatCode (env s). 
End flats_Sig.

Module flats_Def : flats_Sig.

Definition flats s := @flat pcmFlatCode (loc s) \+ 
                      @flat typeFlatCode (shr s) \+ 
                      @flat pcmFlatCode (env s). 

Definition flatsE s := (erefl (flats s)).

End flats_Def.

Export flats_Def.

(* monoidal properties *)

Module StatePCM.
Section StatePCM.

Let svalid s := [&& valid (loc s), valid (shr s) & valid (env s)].

Let sjoin s1 s2 := [loc s1 \+ loc s2, shr s1 \+ shr s2, env s1 \+ env s2].

Let sunit := [Unit, Unit, Unit].

Lemma joinC s1 s2 : sjoin s1 s2 = sjoin s2 s1.
Proof. by congr State; rewrite joinC. Qed.

Lemma joinA s1 s2 s3 : sjoin s1 (sjoin s2 s3) = sjoin (sjoin s1 s2) s3.
Proof. by congr State; rewrite joinA. Qed.

Lemma validL s1 s2 : svalid (sjoin s1 s2) -> svalid s1. 
Proof. by rewrite /svalid /=; case/and3P; do 3![move/validL=>->]. Qed.

Lemma unitL s : sjoin sunit s = s.
Proof. by rewrite /sjoin !unitL st_eta. Qed.

Lemma valid_unit : svalid sunit.
Proof. by []. Qed.

End StatePCM.
End StatePCM.

Definition statePCMMixin := 
  PCMMixin StatePCM.joinC StatePCM.joinA StatePCM.unitL 
           StatePCM.validL StatePCM.valid_unit. 
Canonical statePCM := Eval hnf in PCM state statePCMMixin.

(* boolean test if state is empty *)

Definition sempb (s : state) := 
  [&& fempb (loc s), fempb (shr s) & fempb (env s)].

(* domain of a state *)

Definition sdom (s : state) := 
  [pred x | [|| x \in fdom (loc s), x \in fdom (shr s) | x \in fdom (env s)]].

Lemma sdomUn s1 s2 : 
        valid (s1 \+ s2) -> sdom (s1 \+ s2) =i [predU sdom s1 & sdom s2].
Proof.
move=>V x; apply/idP/idP; [|case/and3P: V=>V1 V2 V3];
rewrite /sdom !inE /= !fdomUn !inE /=.
- by case/or3P; case/andP=>V1; case/orP=>->; rewrite ?orbT. 
by rewrite V1 V2 V3 /=; case/orP; case/or3P=>->; rewrite ?orbT.
Qed.

(* random needed lemmas *)

Lemma flats_unit : flats Unit = Unit. 
Proof. by rewrite flatsE /= !flat_unit !unitR. Qed.

Lemma flatsUn (s1 s2 : state) : 
        valid (s1 \+ s2) -> 
        flats (s1 \+ s2) = flats s1 \+ flats s2.
Proof.
case/and3P=>V1 V2 V3; rewrite !flatsE /= !flatUn V1 V2 V3.
rewrite -!joinA -(joinCA (@flat typeFlatCode (shr s1))).
by rewrite -2!(joinCA (@flat pcmFlatCode (env s1))).
Qed.

Lemma joinKs (s1 s2 s : state) : 
        valid (s1 \+ s) -> s1 \+ s = s2 \+ s -> s1 = s2.
Proof.
case/and3P=>V1 V2 V3; case/stE=>H1 H2 H3; apply/stE.
by rewrite (funKhl V1 H1) (funKhl V2 H2) (funKhl V3 H3).
Qed.

Lemma joinsK (s1 s2 s : state) : 
        valid (s \+ s1) -> s \+ s1 = s \+ s2 -> s1 = s2.
Proof. by rewrite !(joinC s); apply: joinKs. Qed.

(* cancellation modulewise *)
Section ModulewiseCancellation.
Variables (U : encoded_pcm) (V : encoded_type).

Lemma scancelPt x (l1 l2 e1 e2 : U) (t1 t2 : V) :
        x ->> [l1, t1, e1] = x ->> [l2, t2, e2] -> 
        (l1, t1, e1) = (l2, t2, e2).
Proof.
rewrite !fsE; case/stE=>/= /(@fcancelPt pcmCode)=>->.
by move/(@fcancelPt typeCode)=>-> /(@fcancelPt pcmCode) ->.
Qed.

Lemma scancel x (l1 l2 e1 e2 : U) (t1 t2 : V) s1 s2 :
        valid (x ->> [l1, t1, e1] \+ s1) ->
        x ->> [l1, t1, e1] \+ s1 = x ->> [l2, t2, e2] \+ s2 -> 
        [/\ (l1, t1, e1) = (l2, t2, e2), valid s1 & s1 = s2].
Proof.
rewrite !fsE; case/and3P=>/= V1 V2 V3; case/stE=>/=.
case/(fcancel (C:=pcmCode) V1)=>-> {V1} V1 E1. 
case/(fcancel (C:=typeCode) V2)=>-> {V2} V2 E2.
case/(fcancel (C:=pcmCode) V3)=>-> {V3} V3 E3.  
by split=>//; [apply/and3P | apply/stE].
Qed.

Lemma st0E (s1 s2 : state) : s1 \+ s2 = Unit -> s1 = Unit /\ s2 = Unit.
Proof. by rewrite -[s1]st_eta -[s2]st_eta; case; do 3![case/fun0E=>->->]. Qed.

End ModulewiseCancellation.

(* cancellation and flattening by components *)
(* should eventually build a canonical struct for these *)

Section ComponentsCancellation.
Variables (U : encoded_pcm) (V : encoded_type).

Lemma sfcancelPt x (l1 l2 : U) :
        x@sf -> l1 = x@sf -> l2 -> l1 = l2.
Proof. by rewrite !sfE; case/stE=>/= /(@fcancelPt pcmCode). Qed.

Lemma sfcancel x (l1 l2 : U) s1 s2 :
        valid (x@sf -> l1 \+ s1) ->
        x@sf -> l1 \+ s1 = x@sf -> l2 \+ s2 ->
        [/\ l1 = l2, valid s1 & s1 = s2].
Proof.
rewrite !sfE; case/and3P=>/= V1 V2 V3 /stE [/=]. 
rewrite !unitL in V1 V2 V3 *.
case/(fcancel (C:=pcmCode) V1)=>-> {V1} V1 E1 E2 E3. 
by split=>//; [apply/and3P | apply/stE]. 
Qed.

Lemma jncancelPt x (l1 l2 : V) :
        x@jn -> l1 = x@jn -> l2 -> l1 = l2.
Proof. by rewrite !jnE; case/stE=>/= _ /(@fcancelPt typeCode). Qed.

Lemma jncancel x (l1 l2 : V) s1 s2 :
        valid (x@jn -> l1 \+ s1) ->
        x@jn -> l1 \+ s1 = x@jn -> l2 \+ s2 ->
        [/\ l1 = l2, valid s1 & s1 = s2].
Proof.
rewrite !jnE; case/and3P=>/= V1 V2 V3 /stE [/=].
rewrite !unitL in V1 V2 V3 *.
move=>E1 /(fcancel (C:=typeCode) V2) [->] {V2} V2 E2 E3.
by split=>//; [apply/and3P | apply/stE]. 
Qed.

Lemma otcancelPt x (l1 l2 : U) :
        x@ot -> l1 = x@ot -> l2 -> l1 = l2.
Proof. by rewrite !otE; case/stE=>/= _ _ /(@fcancelPt pcmCode). Qed.

Lemma otcancel x (l1 l2 : U) s1 s2 :
        valid (x@ot -> l1 \+ s1) ->
        x@ot -> l1 \+ s1 = x@ot -> l2 \+ s2 ->
        [/\ l1 = l2, valid s1 & s1 = s2].
Proof.
rewrite !otE; case/and3P=>/= V1 V2 V3 /stE [/=].
rewrite !unitL in V1 V2 V3 *.
move=>E1 E2 /(fcancel (C:=pcmCode) V3) [->] {V3} E3.
by split=>//; [apply/and3P | apply/stE]. 
Qed.

Lemma sfflats x (l : U) : flats (x@sf -> l) = flat (x \-> l).
Proof. by rewrite flatsE sfE !flat_unit !unitR. Qed.

Lemma jnflats x (l : V) : flats (x@jn -> l) = flat (x ::-> l).
Proof. by rewrite flatsE jnE !flat_unit !unitL !unitR. Qed.

Lemma otflats x (l : U) : flats (x@ot -> l) = flat (x \-> l).
Proof. by rewrite flatsE otE !flat_unit !unitL. Qed.

End ComponentsCancellation.

(* in case we're flattening components with heap fields *)
Lemma sfflatsE x (l : heap) : flats (x@sf -> l) = l.
Proof. by rewrite sfflats pcm_flatPtE. Qed.

Lemma jnflatsE x (l : heap) : flats (x@jn -> l) = l.
Proof. by rewrite jnflats tp_flatPtE. Qed.

Lemma otflatsE x (l : heap) : flats (x@ot -> l) = l.
Proof. by rewrite otflats pcm_flatPtE. Qed.

(* filtering *)

Definition sfilter (p : pred nat) (s : state) := 
  locked [ffilter p (loc s), ffilter p (shr s), ffilter p (env s)].

Lemma sfiltUn p (s1 s2 : state) :
        valid (s1 \+ s2) ->
        sfilter p (s1 \+ s2) = sfilter p s1 \+ sfilter p s2.
Proof. by case/and3P=>V1 V2 V3; rewrite /sfilter -!lock !ffiltUn. Qed.

Lemma sfilt0 p : sfilter p Unit = Unit.
Proof. by rewrite /sfilter -lock /= !ffilt0. Qed.

Lemma sfiltPt (U : encoded_pcm) (V : encoded_type) p 
              (x : nat) (lp ep : U) (sh : V) : 
        sfilter p (x ->> [lp, sh, ep]) = 
        if p x then x ->> [lp, sh, ep] else Unit.
Proof.
rewrite fsE /sfilter -lock !(ffiltPt (C:=pcmCode)) (ffiltPt (C:=typeCode)).
by case: ifP.
Qed.

Lemma sfilt_pred0 s : valid s -> sfilter pred0 s = Unit.
Proof. by case/and3P=>V1 V2 V3; rewrite /sfilter -lock !ffilt_pred0. Qed.

Lemma sfilt_predT s : sfilter predT s = s.
Proof. by rewrite /sfilter -lock !ffilt_predT st_eta. Qed.

Lemma sfilt_predI (p1 p2 : pred nat) s : 
        sfilter [predI p1 & p2] s = sfilter p1 (sfilter p2 s).
Proof. by rewrite /sfilter -!lock !ffilt_predI. Qed.

Lemma sfilt_predU (p1 p2 : pred nat) s : 
        sfilter [predU p1 & p2] s = sfilter p1 s \+ sfilter [predD p2 & p1] s.
Proof. by rewrite /sfilter -!lock !ffilt_predU. Qed.

Lemma eq_in_sfilter (p1 p2 : pred nat) s : 
        {in sdom s, p1 =i p2} -> sfilter p1 s = sfilter p2 s.
Proof.
move=>S; rewrite /sfilter -!lock. 
apply/stE; split=>/=; apply: eq_in_ffilter;
by move=>x H; apply: S; rewrite /sdom !inE /= H ?orbT. 
Qed. 

(* i need to iterate filtering, so i need to package the *)
(* validity condition *)

Definition pfilter (p : pred nat) (s : state) (pf : valid s) := sfilter p s.

Notation "[ 'filter' p s ]" := (@pfilter p s _)
  (at level 0, format "[ 'filter'  p  s ]").

Lemma pfiltUn p (s1 s2 : state) (pf : valid (s1 \+ s2)) :
        pfilter p pf = pfilter p (validL pf) \+ pfilter p (validR pf).
Proof. by rewrite /pfilter sfiltUn. Qed.

Lemma pfilt0 p pf : @pfilter p Unit pf = Unit.
Proof. by rewrite /pfilter sfilt0. Qed.

Lemma pfiltPt (U : encoded_pcm) (V : encoded_type) p 
              (x : nat) (lp ep : U) (sh : V) pf : 
        @pfilter p (x ->> [lp, sh, ep]) pf = 
        if p x then x ->> [lp, sh, ep] else Unit.
Proof. by rewrite /pfilter sfiltPt. Qed.

Lemma pfiltPtsf (U : encoded_pcm) p (x : nat) (lp : U) pf :
        @pfilter p (x@sf->lp) pf = 
        if p x then x@sf->lp else Unit.
Proof.
rewrite /pfilter /sfilter -lock sfE (ffiltPt (C:=pcmCode)) !ffilt0. 
by case: ifP.
Qed.

Lemma pfiltPtjn (V : encoded_type) p (x : nat) (sh : V) pf :
        @pfilter p (x@jn->sh) pf = 
        if p x then x@jn->sh else Unit.
Proof.
rewrite /pfilter /sfilter -lock jnE (ffiltPt (C:=typeCode)) !ffilt0. 
by case: ifP.
Qed.

Lemma pfiltPtot (U : encoded_pcm) p (x : nat) (ep : U) pf :
        @pfilter p (x@ot->ep) pf = 
        if p x then x@ot->ep else Unit.
Proof.
rewrite /pfilter /sfilter -lock otE (ffiltPt (C:=pcmCode)) !ffilt0. 
by case: ifP.
Qed.

Lemma filt_conv p s (pf : valid s) : sfilter p s = pfilter p pf.
Proof. by []. Qed.

Definition filtE := (pfiltUn, pfilt0, pfiltPt, 
                     pfiltPtsf, pfiltPtjn, pfiltPtot, 
                     unitL, unitR).

Module SillyExample.

Lemma xx x : [x \-> 3, x ::-> (@Unit heapPCM), x \-> 3] = 
             [x \-> 3, x ::-> (@Unit heapPCM), x \-> 3].
Proof. by []. Qed.

End SillyExample.
