Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype finmap.
Require Import Eqdep prelude idynamic pcm codes unionmap heap feaps state.
Require Import zigzag worlds inject rely retract actions auto_split.
Require Import privstate footprint model lemmas locks.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(***********************************************)
(* Modules for ticketed lock                   *)
(***********************************************)

(***************************** 
  Reference implementation: 
 ***************************** 

lock(x) {
<int tck := INCR_NEXT(x.next);> 
while(! TEQ(tck, x.owner)) skip;
}

// Changes ghosts of the counting module
INCR_NEXT(z) = <tmp = z; z := !z + 1; tmp>

// TEQ - Ticked EQual
// can be verified only if i's value is in the local ghost map
// If positive, transfers the shared 
TEQ(i, z) = (i == z)

unlock(x) { 
// Works as unlock, transfers shared state back and increments the owner of the lock.
INCR_OWNER(x.owner);
}

makelock(n) {
local x := alloc(n+2); 
(x+1).owner := 0; 
(x+1).next := 0; 
return (x+1);
}

*****************************)

Structure tlockT := TLockT {
  owner_of : ptr; 
  next_of : ptr;
  lockish :> lockLike
}.

Notation nat_set := (fset [ordType of nat]).

Module TLockCoh.
Section TLockCoh.
Variable lkT: tlockT. 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)

(* A valid ticket dispenser *)
Definition validDispenser P s :=
  exists (n1 n2 : nat) (fL fE : nat_set) (gL gE : U) (b : bool) (h : heap),     
    [/\ s = dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, (fE, gE)], 
        n1 <= n2 & P n1 n2 (fL \+ fE) (gL \+ gE) b h].  

Definition cohRes P := 
  [Pred s : state | [/\ validH s, validSO s & validDispenser P s]].

(* locked state: *)
(* - the called ticket is taken (n1 \in dom f), *)
(* - and is being served (b = true), *)
(* - the resource heap is given away (h = Unit) *)
Definition locked n1 (f : nat_set) b (h : heap) := 
  [/\ n1 \in dom f, b = true & h = Unit].

(* transitional state: *)
(* - the called ticket is taken (n1 \in dom f), *)
(* - but is not served yet (b = false and n1 < n2), *)
(* - the resource heap is not given away (I g h) *)
Definition transit n1 n2 (f : nat_set) (g : U) b h := 
  [/\ n1 \in dom f, b = false, n1 < n2 & I g h].

(* empty queue: *)
(* - the called ticked hasn't been issued yet (n1 = n2) *)
(* - the state must unlocked (b = false) *)
(* - the resource heap is not given away (I g h) *)
Definition unlocked (n1 n2 : nat) (g : U) b h := 
  [/\ n1 = n2, b = false & I g h].

Definition either n1 n2 (f : nat_set) g b h :=
  dom f =i [pred x | n1 <= x < n2] /\
  [\/ locked n1 f b h | if n1 < n2 then transit n1 n2 f g b h 
                        else unlocked n1 n2 g b h]. 

Definition coh := cohRes either. 

Lemma cohSO P s : s \In cohRes P -> validSO s. 
Proof. by case. Qed.

Lemma cohH P s : s \In cohRes P -> validH s. 
Proof. by case. Qed.

Lemma cohfV P s : s \In cohRes P -> footV_eq s (mem_seq' [:: dl]).
Proof. by case=>_ _ [n1][n2][fL][fE][gL][gE][b][h][->]; rewrite fsE. Qed.

Lemma coh_weq P s s' : 
        s \In cohRes P -> s' \In cohRes P -> weq typeCode (shr s) (shr s').
Proof.
case=>_ _ [n1][n2][fL][fE][gL][gE][b][h][->] _ _.
case=>_ _ [n1'][n2'][fL'][fE'][gL'][gE'][b'][h'][->] _ _.
by rewrite !fsE /=.
Qed.

Lemma coh_sig P s : 
        s \In cohRes P -> dl \-> (Unit : nat_set * U) = [zunit loc s].
Proof. 
by case=>_ _ [n1][n2][fL][fE][gL][gE][b][h][->]; rewrite fsE zunitPt. 
Qed.

Lemma coh_zigzag P s p : s \< p \In cohRes P <-> s \> p \In cohRes P.
Proof.
suff L : validDispenser P (s \< p) <-> validDispenser P (s \> p).
- split; case=>Vh Vso Vp.
  - by split; rewrite -?validH_zigzag // -?validSO_zigzag -?L.
  by split; rewrite ?validH_zigzag // ?validSO_zigzag ?L.
split; case=>n1 [n2][fL][fE][gL][gE][b][h][]; 
rewrite -[s]st_eta zigE zagE fsE; case. 
- case/zipPtf=>/= [[f1 g1]][[f2 g2]][[->->]] ->->->-> H1 H2.
  rewrite zipPt -fsE; exists n1, n2, f2, (fE \+ f1), g2, (gE \+ g1), b, h. 
  by split=>//; rewrite !joinA -(joinC f1) -(joinC g1) !joinA.
move=>->-> /zipPtf [[f1 g1]] [[f2 g2]][[->->]] ->-> H1 H2. 
rewrite zipPt -fsE; exists n1, n2, (f2 \+ fL), f1, (g2 \+ gL), g1, b, h. 
by split=>//; rewrite -!joinA (joinC f2) (joinC g2) -!joinA.
Qed.

End TLockCoh.


Module Exports.
Section Exports.
Variable lkT : tlockT. 
Notation dl := (@label_of lkT).       (* feap label *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation either := (@either lkT).

Definition TLockCohMixin :=
  CohPredMixin (@cohSO lkT either) (@cohH lkT either)
               (@cohfV lkT either) (@coh_weq lkT either) 
               (@coh_sig lkT either) (@coh_zigzag lkT either).

Definition tlock_coh := CohPred (coh lkT) TLockCohMixin.

Lemma tlock_flatsE (fgL fgE : nat_set * U) (h : heap) :
        flats (dl ->> [fgL, h, fgE]) = h.
Proof. by rewrite fsE flatsE !pcm_flatPt tp_flatPtE unitL unitR. Qed.

Lemma tlockVF (fL fE : nat_set) (gL gE : U) (h : heap) : 
        dl ->> [(fL, gL), h, (fE, gE)] \In tlock_coh -> valid (fL \+ fE).
Proof. by move/cohSO; rewrite /validSO fsE /= zipPt allvalidPt; case/andP. Qed.

Lemma tlockVG (fL fE : nat_set) (gL gE : U) (h : heap) : 
        dl ->> [(fL, gL), h, (fE, gE)] \In tlock_coh -> valid (gL \+ gE).
Proof. by move/cohSO; rewrite /validSO fsE /= zipPt allvalidPt; case/andP. Qed.

Lemma tlockI (fgL fgE : nat_set * U) (h : heap) : 
        dl ->> [fgL, h, fgE] \In tlock_coh -> 
        exists (n1 n2 : nat) (b : bool) (h' : heap), 
          h = owner :-> (n1, b) \+ next :-> n2 \+ h'.
Proof.
case=>_ _ [n1][n2][fL][fE][gL][gE][b][k][].
rewrite !fsE; case/stE=>/= _ H _ _ _; move/(fcancelPt (C:=typeCode)): H.
by exists n1, n2, b, k.
Qed.

Lemma tlock_prec (fL fL' fE fE' : nat_set) (gL gL' gE gE'  : U) 
                 (n1 n1' n2 n2' : nat) (b b' : bool) (h h' : heap) : 
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                (fE, gE)] \In tlock_coh ->
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                (fE, gE)] = 
        dl ->> [(fL', gL'), owner :-> (n1', b') \+ next :-> n2' \+ h', 
                (fE', gE')] ->
        [/\ (fL, gL) = (fL', gL'), (fE, gE) = (fE', gE'), 
            n1 = n1', b = b', n2 = n2' & h = h'].
Proof.
move/cohH; rewrite /validH tlock_flatsE -!joinA !fsE => V /stE /= []. 
case/(fcancelPt (C:=pcmCode))=>->-> /(fcancelPt (C:=typeCode)).
case/(hcancelV V)=>{V} [[->->]] V /(hcancelV V) [-> _ ->].
by case/(fcancelPt (C:=pcmCode))=>->->. 
Qed.

Lemma tlock_prec_emp (fL fL' fE fE' : nat_set) (gL gL' gE gE'  : U) 
                 (n1 n1' n2 n2' : nat) (b b' : bool) : 
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2, 
                (fE, gE)] \In tlock_coh ->
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2, 
                (fE, gE)] = 
        dl ->> [(fL', gL'), owner :-> (n1', b') \+ next :-> n2', 
                (fE', gE')] ->
        [/\ (fL, gL) = (fL', gL'), (fE, gE) = (fE', gE'), 
            n1 = n1', b = b' & n2 = n2'].
Proof.
rewrite -[_ \+ next :-> n2]unitR -[_ \+ next :-> n2']unitR.
by move=>C; case/(tlock_prec C). 
Qed.

Lemma tlock_inv (n1 n2 : nat) (fL fE : nat_set) b (gL gE : U) (h : heap) :
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
               (fE, gE)] \In tlock_coh ->
        if b then h = Unit else I (gL \+ gE) h.
Proof.
move=>C; case: (C)=>Vh Vs [n1'][n2'][fL'][fE'][gL'][gE'][b'][h'][].
case/(tlock_prec C); case=><-<- [<-<-] <-<-<-<-; rewrite leq_eqVlt => H. 
by case=>D []; [case | case: ltngtP H=>//= H _; case]; move=>_ ->.
Qed.

(* for entangling *)
Lemma priv_tlock_prec hp (hL hL' hE hE' : heap) (fL fL' fE fE' : nat_set) 
                      (gL gL' gE gE'  : U) (n1 n1' n2 n2' : nat) 
                      (b b' : bool) (h h' : heap) : 
        hp ->> [hL, tt, hE] \+
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                (fE, gE)] \In star_coh (priv_coh hp) tlock_coh ->
        hp ->> [hL, tt, hE] \+
        dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                (fE, gE)] = 
        hp ->> [hL', tt, hE'] \+
        dl ->> [(fL', gL'), owner :-> (n1', b') \+ next :-> n2' \+ h', 
                (fE', gE')] ->
        [/\ hL = hL' /\ hE = hE', n1 = n1' /\ n2 = n2', b = b', 
            (fL, gL) = (fL', gL'),
            (fE, gE) = (fE', gE') & h = h'].
Proof.
move=>C; case/split_coh: (C)=>/= _ _ Ch Ct.
by case/(scancel (cohV C)); case=>->-> _; case/(tlock_prec Ct).
Qed.

End Exports.
End Exports.
End TLockCoh.

Export TLockCoh.Exports.


(* Defining guarantee transitions *)
Module TLockGuar.
Section TLockGuar.
Variable lkT: tlockT. 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation coh := (tlock_coh lkT).

(* internal transition: counters can only grow monotonically *)

Definition int_rel s s' := 
  [/\ exists (n1 n2 n2' : nat) (fL fE fL' : nat_set) 
             (gL gE : U) (b : bool) (h : heap), 
        [/\ s = dl  ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                         (fE, gE)],
            s' = dl ->> [(fL', gL), owner :-> (n1, b) \+ next :-> n2' \+ h, 
                         (fE, gE)] &
            n2 <= n2'] &
        s \In coh /\ s' \In coh].

Lemma intguar_coh s s' : int_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case. Qed.

Lemma intguar_refl s : s \In coh -> int_rel s s. 
Proof. 
move=>C; split=>//; case: C=>_ _ [n1][n2][fL][fE][gL][gE][b][h][->] H1 H2. 
by exists n1, n2, n2, fL, fE, fL, gL, gE, b, h.
Qed.

Lemma intguar_footH s s' : int_rel s s' -> footeqH (flats s) (flats s').
Proof.
case; case=>n1 [n2][n2'][fL][fE][fL'][gL][gE][b][h][->->] _ _.
rewrite !tlock_flatsE; apply: footeqHUnL; apply: footeqHUnR.
by apply: footeqHPt.
Qed.

Lemma intguar_env s s' : int_rel s s' -> env s = env s'.
Proof. 
case; case=>n1 [n2][n2'][fL][fE][fL'][gL][gE][b][h][->->] _ _. 
by rewrite !fsE /=.
Qed.

Lemma intguar_zagzig s s' p : 
        env s = env s' ->
        int_rel (s \> p) (s' \> p) -> int_rel (s \< p) (s' \< p).
Proof.
move=>E; case; case=>[n1][n2][n2'][fL][fE][fL'][gL][gE][b][h][F F' H][]. 
move/coh_zigzag=>C /coh_zigzag C'; split=>// {C C'}; exists n1, n2, n2'.
rewrite -[s]st_eta -[s']st_eta !fsE !zigE !zagE in F F' *.
case: F F' E => ->-> /zipPtf [[f1 g1]][[f2 g2]][[->->{fE gE}]] ->->{p}.
case=>->-> /zipPtf [[f1' g1']][_][_] -> _; rewrite !zipPt -!fsE.
case/(fcancelPt (C:=pcmCode))=><-<-{f1' g1'} . 
by exists (f2 \+ fL), f1, (f2 \+ fL'), (g2 \+ gL), g1, b, h.
Qed.

Section Acquire.
Variable h : heap.

(*
acquire transition: 
- return the current ticket
- switch the flag b from true to false
- acquire the resource heap
- call the next ticket
*)
Definition acq_rel s s' :=  
  [/\ exists (n1 n2 : nat) (fL fE : nat_set) (gL gE gL' : U), 
        s = dl ->> [(fL \+ #n1, gL), 
                    owner :-> (n1, true) \+ next :-> n2, (fE, gE)] /\
        s' = dl->> [(fL, gL'), 
                    owner :-> (n1.+1, false) \+ next :-> n2 \+ h, 
                    (fE, gE)] & 
       s \In coh /\ s' \In coh].

Lemma acqguar_coh s s' : acq_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case. Qed.

Lemma acqguar_env s s' : acq_rel s s' -> env s = env s'.
Proof. by case; case=>n1 [n2][fL][fE][gL][gE][gL'][->->] _; rewrite !fsE. Qed.

Lemma acqguar_zagzig s s' p : 
        env s = env s' -> 
        acq_rel (s \> p) (s' \> p) -> acq_rel (s \< p) (s' \< p).
Proof.
move=>E; case; case=>n1 [n2][fL][fE][gL][gE][gL'][F F'][].
move/coh_zigzag=>C /coh_zigzag C'; split=>// {C C'}; exists n1, n2.
rewrite -[s]st_eta -[s']st_eta !fsE !zigE !zagE in F F' *.
case: F F' E => ->-> /zipPtf [[f1 g1]][[f2 g2]][[->->{fE gE}]] ->->{p}.
case=>->-> /zipPtf [[f1' g1']][_][_] -> _; rewrite !zipPt -!fsE.
case/(fcancelPt (C:=pcmCode))=><-<-{f1' g1'}. 
exists (f2 \+ fL), f1, (g2 \+ gL), g1, (g2 \+ gL'). 
by rewrite -!joinA.  
Qed.

Lemma acqguar_footH s s' : acq_rel s s' -> footeqH (flats s \+ h) (flats s').
Proof.
case; case=>n1 [n2][fL][fE][gL][gE][gL'][->->] _. 
by rewrite !tlock_flatsE; do 2![apply: footeqHUnL]; apply: footeqHPt.
Qed.

End Acquire.


Section Release.
Variable h : heap.

(*
release transition:
- the called ticked is owned (n1 \in dom fL)
- the lock is not acquired yet (b = false)
- switch b to true, and release the heap h
*)
Definition rel_rel s s' :=  
 [/\ exists (n1 n2 : nat) (fL fE : nat_set) (gL gE: U), 
       s = dl ->> [(fL \+ #n1, gL), owner :-> (n1, false) \+ 
                                    next :-> n2 \+ h, (fE, gE)] /\
       s' = dl->> [(fL \+ #n1, gL), owner :-> (n1, true) \+ 
                                    next :-> n2, (fE, gE)] &
      s \In coh /\ s' \In coh].

Lemma relguar_coh s s' : rel_rel s s' -> s \In coh /\ s' \In coh.
Proof. by case. Qed.

Lemma relguar_env s s' : rel_rel s s' -> env s = env s'.
Proof. by case; case=>n1 [n2][fL][fE][gL][gE][->->]; rewrite !fsE. Qed.

Lemma relguar_zagzig s s' p : 
        env s = env s' -> 
        rel_rel (s \> p) (s' \> p) -> rel_rel (s \< p) (s' \< p).
Proof.
move=>E. case; case=>n1 [n2][fL][fE][gL][gE][F F'][C C'].
have V : valid (fL \+ #n1 \+ fE) by rewrite F in C; apply: tlockVF C.
move/coh_zigzag: C C'=>C /coh_zigzag C'; split=>// {C C'}; exists n1, n2.
rewrite -[s]st_eta -[s']st_eta !fsE !zigE !zagE in F F' *.
case: F F' E V => ->-> /zipPtf [[f1 g1]][[f2 g2]][[->->{fE gE}]] ->->{p}.
case=>->-> /zipPtf [[f1' g1']][_][_] -> _; rewrite !zipPt -!fsE.
case/(fcancelPt (C:=pcmCode))=><-<-{f1' g1'} V. 
by exists (f2 \+ fL), f1, (g2 \+ gL), g1; rewrite -!joinA. 
Qed.

Lemma relguar_footH s s' : rel_rel s s' -> footeqH (flats s) (h \+ flats s').
Proof.
case; case=>n1 [n2][fL][fE][gL][gE][->->] _; rewrite !tlock_flatsE (joinC h).
by do 2![apply: footeqHUnL]; apply: footeqHPt.
Qed.

End Release.

End TLockGuar.


Module Exports.
Section Exports.
Variable lkT : tlockT. 

Definition TLockGuarMixin := 
  GuarMixin (@intguar_coh lkT) (@intguar_env lkT) (@intguar_zagzig lkT).
Canonical tlock_int_guar := Eval hnf in Guar (int_rel lkT) TLockGuarMixin.

Definition TLockIntGuarMixin := IntGuarMixin (@intguar_footH lkT).
Canonical TLockIntGuar := Eval hnf in IntGuar tlock_int_guar TLockIntGuarMixin.

Definition TLockReflGuarMixin := ReflGuarMixin (@intguar_refl lkT).
Canonical TLockReflGuar := Eval hnf in ReflGuar tlock_int_guar TLockReflGuarMixin.

Definition baseTLockAcqGuarMixin h := 
  GuarMixin (@acqguar_coh lkT h) (@acqguar_env lkT h) (@acqguar_zagzig lkT h).
Canonical tlock_acq_guar h := 
  Eval hnf in Guar (acq_rel lkT h) (baseTLockAcqGuarMixin h).

Definition TLockAcqGuarMixin := AcqGuarMixin (@acqguar_footH lkT).
Canonical TLockAcqGuar := Eval hnf in AcqGuar tlock_acq_guar TLockAcqGuarMixin.

Definition baseTLockRelGuarMixin h := 
  GuarMixin (@relguar_coh lkT h) (@relguar_env lkT h) (@relguar_zagzig lkT h).
Canonical tlock_rel_guar h := 
  Eval hnf in Guar (rel_rel lkT h) (baseTLockRelGuarMixin h).

Definition TLockRelGuarMixin := RelGuarMixin (@relguar_footH lkT).
Canonical TLockRelGuar := Eval hnf in RelGuar tlock_rel_guar TLockRelGuarMixin.

Definition TLockChanCore := ChanCore TLockAcqGuar TLockRelGuar.
Canonical TLockCore : mod_core := @Mod.Core (tlock_coh lkT) TLockReflGuar. 
Canonical TLock : mod := @Mod.Make TLockCore [:: TLockChanCore]. 

End Exports.
End Exports.

End TLockGuar.

Export TLockGuar.Exports.

(*****************************************************)
(* Instances of decorate and mod_unit structures for *)
(* ticketed locks                                    *)
(*****************************************************)

Module TLockDecorate.
Section TLockDecorate.
Variable lkT : tlockT.
Notation dl := (@label_of lkT). 
Notation owner := (@owner_of lkT).
Notation next := (@next_of lkT).  
Notation U := (@pcm_of lkT).      
Notation I := (@inv_of lkT).
Notation coh := (tlock_coh lkT).
Notation T := ((nat_set * U)%type).
Notation pf_prec := (@precise_of lkT).

Definition tlock_dec (g : T) s := 
  s = dl ->> [g, flats s, Unit] /\ s \In tlock_coh lkT.

Lemma dec_prec g h1 h2 s1 s2 :
        valid (flats s1 \+ h1) -> 
        flats s1 \+ h1 = flats s2 \+ h2 ->
        tlock_dec g s1 -> tlock_dec g s2 ->
        s1 = s2.
Proof. 
case: g=>g1 g2 V E [-> /= C1][-> /= C2].
case: C1=>_ _ [n1][n1'][fL1][fE1][gL1][gE1][b1][x1][] E1 N1 H1. 
case: C2=>_ _ [n2][n2'][fL2][fE2][gL2][gE2][b2][x2][] E2 N2 H2. 
case/scancelPt: E1 E2 V E H1 H2=>->->-><-<- /scancelPt [->->->->->] V.
rewrite -!joinA in V *.
case/(hcancelV V); case=><-<-{b2} {V} V /(hcancelV V) [<-]{V} V E H1 H2. 
case: H1 H2 V E=>_; case.
- by case=>_ ->->{b1 x1} [_]; case; [case=>_ _ -> | case: ifP=>_; case].
case: ifP=>_.
- case=>_ -> N1' inv1 [_]; case; first by case. 
  by rewrite N1'; case=>_ _ _ inv2 V E; rewrite (pf_prec V E inv1 inv2). 
case=>N1' -> inv1 [_]; case; first by case.
by rewrite N1' ltnn; case=>_ _ inv2 V E; rewrite (pf_prec V E inv1 inv2).
Qed.

Lemma dec_inj g1 g2 s : 
        tlock_dec g1 s -> tlock_dec g2 s -> g1 = g2.
Proof. 
by case: g1 g2=>g11 g12 [g21 g22][-> /= _][]; case/scancelPt=>->->->. 
Qed.

Lemma dec_coh g s : tlock_dec g s -> s \In Mod.coh (TLock lkT).
Proof. by case. Qed.

Lemma dec_onto g1 s1 s2 : 
        tlock_dec g1 s1 -> s2 \In tlock_coh lkT -> 
        env s1 = env s2 ->  exists g2, tlock_dec g2 s2. 
Proof. 
case=>-> /= C1 C2.
case: C2 (C2)=>_ _ [n1][n2][fL][fE][gL][gE][b][h][->{s2} _ _] C2 H.
rewrite !fsE /= in H.
move/(fcancelPt (C:=pcmCode)): H C2=><- C2.
by exists (fL, gL); split=>//; rewrite tlock_flatsE.
Qed.

(*
Lemma dec_onto s : 
        s \In Mod.coh (TLock lkT) -> exists g h, tlock_dec g h s.
Proof. 
move=>C; case: (C)=>_ _ [n1][n2][fL][fE][gL][gE][b][h][E] _ _.
by exists ((fL, gL), (fE, gE)), (owner :-> (n1, b) \+ next :->n2 \+ h).
Qed.

Lemma dec_flat g h s : tlock_dec g h s -> flats s = h.
Proof. by case=>->; rewrite tlock_flatsE. Qed.
*)

Lemma dec_env g1 g2 s1 s2 :
        tlock_dec g1 s1 -> tlock_dec g2 s2 -> env s1 = env s2.
Proof. by rewrite /tlock_dec !fsE; case=>-> C1 [-> C2]; split=>[->|]. Qed.

End TLockDecorate.

Module Exports.
Notation tloc_dec := tlock_dec.

Section Exports.
Variable lkT : tlockT.
Notation dl := (@label_of lkT). 
Notation owner := (@owner_of lkT).
Notation next := (@next_of lkT).  
Notation U := (@pcm_of lkT).      
Notation I := (@inv_of lkT).
Notation coh := (tlock_coh lkT).
Notation T := ((nat_set * U)%type).
Notation pf_prec := (@precise_of lkT).

Definition TLockDecorateMixin := 
  DecorateMixin (@dec_prec lkT) (@dec_inj lkT) (@dec_coh lkT) 
                (@dec_onto lkT) (@dec_env lkT).
Canonical TLockDecorate := 
  Eval hnf in Decorate (@tlock_dec lkT) TLockDecorateMixin.

End Exports.
End Exports.

End TLockDecorate.

Export TLockDecorate.Exports.


(***************************************************************)
(***               Actions with Ticketed Locks               ***)
(***************************************************************)

(* Internal action: draw a ticket *)
(* - atomically increments the next pointer *)

Module TickDrawAct.
Section TickDrawAct.
Variable lkT: tlockT. 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation coh := (tlock_coh lkT). 
Notation guar := (tlock_int_guar lkT). 

(* We can always draw a ticket *)
Definition draw_safe s := s \In coh.

(* Draw a ticked -- atomic INCR *)
Definition draw_step s s' (r : nat) := 
  [/\  exists (n1 n2 : nat) (fL fE : nat_set) (gL gE : U) (b : bool)(h : heap), 
         [/\ s = dl ->> [(fL, gL), 
                          owner :-> (n1, b) \+ next :-> n2 \+ h, (fE, gE)], 
             s' = dl->> [(fL \+ #n2, gL), 
                         owner :-> (n1, b) \+ next :-> n2.+1 \+ h, (fE, gE)] &
             r = n2] &
        s \In coh /\ s' \In coh].

Lemma draw_safe_coh : safe_coh coh draw_safe.
Proof. by []. Qed.

Lemma draw_safe_zagzig : safe_zagzig draw_safe.
Proof. by move=>s p /coh_zigzag. Qed.

Lemma draw_step_safe : step_safe draw_safe draw_step.
Proof. by move=>s1 s2 r [_][]. Qed.

Lemma draw_step_guar : step_guar guar draw_step.
Proof. 
move=>s s' r /= [G][C C']; split=>// {C C'}. 
case: G=>n1 [n2][fL][fE][gL][gE][b][h][E E' _]. 
by exists n1, n2, n2.+1, fL, fE, (fL \+ #n2), gL, gE, b, h. 
Qed.

Lemma draw_step_frame : step_frame draw_safe draw_step.
Proof. 
move=>s1 s2 p r _.
case; case=>n1 [n2][fL][fE][gL][gE][b][h][E ->->][/coh_zigzag].
rewrite -[s1]st_eta !fsE in E *. 
case: E=>/zipPtf [[f2 g2]][[f1 g1]][[->->{fL gL}]] ->{p} ->->-> C C'.
exists (dl ->> [(f1 \+ #n2, g1), owner :-> (n1, b) \+ 
                                 next :-> n2.+1 \+ h, (fE, gE)]).
rewrite !fsE zigE zagE !zipPt -!fsE in C C' *. 
split; first by rewrite (pcm_joinE (f2, g2)) /= joinA.
split.
- exists n1, n2, f1, (fE \+ f2), g1, (gE \+ g2), b, h.
  by split=>//; rewrite !fsE !zagE !zipPt -!fsE. 
split=>//; apply/coh_zigzag.
by rewrite fsE zigE zipPt -fsE (pcm_joinE (f2, g2)) /= joinA. 
Qed.

Lemma draw_step_total : step_total draw_safe draw_step.
Proof.
move=>s C. case: C (C)=>_ _ [n1][n2][fL][fE][gL][gE][b][h][E H C'] C.
subst s.
exists (dl ->> [(fL \+ #n2, gL), owner :-> (n1, b) \+ next :-> n2.+1 \+ h, (fE, gE)]).
exists n2. split=>//.
- by eexists _, _, _, _, _, _, _, _.
split=>//.
have [V1 [V2 V3]] : valid (fL \+ fE) /\ valid (fL \+ fE \+ #n2) /\ valid (gL \+ gE).
- move/cohSO: C. 
  rewrite /validSO !fsE /= zipPt allvalidPt.
  case/andP=>/= V1 V2; do !split=>//. 
  case: validUn=>//.
  - by rewrite V1.
  - by rewrite um_validPt. 
  simpl. 
  move=>k H'. rewrite um_domPt inE. 
  case: eqP H'=>// <-. 
  case: (C') => D _.
  move/(_ n2): D.  move=>->. 
  rewrite inE. rewrite ltnn andbF.  by [].
split. 
- move/cohH: C.
  rewrite /validH. 
  rewrite !tlock_flatsE.
  rewrite -!joinA.
  rewrite -!(joinCA (next :-> _)).
  by rewrite !hvalidPtUn.
- rewrite /validSO !fsE /=.
  rewrite !zipPt.
  rewrite !allvalidPt. 
  apply/andP=>/=. rewrite joinAC. by []. 
exists n1, n2.+1.
exists (fL \+ #n2), fE. exists gL, gE. exists b. exists h.
split=>//.
- by apply: leq_trans H _. 
case: C'=>D H'.
split. 
- move=>/= k. 
  rewrite joinAC.
  rewrite domUn !inE /=.
  rewrite V2. simpl.
  rewrite D. rewrite um_domPt. rewrite !inE /=. 
  rewrite ltnS. rewrite (leq_eqVlt k n2) eq_sym.
  rewrite orbC.
  case: eqP=>//=.
  - move=>->.  rewrite H. by [].
case: H'=>H'.
- left. rewrite /TLockCoh.locked in H' *.
  case: H'=>H' ->->.
  split=>//.
  rewrite joinAC. rewrite domUn !inE /= V2 /= H'. by [].
case: ifP H'=>H' H''.
- right. rewrite ltnS. rewrite leq_eqVlt H' orbT.
  case: H''=>H'' -> _ I.
  split=>//.
  rewrite joinAC. rewrite domUn !inE /= V2 /= H''. by [].

rewrite leq_eqVlt H' orbF in H.
rewrite (eqP H).
rewrite ltnSn. simpl. 
right.
case: H''=>_ -> I.
split=>//.
rewrite joinAC. rewrite domUn inE /= V2 /= um_domPt inE /= eq_refl orbT.
by [].
Qed.

Definition draw_pre := 
  Action draw_safe_coh draw_safe_zagzig draw_step_safe draw_step_guar
         draw_step_frame draw_step_total.

Lemma draw_step_erase : int_step_erase draw_pre.
Proof. 
move=>s h t r s' h' t' r' V E St1 St2; move: St1 St2 V E.
case; case=>n11 [n21][fL1][fE1][gL1][gE1][b1][h1][->->->][C1 C1'].
case; case=>n12 [n22][fL2][fE2][gL2][gE2][b2][h2][->->->][C2 C2'].
rewrite !tlock_flatsE -!joinA => V.
by case/(hcancelV V)=>{V} -> V; case/(hcancelV V)=>-> _ ->.
Qed.

End TickDrawAct.

Module Exports.
Definition tdraw lkT := IntAction (@draw_step_erase lkT).
End Exports.

End TickDrawAct. 

Export TickDrawAct.Exports.

(* release action: show that you have *the called* ticket *)
(* will be entangled into an action which tries to lock *)
Module TickRelAct. 
Section TickRelAct. 
Variables (lkT : tlockT) (n : nat). 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation coh := (tlock_coh lkT). 

Section Internal.
Variable h : heap.
Notation guar := (tlock_rel_guar lkT h). 

Definition rel_safe s := 
  [/\ exists (n2 : nat) (fL fE : nat_set) (gL gE : U),
        s = dl ->> [(fL \+ #n, gL), owner :-> (n, false) \+ next :-> n2 \+ h, 
                    (fE, gE)] &
        s \In coh].

Definition rel_step s s' (r: unit) := 
  [/\ exists (n2 : nat) (fL fE : nat_set) (gL gE : U), 
        s = dl ->> [(fL \+ #n, gL), owner :-> (n, false) \+ next :-> n2 \+ h,
                    (fE, gE)] /\
        s' = dl->> [(fL \+ #n, gL), owner :-> (n, true) \+ next :-> n2,
                    (fE, gE)] &
        s \In coh /\ s' \In coh].

Lemma rel_safe_coh : safe_coh coh rel_safe.
Proof. by move=>s []. Qed.

Lemma rel_safe_zagzig : safe_zagzig rel_safe.
Proof. 
move=>s p [G /coh_zigzag C]; split=>// {C}. 
case: G=>n2 [fL][fE][gL][gE]; rewrite fsE -[s]st_eta.
case=>->-> /zipPtf [[f1 g1]][[f2 g2]][_ ->] ->{p}. 
rewrite zigE zipPt -fsE; exists n2, (f2 \+ fL), f1, (g2 \+ gL), g1. 
by rewrite -!joinA. 
Qed.

Lemma rel_step_safe : step_safe rel_safe rel_step.
Proof. 
move=>s s' [][]; case=>n2 [fL][fE][gL][gE][->{s} _][C _]. 
by split=>//; exists n2, fL, fE, gL, gE. 
Qed.

Lemma rel_step_guar : step_guar guar rel_step. 
Proof. 
move=>s s' [][G][C C']; split=>// {C'}. 
case: G C=>n2 [fL][fE][gL][gE][->->{s s'}] C. 
by exists n, n2, fL, fE, gL, gE. 
Qed.

Lemma rel_step_frame : step_frame rel_safe rel_step.
Proof. 
move=>s s' p [] C1 [G]; case: G C1=>n2 [fL][fE][gL][gE][].
rewrite -[s]st_eta !fsE.
case=>/zipPtf [[f1 g1]][[f2 g2]][E] ->->->->->{p s'}.
move=>C1 [/coh_zigzag C C'].
exists (dl ->> [(f2, g2), owner :-> (n, true) \+ next :-> n2, (fE, gE)]).
rewrite fsE zigE !zipPt E; split=>//. 
split; last by split=>//; apply/coh_zigzag; rewrite zigE zipPt -E. 
case: C1; case=>n2' [fL'][fE'][gL'][gE'] {E C C'}. 
rewrite zagE zipPt -fsE=>E C.
case/(tlock_prec C): {C} E=>-> _ _ _ _ _. 
exists n2, fL', (fE \+ f1), gL', (gE \+ g1).
by split=>//; rewrite fsE zagE zipPt -fsE. 
Qed.

Lemma rel_step_total : step_total rel_safe rel_step.
Proof.
move=>s. 
case; case=>n2 [fL][fE][gL][gE] ->{s} C.
case: C (C)=>_ _ [n1'][n2'][fL'][fE'][gL'][gE'][b'][h'][E] H E' C.
move/(tlock_prec C): E. case. case=>??[??]????. subst fL' fE' gL' gE' n1' b' n2' h'.
exists (dl ->> [(fL \+ #n, gL), owner :-> (n, true) \+ next :-> n2, (fE, gE)]), tt.
split. 
- by eexists _, _, _, _, _.
split=>//. 
split.
- move/cohH: C.
  rewrite /validH. 
  rewrite !tlock_flatsE.
  move/validL.
  by rewrite !hvalidPtUn.
- rewrite /validSO !fsE /=.
  rewrite !zipPt.
  rewrite !allvalidPt. 
  apply/andP=>/=. 
  by rewrite (tlockVF C) (tlockVG C).
rewrite /TLockCoh.validDispenser. simpl.
exists n. exists n2. exists (fL \+ #n), fE. exists gL, gE. 
exists true. exists Unit. rewrite unitR. split=>//.
case: E'=>D H'.
split=>//.
case: H'=>H'.
- by case: H'.
case: ifP H'=>H' H''. 
- case: H''=>H1 _ _ I.
  by left. 
rewrite leq_eqVlt H' orbF in H.
case: eqP H H' C=>// -> _ _ C.
left. split=>//.
rewrite joinAC domUn inE /= -joinAC. 
rewrite (tlockVF C) /= um_domPt inE /= eq_refl orbT.
by [].
Qed.

Definition TickRel_pre :=
  Action rel_safe_coh rel_safe_zagzig rel_step_safe rel_step_guar
         rel_step_frame rel_step_total.

End Internal.

Lemma rel_st_erase : rel_step_erase TickRel_pre.
Proof.
move=>s1 h1 t1 [] k1 s2 h2 t2 [] k2 V E St1 St2; move: St1 St2 V E.
case; case=> n2 [fL][fE][gL][gE][->->{s1 t1}] _. 
case; case=> n2' [fL'][fE'][gL'][gE'][->->{s2 t2}] _. 
rewrite !tlock_flatsE -!joinA => V.
by case/(hcancelV V)=>{V} _ V; case/(hcancelV V)=><- _ <-.
Qed.

Lemma rel_sf_owns : rel_safe_owns TickRel_pre.
Proof.
move=>s h []; case=>n2 [fL][fE][gL][gE] ->{s} C.
by rewrite tlock_flatsE joinC; eexists _.  
Qed.

End TickRelAct.
 
Module Exports.
Definition TLockRel lkT n := RelAction (@rel_st_erase lkT n) (@rel_sf_owns lkT n).
End Exports.

End TickRelAct.

Export TickRelAct.Exports.


(* acquire action: release the lock, if you own it *)
Module TickAcquireAct.
Section TickAcqAct.
Variable lkT : tlockT. 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation coh := (tlock_coh lkT). 
Variables (f : U -> U) (pf_loc : local f).

Section Internal.
Variable h : heap.
Notation guar := (tlock_acq_guar lkT h). 

Definition acq_safe s := 
  [/\ exists (n1 n2 : nat) (fL fE : nat_set) (gL gE : U), 
        [/\ s = dl ->> [(fL \+ #n1, gL), owner :-> (n1, true) \+ next :-> n2, 
                        (fE, gE)],
            valid (f gL \+ gE), 
            valid (owner :-> (n1.+1, false) \+ next :-> n2 \+ h) &
            I (f gL \+ gE) h] &
      s \In coh].
  
Definition acq_step s s' (r : unit) := 
  [/\ exists (n1 n2 : nat) (fL fE : nat_set) (gL gE : U), 
        s = dl ->> [(fL \+ #n1, gL), owner :-> (n1, true) \+ next :-> n2, 
                    (fE, gE)] /\
        s' = dl->> [(fL, f gL), owner :-> (n1.+1, false) \+ next :-> n2 \+ h,
                    (fE, gE)] &
      s \In coh /\ s' \In coh].

Lemma acq_safe_coh : safe_coh coh acq_safe.
Proof. by move=>s []. Qed.

Lemma acq_safe_zagzig : safe_zagzig acq_safe.
Proof. 
move=>s p [G /coh_zigzag C]; split=>//. 
case: G C=>n1 [n2][fL][fE][gL][gE][]; rewrite fsE -[s]st_eta.
case=>->-> /zipPtf [[f1 g1]][[f2 g2]][[_ ->] ->] ->{p} V1 V2 inv C. 
have {C} V3 : valid (gL \+ g2). 
- by rewrite zigE zipPt -fsE joinC in C; move/tlockVG/validL: C. 
have V4 : valid (f gL) by move/validL: V1. 
rewrite zigE zipPt -fsE; exists n1, n2, (f2 \+ fL), f1, (g2 \+ gL), g1. 
split=>//; first by [rewrite -!joinA];
by rewrite -(joinC gL) pf_loc // joinAC -joinA.
Qed.

Lemma acq_step_safe : step_safe acq_safe acq_step.
Proof. 
move=>s s' [][][n1][n2][fL][fE][gL][gE][->->][C C'].
split=>//; exists n1, n2, fL, fE, gL, gE.  
case: (C')=>_ _ [n1'][n2'][fL'][fE'][gL'][gE'][b'][h'][].
case/(tlock_prec C')=>[[<-<-]][<-<-] <-<-<-<- _ _.
split=>//; first by apply: tlockVG C'.
- by move/cohH: C'; rewrite /validH tlock_flatsE. 
by apply: tlock_inv C'.
Qed.

Lemma acq_step_guar : step_guar guar acq_step. 
Proof. 
move=>s s' [][][n1][n2][fL][fE][gL][gE][->->] C.
by split=>//; exists n1, n2, fL, fE, gL, gE, (f gL). 
Qed.

Lemma acq_step_frame : step_frame acq_safe acq_step.
Proof. 
move=>s s' p [] C1 [G]; case: G C1=>n1 [n2][fL][fE][gL][gE][].
rewrite -[s]st_eta !fsE.
case=>/zipPtf [[f1 g1]][[f2 g2]][E] ->->->->->{p s'} C1 [/coh_zigzag C C'].
have [V1 V2] : valid (fL \+ #n1) /\ valid gL.
- move/acq_safe_coh/coh_zigzag: C1; rewrite zigE zipPt -fsE -E=>C1.
  by split; [move/tlockVF/validL: C1 | move/tlockVG/validL: C1]. 
case: C1; case=>n1' [n2'][fL'][fE'][gL'][gE'][].
rewrite {1 2}zagE zipPt -{1 2}fsE => X /validL V3 _ _ Cx. 
case/(tlock_prec_emp Cx): {Cx} X E C=>/= -> _ <-{n1'} _ _. 
case; rewrite joinA; move/(joinKf V1)=>?? C; subst fL gL.
rewrite (joinC g1) in V2 C' *; rewrite pf_loc // -(joinC g1) in C' *.
exists (dl ->> [(fL', f gL'), owner :-> (n1.+1, false) \+ next :-> n2 \+ h, 
                (fE, gE)]).
split; first by rewrite fsE zigE zipPt -!fsE. 
split; last by split=>//; apply/coh_zigzag; rewrite fsE zigE zipPt. 
exists n1, n2, fL', (fE \+ f1), gL', (gE \+ g1).
by rewrite !fsE !zagE !zipPt -!fsE. 
Qed.

Lemma acq_step_total : step_total acq_safe acq_step.
Proof.
move=>s.
case. case=>n1 [n2][fL][fE][gL][gE][->{s} V1 V2 I] C.
case: C (C)=>_ _ [n1'][n2'][fL'][fE'][gL'][gE'][b'][h'][E H H'] C.
rewrite -[_ \+ next :-> _]unitR in E C.
move/(tlock_prec C): E. case. case=>??[??]????. subst fL' fE' gL' gE' n1' b' n2' h'.
rewrite unitR in C.
exists (dl ->> [(fL, f gL), owner :-> (n1.+1, false) \+ next :-> n2 \+ h, (fE, gE)]), tt.
split. 
- by eexists _, _, _, _, _, _.
split=>//. 
have V1': valid (fL \+ fE \+ #n1).
- rewrite joinAC. rewrite (tlockVF C). by [].
have V2' : n1 \notin dom (fL \+ fE).
- case: validUn V1'=>// _ _ H2 _.
  apply: contra (H2 n1) _.
  by rewrite negbK um_domPt inE eq_refl.
have {H} H: n1 < n2.
- case: H'. move/(_ n1). rewrite joinAC.
  rewrite domUn !inE /= V1' /= (negbTE V2') /=. 
  rewrite um_domPt inE eq_refl. move/esym. case/andP. by [].
split.
- by rewrite /validH !tlock_flatsE.
- rewrite /validSO !fsE /=.
  rewrite !zipPt.
  rewrite !allvalidPt. 
  apply/andP=>/=. 
  move: (tlockVF C). rewrite joinAC. move/validL. move=>->. by [].
rewrite /TLockCoh.validDispenser. simpl.
exists n1.+1. exists n2. exists fL, fE. exists (f gL), gE. 
exists false. exists h. split=>//.
case: H'=>D H'.
have D' : dom (fL \+ fE) =i [pred x | n1 < x < n2].
- move=>/= k. 
  rewrite inE /=.
  move: (D k). rewrite joinAC. rewrite domUn !inE /= V1' /=.
  rewrite um_domPt inE /=.
  rewrite orbC leq_eqVlt.
  case: eqP=>//= <-. move=>_. rewrite (negbTE V2').
  by rewrite ltnn.
rewrite H in H'.
split=>//.
case: H'=>H'.
- right. 
  case: ltnP=>N.
  - split=>//. rewrite D' inE /= N. rewrite ltnSn. by [].
  split=>//.
  rewrite leq_eqVlt ltnS leqNgt H orbF in N.
  by rewrite (eqP N).
right.
case: ltnP=>N.
- split=>//. 
  by rewrite D' inE /= ltnSn N.
split=>//. 
rewrite leq_eqVlt ltnS leqNgt H orbF in N.
by rewrite (eqP N).
Qed.
  
Definition TickAcq_pre :=
  Action acq_safe_coh acq_safe_zagzig acq_step_safe acq_step_guar
         acq_step_frame acq_step_total.

End Internal.

Lemma acq_st_erase : acq_step_erase TickAcq_pre.
Proof.
move=>s1 h1 t1 [] k1 s2 h2 t2 [] k2 V E St1 St2; move: St1 St2 V E.
case; case=>n1 [n2][fL][fE][gL][gE][->->] _.
case; case=>n1' [n2'][fL'][fE'][gL'][gE'][->->] _.
rewrite !tlock_flatsE -!joinA => V.
by case/(hcancelV V)=>{V} [[<-]] V /(hcancelV V) [<- _ <-]. 
Qed.

Lemma acq_st_owns : acq_step_owns TickAcq_pre.
Proof.
move=>s s' [] k []; case=>n1 [n2][fL][fE][gL][gE][->->] _.
by exists (owner :-> (n1.+1, false) \+ next :-> n2); rewrite tlock_flatsE joinC.
Qed.

End TickAcqAct.

Module Exports.
Definition TLockAcq lkT f pf := 
  AcqAction (@acq_st_erase lkT f pf) (@acq_st_owns lkT f pf).
End Exports.

End TickAcquireAct.

Export TickAcquireAct.Exports.


(* The *real* unlocking action *)
Module TickUnlockAct.
Section TickUnlockAct.
Variables (hp : nat) (lkT : tlockT).
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (TLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

(* entangling actions yields cumbersome pre- and post-conditions *)
(* typically, we want to simplify them, e.g., in the stability lemmas *)
(* here're the ones for unlock *)

Definition unlock_safe s := 
  [/\ exists (hL hE h : heap) (n1 n2 : nat) (fL fE : nat_set) (gL gE : U), 
        [/\ s = hp ->> [hL \+ h, tt, hE] \+
                dl ->> [(fL \+ #n1, gL), owner :-> (n1, true) \+ next :-> n2, 
                        (fE, gE)],
            valid (f gL \+ gE) & I (f gL \+ gE) h] &
      s \In coh].

Definition unlock_step s s' (r : unit) := 
  [/\ exists (hL hE h : heap) (n1 n2 : nat) (fL fE : nat_set) (gL gE : U), 
        s = hp ->> [hL \+ h, tt, hE] \+
            dl ->> [(fL \+ #n1, gL), owner :-> (n1, true) \+ next :-> n2, 
                    (fE, gE)] /\
        s' = hp->> [hL, tt, hE] \+
             dl->> [(fL, f gL), owner :-> (n1.+1, false) \+ next :-> n2 \+ h,
                        (fE, gE)] &
      s \In coh /\ s' \In coh].

(* Unique transfer lemma *)
Lemma uniq_xfer_pf : uniq_xfer (TLockAcq pf) (PrivRel hp). 
Proof.
move=>s h h' [s1][s2][-> Vh][Sf1 Sf2][s1'][s2'][E _][Sf1' Sf2'].
move: (coh_prec (validHV Vh) E (asf_coh Sf1) (asf_coh Sf1'))=>E'.
rewrite -{s1'}E' in E Sf1' *; rewrite -{E s2'}(joinsK (validHV Vh) E) in Sf2'.
case: Sf1 Sf1'; case=>n1 [n2][fL][fE][gL][gE][->] _ _ inv C []. 
case=>n1' [n2'][fL'][fE'][gL'][gE'][E] _ _ inv'; move: E inv inv'.
case/(tlock_prec_emp C); case=>_ <- [_ <-] _ _ _ inv inv' _. 
move: (validHR Vh); rewrite /validH.
case/rel_asf_owns: Sf2'; case/rel_asf_owns: Sf2=>h2 -> h2' E' V'.
by apply: pf_prec V' E' inv inv'.
Qed.

Definition unlock_ent :=
  @EntAcqRel unit unit unit (fun _ => tt) (TLock lkT) (Priv hp)
             (TLockChanCore lkT) (PrivChanCore hp) 
             (or_introl (erefl _)) (or_introl (erefl _)) _ _  
             uniq_xfer_pf. 

(* the simplified precondition is equivalent to the default one *)
Lemma unlockP s : unlock_safe s <-> s \In asafe unlock_ent. 
Proof.
split; last first.
- move=>S; split; last by move/asf_coh: S; rewrite /= starC.
  case: S=>h [s1][s2][->{s} _][Sf1 Sf2]; rewrite joinC.
  case: Sf1 Sf2; case=>n1 [n2][fL][fE][gL][gE][-> ????][hL][hE][-> _].
  by exists hL, hE, h, n1, n2, fL, fE, gL, gE. 
case; case=>hL [hE][h][n1][n2][fL][fE][gL][gE][-> V inv].
set s1 := (hp ->> [_, _, _]); set s2 := (dl ->> [_, _, _]).
case/split_coh=>/= _ Vh C1 C2; exists h, s2, s1; rewrite (joinC s2). 
split=>//; split; last by [exists hL, hE]; split=>//.  
exists n1, n2, fL, fE, gL, gE; split=>//.
move: Vh; rewrite validHUn; case/andP=>_. 
rewrite priv_flatsE tlock_flatsE !joinA !(joinAC _ h) -!joinA. 
by move/validR/validR; rewrite !hvalidPtUn. 
Qed.

(* the simplified postcondition *)
Lemma unlockQ s s' : unlock_step s s' tt <-> astep s unlock_ent s' tt.
Proof.
split.
- case; case=>hL [hE][h][n1][n2][fL][fE][gL][gE][->->{s s'}][].
  set s1 := (hp ->> [_, _, _]); set s2 := (dl ->> [_, _, _]).
  set s1' := (hp ->> [_, _, _]); set s2' := (dl ->> [_, _, _]).
  case/split_coh=>/= _ Vh Cp Ct; case/split_coh=>/= _ Vh' Cp' Ct'.
  exists s2, s1, s2', s1', tt, tt; rewrite (joinC s2) (joinC s2').
  split=>//; exists h; split; last by exists hL, hE.
  by split=>//; exists n1, n2, fL, fE, gL, gE. 
case=>s1 [s2][s1'][s2'][[]][[]][->->{s s'} _] Vh [h][St1 St2]; split.
- case: St1 Vh; case=>n1 [n2][fL][fE][gL][gE][->->{s1 s1'}][C1 C1'] Vh.
  case: St2 Vh=>hL [hE][->->{s2 s2'}] C2 Vh.
  by exists hL, hE, h, n1, n2, fL, fE, gL, gE; split; rewrite joinC.
split. 
- split=>//; exists s2, s1; rewrite joinC.
  by case/ast_coh: St1; case/ast_coh: St2.
split; last first.
- exists s2', s1'; rewrite joinC.
  by case/ast_coh: St1; case/ast_coh: St2.
move: Vh; rewrite !validHUn. 
case: St1; case=>n1 [n2][fL][fE][gL][gE][->->] _.
case: St2; move=>hL [hE][->->] _.
rewrite !tlock_flatsE !priv_flatsE -!joinA (joinCA h).
by rewrite !hvalidPtUn !pcm_validE /= !fsE /= !fvalidPtUn. 
Qed.

Lemma unlock_safe_coh : safe_coh coh unlock_safe.
Proof. by move=>s /unlockP /asf_coh; rewrite /= starC. Qed.

Lemma unlock_safe_zagzig : safe_zagzig unlock_safe.
Proof. by move=>s p /unlockP /asf_zagzig /unlockP. Qed.

Lemma unlock_step_safe : step_safe unlock_safe unlock_step.
Proof. by move=>s1 s2 [] /unlockQ /ast_safe /unlockP. Qed.

Lemma unlock_step_guar : step_guar guar unlock_step.
Proof. by move=>s1 s2 [] /unlockQ /ast_guar; rewrite entangle_intC. Qed.

Lemma unlock_step_frame : step_frame unlock_safe unlock_step.
Proof. 
move=>s s' p [] /unlockP H1 /unlockQ.
by case/(ast_frame H1)=>s'' [->] /unlockQ; exists s''. 
Qed.

Lemma unlock_step_total : step_total unlock_safe unlock_step.
Proof.
by move=>s /unlockP /ast_total; case=>s' [[]] /unlockQ; exists s', tt.
Qed.

Definition TUnlock_pre := 
  Action unlock_safe_coh unlock_safe_zagzig unlock_step_safe unlock_step_guar 
         unlock_step_frame unlock_step_total.

Lemma unlock_step_erase : int_step_erase TUnlock_pre.
Proof. 
move=>s h t [] s' h' t' [] V E /unlockQ St /unlockQ.
by move/(int_ast_erase V E St). 
Qed.

End TickUnlockAct.

Module Exports.
Definition tunlock hp lkT f pf := IntAction (@unlock_step_erase hp lkT f pf).
End Exports. 

End TickUnlockAct.

Export TickUnlockAct.Exports.


(* An auxiliary helper module for reading from ticketed lock's owner field. *)
(* Actually, it's begging to be unified with SCSL lock's ReadLock module. *)
(* both are identity actions, and the only interesting proofs about them *)
(* involve the safety conditions zig-zaging properties *)

Module ReadOwner.
Section ReadOwner.
Variables (lkT : tlockT) (n n1 : nat) (b : bool). 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation coh := (tlock_coh lkT). 
Notation guar := (tlock_int_guar lkT). 

Definition read_owner_safe s := 
  [/\ exists (n2 : nat) (fL fE : nat_set) (gL gE : U) h, 
        s = dl ->> [(fL \+ #n, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                    (fE, gE)] &
      s \In coh]. 

Definition read_owner_step s s' (v : unit) := s = s' /\ read_owner_safe s.

Lemma readowner_safe_coh : safe_coh coh read_owner_safe.
Proof. by move=>s []. Qed.

Lemma readowner_safe_zagzig : safe_zagzig read_owner_safe.
Proof. 
move=>s p []; case=>n2 [fL][fE][gL][gE][h]; rewrite -[s]st_eta fsE.
case=>->-> /zipPtf [[f1 g1]][[f2 g2]][_ ->->{p}] /coh_zigzag C. 
split=>//; exists n2, (f2 \+ fL), f1, (g2 \+ gL), g1, h. 
by rewrite zigE zipPt -fsE -!joinA. 
Qed.

Lemma readowner_step_safe : step_safe read_owner_safe read_owner_step.
Proof. by move=>s s' [][]. Qed.

Lemma readowner_step_frame : step_frame read_owner_safe read_owner_step.
Proof. by move=>s s' p [Sf][<-{s'} _]; exists s. Qed.

Lemma readowner_step_guar : step_guar guar read_owner_step.
Proof. 
move=>s s' [] St; case: (St)=><- _; apply: refl_guar_refl. 
by move/readowner_safe_coh: (readowner_step_safe St).
Qed.

Lemma readowner_step_total : step_total read_owner_safe read_owner_step.
Proof. by move=>s C; exists s, tt. Qed.

Definition ReadOwner_pre := 
  Action readowner_safe_coh readowner_safe_zagzig readowner_step_safe 
         readowner_step_guar readowner_step_frame readowner_step_total.

End ReadOwner.

Lemma readowner_step_erase lkT n n1 b : 
        int_step_erase (@ReadOwner_pre lkT n n1 b).
Proof. by move=>s1 h1 t1 [] s2 h2 t2 [] _ E [<- _][<- _]. Qed.

Module Exports.

Definition ReadOwner lkT n n1 b := 
  IntAction (@readowner_step_erase lkT n n1 b).

End Exports.
End ReadOwner.

Export ReadOwner.Exports.


Module TickTryLockAct.
Section TickTryLockAct.
Variables (hp : nat) (lkT : tlockT) (n : nat). 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (TLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Definition trylock_safe s := 
  [/\ exists (hL hE : heap) (n1 n2 : nat) (fL fE : nat_set) 
             (gL gE : U) (b : bool) h, 
        s = hp ->> [hL, tt, hE] \+
            dl ->> [(fL \+ #n, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                    (fE, gE)] /\ 
        (n1 = n -> ~~ b) &
      s \In coh].

Definition trylock_step s s' (r : bool) := 
  [/\ exists (hL hE : heap) (n1 n2 : nat) (fL fE : nat_set) 
             (gL gE : U) (b : bool) h,
        s = hp ->> [hL, tt, hE] \+
            dl ->> [(fL \+ #n, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                    (fE, gE)] /\
        if n1 == n then 
           [/\ b = false,
               s' = hp->> [hL \+ h, tt, hE] \+
                    dl->> [(fL \+ #n, gL), owner :-> (n1, true) \+ next :-> n2,
                           (fE, gE)] & r = true]
        else s' = s /\ r = false &
        s \In coh /\ s' \In coh].

(* Unique transfer lemma *)
Lemma uniq_xfer_pf : uniq_xfer (PrivAcq hp) (TLockRel lkT n).
Proof.
move=>s h h' [s1][s2][-> Vh][Sf1 Sf2][s1'][s2'][E _][Sf1' Sf2'].
move: (coh_prec (validHV Vh) E (asf_coh Sf1) (asf_coh Sf1'))=>E'.
rewrite -{s1'}E' in E Sf1' *; rewrite -{E s2'}(joinsK (validHV Vh) E) in Sf2'.
case: Sf2 Sf2'; case=>n2 [fL][fE][gL][gE] -> C [[n2']][fL'][fE'][gL'][gE']. 
by case/(tlock_prec C).
Qed.

Definition trylock_ent := 
  @EntAcqRel unit unit unit (fun _ => tt) (Priv hp) (TLock lkT)
             (PrivChanCore hp) (TLockChanCore lkT)
             (or_introl (erefl _)) (or_introl (erefl _)) _ _
             uniq_xfer_pf. 

Definition check_owner n1 b := 
  FrameAct (ReadOwner lkT n n1 b) (leq_guar_entangle_inr (Priv hp) _). 

(* simplification lemmas for the helper actions *)
Lemma check_ownerP n1 b s : 
        s \In asafe (check_owner n1 b) <->
        [/\ exists (hL hE : heap) (n2 : nat) (fL fE : nat_set) 
                   (gL gE : U) h, 
              s = hp ->> [hL, tt, hE] \+
                  dl ->> [(fL \+ #n, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, 
                          (fE, gE)] &
              s \In coh].
Proof.
split; last first.
- case; case=>hL [hE][n2][fL][fE][gL][gE][h] ->{s} C.
  split=>//; case/split_coh: C=>/= _ _ C1 C2.
  by rewrite joinC; eexists _, _; do !split=>//; eauto 20.
move=>S; split; last by move/asf_coh: S.
case: S; case=>s1 [s2][->{s}] S1 C; rewrite starC in C.
move/(star_coh_prec C): (asf_coh S1)=>C2.
case: C2 C=>_ [hL][hE] ->; case: S1; case=>n2 [fL][fE][gL][gE][h] ->.
by exists hL, hE; rewrite joinC; eauto 20. 
Qed.

(* and for postcondition *)  
Lemma check_ownerQ n1 b s s' : 
        astep s (check_owner n1 b) s' tt <->
        s' = s /\ s \In asafe (check_owner n1 b).
Proof.
split.
- case; case=>s1 [s2][s1'][->->{s s'}][<-{s1'}] H C _.
  by do !split=>//; exists s1, s2. 
case=><- S; move: (asf_coh S)=>C; split=>//.
by case: S; case=>s1 [s2][->] H _; exists s1, s2, s1. 
Qed.

(* and for trylock's pre (we won't need one for trylock's post *)
(* as it's used only once *)
Lemma entP s : s \In asafe trylock_ent <-> s \In asafe (check_owner n false).
Proof.
split. 
- move=>G; split; last by move/asf_coh: G.
  case: G=>h [s1][s2][->{s} Vh][St1 St2].
  exists s2, s1; rewrite joinC; split=>//. 
  case: St2; case=>n2 [fL][fE][gL][gE] -> C.
  by split=>//; eauto 20.
case/check_ownerP; case=>hL [hE][n2][fL][fE][gL][gE][h] -> C.
case/split_coh: C=>/= _ Vh C1 C2; eexists h, _, _; split=>//.
split; last by split=>//; eexists _, _, _, _, _. 
exists hL, hE; split=>//. 
move: Vh; rewrite validHUn priv_flatsE tlock_flatsE.
case/andP=>_; rewrite -(joinC h) -!joinA (joinCA hE) 3!joinA.  
by move/validL. 
Qed.

Lemma check_owner_prec s1 s2 h1 h2 n1 n2 b1 b2 : 
        valid (flats s1 \+ h1) ->
        flats s1 \+ h1 = flats s2 \+ h2 ->
        s1 \In asafe (check_owner n1 b1) -> 
        s2 \In asafe (check_owner n2 b2) -> n1 = n2 /\ b1 = b2.
Proof.
move=>V E Sf1 Sf2; move: Sf1 Sf2 V E.
case; case=>t1 [t1'][-> Sf1] C1; case; case=>t2 [t2'][-> Sf2] C2.
rewrite (flatsUn (cohV C1)) (flatsUn (cohV C2)) {C1 C2}.
case: Sf1; case=>n21 [fL1][fE1][gL1][gE1][k1] ->{t1} C1. 
case: Sf2; case=>n22 [fL2][fE2][gL2][gE2][k2] ->{t2} C2. 
by rewrite !tlock_flatsE -!joinA => V; case/(hcancelV V); case.
Qed.

Lemma check_owner_eq s n1 n2 b1 b2 :  
        s \In asafe (check_owner n1 b1) -> 
        s \In asafe (check_owner n2 b2) -> n1 = n2 /\ b1 = b2.
Proof.
move=>S1 S2; move/cohH: (asf_coh S1)=>Vh.
apply: (@check_owner_prec _ _ Unit Unit) S1 S2; last first.
- by rewrite unitR.
by rewrite /validH unitR in Vh *.
Qed.

(* now for the main action *)
Lemma trylockP s : 
        trylock_safe s <-> 
        exists n1 b, s \In asafe (check_owner n1 b) /\ (n1 = n -> ~~ b). 
Proof.
split; last first.
- case=>n1 [b][] /check_ownerP []; case=>hL [hE][n2][fL][fE][gL][gE][h] -> C J.
  by split=>//; eauto 15.
case; case=>hL [hE][n1][n2][fL][fE][gL][gE][b][h][-> J] C.
by exists n1, b; split=>//; apply/check_ownerP; split=>//; eauto 15. 
Qed.

Lemma trylockQ s s' r : 
        trylock_step s s' r <-> 
        if r then astep s trylock_ent s' tt 
        else exists n1 b, n1 <> n /\ astep s (check_owner n1 b) s' tt.
Proof.
split.
- case; case=>hL [hE][n1][n2][fL][fE][gL][gE][b][h][->{s}]. 
  case: eqP=>[->{n1}|]; last first.
  - move=>H [->->] [C C']; exists n1, b; split=>//. 
    apply/check_ownerQ; split=>//; apply/check_ownerP. 
    by split=>//; eauto 15.
  case=>->->-> [C C']. 
  eexists _, _, _, _, tt, tt; split=>//; first by move/cohH: C.
  exists h; split; first by exists hL, hE; split=>//; case/split_coh: C'.
  split; first by eauto 15. 
  by case/split_coh: C; case/split_coh: C'.
case: r; last first.
- case=>n1 [b][H] /check_ownerQ [-> Sf]; split; last by case/asf_coh: Sf.
  case/check_ownerP: Sf; case=>hL [hE][n2][fL][fE][gL][gE][h] -> C. 
  by eexists _, _, _, _, _, _, _, _, _, _; split=>//; case: eqP H. 
move=>St; split; last by case/ast_coh: St. 
case: St=>s1 [s2][s1'][s2'][[]][[]][->->{s s'}] _ Vh [h][St1 St2]. 
case: St1 St2 Vh=>hL [hE][->->{s1 s1'}] Cp [].
case=>n2 [fL][fE][gL][gE][->->{s2 s2'}] Ct Vh.
by eexists _, _, _, _, _, _, _, _, _, _; split=>//; rewrite eq_refl.
Qed.

(* the usual lemmas *)
Lemma trylock_safe_coh : safe_coh coh trylock_safe.
Proof. by move=>s []. Qed.

Lemma trylock_safe_zagzig : safe_zagzig trylock_safe.
Proof.
move=>s p /trylockP [n1][b][] /asf_zagzig H J.
by apply/trylockP; exists n1, b. 
Qed.

Lemma trylock_step_safe : step_safe trylock_safe trylock_step.
Proof. 
move=>s s' b /trylockQ G; apply/trylockP.
case: b G; last by case=>n1 [b][H1] /ast_safe; exists n1, b.  
by move/ast_safe/entP; exists n, false. 
Qed.

Lemma trylock_step_guar : step_guar guar trylock_step.
Proof. by move=>s s' b /trylockQ; case: b=>[|[n1][n2][H]] /ast_guar. Qed.

Lemma trylock_step_frame : step_frame trylock_safe trylock_step.
Proof.
move=>s s' p v /trylockP [n1][b][H J] /trylockQ.
case: v; last first.
- case=>n2 [b'][H1] H2.
  case: (check_owner_eq (asf_zagzig H) (ast_safe H2))=>??; subst n1 b.
  case/(ast_frame H): H2=>s'' [->{s'}] St.
  by exists s''; split=>//; apply/trylockQ; exists n2, b'. 
move=>H2; move/entP: (ast_safe H2). 
case/(check_owner_eq (asf_zagzig H))=>??; subst n1 b.
move/entP: H=>H; case/(ast_frame H): H2=>s'' [->{s'}] St.
by exists s''; split=>//; apply/trylockQ.
Qed.

Lemma trylock_step_total : step_total trylock_safe trylock_step.
Proof.
move=>s /trylockP [n1][b][H].
case: (n1 =P n)=>E.
- rewrite {n1}E in H *. move/(_ (erefl _)). case: b H=>//.
  move/entP. move/ast_total. case=>s2 [[]] H _.
 exists s2, true. by apply/trylockQ.
case: b H=>H; last first.
- move=>_. 
  move/ast_total: H. case=>s2 [[]] H.
  exists s2, false. apply/trylockQ. exists n1. exists false. by [].
move=>_.
move/ast_total: H. case=>s2 [[]] H.
exists s2, false. apply/trylockQ. exists n1. exists true. by [].
Qed.

Definition trylock_pre := 
  Action trylock_safe_coh trylock_safe_zagzig trylock_step_safe 
         trylock_step_guar trylock_step_frame trylock_step_total.

Lemma trylock_step_erase : int_step_erase trylock_pre.
Proof.
move=>s h t r s' h' t' r' V E /=.
(move/trylockQ; case: r=>[H1|[n1][b1][N1] /check_ownerQ [-> H1]]);
(move/trylockQ; case: r'=>[H2|[n2][b2][N2] /check_ownerQ [-> H2]])=>//.
- by case: (int_ast_erase V E H1 H2). 
- by move/ast_safe/entP: H1=>H1; case: (check_owner_prec V E H1 H2) N2=>->.
by move/ast_safe/entP: H2; case/(check_owner_prec V E H1).
Qed.

End TickTryLockAct.

Module Exports.

Definition ttryLock hp lkT n := IntAction (@trylock_step_erase hp lkT n).

End Exports.
End TickTryLockAct.

Export TickTryLockAct.Exports.

(*************)
(* Stability *)
(*************)

(* the mutual exclusion lemmas for ticketed locks *)
Section TLockStableBase.
Variable (hp : nat) (lkT : tlockT).
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation V := (TLock lkT).
Notation W := (entangle (Priv hp) V).

(* if we own the called ticket (with or without owning the lock) *)
(* the environment can't change gE, n1, h *)
(* if we don't own the current ticket, and the next served ticket is ours *)
(* then the lock will be unlocked, waiting in the transitory stage *)
(* on us to lock it *)

(* Note: does the existence of this transitory stage look like "helping" *)
(* in the sense of Turon+Dreyer? That is, the process that releases its lock *)
(* and increases the owner counter, in effect gives the lock ownership *)
(* to the next process in line. But, in SCLP we still demand that the *)
(* process that wants to acquire the lock executes the acquire action *)
(* if for nothing else, then to perform the realignment of the ghosts *)

Lemma mtx_tlock_mrely (n1 n2 : nat) (fL fE : nat_set) 
                  (gL gE : U) (b : bool) (h : heap) s : 
        mrely V (dl ->> [(fL, gL), owner :-> (n1, b) \+ 
                                   next :-> n2 \+ h, (fE, gE)]) s ->
        exists (n1' n2' : nat) (b' : bool) h' fE' gE' , 
          s = dl ->> [(fL, gL), owner :-> (n1', b') \+ 
                                next :-> n2' \+ h', (fE', gE')] /\
          if valid (fL \+ #n1) then b' -> valid (fL \+ #n1')
          else [/\ n1' = n1, b' = b, h' = h & gE' = gE].
Proof.
rewrite -[s]st_eta; move=>M; case: (mrely_coh M)=>C _; case: M=>M. 
- case: M; case=>n1' [n2'][n3'][fL'][fE'][fE''][gL'][gE'][b'][h'][].
  move/transp_coh: C; rewrite /transp !fsE /= -2!fsE => C.
  case/(tlock_prec C)=>[[_ <-]] -> <-<- <- <- [->->->] H _.
  by exists n1, n3', b, h, fE'', gE; rewrite -fsE; case: ifP.
case: M=>_ [h'][/Mem_map_inv][ct][->]; rewrite InE /= => ->{ct}.
case. 
- case; case=>n1' [n2'][fL'][fE'][gL'][gE'][gE''][].
  move/transp_coh: C; rewrite /transp !fsE /= -2!fsE => C.  
  rewrite -[_ \+ next :-> n2']unitR.
  case/(tlock_prec C)=>[[??]][??]????; subst fE gL' fE' gE' n1' b n2' h.
  case=>->->-> _. 
  exists n1.+1, n2, false, h', fL', gE''; rewrite fsE; split=>//.
  suff: valid (fL \+ #n1) by move=>->.
  by move: (tlockVF C); rewrite -joinA; move/validR; rewrite joinC.
case; case=>n1' [n2'][fL'][fE'][gL'][gE'][].
move/transp_coh: C; rewrite /transp !fsE /= -2!fsE => C.  
case/(tlock_prec C)=>[[??]][??]????; subst fE gL' fE' gE' n1' b n2' h'.
case=>->->-> _.
exists n1, n2, true, Unit, (fL' \+ #n1), gE; rewrite fsE unitR; split=>//.
suff: valid (fL \+ #n1) by move=>->.
by move: (tlockVF C); rewrite -joinA; move/validR; rewrite joinC.
Qed.

(* mutual exclusion *)
Lemma mtx_tlock (n1 n2 : nat) (fL fE : nat_set) 
                (gL gE : U) (b : bool) (h : heap) s : 
        menv_steps V (dl ->> [(fL, gL), owner :-> (n1, b) \+ 
                                        next :-> n2 \+ h, (fE, gE)]) s ->
        exists (n1' n2' : nat) (b' : bool) h' fE' gE', 
          s = dl ->> [(fL, gL), owner :-> (n1', b') \+ 
                                next :-> n2' \+ h', (fE', gE')] /\
          if valid (fL \+ #n1) then b' -> valid (fL \+ #n1')
          else [/\ n1' = n1, b' = b, h' = h & gE' = gE].
Proof.
case=>n; elim: n n1 n2 b h fE gE=>[|n IH] /= n1 n2 b h fE gE.
- by case; exists n1, n2, b, h, fE, gE; split=>//; case: ifP=>// ->.
case=>s'' [] /mtx_tlock_mrely [n1'][n2'][b'][h'][fE'][gE'][->{s''}] T.
case/IH=>n1'' [n2''][b''][h''][fE''][gE''][->{s IH}] T'.
exists n1'', n2'', b'', h'', fE'', gE''; split=>//.
case: ifP T T'=>E1; last by case=>->->->->; rewrite E1. 
by case: ifP=>E2 // H [_ -> _ _] /H.
Qed.

(* for entangled actions *)
Lemma mtx_priv_tlock (n1 n2 : nat) (hL hE : heap) (fL fE : nat_set) 
                     (gL gE : U) (b : bool) (h : heap) s : 
        menv_steps W 
          (hp ->> [hL, tt, hE] \+
           dl ->> [(fL, gL), owner :-> (n1, b) \+ 
                             next :-> n2 \+ h, (fE, gE)]) s ->
        exists hE' (n1' n2' : nat) (b' : bool) h' fE' gE', 
          s = hp ->> [hL, tt, hE'] \+
              dl ->> [(fL, gL), owner :-> (n1', b') \+ 
                               next :-> n2' \+ h', (fE', gE')] /\
          if valid (fL \+ #n1) then b' -> valid (fL \+ #n1')
          else [/\ n1' = n1, b' = b, h' = h & gE' = gE].
Proof.
move=>M; case: (menvs_coh M); case/split_coh=>/= _ _ Cp Ct C.
case: C M=>_ [s1][s2][->{s}] C1 _. 
case/(menvs_split (entangle_extL _ _) Cp C1)=>Cp' /= Cl' {C1}.
case/stb_ex1: Cp'; case; case=>hE' ->{s1}. 
case/mtx_tlock: Cl'=>n1' [n2'][b'][h'][fE'][gE'][->{s2}] T.
by eexists _, _, _, _, _, _,_.
Qed.

(* a simplification for the common special case *)
Lemma mtx_priv_tlock' (n1 n2 : nat) (hL hE : heap) (fL fE : nat_set) 
                     (gL gE : U) (b : bool) (h : heap) s : 
        menv_steps W 
          (hp ->> [hL, tt, hE] \+
           dl ->> [(fL \+ #n1, gL), owner :-> (n1, b) \+ 
                                    next :-> n2 \+ h, (fE, gE)]) s ->
        exists hE' (n2' : nat) fE', 
          s = hp ->> [hL, tt, hE'] \+
              dl ->> [(fL \+ #n1, gL), owner :-> (n1, b) \+ 
                                       next :-> n2' \+ h, (fE', gE)].
Proof.
case/mtx_priv_tlock=>hE' [n1'][n2'][b'][h'][fE'][gE'][->{s}] T.
exists hE', n2', fE'; case: ifP T; last by move=>_ [->->->->]. 
by rewrite -joinA; move/validR; rewrite validUnEb um_ptsU empbU.
Qed.

End TLockStableBase.


(* for draw action *)
Section TDrawStable.
Variable lkT : tlockT.
Notation dl := (@label_of lkT).        (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)

Lemma tdraw_pre_stable s s' : 
        menv_steps (TLock lkT) s s' -> s' \In asafe (tdraw lkT).
Proof.
move=>M; case: (menvs_coh M)=>_ [Vh Vso].
case=>n1 [n2][fL][fE][gL][gE][b][h][E H1 H2]. 
by split=>//; eexists n1, n2, fL, fE, gL, gE, b, h. 
Qed.

Let Q s r s' := 
  forall (fL : nat_set) (gL : U) (n1 n2 : nat) (b : bool) h fE gE, 
    s = dl ->> [(fL, gL), owner :-> (n1, b) \+ next :-> n2 \+ h, (fE, gE)] -> 
    exists (n1' n2' : nat) (b' : bool) h' fE' gE',
      s' = dl ->> [(fL \+ #r, gL), owner :-> (n1', b') \+ next :-> n2' \+ h', 
                   (fE', gE')] /\ 
      (n1' = r -> ~~ b').

Lemma tdraw_post_conseq s s' r : 
        astep s (tdraw lkT) s' r -> Q s r s'.
Proof.
case; case=>n1 [n2][fL][fE][gL][gE][b][h][->->{s s'}] ->{r} [C _]. 
move=>_ _ _ _ _ _ _ _ /(tlock_prec C) [[<-<-]] _ _ _ _ _. 
exists n1, (n2.+1), b, h, fE, gE; split=>//.
case: (C)=>_ _ [_][_][_][_][_][_][_][_][] 
/(tlock_prec C) [[<-<-]][<-<-]<-<-<-<- H1 [H2].
case=>H E; rewrite -{n2}E in H H1 H2 C *.
- by case: H; rewrite H2 inE ltnn andbF.
by move: H; rewrite ltnn; case=>_ ->. 
Qed.

Lemma tdraw_post_stable s s' s'' r : 
        Q s r s' -> menv_steps (TLock lkT) s' s'' -> Q s r s''.
Proof.
move=>H M fL gL n1 n2 b h fE gE E; rewrite {s}E in H.
case: (H _ _ _ _ _ _ _ _ (erefl _))=>n1' [n2'][b'][h'][fE'][gE'][E' T']. 
rewrite {s'}E' in H M.
case/mtx_tlock: M=>n1'' [n2''][b''][h''][fE''][gE''][->{s''}] T.
eexists _, _, _, _, _, _; split=>// E; rewrite {n1''}E in T.
case: ifP T=>_; last by case; move/esym/T'/negbTE=>->->. 
case: b''=>// /(_ (erefl _)).
by rewrite -joinA; move/validR; rewrite validUnEb um_empbPt. 
Qed.

Lemma val_tdraw (fL : nat_set) (gL : U) (h : heap) fE gE (r : cont nat) :
        (forall n (n1' n2' : nat) (b' : bool) h' fE' gE', 
           (n1' = n -> ~~ b') ->
           dl ->> [(fL \+ #n, gL), owner :-> (n1', b') \+ next :-> n2' \+ h', 
                   (fE', gE')] \In Mod.coh (TLock lkT) ->
           r n (dl ->> [(fL \+ #n, gL), owner :-> (n1', b') \+ next :-> n2' \+ h', 
                        (fE', gE')])) ->
        verify (dl ->> [(fL, gL), h, (fE, gE)]) (act (tdraw lkT)) r.
Proof.
move=>H; apply: val_act=>// s' M; split; first by apply: tdraw_pre_stable M.
case/tlockI: {M} (proj1 (menvs_coh M)) (M)=>n1 [n2][b][k] ->{h} M.
move=>y s'' m /tdraw_post_conseq /tdraw_post_stable H' M'.
move/(_ _ M'): {M'} H' (proj2 (menvs_coh M')).
case/mtx_tlock: M=>n1' [n2'][b'][h'][fE'][gE'][->{s'}] T.
case/(_ _ _ _ _ _ _ _ _ (erefl _))=>n1'' [n2''][b''][h''][fE''][gE''][->{m}].
by apply: H.
Qed.

End TDrawStable.


Section TUnlockStable. 
Variables (hp : nat) (lkT : tlockT).
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (TLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Lemma tunlock_pre_stable s s' : 
        s \In asafe (tunlock hp pf) ->
        menv_steps W s s' -> 
        s' \In asafe (tunlock hp pf). 
Proof.
case; case=>hL [hE][h][n1][n2][fL][fE][gL][gE][->{s}] V inv C M.
split; last by case/menvs_coh: M.
rewrite -[_ \+ next :-> _]unitR in M.
case/mtx_priv_tlock': M=>hE' [n2'][fE'] ->{s'}.
by rewrite unitR; eexists _, _, _, _, _, _, _, _, _.
Qed.

Let Q s (r : unit) s' := 
  forall (hL hE : heap) (fL : nat_set) (gL : U) (n n2 : nat) h fE gE, 
    s = hp ->> [hL \+ h, tt, hE] \+ 
        dl ->> [(fL \+ #n, gL), owner :-> (n, true) \+ 
                                next :-> n2, (fE, gE)] -> 
    I (f gL \+ gE) h ->
    exists hE' (n1' n2' : nat) (b' : bool) h' fE' gE',
      s' = hp ->> [hL, tt, hE'] \+
           dl ->> [(fL, f gL), owner :-> (n1', b') \+ 
                               next :-> n2' \+ h', (fE', gE')]. 

Lemma tunlock_post_conseq s s' v : 
        astep s (tunlock hp pf) s' v -> Q s v s'.
Proof.
case; case=>hL [hE][h][n1][n2][fL][fE][gL][gE][->->{s s'}][C C'].
move=>hL' hE' fL' gL' n1' n2' h' fE' gE'.
rewrite -[_ \+ next :-> n2]unitR -[_ \+ next :-> n2']unitR in C *.
case/(priv_tlock_prec C); case=>E1 _ [<- _] _ [E2 <-][_ <-] _ inv.
exists hE, n1.+1, n2, false, h, fE, gE. 
have [V1 V2] : valid (hL \+ h) /\ valid (fL \+ #n1).
- case/split_coh: C=>/= _ _ /cohH; rewrite /validH priv_flatsE.
  by move/validL=>-> /tlockVF/validL.
have inv' : I (f gL \+ gE) h by case/split_coh: C'=>/= _ _ _ /tlock_inv.
rewrite -(joinC h) -(joinC h') in E1 V1.
move: (pf_prec V1 E1 inv' inv)=>?; subst h'.
by rewrite -{E1}(joinfK V1 E1) -(joinKf V2 E2). 
Qed.

Lemma tunlock_post_stable s s' s'' r : 
        Q s r s' -> menv_steps W s' s'' -> Q s r s''.
Proof.
move=>H M hL hE fL gL n1 n2 h fE gE E inv; rewrite {s}E in H.
move: (H hL hE fL gL n1 n2 h fE gE (erefl _) inv). 
case=>hE' [n1'][n2'][b'][h'][fE'][gE'] E'; rewrite {s'}E' in H M.
case/mtx_priv_tlock: M=>hE'' [n1''][n2''][b''][h''][fE''][gE''][->{s''}].
by eexists _, _, _, _, _, _, _.
Qed.

Lemma val_tunlock (hL hE h : heap) (fL fE : nat_set) (gL gE : U) 
                  (n n2 : nat) (r : cont unit):
      valid (f gL \+ gE) ->
      I (f gL \+ gE) h -> 
      (forall hE' (n1' n2' : nat) (b' : bool) h' fE' gE', 
          hp ->> [hL, tt, hE'] \+ 
          dl ->> [(fL, f gL), owner :-> (n1', b') \+ 
                              next :-> n2' \+ h', (fE', gE')] \In Mod.coh W -> 
          r tt (hp ->> [hL, tt, hE'] \+ 
               dl ->> [(fL, f gL), owner :-> (n1', b') \+ 
                                   next :-> n2' \+ h', (fE', gE')])) ->
       verify (hp ->> [hL \+ h, tt, hE] \+
               dl ->> [(fL \+ #n, gL), owner :-> (n, true) \+ 
                                       next :-> n2, (fE, gE)]) 
              (act (tunlock hp pf)) r.
Proof.
move=>V inv H; apply: val_act=>// s' M; split. 
- apply: tunlock_pre_stable (M); split; last by case: (menvs_coh M). 
  by exists hL, hE, h, n, n2, fL, fE, gL, gE. 
case=>s'' m /tunlock_post_conseq /tunlock_post_stable H' M'.
move/(_ _ M'): {M'} H' (proj2 (menvs_coh M')).
rewrite -[_ \+ next :-> _]unitR in M.
case/mtx_priv_tlock': M=>hE' [n2'][fE'] ->{s'}; rewrite unitR.
case/(_ _ _ _ _ _ _ _ _ _ (erefl _) inv). 
by move=>hE'' [n1''][n2''][b''][h''][fE''][gE''] ->{m}; apply: H.
Qed.

End TUnlockStable.


Section TTryLockStable. 
Variables (hp : nat) (lkT : tlockT) (n : nat). 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (TLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Lemma ttrylock_pre_stable s s' : 
        s \In asafe (ttryLock hp lkT n) ->
        menv_steps W s s' -> 
        s' \In asafe (ttryLock hp lkT n). 
Proof.
case; case=>hL [hE][n1][n2][fL][fE][gL][gE][b][h][->{s} J] C M. 
split; last by case/menvs_coh: M. 
case/mtx_priv_tlock: M=>hE' [n1'][n2'][b'][h'][fE'][gE'][->{s'}] T.
eexists _, _, _, _, _, _, _, _, _, _; split=>// H.
move: H T J=>->{n1'}; case: ifP=>_; last by case=>->-> _ _; apply.
case: b'=>// /(_ (erefl _)). 
by rewrite -joinA; move/validR; rewrite validUnEb um_ptsU empbU. 
Qed.

Let Q s (r : bool) s' := 
  forall (hL hE : heap) (n1 n2 : nat) (fL fE : nat_set) 
         (gL gE : U) (b : bool) (h : heap), 
    s = hp ->> [hL, tt, hE] \+ 
        dl ->> [(fL \+ #n, gL), owner :-> (n1, b) \+ 
                                next :-> n2 \+ h, (fE, gE)] -> 
    if n1 == n then 
      [/\ b = false, r = true &
          exists hE' (n2' : nat) fE',
            s' = hp ->> [hL \+ h, tt, hE'] \+
               dl ->> [(fL \+ #n, gL), owner :-> (n1, true) \+ next :-> n2',
                       (fE', gE)]]
    else 
      r = false /\ exists hE' (n1' n2' : nat) (b' : bool) h' fE' gE',
        s' = hp ->> [hL, tt, hE'] \+
             dl ->> [(fL \+ #n, gL), owner :-> (n1', b') \+ next :-> n2' \+ h',
                     (fE', gE')] /\
             (n1' = n -> ~~ b').

Lemma ttrylock_post_conseq s b s' : 
        astep s (ttryLock hp lkT n) s' b -> Q s b s'.
Proof.
case; case=>hL [hE][n1][n2][fL][fE][gL][gE][r][/= h][->{s}] D [C] _. 
move=>_ _ _ _ _ _ _ _ _ _ /(priv_tlock_prec C) [[<-]]_[<-_]<-[<-<-][_<-]<-.
case: ifP D {C}=>N. 
- by case=>->->->; split=>//; exists hE, n2, fE. 
by case=>->->; split=>//; exists hE, n1, n2, r, h, fE, gE; case: eqP N.
Qed.

Lemma ttrylock_post_stable s s' s'' r : 
        Q s r s' -> menv_steps W s' s'' -> Q s r s''.
Proof.
move=>H M hL hE n1 n2 fL fE gL gE b h E; rewrite {s}E in H.
move: (H hL hE n1 n2 fL fE gL gE b h (erefl _))=>{H} H. 
case: ifP H M=>N.
- rewrite {N n1}(eqP N); case=>->-> [hE'][n2'][fE'] ->{s'}.
  rewrite -[_ \+ next :-> _]unitR.
  case/mtx_priv_tlock'=>hE'' [n2''][fE''] ->{s''}; rewrite unitR.
  by split=>//; eauto 15.
case=>-> [hE'][n1'][n2'][b'][h'][fE'][gE'][->{s'}] T.
case/mtx_priv_tlock=>hE'' [n1''][n2''][b''][h''][fE''][gE''][->{s''}] T'.
split=>//; eexists _, _, _, _, _, _, _; split=>// E; rewrite {n1''}E in T'.
case: ifP T'=>_; last by case; move/esym/T/negbTE=>->->. 
case: b''=>// /(_ (erefl _)).
by rewrite -joinA; move/validR; rewrite validUnEb um_empbPt. 
Qed.

Lemma val_ttrylock (hL hE : heap) (fL fE : nat_set) (gL gE : U) 
                   (n1 n2 : nat) (b : bool) h (r : cont bool) : 
        (* we're not trying to grab while already holding the lock *)
        (n1 = n -> ~~ b) ->
        (* if we succeed *)
        (forall hE' fE' gE' (n2' : nat) h', 
           valid (gL \+ gE') ->
           I (gL \+ gE') h' ->
           hp ->> [hL \+ h', tt, hE'] \+ 
           dl ->> [(fL \+ #n, gL), owner :-> (n, true) \+ 
                                   next :-> n2', (fE', gE')] \In Mod.coh W ->
           r true (hp ->> [hL \+ h', tt, hE'] \+ 
                   dl ->> [(fL \+ #n, gL), owner :-> (n, true) \+ 
                                            next :-> n2', (fE', gE')])) ->
        (* if we fail *)
        (forall hE' fE' gE' (n1' n2' : nat) (b' : bool) h', 
           (n1' = n -> ~~ b') ->
           hp ->> [hL, tt, hE'] \+ 
           dl ->> [(fL \+ #n, gL), owner :-> (n1', b') \+ 
                   next :-> n2' \+ h', (fE', gE')] \In Mod.coh W ->
           r false (hp ->> [hL, tt, hE'] \+ 
                    dl ->> [(fL \+ #n, gL), owner :-> (n1', b') \+ 
                            next :-> n2' \+ h', (fE', gE')])) ->
        verify (hp ->> [hL, tt, hE] \+ 
                dl ->> [(fL \+ #n, gL), owner :-> (n1, b) \+ 
                                        next :-> n2 \+ h, (fE, gE)])
               (act (ttryLock hp lkT n)) r.
Proof.
move=>V H1 H2; apply: val_act=>// s' M; split. 
- apply: ttrylock_pre_stable (M); split; last by case: (menvs_coh M).
  by eauto 15.
move=>y s'' m /ttrylock_post_conseq /ttrylock_post_stable H' M'.
move/(_ _ M'): {M'} H' (proj2 (menvs_coh M')).
case/mtx_priv_tlock: M (M)=>hE' [n1'][n2'][b'][h'][fE'][gE'][->{s'}] _ M.
move/(_ _ _ _ _ _ _ _ _ _ _ (erefl _)); case: ifP=>E; last first.
- by case=>-> [hE''][n1''][n2''][b''][h''][fE''][gE''][->{m} T] C'; apply: H2.
case/menvs_coh: M=>_; rewrite {E n1'}(eqP E); case/split_coh=>/= _ _ _ C H. 
case: H C=>->-> [hE''][n2''][fE''] ->{m} C.
by apply: H1; [apply: tlockVG C | apply: tlock_inv C]. 
Qed.

End TTryLockStable. 


(**********************************************)
(************* More abstract lemmas ***********)
(**********************************************)
Section TLockProperties. 
Variables (hp : nat) (lkT : tlockT) (n : nat). 
Notation l := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (TLock lkT)). 
Notation coh := (Mod.coh W).
Notation guar := (Mod.internal W). 

Definition t_is_lock (gl ge : U) s := 
  exists (fl fe: nat_set) (n1 n2: nat) (b: bool) (h: heap),
    s = l ->> [(fl, gl), owner :-> (n1, b) \+ next :-> n2 \+ h, (fe, ge)] /\  
    s \In Mod.coh (TLock lkT).

Lemma t_is_lock_coh gl ge s: t_is_lock gl ge s -> s \In Mod.coh (TLock lkT).
Proof. by case=>fl[fe][n1][n2][b][h]; case. Qed.

Lemma t_islock_sym gl ge s: t_is_lock gl ge s <-> t_is_lock ge gl (transp s).
Proof. 
split.
* case=>fl[fe][n1][n2][b][h][->{s}] C; exists fe, fl, n1, n2, b, h.
  move/transp_coh: C=> C; split=>//; last by rewrite fsE /transp /= -fsE.
* case=>fl[fe][n1][n2][b][h][T]C. exists fe, fl, n1, n2, b, h.
  move/transp_coh: C=> C; split=>//{C}. 
  rewrite fsE /transp /= in T. rewrite -[s]st_eta. 
  by case: T=>->->->; rewrite -fsE.
Qed.

Definition t_locked (gl ge: U) s :=
  (exists (fl fe : nat_set) (n1 n2: nat),
    s = l->>[(fl \+ #n1, gl), owner:->(n1, true) \+ next:->n2, (fe, ge)]) /\
  s \In Mod.coh (TLock lkT).

Lemma t_locked_is_lock s gl ge : t_locked gl ge s -> t_is_lock gl ge s.
Proof.
case=>H1 H2'; case: H1=>fl[fe][n1][n2]D; subst s.
by exists (fl \+ #n1), fe, n1, n2, true, Unit; rewrite unitR; split=>//.
Qed.
 
(* Mutual exclusion lemma *) 
Lemma t_mtx s (gl ge: U) :
  t_locked gl ge s -> t_locked ge gl (transp s) -> False.
Proof.
case; case=> fl[fe][n1][n2][->{s}] C.
case; case=>  fl'[fe'][n1'][n2']E C'.
rewrite !fsE /transp /= -!fsE in C' E.
case/scancelPt: E=> D1 D2 D3. subst fe' fe. 
case: (C)=>Vh _ _; rewrite /validH tlock_flatsE in Vh.
case/(hcancelV Vh): D2; case=> D1 D2 _; subst n1'; rewrite -D2 in C C'=> {D2}.
case: C=> _ Vs _. rewrite /validSO !fsE zipPt allvalidPt /= /valid /= in Vs.
case/andP: Vs. rewrite joinAC -2!joinA; move/validR /validR. 
by rewrite validUnEb um_ptsU empbU.
Qed.

Definition t_can_try (n: nat)(gl: U) s := 
  (exists (fl fe: nat_set)(ge: U)(n1 n2: nat)(b: bool) h,
   s = l->>[(fl \+ #n, gl), owner:->(n1, b) \+ next:->n2 \+ h, (fe, ge)] /\
   (n1 = n -> ~~b)) /\
  s \In Mod.coh (TLock lkT).

Lemma val_tdraw' (gl ge: U) i (r : cont nat) :
        t_is_lock gl ge i -> 
        (forall n j, t_can_try n gl j -> r n j) ->
        verify i (act (tdraw lkT)) r.
Proof.
case=>fl[fe][n1][n2][b][h][->{i}] C'.
case:(C')=>_ _ [n1'][n2'][fl'][fe'][gl'][ge'][b'][h']; case.
case/tlock_prec=>//[[<-][<-]][[<-][<-]]<-<-<-<-_ Eit H1 {n1' n2' fl' fe' gl' ge' b' h'}.
apply: val_tdraw=>n' n1' n2' b' h' fe' ge' T C. 
by apply: (H1 n'); split=>//; eauto 15. 
Qed.
  
Lemma val_ttrylock' (hl he: heap) (gl: U) i (r : cont bool) : 
        t_can_try n gl i ->
        (* if we succeed *)
        (forall (ge': U) he' h' s,
           valid (gl \+ ge') ->
           I (gl \+ ge') h' ->
           t_locked gl ge' s ->
           hp->>[hl \+ h', tt, he'] \+ s \In Mod.coh W ->
           r true (hp->>[hl \+ h', tt, he'] \+ s)) ->
        (* if we fail *)
        (forall j he',
           t_can_try n gl j ->
           hp->>[hl, tt, he'] \+ j \In Mod.coh W ->
           r false (hp->>[hl, tt, he'] \+ j)) ->        
        verify (hp->>[hl, tt, he] \+ i) 
               (act (ttryLock hp lkT n)) r.
Proof.
case; case=>fl'[fe][ge][n1][n2][b][h][->{i}] E C H1 H2.
apply: val_ttrylock=>//. 
- move=>he' fe' ge' n2' h' V inv C'; apply: H1 V inv _ (C').
  by case/split_coh: C'=>/= _ _ _ Ct; split=>//; eauto 10. 
move=>he' fe' ge' n1' n2' b' h' T C'; apply: H2=>//. 
by case/split_coh: C'=>/= _ _ _ Ct; split=>//; eauto 10. 
Qed.

Lemma val_tunlock' (hL hE h : heap)(gL gE : U) s (r : cont unit):
      valid (f gL \+ gE) ->
      I (f gL \+ gE) h -> 
      t_locked gL gE s ->
      (forall i hE' gE',
          t_is_lock (f gL) gE' i -> 
          r tt (hp ->> [hL, tt, hE'] \+ i)) ->
       verify (hp ->> [hL \+ h, tt, hE] \+ s) 
              (act (tunlock hp pf)) r.
Proof.
move=> V inv L H.
case: L; case=> fl[fe][n1][n2][->{s}]C.
apply: val_tunlock=>//. 
move=> he' n1' n2' b' h' fe' ge' C'.
apply: (H _ _ ge'); exists fl, fe', n1', n2', b', h'.
split=>//;last by case/split_coh: C'.
Qed.

End TLockProperties.

(* Programming the lock procedure *)
Section TLockProgram.
Variables (hp : nat) (lkT : tlockT) (n : nat). 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Notation W := (entangle (Priv hp) (TLock lkT)). 
Notation coh := (Mod.coh W).

(***************************************************)
(************** Locking procedure ******************)
(***************************************************)

Definition tlock_spec : spec unit := lock_spec hp (@t_is_lock lkT) (@t_locked lkT).

(* Inner loop invariant *)
Definition t_loop_spec (n: nat) : spec unit := 
  (fun i => exists (hl he : heap)(gl: U) i', 
     i = hp ->> [hl, tt, he] \+ i' /\ t_can_try n gl i', 
   fun i y m => forall hl he (gl: U) i', 
     i = hp ->> [hl, tt, he] \+ i' -> t_can_try n gl i' -> 
     exists h ge' he' m',
       [/\ m = hp ->> [hl \+ h, tt, he'] \+ m', 
           t_locked gl ge' m',
           valid (gl \+ ge') & I (gl \+ ge') h]).

(* Locking program itself *)
Program Definition tlock : STsep W (tlock_spec) :=
 Do (n <-- extend _ (act (tdraw lkT)); 
     ffix (fun (loop : unit -> STsep W (t_loop_spec n)) (_ : unit) =>
       Do (b <-- act (ttryLock hp lkT n);
           if b then ret tt else loop tt)) tt).

Next Obligation. by apply: entangle_extR. Defined.

Next Obligation. 
rewrite !ghs; apply: ghC=>i [[[hl he]] gl] m /=[->{i} T]; apply: step.
apply: val_ttrylock' T _ _=>[gE' hE' h' s V inv T _ | j he' Ct C]; last first.
- apply: val_do=>[|[] m' /(_ hl he' gl j)]; first by exists hl, he', gl, j.
  case=>// h [ge'][he''][m''][->{m'}] L V inv C'.
  by exists h, ge', he'', m''. 
case: T; case=>fl [fe][n1][n2] ->{s} _.
apply: val_ret=>m'; rewrite -(unitR (next:->_)) joinA => M.
case/mtx_priv_tlock' : M (M)=> he'' [n2''][fe''] ->{m'} M.
eexists h', gE', he'', _; split=>//.
case/menvs_coh: M=>_ /split_coh [] /= _ _ _ Ct.
by split=>//; rewrite unitR; eauto 10.
Qed.

Next Obligation.
rewrite !ghs; apply: ghC=>i[[[[hl he] gl] ge] m] /=[->{i}] H. 
case:(H)=>fl[fe][n1][n2][b][h][D] C; subst m.
apply: step; rewrite joinC.
apply: val_extend=>//; apply (@val_tdraw' _ gl ge)=>// n' j T j' _ /= M. 
case/mtx_priv: M=>hE' ->; rewrite joinC. 
case: T (T); case=> fl'[fe'][ge'][n1'][n2'][b'][h'][->{j}] _ _ T.
apply: val_do=>[|[] m' /(_ hl hE' gl _ (erefl _)) H' _]; last by apply: H'.
by eexists _, _, gl, _.
Qed.

(***************************************************)
(************** Unlocking procedure ****************)
(***************************************************)


(* Unlocking procedure specification *)
Definition tunlock_spec : spec unit := @unlock_spec hp lkT f (@t_is_lock lkT) (@t_locked lkT).

(* Unlocking procedure itself *)
Program Definition tunlock : STsep W (tunlock_spec) :=
 Do (act (tunlock hp pf)).

Next Obligation.
rewrite !ghs /=. apply: ghC=>i[[[[[gl ge] hl] he] h] s] /=.
case=> V inv L ->{i}. 
apply: (val_tunlock' V)=>// i he' ge' T.
by exists he', ge', i.
Qed.

End TLockProgram.


(* Instantiating the lock interface *)
Section TLockSpec.
Variables (hp : nat) (lkT : tlockT). 
Notation W := (entangle (Priv hp) (TLock lkT)). 

Definition ticketed_lock := 
  Lockish (@t_is_lock_coh lkT) (@t_islock_sym lkT) (@t_locked_is_lock lkT) 
          (@t_mtx lkT) (@tlock hp lkT) (@tunlock hp lkT).

End TLockSpec.


(*************************  INCOMPLETE BELOW  *****************************)


Section TLockCreate.
Variables (hp : nat) (lkT : tlockT) (n : nat). 
Notation dl := (@label_of lkT).       (* feap label *)
Notation owner := (@owner_of lkT).    (* owner pointer *)
Notation next := (@next_of lkT).      (* next pointer *)
Notation U := (@pcm_of lkT).          (* pcm *) 
Notation I := (@inv_of lkT).          (* resource invariant *)
Notation pf_prec := (@precise_of lkT).
Variables (f : U -> U) (pf : local f).
Variable V : mod.
Notation W := (entangle (entangle (Priv hp) V) (TLock lkT)). 
Notation coh := (Mod.coh W).

(***************************************************)
(************** Creating the lock  *****************)
(***************************************************)
Variables (A : Type) (gL : U) (prog : ST W A).

(*

[Ilya] Some observations:

The lemma val_privhide' accounts for ``additional'' concurroids, when
making the retraction. This affects the following specification: we
need to mention some state i' explicitly, and the following proof is
cumbersome (and is, eventually, incomplete). Moreover, this is not
modular at all: in order to inject the lock, the client *needs to
know* about *all* other locks in the system.

The natural questions are:

- Can we formulate the val_privhide lemma in a small footprint-like way?  
*)

(*
Definition my_pre (h hl he: heap) :=
     pre_of prog (hp ->> [hl, tt, he] \+
     dl ->> [((Unit: nat_set), gl), owner:->(0, false) \+ next:->0 \+ h, ((Unit: nat_set), ge)]).

Definition my_post (h hl he: heap) :=
     post_of prog (hp ->> [hl, tt, he] \+
     dl ->> [((Unit: nat_set), gl), owner:->(0, false) \+ next:->0 \+ h, ((Unit: nat_set), ge)]).
*)

Definition create_tlock_spec : spec A := 
  (fun i => exists (hL hE : heap) (j s : state),
              ((i = hp ->> [hL \+ flats s, tt, hE] \+ j /\ 
                tloc_dec (Unit:nat_set, gL) s) /\
                valid (hp ->> [hL, tt, hE] \+ j \+ s)) /\
              pre_of prog (hp ->> [hL, tt, hE] \+ j \+ s),
   fun i y m => 
     forall hL hE j s, 
       i = hp ->> [hL \+ flats s, tt, hE] \+ j ->
       tloc_dec (Unit:nat_set, gL) s -> 
       valid (hp ->> [hL, tt, hE] \+ j \+ s) ->
       pre_of prog (hp ->> [hL, tt, hE] \+ j \+ s) ->
     exists hL' hE' j' fL' gL' s', 
       m = hp ->> [hL' \+ flats s', tt, hE'] \+ j' /\ 
       tloc_dec (lkT := lkT) (fL', gL') s' /\
       post_of prog 
        (hp ->> [hL, tt, hE] \+ j \+ s)
        y
        (hp ->> [hL', tt, hE'] \+ j' \+ s')).

Program Definition create_tlock : STsep (entangle (Priv hp) V) create_tlock_spec :=
 Do (priv_hide (TLockDecorate lkT) (Unit, gL) prog).

Next Obligation. 
rewrite !ghs. apply: ghC=>i [[[hL hE]]] j s /=.  
case. case.  case.  move=>->{i} T Vjs H. 
apply: vrf_coh=>C.
apply: val_privhide' (T) (Vjs) _.
apply: val_do H _.
move=>x m H Cm i' j' hL' hE' gL' ?. subst m. move=>T'.
exists hL'. exists hE'. exists i'. exists (gL'.1). exists gL'.2.
exists j'.
split=>//.
split=>//.
by case: gL' T'.
Qed.

End TLockCreate.

