Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq path rels ordtype.
Require Import Eqdep prelude pcm unionmap heap codes feaps state idynamic.
Require Import Setoid Morphisms.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* we have three different notions of validity on states *)
(* ordinary valid s checks if the labels of s are disjoint *)
(* validH s checks if the flattening heap of s is defined *)
(* validSO checks if the self and other component of s *)
(* zip to a valid value *)
(* I would have liked to combine all these three into one *)
(* but that doesn't seem possible *)
(* 1. we can't combine valid and validSO, because validSO *)
(* doesn't satisfy the same validL and validR laws *)
(* we can change the definition of state to be not a triple of feaps *)
(* but a feap of triples; then we can make validSO satisfy the right *)
(* laws, but doing so would make it difficult to state the always_par *)
(* lemmas, where we directly work with triples of feaps, not a feap of *)
(* triples *)
(* 2. we can perhaps combine validH with valid, since they do satisfy *)
(* the same laws, but then footprint equality becomes a bit tricky *)
(* do define, because we need to take a conjunction of footeqV *)
(* and footeqH. But footeqH is stated on heaps, not states, to *)
(* acommodate for transfer action, so that's out too *)

(***************)
(* Definitions *)
(***************)

(* zig: union onto self *)
Definition zig s p := nosimpl
  (let: [lp, sh, ep] := s in [p \* lp, sh, ep]).

(* zag: union onto other *)
Definition zag s p := nosimpl
  (let: [lp, sh, ep] := s in [lp, sh, ep \* p]).

Notation "s \< p" := (zig s p) 
  (at level 34, left associativity).

Notation "s \> p" := (zag s p) 
  (at level 34, left associativity).

(* predicates used in the definitions of coherence *)
(* these will be used in addition to the valid s predicate *)
(* which is defined previously *)
Definition validH s := valid (flats s).
Definition validSO s := allvalid (loc s \* env s).

(* three footprint equalities, wrt. valid, validH and validSO *)

Definition footeqV (s1 s2 : state) := 
  [&& fdom_eq (loc s1) (loc s2), 
      fdom_eq (shr s1) (shr s2) &
      fdom_eq (env s1) (env s2)].

(* The next definition states that *heaps* h1 and h2 have the same domain *)
(* a version of this already exists in heaps.v and is called dom_eq *)
(* however, dom_eq uses def and :+ instead of valid and \+ *)
(* also, I wanted to use Prop as a range for footeqH instead of bool *)
(* in order to make it possible to rewrite by setoids *)
(* so i repeat the development *)

Definition footeqH (h1 h2 : heap) := 
  [/\ valid h1 = valid h2 & dom h1 =i dom h2].

(*****************************************)
(* Helper lemmas for zigging and zagging *) 
(*****************************************)

(*
* zigzag pattern, move from Self to Other:
* P (p \> s) -> P (s \< p)
* zagzig pattern, move from Other to Self:
* P (s \< p) -> P (p \> s)
*)

Lemma zigE lp sh ep p : [lp, sh, ep] \< p = [p \* lp, sh, ep].
Proof. by []. Qed.

Lemma zagE lp sh ep p : [lp, sh, ep] \> p = [lp, sh, ep \* p].
Proof. by []. Qed.

Lemma sp_eta lp sh ep : [lp, sh, ep] = [[zunit lp], sh, [zunit ep]] \< lp \> ep.
Proof. by rewrite zigE zagE zunitR zunitL. Qed.

Lemma spSS s p1 p2 : s \< p1 \< p2 = s \< (p1 \* p2).
Proof. by rewrite -[s]st_eta !zigE zipCA zipA. Qed.

Lemma spOO s p1 p2 : s \> p1 \> p2 = s \> (p1 \* p2).
Proof. by rewrite -[s]st_eta !zagE zipA. Qed.

Lemma spSO s p1 p2 : s \< p1 \> p2 = s \> p2 \< p1.
Proof. by rewrite -[s]st_eta zigE !zagE zigE. Qed.

Lemma spOS s p1 p2 : s \> p1 \< p2 = s \< p2 \> p1.
Proof. by rewrite spSO. Qed.

Lemma spSloc s p : loc (s \< p) = p \* loc s.
Proof. by rewrite -[s]st_eta. Qed.

Lemma spSshr s p : shr (s \< p) = shr s.
Proof. by rewrite -[s]st_eta. Qed.

Lemma spSenv s p : env (s \< p) = env s.
Proof. by rewrite -[s]st_eta. Qed.

Lemma spOloc s p : loc (s \> p) = loc s.
Proof. by rewrite -[s]st_eta. Qed.

Lemma spOshr s p : shr (s \> p) = shr s.
Proof. by rewrite -[s]st_eta. Qed.

Lemma spOenv s p : env (s \> p) = env s \* p.
Proof. by rewrite -[s]st_eta. Qed.

Definition spS := (spSloc, spSshr, spSenv).
Definition spO := (spOloc, spOshr, spOenv).
Definition spP := (spS, spO).

Section ZigZagUnion.
Variables (U : encoded_pcm) (V : encoded_type).
Variables (l : nat) (s : state) (p : feap dpcm).
Variables (lp : U) (sh : V) (ep : U) (v : U).

Lemma zigUnPt : 
        (l ->> [lp, sh, ep] \+ s) \< (l \-> v \+ p) =
        (l ->> [lp, sh, ep] \< l \-> v) \+ (s \< p).
Proof. by rewrite fsE -[s]st_eta /= !zigE /= zipUnPt zipPt. Qed.

Lemma zagUnPt :
        (l ->> [lp, sh, ep] \+ s) \> (l \-> v \+ p) =
        (l ->> [lp, sh, ep] \> l \-> v) \+ (s \> p). 
Proof. by rewrite fsE -[s]st_eta /= !zagE /= zipUnPt zipPt. Qed. 

End ZigZagUnion.

(********************************************)
(* Helper lemmas for valid, validH, validSO *)
(********************************************)

Lemma validHV s : validH s -> valid s.
Proof.
rewrite /validH flatsE {2}/valid /= !flatE /=. 
case: (loc s)=>[//|f1]. 
case: (shr s)=>[|f2]; first by rewrite -(joinC um_undef).
by case: (env s)=>//; rewrite -(joinC um_undef). 
Qed.

Lemma validHUn s1 s2 : 
        validH (s1 \+ s2) =
        valid (s1 \+ s2) && valid (flats s1 \+ flats s2).
Proof.
apply/idP/idP; last by case/andP=>V Vf; rewrite /validH flatsUn.
move=>V; rewrite (validHV V) /=.
by move: (V); rewrite /validH (flatsUn (validHV V)).  
Qed.

Lemma validHL s1 s2 : validH (s1 \+ s2) -> validH s1.
Proof. by rewrite validHUn; case/andP=>_; move/validL. Qed.

Lemma validHR s1 s2 : validH (s1 \+ s2) -> validH s2.
Proof. by rewrite joinC; move/validHL. Qed.

(* this one is not iterative; has to be automated by canonical structures *)
Lemma validHUn3 s1 s2 s3 : 
        validH (s1 \+ s2 \+ s3) = 
        valid (s1 \+ s2 \+ s3) && 
        valid (flats s1 \+ flats s2 \+ flats s3).
Proof.
apply/idP/idP.
- by rewrite validHUn; case/andP=>V Vf; rewrite -(flatsUn (validL V)) V Vf. 
by case/andP=>V Vf; rewrite validHUn V (flatsUn (validL V)). 
Qed.

(* can we bake valid, validH, validSO into one predicate *)
(* and have that be the validity for PCM of state? *)

(* unfortunately, this does not seem possible *)
(* if that were to work, we'd need the following lemma *)
(* which is not true *)

Lemma validSOL s1 s2 : validSO (s1 \+ s2) -> validSO s1.
Proof.
rewrite /validSO. simpl.
move=>V. 
move: (V); rewrite -zipUn.
- rewrite allvalidUn /=.
  case/andP. case/andP. by [].
  (* hmm, okay, this doesn't hold *)
  (* the lemma is not true *)
Abort.

(* But we should have the following 'cancellation' property *)
Lemma validSOK s1 s2 : validSO s1 -> validSO (s1 \+ s2) -> validSO s2.
Proof.
rewrite /validSO /=; move/allvalidV=>V A; case: validZP V=>//.
case=>Vl Ve Vz _.
have Wl: valid (loc s2).
- by move/allvalidV: A; case/validZP; move/validR. 
have We: valid (env s2).
- by move/allvalidV: A; case/validZP=>_; move/validR. 
have Wz: [zunit loc s2] = [zunit env s2].
- move/allvalidV: (A); case/validZP=>_ _; rewrite !zunitUn -Vz. 
  have D: valid ([zunit loc s1] \+ [zunit loc s2]).
  - by rewrite -zunitUn valid_zunit; move/allvalidV: A; case/validZP.
  by move/(funhKl D).
rewrite -zipUn in A; last first.
- by apply/validZP; rewrite Wl We Wz.
- by apply/validZP; rewrite Vl Ve Vz.
by move: A; rewrite allvalidUn; case/andP. 
Qed.

Lemma validH_unit : validH Unit.
Proof. by rewrite /validH flats_unit. Qed.

Lemma validSO_unit : validSO Unit.
Proof. by rewrite /validSO /= -{1}zunit_unit zunitL allvalid_unit. Qed.

Hint Resolve validH_unit validSO_unit.

Lemma validSO_dom s : validSO s -> fdom_eq (loc s) (env s).
Proof.
case/allvalidV/validZP=>V1 V2 F.
apply/fdomeqP; rewrite V1 V2; split=>// x.
by rewrite -(fdomZU (loc s)) -(fdomZU (env s)) F.
Qed.

Lemma validSOUn s1 s2 : 
        valid (s1 \+ s2) -> validSO s1 -> validSO s2 ->
        validSO (s1 \+ s2).
Proof.
rewrite /validSO=>/and3P [V _ _] H1 H2.
move: (allvalidV H1) (allvalidV H2)=>V1 V2. 
rewrite -zipUn // allvalidUn H1 H2 !andbT. 
case: fvalidUn=>//; rewrite ?V1 ?V2 //.
move=>x; rewrite !fdomZL V1 V2. 
by case: fvalidUn V=>// _ _ L _; move/L/negbTE=>->.
Qed.
  
(******************************************)
(* Helper lemmas for footprint equalities *)
(******************************************)

(* footeqV *)

Lemma footeqV_refl s : footeqV s s.
Proof. by apply/and3P. Qed.

Hint Resolve footeqV_refl.

Lemma footeqV_sym s1 s2 : footeqV s1 s2 = footeqV s2 s1.
Proof.
rewrite /footeqV (fdomeq_sym (loc s1)). 
by rewrite (fdomeq_sym (shr s1)) (fdomeq_sym (env s1)). 
Qed.

Lemma footeqV_trans s1 s2 s3 : 
        footeqV s1 s2 -> footeqV s2 s3 -> footeqV s1 s3.
Proof.
case/and3P=>E1 E2 E3 /and3P [F1 F2 F3]; apply/and3P; split.
- by apply: fdomeq_trans E1 F1.
- by apply: fdomeq_trans E2 F2.
by apply: fdomeq_trans E3 F3.
Qed.

Lemma footeqV_validUn s1 s2 s1' s2' : 
        footeqV s1 s2 -> footeqV s1' s2' -> 
        valid (s1 \+ s1') = valid (s2 \+ s2').
Proof.
case/and3P=>E1 E2 E3 /and3P [F1 F2 F3]; rewrite /valid /=.
by rewrite (fdomeq_validUn E1 F1) (fdomeq_validUn E2 F2)
           (fdomeq_validUn E3 F3).
Qed.

(* special forms to simplify notation in proofs *)
(* should eventually be replaced by some kind of setoid setting *)

Lemma footeqV_validUnL {s} s1 s2 : 
        footeqV s1 s2 -> valid (s1 \+ s) = valid (s2 \+ s).
Proof. by move=>H; apply: footeqV_validUn H _. Qed.

Lemma footeqV_validUnR {s} s1' s2' : 
        footeqV s1' s2' -> valid (s \+ s1') = valid (s \+ s2').
Proof. by apply: footeqV_validUn. Qed.

(* union is a parametric morphism for footeqV *)
Lemma footeqVUn s1 s2 s1' s2' : 
        footeqV s1 s2 -> footeqV s1' s2' -> 
        footeqV (s1 \+ s1') (s2 \+ s2').
Proof.
case/and3P=>E1 E2 E3 /and3P [F1 F2 F3]; rewrite /footeqV.
by rewrite (fdomeqUn E1 F1) (fdomeqUn E2 F2) (fdomeqUn E3 F3).
Qed.

(* special forms *)
Lemma footeqVUnL {s} s1 s2 : 
        footeqV s1 s2 -> footeqV (s1 \+ s) (s2 \+ s).
Proof. by move=>H; apply: footeqVUn H _. Qed.

Lemma footeqVUnR {s} s1' s2' : 
        footeqV s1' s2' -> footeqV (s \+ s1') (s \+ s2').
Proof. by apply: footeqVUn. Qed.

Lemma footeqV_prec s1 s2 s1' s2' : 
        valid (s1 \+ s1') ->
        s1 \+ s1' = s2 \+ s2' ->
        footeqV s1 s2 -> s1 = s2.
Proof.
case/and3P=>/= V1 V2 V3 /stE [E1 E2 E3] /and3P [].
case/fdomeqP=>_ F1 /fdomeqP [_ F2] /fdomeqP [_ F3]; apply/stE.
by rewrite (fdomE V1 E1 F1) (fdomE V2 E2 F2) (fdomE V3 E3 F3).
Qed.

Lemma footeqVK s1 s2 s1' s2' : 
        valid (s1 \+ s1') ->
        footeqV (s1 \+ s1') (s2 \+ s2') ->
        footeqV s1 s2 -> footeqV s1' s2'.
Proof.
case/and3P=>/= V1 V2 V3 /and3P [E1 E2 E3] /and3P [F1 F2 F3]; apply/and3P.
by rewrite (fdomeqK V1 E1 F1) (fdomeqK V2 E2 F2) (fdomeqK V3 E3 F3).
Qed.

Section CancelV.
Variables (U : encoded_pcm) (V : encoded_type) (x : nat).
Variables (l1 l2 e1 e2 : U) (t1 t2 : V).

Lemma footeqVPt : footeqV (x ->> [l1, t1, e1]) (x ->> [l2, t2, e2]).
Proof. by rewrite !fsE /footeqV !fdomeqPt. Qed.

Lemma cancelV s1 s2 : 
        valid (x ->> [l1, t1, e1] \+ s1) ->
        footeqV (x ->> [l1, t1, e1] \+ s1)
                (x ->> [l2, t2, e2] \+ s2) ->
        valid s1 /\ footeqV s1 s2.
Proof.
move=>W H; split; first by apply: validR W. 
by apply: footeqVK W H footeqVPt. 
Qed.

Lemma congrV s1 s2 :
        footeqV s1 s2 ->
        footeqV (x ->> [l1, t1, e1] \+ s1) (x ->> [l2, t2, e2] \+ s2).
Proof. by apply: footeqVUn footeqVPt. Qed.

End CancelV.

Hint Resolve footeqV_refl footeqVPt.

(* footeqH *)

(* instantiate with h1 := flats s1, h2 := flats s2 to lift to states *)

Lemma footeqH_refl h : footeqH h h.
Proof. by []. Qed.

Lemma footeqH_sym h1 h2 : footeqH h1 h2 -> footeqH h2 h1.
Proof. by case; split. Qed.

Lemma footeqH_trans h1 h2 h3 : 
        footeqH h1 h2 -> footeqH h2 h3 -> footeqH h1 h3.
Proof. 
case=>V1 E1 [V2 E2]; split; first by rewrite V1 V2. 
by move=>x; rewrite E1 E2. 
Qed.

Add Parametric Relation : heap footeqH
  reflexivity proved by footeqH_refl
  symmetry proved by footeqH_sym
  transitivity proved by footeqH_trans as footeqH_rel.

Lemma footeqH_validUn h1 h2 h1' h2' : 
        footeqH h1 h2 -> footeqH h1' h2' -> 
        valid (h1 \+ h1') = valid (h2 \+ h2').
Proof. by move=>H1 H2; apply: domeq_validUn; apply/domeqP. Qed.

Lemma footeqH_validUnL {h} h1 h2 : 
        footeqH h1 h2 -> valid (h1 \+ h) = valid (h2 \+ h).
Proof. by move=>H; apply: footeqH_validUn H _. Qed.

Lemma footeqH_validUnR {h} h1' h2' : 
        footeqH h1' h2' -> valid (h \+ h1') = valid (h \+ h2').
Proof. by apply: footeqH_validUn. Qed.

Lemma footeqHUn h1 h2 h1' h2' : 
        footeqH h1 h2 -> footeqH h1' h2' -> 
        footeqH (h1 \+ h1') (h2 \+ h2').
Proof. by move=>H1 H2; apply/domeqP; apply: domeqUn; apply/domeqP. Qed.

Lemma footeqHUnL {h} h1 h2 : 
        footeqH h1 h2 -> footeqH (h1 \+ h) (h2 \+ h).
Proof. by move=>H; apply: footeqHUn H _. Qed.

Lemma footeqHUnR {h} h1' h2' : 
        footeqH h1' h2' -> footeqH (h \+ h1') (h \+ h2').
Proof. by apply: footeqHUn. Qed.

Add Parametric Morphism : (@PCM.join heapPCM) with
signature footeqH ==> footeqH ==> footeqH as heapjoin_morph.
Proof. by move=>x1 y1 H1 x2 y2 H2; apply: footeqHUn. Qed.

Lemma footeqHK f1 f2 f1' f2' : 
        valid (f1 \+ f1') ->
        footeqH (f1 \+ f1') (f2 \+ f2') ->
        footeqH f1 f2 -> footeqH f1' f2'.
Proof. 
by move=>V /domeqP H1 /domeqP H2; apply/domeqP; apply: domeqK V H1 H2. 
Qed.

Lemma footeqHPt A1 A2 x (v1 : A1) (v2 : A2) : footeqH (x :-> v1) (x :-> v2).
Proof. by split=>[|y]; [rewrite !hvalidPt | rewrite !hdomPt]. Qed.

Hint Resolve footeqHPt.

(*************************************)
(* Additional lemmas for zig/zagging *)
(*************************************)

Lemma validSO_zigzag s p : validSO (s \< p) = validSO (s \> p).
Proof. by rewrite /validSO !spP (zipC p) (zipAC _ p) zipA. Qed.

(* maybe avoid the precondition by baking it into flats *)
(* that is, flats may first check if validSO s is true *)
(* and return undef heap if not *)

(* the lemma is perhaps strenghtened and made symmetrict *)
(* by using validSO s, not validSO (s \< p) *)
(* but probably, the best thing is to just test for validity *)
(* in the definition of flats; then move these lemmas to state.v *)

Lemma flats_zigzag s p : validSO (s \< p) -> flats (s \< p) = flats (s \> p).
Proof.
move/allvalidV; rewrite !flatsE !spP=>V.
have V1: valid (p \* loc s). 
- by move: V; rewrite validZE; case/and3P.
have V2: valid (env s \* p).   
- by move: V; rewrite (zipC p) (zipAC _ p) -zipA validZE; case/and3P. 
by rewrite flatZ // flatZ // -!joinA joinC !joinA.
Qed.

Lemma validH_zigzag s p : 
        validSO (s \< p) -> validH (s \< p) = validH (s \> p).
Proof. by move=>V; rewrite /validH flats_zigzag. Qed.

Lemma valid_zigzag s p : 
        validSO (s \< p) -> valid (s \< p) = valid (s \> p).
Proof.
move/allvalidV; rewrite -[s]st_eta zigE zagE /= => V.
apply/idP/idP.
- case/and3P=>/= V1 V2 V3; apply/and3P=>/=.
  by move: V; rewrite zipC zipA; case/validZP.
case/and3P=>/= V1 V2 V3; apply/and3P=>/=.
by case/validZP: V.
Qed.

Lemma valid_lzig s p : valid (loc (s \< p)) -> valid (loc s) /\ valid p.
Proof.
rewrite -(st_eta s) zigE /=.
by case: validZP=>//; case.
Qed.

Lemma valid_ezag s p : valid (env (s \> p)) -> valid (env s) /\ valid p.
Proof.
rewrite -(st_eta s) zagE /=.
by case: validZP=>//; case.
Qed.

Lemma valid_zig s p : valid (s \< p) -> valid s /\ valid p.
Proof.
case/and3P; rewrite -[s]st_eta {4}/valid /=.
by case/validZP=>->-> _ ->->. 
Qed.

Lemma valid_zag s p : valid (s \> p) -> valid s /\ valid p.
Proof.
case/and3P; rewrite -[s]st_eta {4}/valid /= => ->->.
by case/validZP=>->->.
Qed.

Lemma fdom_zigs s p : 
        valid (loc (s \< p)) -> fdom (loc (s \< p)) =i fdom (loc s).
Proof.
rewrite -(st_eta s) zigE /= => V x.
by rewrite -fdomZU zunitZR V fdomZU.
Qed.

Lemma fdom_zigp s p : 
        valid (loc (s \< p)) -> fdom (loc (s \< p)) =i fdom p.
Proof.
rewrite -(st_eta s) zigE /= => V x.
by rewrite -fdomZU zunitZL V fdomZU.
Qed.

Lemma fdom_zags s p : 
        valid (env (s \> p)) -> fdom (env (s \> p)) =i fdom (env s).
Proof.
rewrite -(st_eta s) zagE /= => V x.
by rewrite -fdomZU zunitZL V fdomZU.
Qed.

Lemma fdom_zagp s p : 
        valid (env (s \> p)) -> fdom (env (s \> p)) =i fdom p.
Proof.
rewrite -(st_eta s) zagE /= => V x.
by rewrite -fdomZU zunitZR V fdomZU.
Qed.

(* now i need lemmas for interaction of \+ and zig-zag *)

(* the lemmas are stated with preconditions of validity *)
(* but in practice often what we have is coherence *)
(* so consider strenghtening the preconditions to coherence *)

Lemma zig_unitf s p : 
        s \< p = Unit -> s = Unit /\ p = Unit.
Proof. by rewrite -[s]st_eta; case/stE; case/zip_unitE=>->->->->. Qed.

Lemma zig_plusf s s1 s2 p : 
        valid (loc (s \< p)) ->
        s \< p = s1 \+ s2 ->  
        exists s1', exists s2', exists p1, exists p2, 
        [/\ s = s1' \+ s2', p = p1 \+ p2,
            s1 = s1' \< p1 & s2 = s2' \< p2].
Proof.
move=>V E; rewrite E in V; move: E; rewrite -(st_eta s) zigE.
case/stE=>/= /(zipUnf V) [f1][g1][f2][g2][E1 E2 ->{p} ->] ->->.
exists [g1, shr s1, env s1], [g2, shr s2, env s2], f1, f2.
by rewrite -(st_eta s1) -(st_eta s2) /= !zigE E1 E2.
Qed.

Lemma zag_unitf s p : 
        s \> p = Unit -> s = Unit /\ p = Unit.
Proof. by rewrite -[s]st_eta; case/stE=>/= ->-> /zip_unitE [->->]. Qed.

Lemma zag_plusf s s1 s2 p : 
        valid (env (s \> p)) ->
        s \> p = s1 \+ s2 ->  
        exists s1', exists s2', exists p1, exists p2, 
        [/\ s = s1' \+ s2', p = p1 \+ p2,
            s1 = s1' \> p1 & s2 = s2' \> p2].
Proof.
move=>V E; rewrite E in V; move: E; rewrite -(st_eta s) zagE.
case/stE=>/= ->-> /(zipUnf V) [f1][g1][f2][g2][E1 E2 -> ->{p}].
exists [loc s1, shr s1, f1], [loc s2, shr s2, f2], g1, g2.
by rewrite -(st_eta s1) -(st_eta s2) /= !zagE E1 E2.
Qed.

Lemma zig_unit s : loc s = Unit -> s \< Unit = s.
Proof. by move=>E; rewrite -[s]st_eta /= zigE E unit_zipE. Qed.

Lemma zig_plus s1 s2 p1 p2 : 
        valid (loc (s1 \< p1)) -> 
        valid (loc (s2 \< p2)) ->
        (s1 \+ s2) \< (p1 \+ p2) = (s1 \< p1) \+ (s2 \< p2).
Proof.
move=>V1 V2; apply/stE; rewrite /= !spP in V1 V2 *. 
by rewrite zipUn.
Qed.

Lemma zag_unit s : env s = Unit -> s \> Unit = s.
Proof. by move=>E; rewrite -[s]st_eta /= zagE E unit_zipE. Qed.

Lemma zag_plus s1 s2 p1 p2 : 
        valid (env (s1 \> p1)) -> 
        valid (env (s2 \> p2)) ->
        (s1 \+ s2) \> (p1 \+ p2) = (s1 \> p1) \+ (s2 \> p2).
Proof.
move=>V1 V2; apply/stE; rewrite /= !spP in V1 V2 *.
by rewrite zipUn.
Qed.

Lemma zagzig_eq s1 s2 p : 
        env s1 = env s2 -> s1 \> p = s2 \> p -> s1 = s2. 
Proof.
move=>V; move/stE; rewrite !spP; case=>V1 V2 V3.
by apply/stE; rewrite V1 V2 V. 
Qed.

Lemma zigzag_eq s1 s2 p : 
        loc s1 = loc s2 -> s1 \< p = s2 \< p -> s1 = s2. 
Proof.
move=>V; move/stE; rewrite !spP; case=>V1 V2 V3.
by apply/stE; rewrite V V2 V3. 
Qed.

(* the next lemma has a stronger postcondition than necessary *)
(* (all that's needed is valid (loc (si \< pi))), but this is what *)
(* I end up having in proofs where this is used, so that's what I go with *)
Lemma join_zagzig s1 s2 p1 p2 : 
        validSO (s1 \> p1) -> validSO (s2 \> p2) ->
        (s1 \+ s2) \> (p1 \+ p2) = s1 \> p1 \+ s2 \> p2 ->
        (s1 \+ s2) \< (p1 \+ p2) = s1 \< p1 \+ s2 \< p2.
Proof.
rewrite -!validSO_zigzag => Vso1 Vso2 E.
rewrite zig_plus //.
- by move/allvalidV: Vso1; case/validZP. 
by move/allvalidV: Vso2; case/validZP. 
Qed.

(****************************)
(* transposition of a state *)
(****************************)

Definition transp s := [env s, shr s, loc s].

Lemma transpI s : transp (transp s) = s.
Proof. by case: s. Qed.

Lemma transp0 : transp Unit = Unit.
Proof. by []. Qed.

Lemma transpUn s1 s2 : transp (s1 \+ s2) = transp s1 \+ transp s2.
Proof. by []. Qed.

Lemma transp_zig s p : transp (s \< p) = transp s \> p.
Proof. by rewrite -[s]st_eta /= zigE zagE zipC. Qed.

Lemma transp_zag s p : transp (s \> p) = transp s \< p.
Proof. by rewrite -[s]st_eta /= zigE zagE zipC. Qed.

Lemma transp_footV s1 s2 : footeqV s1 s2 = footeqV (transp s1) (transp s2).
Proof.
case: s1 s2=>lp1 sh1 ep1 [lp2 sh2 ep2]. 
by rewrite /footeqV andbA andbC andbA -andbAC -andbA. 
Qed.

Lemma transp_flatsE s : flats s = flats (transp s).
Proof. by case: s=>lp sh ep; rewrite !flatsE -!joinA joinC -joinA joinCA. Qed.

Lemma transp_inj s1 s2 : transp s1 = transp s2 <-> s1 = s2.
Proof. by split=>[|->//]; rewrite -{2}[s1]transpI -{2}[s2]transpI => ->. Qed.

Lemma env_transp s : env (transp s) = loc s. Proof. by []. Qed.

Lemma loc_transp s : loc (transp s) = env s. Proof. by []. Qed.

Lemma valid_transp s : valid (transp s) = valid s. 
Proof. by rewrite /valid /= andbC -(andbC (valid (loc s))) andbA. Qed.

